Package: HLAtools
Type: Package
Title: Toolkit for HLA Immunogenomics
Version: 1.11.0
Authors@R: 
    c(person("Livia", "Tran", , "livia.tran@ucsf.edu", role = c("aut")),
    person("Ryan", "Nickens", , "rynickens@gmail.com", role = c("aut")),
    person("Leamon", "Crooms IV", , "lcroomsiv@gmail.com", role = "aut"),
    person("Derek", "Pappas", , "djpappas75@gmail.com", role = "aut"),
    person("Vinh", "Luu", , "vinhluu864@berkeley.edu", role = "ctb"),
    person("Josh", "Bredeweg", , role = "ctb"),
    person("Steven", "Mack", , "Steven.Mack@ucsf.edu", role = c("aut", "cre"), 
           comment = c(ORCID = "0000-0001-9820-9547")))
Maintainer: Steven Mack <Steven.Mack@ucsf.edu>
Description: A toolkit for the analysis and management of data for genes in the so-called "Human Leukocyte Antigen" (HLA) region. Functions extract reference data from the Anthony Nolan HLA Informatics Group/ImmunoGeneTics HLA 'GitHub' repository (ANHIG/IMGTHLA) <https://github.com/ANHIG/IMGTHLA>, validate Genotype List (GL) Strings, convert between UNIFORMAT and GL String Code (GLSC) formats, translate HLA alleles and GLSCs across ImmunoPolymorphism Database (IPD) IMGT/HLA Database release versions, identify differences between pairs of alleles at a locus, generate customized, multi-position sequence alignments, trim and convert allele-names across nomenclature epochs, and extend existing data-analysis methods. Tran et al., (2025) <doi:10.1111/iji.70013>. 
Encoding: UTF-8
Imports: DescTools, dplyr, fmsb, rvest, stats, stringr, tibble, utils,
        xfun
Depends: R (>= 3.6.0)
LazyData: true
LazyDataCompression: xz
Suggests: knitr, rmarkdown, testthat (>= 3.0.0)
Config/testthat/edition: 3
VignetteBuilder: knitr
License: GPL (>= 3)
URL: <https://github.com/sjmack/HLAtools>
Config/roxygen2/version: 8.0.0
NeedsCompilation: no
Packaged: 2026-07-21 00:33:19 UTC; sjmack
Author: Livia Tran [aut],
  Ryan Nickens [aut],
  Leamon Crooms IV [aut],
  Derek Pappas [aut],
  Vinh Luu [ctb],
  Josh Bredeweg [ctb],
  Steven Mack [aut, cre] (ORCID: <https://orcid.org/0000-0001-9820-9547>)
Repository: CRAN
Date/Publication: 2026-07-22 08:10:14 UTC
Built: R 4.5.2; ; 2026-07-22 12:27:44 UTC; unix

Package: HLAtools
Type: Package
Title: Toolkit for HLA Immunogenomics
Version: 1.6.3
Authors@R: 
    c(person("Livia", "Tran", , "livia.tran@ucsf.edu", role = c("aut")),
    person("Ryan", "Nickens", , "rynickens@gmail.com", role = c("aut")),
    person("Leamon", "Crooms IV", , "lcroomsiv@gmail.com", role = "aut"),
    person("Derek", "Pappas", , "djpappas75@gmail.com", role = "aut"),
    person("Vinh", "Luu", , "vinhluu864@berkeley.edu", role = "ctb"),
    person("Josh", "Bredeweg", , role = "ctb"),
    person("Steven", "Mack", , "Steven.Mack@ucsf.edu", role = c("aut", "cre"), 
           comment = c(ORCID = "0000-0001-9820-9547")))
Maintainer: Steven Mack <Steven.Mack@ucsf.edu>
Description: A toolkit for the analysis and management of data for genes in the so-called "Human Leukocyte Antigen" (HLA) region. Functions extract reference data from the Anthony Nolan HLA Informatics Group/ImmunoGeneTics HLA 'GitHub' repository (ANHIG/IMGTHLA) <https://github.com/ANHIG/IMGTHLA>, validate Genotype List (GL) Strings, convert between UNIFORMAT and GL String Code (GLSC) formats, translate HLA alleles and GLSCs across ImmunoPolymorphism Database (IPD) IMGT/HLA Database release versions, identify differences between pairs of alleles at a locus, generate customized, multi-position sequence alignments, trim and convert allele-names across nomenclature epochs, and extend existing data-analysis methods.
Encoding: UTF-8
RoxygenNote: 7.3.2
Imports: DescTools, dplyr, fmsb, rvest, stats, stringr, tibble, utils,
        xfun
Depends: R (>= 3.6.0)
LazyData: true
LazyDataCompression: xz
Suggests: knitr, rmarkdown, testthat (>= 3.0.0)
Config/testthat/edition: 3
VignetteBuilder: knitr
License: GPL (>= 3)
URL: <https://github.com/sjmack/HLAtools>
NeedsCompilation: no
Packaged: 2025-07-21 05:06:52 UTC; sjmack
Author: Livia Tran [aut],
  Ryan Nickens [aut],
  Leamon Crooms IV [aut],
  Derek Pappas [aut],
  Vinh Luu [ctb],
  Josh Bredeweg [ctb],
  Steven Mack [aut, cre] (ORCID: <https://orcid.org/0000-0001-9820-9547>)
Repository: CRAN
Date/Publication: 2025-07-21 05:50:02 UTC
Built: R 4.6.0; ; 2026-04-30 12:24:08 UTC; unix

# Generated by roxygen2: do not edit by hand

export(BDstrat)
export(BDtoPyPop)
export(GIANT)
export(GLStoUNI)
export(GLV)
export(GLV2)
export(GLVhelper)
export(GLtoUN)
export(GLupdate)
export(GLvalidate)
export(UNItoGLS)
export(UNtoGL)
export(addCodonLine)
export(alignmentFull)
export(alignmentSearch)
export(alleleTrim)
export(atlasFull)
export(atlasMaker)
export(buildAlignments)
export(buildGazetteer)
export(buildGenomicReferences)
export(buildIMGTHLAGeneTypes)
export(checkAlignType)
export(checkSource)
export(checkVersion)
export(checkgDNAstart)
export(compareSequences)
export(convertAny)
export(countSpaces)
export(customAlign)
export(expandVersion)
export(ffN)
export(formatHead)
export(getAlignmentNames)
export(getField)
export(getLatestVersion)
export(motifMatch)
export(multiAlign)
export(multiAlleleTrim)
export(multiGLStoUNI)
export(multiLocusValidation)
export(multiQueryRelease)
export(multiSearch)
export(multiTranslateGLstring)
export(multiUNItoGLS)
export(multiUpdateGL)
export(numFields)
export(parseAlignmentHead)
export(posSort)
export(pypopHeaders)
export(queryPositions)
export(queryRelease)
export(redec)
export(relRisk)
export(repoVersion)
export(squashVersion)
export(translateAllele)
export(translateGLstring)
export(typeToSource)
export(uniAlign)
export(uniSearch)
export(updateAll)
export(updateAlleleListHistory)
export(updateGL)
export(validateAllele)
export(validateGLstring)
export(validateLocus)
export(validateMotif)
export(validatePositions)
export(validateUniformat)
export(validateVersion)
export(variantTable)
export(verifyAllele)
import(dplyr)
importFrom(DescTools,IsOdd)
importFrom(DescTools,StrLeft)
importFrom(dplyr,"%>%")
importFrom(dplyr,filter)
importFrom(fmsb,riskratio)
importFrom(rvest,html_elements)
importFrom(rvest,html_nodes)
importFrom(rvest,html_table)
importFrom(rvest,read_html)
importFrom(stats,setNames)
importFrom(stringr,"%>%")
importFrom(stringr,fixed)
importFrom(stringr,str_detect)
importFrom(stringr,str_replace)
importFrom(stringr,str_replace_all)
importFrom(stringr,str_squish)
importFrom(tibble,add_column)
importFrom(utils,capture.output)
importFrom(utils,head)
importFrom(utils,read.table)
importFrom(utils,tail)
importFrom(utils,write.table)
importFrom(xfun,numbers_to_words)

## HLAtools: Functions and Datasets for HLA Informatics

## Version 1.11.0

The HLA region is the most polymorphic section of the human genome, with 45,421 allelic variants identified across 46 loci. The key roles played by the class I and class II HLA genes in stem-cell therapy and transplantation, disease association research, evolutionary biology, and population genetics result in constant discovery of new allele variants. These data are curated and maintained by the [ImmunoPolymorphism Database-IMmunoGeneTics/HLA (IPD-IMGT/HLA) Database](https://www.ebi.ac.uk/ipd/imgt/hla/) and made available on the [Anthony Nolan HLA Informatics Group (ANHIG)/IMGTHLA GitHub repository](https://github.com/ANHIG/IMGTHLA) as static text files, which are updated every three months. Standardized use of the data in this key resource can be challenging. To address this, we have developed HLAtools, an R package that automates the consumption of IPD-IMGT/HLA resources, renders them computable, and makes them available alongside tools for data analysis, visualization and investigation. The package is compatible with all IPD-IMGT/HLA Database release versions up to release 3.65.0. 

HLAtools version 1.6.3 is available on the CRAN repository at [https://cran.r-project.org/package=HLAtools](https://cran.r-project.org/package=HLAtools).

Developmental versions of the package (in this repository) can be installed using the *[xfun](https://CRAN.R-project.org/package=xfun)::install_github(sjmack/HLAtools)* command.

A paper describing HLAtools was published in the [International Journal of Immunogenetics](https://onlinelibrary.wiley.com/doi/10.1111/iji.70013) in September of 2025. 

### Data Resources
The package includes six data objects that foster computation on IPD-IMGT/HLA resources. 

- 'IMGTHLAGeneTypes' describes the [named genes in the HLA region](https://hla.alleles.org/pages/genes/genes_list/).
- 'IMGTGenomicReferences' describes the [genomic reference allele](https://www.ebi.ac.uk/ipd/imgt/hla/help/genomics.html) for each gene supported in the IMGT/HLA Database and its GRCh38 reference position (when known).
- 'HLAgazetteer' defines specific categories of genes supported by the IPD-IMGT/HLA Database. For example:
   - gene fragments (HLAgazetteer$frag : "N" "P" "R" "S" "T" "U" "V" "W" "X" "Z") 
   - non-classical HLA genes (HLAgazetteer$nonclassical : "DMA"  "DMB"  "DOA"  "DOB"  "DPA2" "DPB2" "DQA2" "DQB2" "E"    "F"    "G")
   - genes specific to a particular conserved ancestral HLA haplotype block (HLAgazetteer$beta : "B"    "C"    "MICA" "MICB" "S"    "X")
- 'HLAatlas' identifies the boundaries between gene features (exons, introns and untranslated regions) at each gene, pseudogene and gene fragment.
- 'fragmentFeatureNames' identifies and annotates the non-standard features found in some gene fragments, based on the positions of feature boundaries ("|") in the sequence.
   - For example, the HLA-U gene fragment includes three features (fragmentFeatureNames\$U\$features : "N.1" "J.1" "S.1") that do not align to other class I gene features. The "N.1" feature is 54 nucleotides of novel (N) sequence, the "J.1" feature is a join (J) of the 3' end of Exon 3, and the 5' end of Intron 3, and the "S.1" feature is a segment (S) of Intron 3.
- 'alleleListHistory' is a computable index of the [names of all HLA alleles](https://github.com/ANHIG/IMGTHLA/tree/Latest/allelelist) for all IPD-IMGT/HLA Database release versions. 

'HLAgazetteer', 'HLAatlas', and 'alleleListHistory' can be updated with each IPD-IMGT/HLA Database release. 

In addition, the _alignmentFull()_ function builds the 'HLAalignments' object, which includes computable versions of the protein, codon, coding nucleotide and genomic alignments available in the [IMGTHLA GitHub repository](https://github.com/anhig/IMGTHLA), as specified by the user. 'HLAalignments' is not included the package, but can be built for IPD-IMGT/HLA Database releases 3.00.0 to 3.64.0.

### Trim, Search and Query Functions
The package includes a suite of functions for dissecting and describing similarities and differences between alleles and across loci.

- alleleTrim() trims HLA allele-names by fields or digits, with the option to include expression variant suffixes in truncated versions of full-length allele names that include these suffixes. multiAlleleTrim() extends alleleTrim() to a vector of allele names in the same nomenclature epoch.

```
alleleTrim(allele = "A*03:01:01", resolution = 2)
[1] "A*03:01"

alleleTrim(allele = "A*030101", resolution = 2,version = 2)
[1] "A*0301"

alleles <- c("A*01010102L","DRB1*1613N","HLA-Cw*0322Q")
multiAlleleTrim(alleles,1,2,TRUE
[1] "A*01L"      "DRB1*16N"   "HLA-Cw*03Q"
```
- compareSequences() identifies the positions that differ between a pair of alleles.
```
compareSequences(alignType = "gen", alleles = c("DPA1*01:03:01:04","DPA1*01:03:38:01"))
       allele_name 51 1544 1723 3318 4149
1 DPA1*01:03:01:04  C    A    A    G    C
2 DPA1*01:03:38:01  T    G    G    C    G
```

```
compareSequences("prot",c("DRB1*01:01:01:01","DRB3*01:01:02:01"))
       allele_name -19 -18 -17 -5 -2 9 10 11 13 26 28 30 31 32 37 38 57 60 71 73 74 77 96 98 104 183
1 DRB1*01:01:01:01   C   M   T  P  L W  Q  L  F  L  E  C  I  Y  S  V  D  Y  R  A  A  T  E  K   S   P
2 DRB3*01:01:02:01   S   L   A  R  F E  L  R  S  Y  D  Y  F  H  F  L  V  S  K  G  R  N  H  Q   A   A
```

- customAlign() builds customized peptide, codon, nucleotide and genomic alignments for specific alleles at user-specified positions, for alleles at different loci, and for different positions at each locus. 
```
customAlign("prot",c("DQA1*01:01:01:01","DQB1*05:01:01:01","DPB1*01:01:01:01"),list(1:4,5:8,5:8))
              DQA1 1 2 3 4
1 DQA1*01:01:01:01 E D I V
2             DQB1 5 6 7 8
3 DQB1*05:01:01:01 E D F V
4             DPB1 5 6 7 8
5 DPB1*01:01:01:01 E N Y V

customAlign("codon",c("DRB1*01:01","DQB1*02:01","DPB1*01:01"),c(1,2,3,7,8,9,13,14,15))
      Allele   1   2   3   7   8   9  13  14  15
1 DRB1*01:01 GGG GAC ACC TTC TTG TGG TTT GAA TGT
2 DQB1*02:01 AGA GAC TCT TTC GTG TAC GGC ATG TGC
3 DPB1*01:01 AGG GCC ACT TAC GTG TAC CAG GAA TGC

customAlign("prot",c("DPB1*01:01:01:01","DQA1*01:01:01:01","DQB1*05:01:01:01"),list(19:35,1:4,5:8))
              DPB1 19 20 21 22 23 24 25 26 27 28 29 30 31 32 33 34 35
1 DPB1*01:01:01:01  N  G  T  Q  R  F  L  E  R  Y  I  Y  N  R  E  E  Y
2             DQA1  1  2  3  4                                       
3 DQA1*01:01:01:01  E  D  I  V                                       
4             DQB1  5  6  7  8                                       
5 DQB1*05:01:01:01  E  D  F  V                                       
```

- motifMatch() searches the 'HLAalignments' object to identify full-length or two-field alleles that share user-specified nucleotide or peptide motifs.
```
motifMatch("A*-21M~2P","prot")
[1] "A*02:774"  "A*11:284"  "A*11:417N" "A*68:216N"

motifMatch("A*196G~301A~3046T","gen",FALSE)
[1] "A*01:09"

motifMatch("A*196G~301A~3046T","gen",TRUE)
[1] "A*01:09:01:01" "A*01:09:01:02"
```

- queryPositions() searches the 'HLAalignments' object to identify all of the variants at a specified position, and optionally returns a table of counts and frequencies for each variant.
```
queryPositions("codon","DRB1",c(56,86))
$DRB1_56
 [1] "ACT" "C.T" "CAT" "CCA" "CCC" "CCG" "CCT" "CGT" "CTT" "GCT" "TCT"

$DRB1_86
 [1] "AGT" "ATG" "GAT" "GCG" "GCT" "GGC" "GGG" "GGT" "GTG" "GTT" "TGT"
 
 queryPositions("codon","DRB1",c(56,86),TRUE)
$DRB1_56
   Variant Count    Frequency
1      ACT     1 0.0002614379
2      C.T     1 0.0002614379
3      CAT     3 0.0007843137
4      CCA     1 0.0002614379
5      CCC     2 0.0005228758
6      CCG     1 0.0002614379
7      CCT  3807 0.9952941176
8      CGT     1 0.0002614379
9      CTT     4 0.0010457516
10     GCT     2 0.0005228758
11     TCT     2 0.0005228758

$DRB1_86
   Variant Count    Frequency
1      AGT     2 0.0005228758
2      ATG     4 0.0010457516
3      GAT     5 0.0013071895
4      GCG     2 0.0005228758
5      GCT     4 0.0010457516
6      GGC     1 0.0002614379
7      GGG     3 0.0007843137
8      GGT  1931 0.5048366013
9      GTG  1869 0.4886274510
10     GTT     3 0.0007843137
11     TGT     1 0.0002614379
```

- queryRelease() searches the 'alleleListHistory' object for user-defined allele name elements in a specific IPD-IMGT/HLA release, identifying the number of alleles that match the query term, or a vector of allele names that match the query term. multiQueryRelease() extends queryRelease() to search for multiple allele name elements in a specific IPD-IMGT/HLA release.

```
queryRelease("3.30.0","DRB9",FALSE) 
[1] 1

queryRelease("3.30.0","DRB9",TRUE) 
[1] "DRB9*01:01"

queryRelease("3.31.0","DRB9",FALSE) 
[1] 6

queryRelease("1.05.0","304",TRUE) 
[1] "A*0304"    "A*3304"    "B*1304"    "Cw*03041"  "Cw*03042"  "DQB1*0304" "DRB1*0304" "DRB1*1304" "B*5304"    "A*2304"   

multiQueryRelease("3.58.0",c("DPB","2N"),TRUE)
[1] "DPB1*786:01:02N" "DPB1*401:01:02N"

multiQueryRelease("3.59.0",c("DRB","19","N"),TRUE)
[1] "DRB1*04:119N" "DRB5*02:19N"  "DRB1*14:195N" "DRB1*14:197N" "DRB1*13:319N"
```

- Additional functions include *validateAllele()*, which determines if the specified allele-name is present in the 'HLAalignments' object that has been loaded in the R environment, and *verifyAllele()*, which determines if the specified allele-name is present in the 'AlleleListHistory' object, and optionally identifies the IPD-IMGT/HLA Database releases including that allele, and the allele's IPD-IMGT/HLA accession number.

```
validateAllele("A*01:01:01:117")
[1] TRUE

validateAllele("A*01:01:01")
A*01:01:01 is not found in version 3.55.0 alignments.
[1] FALSE

verifyAllele("A*01:01:01:01")
[1] TRUE

verifyAllele("A*01:01:01:01",TRUE)
[1] "TRUE"   "3.55.0"

verifyAllele("A*010101",TRUE)
[1] "TRUE"   "2.09.0"

verifyAllele("A*010101",TRUE,TRUE)
 [1] "TRUE"   "2.09.0" "2.08.0" "2.07.0" "2.06.0" "2.05.0" "2.04.0" "2.03.0" "2.02.0" "2.01.0" "2.00.0" "1.16.0"
 
verifyAllele("A*010101",TRUE,TRUE,FALSE,TRUE)
 [1] "TRUE"    "2.09.0"   "2.08.0"   "2.07.0"   "2.06.0"   "2.05.0"   "2.04.0"   "2.03.0"   "2.02.0"   "2.01.0"   "2.00.0"   "1.16.0"   "HLA00001"

verifyAllele("A*010101",TRUE,TRUE,TRUE)
[1] "TRUE"   "1.16.0"

verifyAllele("A*0101",TRUE)
[1] "TRUE"   "1.06.0"

```

### Data Format Conversion and Translation Functions
The package includes functions that convert [Genotype List (GL) String Codes](https://glstring.org) across IPD/IMGT-HLA Database release versions and nomenclature epochs, inter-convert between [GL String](https://glstring.org) and [UNIFORMAT](https://hla-net.eu/tools/uniformate/) formats, and translate data frames and vectors of HLA allele name data across IPD-IMGT/HLA Database release versions.  

- updateGL() converts HLA allele names in GL String Code objects across IPD/IMGT-HLA Database release versions. 

```
updateGL("hla#1.07.0#HLA-B*35011", "2.05.0")
[1] "hla#2.05.0#HLA-B*350101"

updateGL("hla#2.05.0#HLA-B*350101", "3.05.0")
[1] "hla#3.05.0#HLA-B*35:01:01:01"

updateGL("hla#3.55.0#HLA-B*44:02:01:01","1.05.0")
[1] "hla#1.05.0#HLA-B*4402"

updateGL("hla#3.25.0#HLA-A*01:01:01:01/HLA-A*01:02+HLA-A*24:02:01:01","3.55.0")
[1] "hla#3.55.0#HLA-A*01:01:01:01/HLA-A*01:02:01:01+HLA-A*24:02:01:01"

updateGL("hla#3.25.0#HLA-A*01:01:01:01/HLA-A*01:02+HLA-A*24:02:01:01","1.05.0")
[1] "hla#1.05.0#HLA-A*0101/HLA-A*0102+HLA-A*2402101"
```

- GLStoUNI() and UNItoGLS() convert between the GL String and UNIFORMAT grammars
```
GLStoUNI("HLA-A*02:01/HLA-A*02:02+HLA-A*03:01/HLA-A*03:02")
[1] "A*02:01,A*03:01|A*02:01,A*03:02|A*02:02,A*03:01|A*02:02,A*03:02"

UNItoGLS("A*02:01,A*03:01|A*02:01,A*03:02|A*02:02,A*03:01|A*02:02,A*03:02")
[1] "HLA-A*02:01/HLA-A*02:02+HLA-A*03:01/HLA-A*03:02"
```

- GIANT() translates vectors and data frames of HLA allele names across specified IPD-IMGT/HLA Database release versions. 
```
GIANT(c("A*01:01:01:01","DQA1*01:01:01:01"),"3.56.0","3.00.0")
[1] "A*01:01:01:01" "DQA1*01:01:01"

GIANT(c("A*01:01:01:01","DQA1*01:01:01:01"),"3.56.0","2.20.0")
[1] "A*01010101"  "DQA1*010101"

GIANT(c("A*01:01:01:01","DQA1*01:01:01:01"),"3.56.0","1.09.0")
[1] "A*01011"   "DQA1*0101"

GIANT(sHLAdata,"3.56.0","2.23.0")[1:5,1:14]
   Subject Status        A      A.1      C    C.1      B    B.1   DRB1 DRB1.1   DQA1 DQA1.1   DQB1 DQB1.1
1 UT900-23      0     <NA>     <NA> 010201 020205   1301 180102 160201   0404 050101 050102 030101   0202
2 UT900-24      0 01010101 02010101   0307   0605   1401 390201   0402 160201 040101 050101 060101 030101
3 UT900-25      0     0210   030102   0712 010201   1520   1301 080201 040701   0201 050102 030302   0202
4 UT900-26      0 01010101     0218   0804 120201 350901   4005 090102 080201 030101   0201   0304 030302
5 UT910-01      0   250101 02010101   1507   0307 510103   1401 110101 090102 050102 030101 050301   0304
```

### Data Analysis Functions
The package includes three data-analysis functions that accept Bridging ImmunoGenomic Data-Analysis Workflow Gaps (BIGDAWG)-formatted genotype datasets as input. 

To facilitate testing and experimentation with these functions, the 'sHLAdata' data object is included in the package. This BIGDAWG-formatted HLA genotype data object represents completely synthetic IPD-IMGT/HLA release 3.56.0 HLA-A, -C, -B, -DRB1, -DQA1, -DQB1, -DPA1, and -DPB1 genotype data for 24 control subjects and 23 case subjects, and does not represent any true human population.

- relRisk() calculates relative risk (RR) values, confidence interval (CI) values and p-values for BIGDAWG-formatted non-case-control genotype datasets. For these analyses, two subject categories are required, but should not be affected/patient and unaffected/control categories; instead, the categories may be, e.g., either of two disease states, where one disease state is coded as 0 and the other is coded as 1 in the second column of the dataset.

```
rr <- relRisk(sHLAdata)

rr$alleles[[1]][c(1:3),]
  Locus     Variant Status_1 Status_0      RelativeRisk            CI.low          CI.high    p.value Significant
1     A 01:01:01:01        5        7 0.833333333333333 0.411641930809013 1.68701094924813 0.59291160            
2     A 02:01:01:01        8       11 0.830409356725146 0.467284553053116  1.4757168736504 0.50779167            
3     A       02:10        4        5               0.9 0.419645026972693 1.93020278553833 0.77978869            

rr$genotypes[[1]][c(1:3),]
  Locus                 Variant Status_1 Status_0 RelativeRisk     CI.low     CI.high      p.value Significant
1     A 01:01:01:01+02:01:01:01        0        2            0          0         NaN  0.161777420            
2     A       01:01:01:01+02:10        1        0   2.0952380   1.5379358  2.85449005  0.306556273            
3     A       01:01:01:01+02:18        0        4            0          0         NaN  0.042730188           *
```

- BDstrat() stratifies BIGDAWG-formatted case-control datasets for individual alleles or multiple alleles at multiple loci, and generates two BIGDAWG-formatted datasets; one for the case and control subjects that have those alleles, and one for the case and control subjects that do not.

```
HLA_data.multi.strat <- BDstrat(sHLAdata,c("DRB1*16:02:01:01","DRB1*04:07:01:01","A*25:01:01:01"),15)

HLA_data.multi.strat$`DRB1*16:02:01:01+DRB1*04:07:01:01+A*25:01:01:01-positive`[1:2,1:8]
   Subject Status           A         A.1           C      C.1           B         B.1
1 UT900-23      0        <NA>        <NA> 01:02:01:01 02:10:06 13:01:01:01    18:01:02
2 UT900-24      0 01:01:01:01 02:01:01:01 03:07:01:01    06:05 14:01:01:01 39:02:01:01

HLA_data.multi.strat$`DRB1*16:02:01:01+DRB1*04:07:01:01+A*25:01:01:01-negative`[1:2,1:8]
   Subject Status           A   A.1           C         C.1           B         B.1
4 UT900-26      0 01:01:01:01 02:18 08:04:01:01    12:02:01 35:09:01:01 40:05:01:01
6 UT910-02      0       02:10 32:04 18:01:01:01 01:02:01:01    78:02:01 13:01:01:01
```

- BDtoPyPop() converts a BIGDAWG-formatted case-control dataset into two PyPop version 1.\*.\* formatted datasets -- one for all 'case' subjects and one for all 'control' subjects.  

```
HLAdata.PP <- BDtoPyPop(sHLAdata,"sHLA",FALSE)

HLAdata.PP$sHLA.neagtive[1:3,1:10]
   Subject Status         A_1         A_2         C_1         C_2         B_1         B_2      DRB1_1      DRB1_2
1 UT900-23      0        ****        **** 01:02:01:01    02:10:06 13:01:01:01    18:01:02 16:02:01:01 04:04:01:01
2 UT900-24      0 01:01:01:01 02:01:01:01 03:07:01:01       06:05 14:01:01:01 39:02:01:01    04:02:01 16:02:01:01
3 UT900-25      0       02:10    03:01:02       07:12 01:02:01:01       15:20 13:01:01:01 08:02:01:01 04:07:01:01

HLAdata.PP$sHLA.positive[1:3,1:10]
    Subject Status      A_1      A_2         C_1         C_2         B_1         B_2      DRB1_1      DRB1_2
25 UT910-30      1    02:18    02:18 01:02:01:01 03:07:01:01 40:05:01:01 14:01:01:01 11:01:01:01 04:07:01:01
26 UT910-31      1     ****     **** 03:07:01:01        **** 14:01:01:01    78:02:01    04:02:01 08:02:01:01
27 UT910-32      1 03:01:02 03:01:02       06:05 03:07:01:01 39:02:01:01 14:01:01:01 08:02:01:01 08:02:01:01

```

## HLAtools: Functions and Datasets for HLA Informatics

## Version 1.6.3

The HLA region is the most polymorphic section of the human genome, with 43,225 allelic variants identified across 46 loci. The key roles played by the class I and class II HLA genes in stem-cell therapy and transplantation, disease association research, evolutionary biology, and population genetics result in constant discovery of new allele variants. These data are curated and maintained by the [ImmunoPolymorphism Database-IMmunoGeneTics/HLA (IPD-IMGT/HLA) Database](https://www.ebi.ac.uk/ipd/imgt/hla/) and made available on the [Anthony Nolan HLA Informatics Group (ANHIG)/IMGTHLA GitHub repository](https://github.com/ANHIG/IMGTHLA) as static text files, which are updated every three months. Standardized use of the data in this key resource can be challenging. To address this, we have developed HLAtools, an R package that automates the consumption of IPD-IMGT/HLA resources, renders them computable, and makes them available alongside tools for data analysis, visualization and investigation. The package is compatible with all IPD-IMGT/HLA Database release versions up to release 3.60.0. 

HLAtools version 1.6.2 is available on the CRAN repository at [https://cran.r-project.org/package=HLAtools](https://cran.r-project.org/package=HLAtools). 

Developmental versions of the package (in this repository) can be installed using the *[xfun](https://CRAN.R-project.org/package=xfun)::install_github(sjmack/HLAtools)* command.

### Data Resources
The package includes five data objects that foster computation on IPD-IMGT/HLA resources. 

- 'IMGTHLAGeneTypes' describes the [named genes in the HLA region](https://hla.alleles.org/pages/genes/genes_list/).
- 'HLAgazetteer' defines specific categories of genes supported by the IPD-IMGT/HLA Database. For example:
   - gene fragments (HLAgazetteer$frag : "N" "P" "R" "S" "T" "U" "V" "W" "X" "Z") 
   - non-classical HLA genes (HLAgazetteer$nonclassical : "DMA"  "DMB"  "DOA"  "DOB"  "DPA2" "DPB2" "DQA2" "DQB2" "E"    "F"    "G")
- 'HLAatlas' identifies the boundaries between gene features (exons, introns and untranslated regions) at each gene, pseudogene and gene fragment.
- 'fragmentFeatureNames' identifies and annotates the non-standard features found in some gene fragments, based on the positions of feature boundaries ("|") in the sequence.
   - For example, the HLA-U gene fragment includes three features (fragmentFeatureNames\$U\$features : "N.1" "J.1" "S.1") that do not align to other class I gene features. The "N.1" feature is 54 nucleotides of novel (N) sequence, the "J.1" feature is a join (J) of the 3' end of Exon 3, and the 5' end of Intron 3, and the "S.1" feature is a segment (S) of Intron 3.
- 'alleleListHistory' is a computable index of the [names of all HLA alleles](https://github.com/ANHIG/IMGTHLA/tree/Latest/allelelist) for all IPD-IMGT/HLA Database release versions. 

HLAgazetteer, HLAatlas, and alleleListHistory can be updated with each IPD-IMGT/HLA Database release. 

In addition, the _alignmentFull()_ function builds the 'HLAalignments' object, which includes computable versions of the protein, codon, coding nucleotide and genomic alignments available in the [IMGTHLA GitHub repository](https://github.com/anhig/IMGTHLA), as specified by the user. 'HLAalignments' is not included the package, but can be built for IPD-IMGT/HLA Database releases 3.00.0 to 3.60.0.

### Trim, Search and Query Functions
The package includes a suite of functions for dissecting and describing similarities and differences between alleles and across loci.

- alleleTrim() trims HLA allele-names by fields or digits, with the option to include expression variant suffixes in truncated versions of full-length allele names that include these suffixes. multiAlleleTrim() extends alleleTrim() to a vector of allele names in the same nomenclature epoch.

```
alleleTrim(allele = "A*03:01:01", resolution = 2)
[1] "A*03:01"

alleleTrim(allele = "A*030101", resolution = 2,version = 2)
[1] "A*0301"

alleles <- c("A*01010102L","DRB1*1613N","HLA-Cw*0322Q")
multiAlleleTrim(alleles,1,2,TRUE)
[1] "A*01L"      "DRB1*16N"   "HLA-Cw*03Q"
```
- compareSequences() identifies the positions that differ between a pair of alleles at a locus.
```
compareSequences(alignType = "gen", alleles = c("DPA1*01:03:01:04","DPA1*01:03:38:01"))
       allele_name 51 1544 1723 3318 4149
1 DPA1*01:03:01:04  C    A    A    G    C
2 DPA1*01:03:38:01  T    G    G    C    G
```
- customAlign() builds customized peptide, codon, nucleotide and genomic alignments for specific alleles at user-specified positions, for alleles at different loci, and for different positions at each locus. 
```
customAlign("prot",c("DQA1*01:01:01:01","DQB1*05:01:01:01","DPB1*01:01:01:01"),list(1:4,5:8,5:8))
              DQA1 1 2 3 4
1 DQA1*01:01:01:01 E D I V
2             DQB1 5 6 7 8
3 DQB1*05:01:01:01 E D F V
4             DPB1 5 6 7 8
5 DPB1*01:01:01:01 E N Y V

customAlign("codon",c("DRB1*01:01","DQB1*02:01","DPB1*01:01"),c(1,2,3,7,8,9,13,14,15))
      Allele   1   2   3   7   8   9  13  14  15
1 DRB1*01:01 GGG GAC ACC TTC TTG TGG TTT GAA TGT
2 DQB1*02:01 AGA GAC TCT TTC GTG TAC GGC ATG TGC
3 DPB1*01:01 AGG GCC ACT TAC GTG TAC CAG GAA TGC

customAlign("prot",c("DPB1*01:01:01:01","DQA1*01:01:01:01","DQB1*05:01:01:01"),list(19:35,1:4,5:8))
              DPB1 19 20 21 22 23 24 25 26 27 28 29 30 31 32 33 34 35
1 DPB1*01:01:01:01  N  G  T  Q  R  F  L  E  R  Y  I  Y  N  R  E  E  Y
2             DQA1  1  2  3  4                                       
3 DQA1*01:01:01:01  E  D  I  V                                       
4             DQB1  5  6  7  8                                       
5 DQB1*05:01:01:01  E  D  F  V                                       
```

- motifMatch() searches the HLAalignments object to identify full-length or two-field alleles that share user-specified nucleotide or peptide motifs.
```
motifMatch("A*-21M~2P","prot")
[1] "A*02:774"  "A*11:284"  "A*11:417N" "A*68:216N"

motifMatch("A*196G~301A~3046T","gen",FALSE)
[1] "A*01:09"

motifMatch("A*196G~301A~3046T","gen",TRUE)
[1] "A*01:09:01:01" "A*01:09:01:02"
```

- queryPositions() searches the HLAalignment object to identify all of the variants at a specified position, and optionally returns a table of counts and frequencies for each variant.
```
queryPositions("codon","DRB1",c(56,86))
$DRB1_56
 [1] "ACT" "C.T" "CAT" "CCA" "CCC" "CCG" "CCT" "CGT" "CTT" "GCT" "TCT"

$DRB1_86
 [1] "AGT" "ATG" "GAT" "GCG" "GCT" "GGC" "GGG" "GGT" "GTG" "GTT" "TGT"
 
 queryPositions("codon","DRB1",c(56,86),TRUE)
$DRB1_56
   Variant Count    Frequency
1      ACT     1 0.0002614379
2      C.T     1 0.0002614379
3      CAT     3 0.0007843137
4      CCA     1 0.0002614379
5      CCC     2 0.0005228758
6      CCG     1 0.0002614379
7      CCT  3807 0.9952941176
8      CGT     1 0.0002614379
9      CTT     4 0.0010457516
10     GCT     2 0.0005228758
11     TCT     2 0.0005228758

$DRB1_86
   Variant Count    Frequency
1      AGT     2 0.0005228758
2      ATG     4 0.0010457516
3      GAT     5 0.0013071895
4      GCG     2 0.0005228758
5      GCT     4 0.0010457516
6      GGC     1 0.0002614379
7      GGG     3 0.0007843137
8      GGT  1931 0.5048366013
9      GTG  1869 0.4886274510
10     GTT     3 0.0007843137
11     TGT     1 0.0002614379
```

- queryRelease() searches the alleleListHistory object for user-defined allele name elements in a specific IPD-IMGT/HLA release, identifying the number of alleles that match the query term, or a vector of allele names that match the query term. multiQueryRelease() extends queryRelease() to search for multiple allele name elements in a specific IPD-IMGT/HLA release.

```
queryRelease("3.30.0","DRB9",FALSE) 
[1] 1

queryRelease("3.30.0","DRB9",TRUE) 
[1] "DRB9*01:01"

queryRelease("3.31.0","DRB9",FALSE) 
[1] 6

queryRelease("1.05.0","304",TRUE) 
[1] "A*0304"    "A*3304"    "B*1304"    "Cw*03041"  "Cw*03042"  "DQB1*0304" "DRB1*0304" "DRB1*1304" "B*5304"    "A*2304"   

multiQueryRelease("3.58.0",c("DPB","2N"),TRUE)
[1] "DPB1*786:01:02N" "DPB1*401:01:02N"

multiQueryRelease("3.59.0",c("DRB","19","N"),TRUE)
[1] "DRB1*04:119N" "DRB5*02:19N"  "DRB1*14:195N" "DRB1*14:197N" "DRB1*13:319N"
```

- Additional functions include *validateAllele()*, which determines if the specified allele-name is present in the 'HLAalignments' object that has been loaded in the R environment, and *verifyAllele()*, which determines if the specified allele-name is present in the 'AlleleListHistory' object, and optionally identifies the IPD-IMGT/HLA Database releases including that allele, and the allele's IPD-IMGT/HLA accession number.

```
validateAllele("A*01:01:01:117")
[1] TRUE

validateAllele("A*01:01:01")
A*01:01:01 is not found in version 3.55.0 alignments.
[1] FALSE

verifyAllele("A*01:01:01:01")
[1] TRUE

verifyAllele("A*01:01:01:01",TRUE)
[1] "TRUE"   "3.55.0"

verifyAllele("A*010101",TRUE)
[1] "TRUE"   "2.09.0"

verifyAllele("A*010101",TRUE,TRUE)
 [1] "TRUE"   "2.09.0" "2.08.0" "2.07.0" "2.06.0" "2.05.0" "2.04.0" "2.03.0" "2.02.0" "2.01.0" "2.00.0" "1.16.0"
 
verifyAllele("A*010101",TRUE,TRUE,FALSE,TRUE)
 [1] "TRUE"    "2.09.0"   "2.08.0"   "2.07.0"   "2.06.0"   "2.05.0"   "2.04.0"   "2.03.0"   "2.02.0"   "2.01.0"   "2.00.0"   "1.16.0"   "HLA00001"

verifyAllele("A*010101",TRUE,TRUE,TRUE)
[1] "TRUE"   "1.16.0"

verifyAllele("A*0101",TRUE)
[1] "TRUE"   "1.06.0"

```

### Data Format Conversion and Translation Funtions
The package includes functions that convert [Genotype List (GL) String Codes](https://glstring.org) across IPD/IMGT-HLA Database release versions and nomenclature epochs, inter-convert between [GL String](https://glstring.org) and [UNIFORMAT](https://hla-net.eu/tools/uniformate/) formats, and translate data frames and vectors of HLA allele name data across IPD-IMGT/HLA Database release versions.  

- updateGL() converts HLA allele names in GL String Code objects across IPD/IMGT-HLA Database release versions. 

```
updateGL("hla#1.07.0#HLA-B*35011", "2.05.0")
[1] "hla#2.05.0#HLA-B*350101"

updateGL("hla#2.05.0#HLA-B*350101", "3.05.0")
[1] "hla#3.05.0#HLA-B*35:01:01:01"

updateGL("hla#3.55.0#HLA-B*44:02:01:01","1.05.0")
[1] "hla#1.05.0#HLA-B*4402"

updateGL("hla#3.25.0#HLA-A*01:01:01:01/HLA-A*01:02+HLA-A*24:02:01:01","3.55.0")
[1] "hla#3.55.0#HLA-A*01:01:01:01/HLA-A*01:02:01:01+HLA-A*24:02:01:01"

updateGL("hla#3.25.0#HLA-A*01:01:01:01/HLA-A*01:02+HLA-A*24:02:01:01","1.05.0")
[1] "hla#1.05.0#HLA-A*0101/HLA-A*0102+HLA-A*2402101"
```

- GLStoUNI() and UNItoGLS() convert between the GL String and UNIFORMAT grammars
```
GLStoUNI("HLA-A*02:01/HLA-A*02:02+HLA-A*03:01/HLA-A*03:02")
[1] "A*02:01,A*03:01|A*02:01,A*03:02|A*02:02,A*03:01|A*02:02,A*03:02"

UNItoGLS("A*02:01,A*03:01|A*02:01,A*03:02|A*02:02,A*03:01|A*02:02,A*03:02")
[1] "HLA-A*02:01/HLA-A*02:02+HLA-A*03:01/HLA-A*03:02"
```

- GIANT() translates vectors and data frames of HLA allele names across specified IPD-IMGT/HLA Database release versions. 
```
GIANT(c("A*01:01:01:01","DQA1*01:01:01:01"),"3.56.0","3.00.0")
[1] "A*01:01:01:01" "DQA1*01:01:01"

GIANT(c("A*01:01:01:01","DQA1*01:01:01:01"),"3.56.0","2.20.0")
[1] "A*01010101"  "DQA1*010101"

GIANT(c("A*01:01:01:01","DQA1*01:01:01:01"),"3.56.0","1.09.0")
[1] "A*01011"   "DQA1*0101"

GIANT(sHLAdata,"3.56.0","2.23.0")[1:5,1:14]
   Subject Status        A      A.1      C    C.1      B    B.1   DRB1 DRB1.1   DQA1 DQA1.1   DQB1 DQB1.1
1 UT900-23      0     <NA>     <NA> 010201 020205   1301 180102 160201   0404 050101 050102 030101   0202
2 UT900-24      0 01010101 02010101   0307   0605   1401 390201   0402 160201 040101 050101 060101 030101
3 UT900-25      0     0210   030102   0712 010201   1520   1301 080201 040701   0201 050102 030302   0202
4 UT900-26      0 01010101     0218   0804 120201 350901   4005 090102 080201 030101   0201   0304 030302
5 UT910-01      0   250101 02010101   1507   0307 510103   1401 110101 090102 050102 030101 050301   0304
```

### Data Analysis Functions
The package includes three data-analysis functions that accept Bridging ImmunoGenomic Data-Analysis Workflow Gaps (BIGDAWG)-formatted genotype datasets as input. 

To facilitate testing and experimentation with these functions, the 'sHLAdata' data object is included in the package. This BIGDAWG-formatted HLA genotype data object represents completely synthetic IPD-IMGT/HLA release 3.56.0 HLA-A, -C, -B, -DRB1, -DQA1, -DQB1, -DPA1, and -DPB1 genotype data for 24 control subjects and 23 case subjects, and does not represent any true human population.

- relRisk() calculates relative risk (RR) values, confidence interval (CI) values and p-values for BIGDAWG-formatted non-case-control genotype datasets. For these analyses, two subject categories are required, but should not be affected/patient and unaffected/control categories; instead, the categories may be, e.g., either of two disease states, where one disease state is coded as 0 and the other is coded as 1 in the second column of the dataset.

```
rr <- relRisk(sHLAdata)

rr$alleles[[1]][c(1:3),]
  Locus     Variant Status_1 Status_0      RelativeRisk            CI.low          CI.high    p.value Significant
1     A 01:01:01:01        5        7 0.833333333333333 0.411641930809013 1.68701094924813 0.59291160            
2     A 02:01:01:01        8       11 0.830409356725146 0.467284553053116  1.4757168736504 0.50779167            
3     A       02:10        4        5               0.9 0.419645026972693 1.93020278553833 0.77978869            

rr$genotypes[[1]][c(1:3),]
  Locus                 Variant Status_1 Status_0 RelativeRisk     CI.low     CI.high      p.value Significant
1     A 01:01:01:01+02:01:01:01        0        2            0          0         NaN  0.161777420            
2     A       01:01:01:01+02:10        1        0   2.0952380   1.5379358  2.85449005  0.306556273            
3     A       01:01:01:01+02:18        0        4            0          0         NaN  0.042730188           *
```

- BDstrat() stratifies BIGDAWG-formatted case-control datasets for individual alleles or multiple alleles at multiple loci, and generates two BIGDAWG-formatted datasets; one for the case and control subjects that have those alleles, and one for the case and control subjects that do not.

```
HLA_data.multi.strat <- BDstrat(sHLAdata,c("DRB1*16:02:01:01","DRB1*04:07:01:01","A*25:01:01:01"),15)

HLA_data.multi.strat$`DRB1*16:02:01:01+DRB1*04:07:01:01+A*25:01:01:01-positive`[1:2,1:8]
   Subject Status           A         A.1           C      C.1           B         B.1
1 UT900-23      0        <NA>        <NA> 01:02:01:01 02:10:06 13:01:01:01    18:01:02
2 UT900-24      0 01:01:01:01 02:01:01:01 03:07:01:01    06:05 14:01:01:01 39:02:01:01

HLA_data.multi.strat$`DRB1*16:02:01:01+DRB1*04:07:01:01+A*25:01:01:01-negative`[1:2,1:8]
   Subject Status           A   A.1           C         C.1           B         B.1
4 UT900-26      0 01:01:01:01 02:18 08:04:01:01    12:02:01 35:09:01:01 40:05:01:01
6 UT910-02      0       02:10 32:04 18:01:01:01 01:02:01:01    78:02:01 13:01:01:01
```

- BDtoPyPop() converts a BIGDAWG-formatted case-control dataset into two PyPop version 1.\*.\* formatted datasets -- one for all 'case' subjects and one for all 'control' subjects.  

```
HLAdata.PP <- BDtoPyPop(sHLAdata,"sHLA",FALSE)

HLAdata.PP$sHLA.neagtive[1:3,1:10]
   Subject Status         A_1         A_2         C_1         C_2         B_1         B_2      DRB1_1      DRB1_2
1 UT900-23      0        ****        **** 01:02:01:01    02:10:06 13:01:01:01    18:01:02 16:02:01:01 04:04:01:01
2 UT900-24      0 01:01:01:01 02:01:01:01 03:07:01:01       06:05 14:01:01:01 39:02:01:01    04:02:01 16:02:01:01
3 UT900-25      0       02:10    03:01:02       07:12 01:02:01:01       15:20 13:01:01:01 08:02:01:01 04:07:01:01

HLAdata.PP$sHLA.positive[1:3,1:10]
    Subject Status      A_1      A_2         C_1         C_2         B_1         B_2      DRB1_1      DRB1_2
25 UT910-30      1    02:18    02:18 01:02:01:01 03:07:01:01 40:05:01:01 14:01:01:01 11:01:01:01 04:07:01:01
26 UT910-31      1     ****     **** 03:07:01:01        **** 14:01:01:01    78:02:01    04:02:01 08:02:01:01
27 UT910-32      1 03:01:02 03:01:02       06:05 03:07:01:01 39:02:01:01 14:01:01:01 08:02:01:01 08:02:01:01

```

#### Compare Allele Sequences v1.0.1 May 29, 2026 

################
##compareSequences
#'Identify Sequence Differences Between Two Alleles
#'
#'Compares the sequences of two alleles, and identifies the differences between them at shared polymorphic sequence positions.
#'Alleles at different loci can be compared. 
#'
#'@param alignType A character string identifying the type of alignment being searched. Allowed values are "codon","gen", nuc" and "prot". Only one 'alignType' value is permitted.
#'@param alleles A character vector containing two full-length names for alleles at the two different loci.
#'
#'@return A two-row data frame identifying the positions and sequences at which the two alleles differ. E.g., compareLoci(alignType = "gen", alleles = c("B\*07:02:01:01","C\*02:02:02:01")). Positions are nor shared between the two loci, or for which the sequence of either allele is unknown are ignored. 
#'
#'@examples
#'\dontrun{
#'compareSequences("gen",c("DPA1*01:03:38:01","DPA1*01:03:38:02"))
#'compareSequences("prot",c("DRB1*01:01:01:01","DRB3*01:01:02:01"))
#'compareSequences("prot", c("B*57:01:01:01", "A*24:02:01:01"))
#'}
#'
#'@export
#'
compareSequences <- function(alignType,alleles) {
  # Data checks
  if(length(alleles)!=2) {stop(paste("Please include exactly two alleles."))}
  
  if(length(alignType)!=1) {stop(paste("Please specify only one 'alignType'."))}
  
  alignType <- checkAlignType(alignType)
  
  locus <- c(alleles[1],alleles[2])
  
  locus[1] <- strsplit(alleles[1],"*",fixed=TRUE)[[1]][1]
  locus[2] <- strsplit(alleles[2],"*",fixed=TRUE)[[1]][1]
  
  if(!validateLocus(locus[1],typeToSource(alignType))) { stop()}
  if(!validateLocus(locus[2],typeToSource(alignType))) { stop()}
  
  if(!validateAllele(alleles[1])) {return(paste(alleles[1], "is not found in the", alignType, "alignment for",paste(locus[1],".",sep=""),sep=" "))}
  if(!validateAllele(alleles[2])) {return(paste(alleles[2], "is not found in the", alignType, "alignment for",paste(locus[2],".",sep=""),sep=" "))}
  
  stopMessage <- ""
  stopErr <- FALSE
  
  for(i in 1:2) {
    if(!alleles[i] %in% HLAalignments[[alignType]][[locus[i]]]$allele_name) {
      stopMessage <- paste(stopMessage,alleles[i]," is not a full-length allele name.\n",sep="")
      stopErr <- TRUE
    }
  }
  
  if(stopErr) {stop(paste("\n",stopMessage,"Please use full-length allele names.",sep="\n"))}
  
  if(alleles[1] == alleles[2]) { return(paste(alleles[1],"and",alleles[2],"are identical.",sep=" ")) }
  
  # Function body
  coor1 <- HLAalignments[[alignType]][[locus[1]]]$allele_name == alleles[1]
  coor2 <- HLAalignments[[alignType]][[locus[2]]]$allele_name == alleles[2]
  
  align1 <- HLAalignments[[alignType]][[locus[1]]][coor1,]
  align2 <- HLAalignments[[alignType]][[locus[2]]][coor2,]
  
  align1 <- align1[4:length(align1)]
  align2 <- align2[4:length(align2)]
  
  align1 <- align1[colnames(align1) %in% colnames(align2)]
  align2 <- align2[colnames(align2) %in% colnames(align1)]
  
  noSeq <- (align1[1,] == "*") | (align2[1,] == "*")
  
  combiPair <- rbind(align1,align2)
  rownames(combiPair) <- c(1,2)
  
  combiPair <- combiPair[,noSeq == F]
  combiPair <- combiPair[,combiPair[1,] != combiPair[2,]]
  
  if(length(combiPair[,combiPair[1,] != combiPair[2,]]) == 2) {return(paste("There are no differences between", alleles[1], "and", alleles[2], "in the", switch(alignType, "codon" = "codon","nuc" = "nucleotide","prot"="protein","gen"="genomic"), "alignment.",sep=" "))}
  
  return(combiPair[,combiPair[1,] != combiPair[2,]])
  
}

#### unified alignment search and construction functions v6.0.0 28 May 2026 Ryan Nickens & Steven Mack 

################
#AlignmentSearch
#'Search Alignments for Specific Positions in a Specific Allele
#'
#'Searches ANHIG/IMGT-HLA alignments and returns protein, codons or nucleotide sequences for a submitted allele name and position(s).
#'
#'@param alignType The type of alignment being searched. Allowed values are "codon","gen", nuc" and "prot". Only one 'alignType' value is allowed.
#'@param allelename A full or 2-field HLA allele name, excluding the "HLA-" prefix.
#'@param positions A vector of sequence positions (e.g., c(-17,1,22,130)); in nucleotide and genomic alignments, indel positions are named using decimals. So the first indel between positions 26 and 27 is named 26.1, and the second indel between 26 and 27 is 26.2, etc.
#'@param prefix A logical that indicates if the position number(s) should be included in the result.
#'@param sep The value that will be used to separate the sequences for each position. No value can be specified (sep=""). If prefix=FALSE and sep="", a sequence block is returned.
#'
#'@return A character string containing the corresponding peptide, codon or nucleotide sequence for each position in 'positions' for 'allelename'. A sequence is not returned if 'allelename' is not not found in the pertinent alignment. A position will be empty if 'allelename' does not have a value at the specified position.
#'
#'@importFrom xfun numbers_to_words
#'@export
#'
#'@note This function requires that the HLAalignments object has been populated with alignments via the alignmentFull() function.
#'@note Indel positions must be text-formatted (e.g. "607.12").
#'
#'@examples
#'\dontrun{
#'alignmentSearch("gen","DRB1*15:07:01",11:22)
#'alignmentSearch("nuc","DRB1*15:07:01",11:22)
#'alignmentSearch("prot","DRB1*15:07:01",11:22)
#'alignmentSearch("codon","DRB1*15:07:01",11:22)
#'alignmentSearch("nuc","DRB1*11:250N",c(605,"607.1","607.12",608))
#'}
#'
alignmentSearch <- function(alignType,allelename,positions,prefix=TRUE,sep="~"){

  alignType <- checkAlignType(alignType)
  if(length(alignType)!=1) {stop(paste("Please specify only one 'alignType'."))}
  
  trimmed <- FALSE
  
  #make sure all necessary parameters are provided(allelename and positions)
      if((missing(allelename) || missing(positions))) {
        #message if one or both of the parameters are not provided
        stop("Please provide an allele name and at least one codon position.\n")
      }
    #splits allele at this point, two sections are subsequently assigned to variables for later use
    split <- unlist(gregexpr(pattern="\\*",allelename))
    #name of allele without numbers(ex DRB1*17:07:01 is just DRB1)
    locus <- substr(allelename,1,split-1)
    #opposite of previous (ex DRB1*17:07:01 is just 17:07:01)
    allele <- substr(allelename,split+1,nchar(allelename))
  
    # assess allelename validity(makes sure allele is named in data source)
     if(!locus %in% names(HLAalignments[[alignType]]) || length(split) != 1) {
    #if it is not, return message
    stop(allelename," is not a valid HLA allele name.")
  }
  
  # exclude aa positions that do not exist for a locus
  checkpos <- positions[positions %in% colnames(HLAalignments[[alignType]][[locus]]) == TRUE]
  posdiff <- length(positions)-length(checkpos)
  #if there is a difference between input positions and checked positions, do this
    if(posdiff != 0) {
      message("There ", ifelse(posdiff==1,"is ","are "), xfun::numbers_to_words(posdiff), " listed position",ifelse(posdiff==1,"","s")," (",paste(positions[!positions %in% checkpos],collapse=","), ") ", ifelse(posdiff==1,"that does","that do")," not exist at the ", locus, " locus for the ", alignType," alignment.",sep="")
     }
  #if no positions are found in check, do this
    if(length(checkpos) == 0) {
    return(message("There are no valid positions."))
      }
      positions <- unique(checkpos)
  
  # assess existence of allele names; check two-field column for 2-field truncates
    if(!allele %in% HLAalignments[[alignType]][[locus]]$allele) {
    message(allelename," is not a known full-length allele name.")
    if(sum(charToRaw(allelename) == charToRaw(":")) == 1) {message("Checking 2-field allele names.")
        if(allelename %in% HLAalignments[[alignType]][[locus]]$trimmed_allele) {
          
        trimmed <- TRUE
      }
      else {message(allelename," is not a known 2-field allele name.")
        return(cat(""))
      }
    }
    else {
      return(cat(""))
    }
  }
  # numeric sort of nucleotide positions
      if(any(posSort(positions,alignType,locus) != positions)) {
        positions <- posSort(positions,alignType,locus)
        message("Sorting positions to reflect sequence order.",if(prefix==FALSE){" Setting 'prefix=TRUE' for clarity."})
        prefix=TRUE
        #positions <- as.numeric(positions)
    }
  return(multiSearch(alignType=alignType,locus=locus,allele=allele,positions=positions,prefix=prefix,sep=sep,trimmed=trimmed))
}

################
##multiSearch
#'Search Alignment Sequences at Multiple Positions for a Specified Allele
#'
#'Generates a character string of multiple protein, codon or nucleotide or genomic sequences at the specified positions for the specified allele name.
#'
#'@param alignType The type of alignment being searched. Allowed values are "prot", codon", "nuc" and "gen". Only one 'alignType' value is allowed.
#'@param locus A specific locus.
#'@param allele An allele name.
#'@param positions Nucleotide positions to search.
#'@param prefix A logical that indicates if the position number(s) should be included in the result.
#'@param sep Defines the separator between position sequences.
#'@param trimmed A logical identifying whether 'allele' is a two-field, 'trimmed' allele name (trimmed=TRUE), or a full-length name (trimmed=FALSE). The default value is FALSE.
#'
#'@return codons (possibly with position number) at 'positions' for 'allele'. If 'allele' does not have a codon at 'position', '' will be returned.
#'
#'@note For internal HLAtools use only.
#'
#'@export
#'
multiSearch <- function(alignType, locus, allele, positions, prefix=TRUE,sep="~",trimmed=FALSE){

    alignType <- checkAlignType(alignType)
    if(length(alignType)!=1) {stop(paste("Please specify only one 'alignType'."))}
  
    if(validateAllele(paste(locus,allele,sep="*"))) {
  if(trimmed) {
      if( paste(locus,allele,sep="*") %in% HLAalignments[[alignType]][[locus]]$allele_name) {
                  message(paste(allele,"is a full length alllele name.",sep=" "))
                  trimmed <- FALSE }
      } else {
                if( paste(locus,allele,sep="*") %in% HLAalignments[[alignType]][[locus]]$trimmed_allele && !paste(locus,allele,sep="*") %in% HLAalignments[[alignType]][[locus]]$allele_name) {
                  trimmed <- TRUE
                }
      }
  } else {stop(paste(paste(locus,allele,sep="*"),"is not a full-length or trimmed allele name.",sep=" "))}
  
  motif <- ""
  if(trimmed == TRUE) {
    fullname <- HLAalignments[[alignType]][[locus]][HLAalignments[[alignType]][[locus]]$trimmed_allele %in% paste(locus,allele,sep="*"),4][1]
 
    message("Returning sequence for ",fullname,", the first of ", nrow(HLAalignments[[alignType]][[locus]][HLAalignments[[alignType]][[locus]]$trimmed_allele %in% paste(locus,allele,sep="*"),])," ",paste(locus,allele,sep="*")," alleles.",sep="")

    allele <- strsplit(fullname,"*",fixed = TRUE)[[1]][2]
    trimmed <- FALSE
    }
    #go through array
      for(i in 1:length(positions)) {
        #using single search to find each set of nucleotides at a position
        motif <- paste(motif,uniSearch(alignType,locus,allele,positions[i],prefix,trimmed=trimmed),sep=sep)
    #   motif <- paste(motif,unicDNAsearch(locus,allele,positions[i],prefix,trimmed=trimmed),sep=sep)
     }
    #separates letters w "~"
    substr(motif,nchar(sep)+1,nchar(motif))
}

################
##uniSearch
#'Search Sequences at a Single Position for an Allele
#'
#'Generates a character string of the peptide, codon or nucleotide sequence at the specified position for the specified HLA allele.
#'
#'@param alignType The type of alignment being searched. Allowed values are "prot", "codon", "nuc" and "gen". Only one 'alignType' value is allowed.
#'@param locus A specific HLA locus.
#'@param allele The name of the allele being searched.
#'@param position The specified position.
#'@param prefix A logical that indicates if the position number(s) should be included in the result.
#'@param trimmed A logical identifying whether 'allele' is a two-field, 'trimmed' allele name (trimmed=TRUE), or a full-length name (trimmed=FALSE). The default value is FALSE.
#'
#'@return The peptide residue, codon or nucleotide sequence at specified position for specified allele as available in the ANHIG/IMGTHLA Github Repository.
#'
#'@note For internal HLAtools use.
#'
#'@export
#'
uniSearch <- function(alignType, locus, allele, position, prefix=TRUE, trimmed=FALSE){

    alignType <- checkAlignType(alignType)
    if(length(alignType)!=1) {stop(paste("Please specify only one 'alignType'."))}
  
    sec <- ""
  ## Initial parameter validations -- SJM
  if(length(position) > 1) {position <- position[1]; warning("More than one position was specified. Results will be returned for position ",position," only.")}
  if(trimmed == TRUE && numFields(allele) > 2) {return(warning("Two-field allele names are required when 'trimmed = TRUE'."))}
  ## End -- SJM
  #if input says trimmed=true...
  #finding 3 nucleotides at specified position
  if(trimmed == TRUE) {
    if(alignType == "codon") {
      cDNA_seq <- HLAalignments[[alignType]][[locus]][HLAalignments[[alignType]][[locus]]$trimmed_allele %in% paste(locus,allele,sep="*"),colnames(HLAalignments[[alignType]][[locus]]) %in% as.character(position)][1,]
    } else {
      cDNA_seq <- HLAalignments[[alignType]][[locus]][HLAalignments[[alignType]][[locus]]$trimmed_allele %in% paste(locus,allele,sep="*"),colnames(HLAalignments[[alignType]][[locus]]) %in% as.character(position)][1]
    }
    for(i in 1:length(cDNA_seq)) {
      sec <- paste0(sec,cDNA_seq[[i]])
    }
    #only use the first allele with shortened name
    sec <- sec[1]
    #includes position number in front of nucleotides
    numsec <- paste0(position, sec) } else {
    #if input says allele_name...
    #finding 3 nucleotides at specified position.  ##### SM 2/16/2024 ADDED AS.DATA.FRAME BELOW FOR NUC ALIGNMENTS
    cDNA_seq <- as.data.frame(HLAalignments[[alignType]][[locus]][HLAalignments[[alignType]][[locus]]$allele_name %in% paste(locus,allele,sep="*"),colnames(HLAalignments[[alignType]][[locus]]) %in% as.character(position)])
    #if no codon is found at the position
        if(nrow(cDNA_seq) == 0) {
          return(warning(paste("No value was found for ", allele,". This may not be a full allele name. Please redo your search using trimmed=TRUE.",sep="")))
        }
      for(i in 1:length(cDNA_seq)) {
        sec <- paste0(sec,cDNA_seq[[i]])
      }
    numsec <- paste0(position, sec)
    }
  #prefix decides how much information to return
  if(prefix){numsec}else{sec}
}

#### Alignment Generation Functions

################
##customAlign
#'Generate a Customized Peptide, Codon or Nucleotide Sequence Alignment.
#'
#'Generates a peptide, codon, coding or genomic nucleotide alignment table for user-specified HLA alleles at user-specified positions.
#'
#'@param alignType The type of alignment being searched. Allowed values are "prot", codon", "nuc" and "gen". Only one 'alignType' value is allowed.
#'@param alleles A vector of un-prefixed HLA allele names.
#'@param positions Either a vector of variant positions, against which all loci will be aligned, or a list of vectors of nucleotide positions, exactly one vector for each allele, against which each corresponding allele will be aligned.
#'
#'@return A data frame of allele names and the corresponding nucleotide sequences for each desired nucleotide position. FALSE is returned when 'positions' includes values that do not exist in the specified alignment, and when a specified locus is not available in the HLAalignments object.
#'
#'@note This function requires that the HLAalignments object has been populated with alignments via the alignmentFull() function.
#'
#'@examples
#'\dontrun{
#'customAlign("codon",c("DRB1*01:01","DQB1*02:01","DPB1*01:01"),c(1,2,3,7,8,9,13,14,15))
#'customAlign("codon",c("DPB1*01:01:01:01","DQA1*01:01:01:01","DQB1*05:01:01:01"),list(19:35,1:4,6:9))
#'customAlign("prot",c("DRB1*08:07","DPB1*14:01"),list(c(12,27,30,39,57,84),c(15,16,39,60,67,81)))
#'}
#'
#'@export
#'
customAlign <- function(alignType,alleles,positions){

  alignType <- checkAlignType(alignType)
  if(length(alignType)!=1) {stop(paste("Please specify only one 'alignType'."))}
  
  #makes sure no duplicate alleles will be present in table
    alleles <- unique(alleles)
      if(is.list(positions)) {
      #calls multicDNAalign if multiple sets of positions are given
      multiAlign(alignType,alleles,positions)
        } else {
          #calls unicDNAalign if only one set of positions are given
            uniAlign(alignType,alleles,positions)
          }
    }

################
##uniAlign
#'Generate an Alignment for Specific Alleles at Specific Positions
#'
#'Generates a peptide, codon, coding or genomic nucleotide alignment at a single set of positions for HLA alleles at one or more loci.
#'
#'@param alignType The type of alignment being searched. Allowed values are "prot", codon", "nuc" and "gen". Only one 'alignType' value is allowed.
#'@param alleles A vector of un-prefixed HLA allele names.
#'@param positions A vector of codon positions, against which all loci will be aligned.
#'
#'@return A data frame of allele names and the corresponding codon sequence for specified position. A codon sequence is not returned for alleles that are not present in the loaded HLAalignments object. A sequence position will be populated with whitespace if a specific allele has no sequence at the specified position. If 'positions' includes values that do not exist in the specified alignment, FALSE is returned.
#'
#'@note For internal HLAtools use.
#'
#'@export
#'
uniAlign <- function(alignType, alleles, positions){

  alignType <- checkAlignType(alignType)
  if(length(alignType)!=1) {stop(paste("Please specify only one 'alignType'."))} 
  
  if(!validatePositions(alignType,strsplit(alleles[1],"*",fixed=TRUE)[[1]][1],positions)) {return(FALSE)}

  #making an array with correct dimensions
  align <- data.frame(matrix(" ", nrow = length(alleles), ncol = length(positions)+1),stringsAsFactors = FALSE)
  #naming first column
  colnames(align)[1] <- "Allele"
  #naming remaining columns, running through positions array and naming each column (top row)
  colnames(align)[2:(length(positions)+1)] <- positions
  
  for(i in 1:length(alleles)){
    #setting the names for first column
    align[i,1] <- alleles[i]
    #getting nucleotides using cDNAsearch
    seqStr <- suppressMessages(alignmentSearch(alignType, alleles[i],positions,prefix=FALSE))
    #filling in the table
    #these positions will be filled if seqStr is not empty.
    align[i,2:(length(positions)+1)] <- if(!is.null(seqStr)) {
      strsplit(seqStr,split="~",fixed = TRUE)[[1]]
        } else { #if value is null, space fills spots
            rep(" ",length(positions))
          }
    
      }
      align
    }


################
##multiAlign
#'Generate an Alignment for Specific Alleles at Different Positions
#'
#'Generates a peptide, codon, coding nucleotide or genomic alignment for HLA alleles, allowing each allele to be aligned to a different set of positions.
#'
#'@param alignType The type of alignment being searched. Allowed values are "prot", codon", "nuc" and "gen".  Only one 'alignType' value is allowed.
#'@param alleles A vector of un-prefixed HLA allele names.
#'@param positions A list of vectors of nucleotide positions, exactly one vector for each allele, against which each corresponding allele will be aligned.
#'
#'@return A data frame of 'allele' and the corresponding nucleotide sequence for specified positions designated for an allele. A nucleotide sequence is not returned for a specific allele if input allele is not available in the ANHIG/IMGTHLA Github Repository. A position will be left empty if the specific allele does not have a variant at that position. If 'positions' includes values that do not exist in the specified alignment, FALSE is returned.
#'
#'@note For internal HLAtools use. 
#'
#'@export
#'
multiAlign <- function(alignType,alleles,positions){

  alignType <- checkAlignType(alignType)
  if(length(alignType)!=1) {stop(paste("Please specify only one 'alignType'."))}
  
  for(l in 1:length(positions)) {
            if(!validatePositions(alignType,strsplit(alleles[l],"*",fixed=TRUE)[[1]][1],positions[[l]])) {return(FALSE)}
                    }

    #initial parameter check
      if(length(alleles) != length(positions)) {return(warning("The number of sets of positions must exactly match the number of alleles, please adjust input accordingly."))}
      #creating an array of correct dimensions
      #as many rows as 2*alleles-1, as many columns as positions
      align <- data.frame(matrix(" ", nrow = (length(alleles)*2)-1, ncol = max(lengths(positions))+1),stringsAsFactors = FALSE)
      #first column name is first locus (ex. DQA1*01:01 is column name as DQA1(it cuts off back part of name))
      colnames(align)[1] <- strsplit(alleles[1],split="*",fixed=TRUE)[[1]][1]
      #remaining columns are given positions for first allele(first set in position array)
      colnames(align)[2:(length(positions[[1]])+1)] <- positions[[1]]
  
  #first spot in first column is named the full first allele
      align[1,1] <- alleles[1]
      #using cDNAsearch to search for nucleotides for the first allele at the first set of positions
      seqStr <- suppressMessages(alignmentSearch(alignType,alleles[1],positions[[1]],prefix=FALSE))
      #filling in the remainder of row 1
      align[1,2:(length(positions[[1]])+1)] <- if(!is.null(seqStr)) {
      #if return is not empty, fill in row with the split return from cDNAsearch method
      strsplit(seqStr,split="~",fixed=TRUE)[[1]]
        } else {
          #if array is empty, fill with spaces
          rep(" ",length(positions[[1]]))
        }
        #filling in remaining rows(rows 2 and on)
        for(n in 2:length(alleles)) {
              #effectively the column name, same as above, will be in same row as positions
              align[(n-1)*2,1] <- strsplit(alleles[n],split="*",fixed=TRUE)[[1]][1]
              #filling in row with positions
              align[(n-1)*2,2:(length(positions[[n]])+1)] <- positions[[n]]
              #full allele name, will be in same row as nucleotides
              align[((n-1)*2)+1,1] <- alleles[n]
              #using cDNAsearch to find nucleotides for specified positions
              seqStr <- suppressMessages(alignmentSearch(alignType,alleles[n],positions[[n]],prefix=FALSE))
              #filling in remainder of row
              align[((n-1)*2)+1,2:(length(positions[[n]])+1)] <- if(!is.null(seqStr)) {
              #if function returns something, splits it and fills in row
              strsplit(seqStr,split="~",fixed=TRUE)[[1]]
          } else { #if function returns nothing, fill with space
      rep(" ",length(positions[[n]]))
          }
    
        }
      align
}

################
##validatePositions
#'Validate the Presence of Variant Positions in an Alignment
#'
#'Returns TRUE when variant positions are present in a specified alignment in HLAalignments. Returns false when any variant position is not present in the specified alignment and generates a message identifying such positions.
#'
#'@param alignType The type of alignment being searched. Allowed values are "prot", codon", "nuc" and "gen".  Only one 'alignType' value is allowed.
#'@param locus A locus supported by the IPD-IMGT/HLA Database.
#'@param positions A vector of variant positions.
#'
#'@return A logical identifying if the position is present in the alignment (TRUE) or, if it is not in the alignment or is not valid (FALSE).
#'
#'@note For internal HLAtools use. 
#'
#'@export
#'
validatePositions <- function(alignType,locus,positions){
  
  multiple <- FALSE
  
  if(!length(alignType) == 1) { message("Please provide a single 'alignType' value.")
            return(FALSE)}
  
  if(length(checkAlignType(alignType)) == 0) { return(FALSE)}
  
  if(!validateLocus(locus,typeToSource(alignType))) {return(FALSE)}
  
  if(!all(positions %in% colnames(HLAalignments[[alignType]][[locus]]))) {
    
    falsePos <- positions[positions %in% colnames(HLAalignments[[alignType]][[locus]]) == FALSE]
    
    if(length(falsePos) > 1) { 
      multiple <- TRUE
      if(length(falsePos) > 2) {
            falsePos <- paste(paste(falsePos[length(falsePos)-1],sep=", "),falsePos[length(falsePos)], sep = " and ")
          } else {falsePos <- paste(falsePos,collapse =" and ")}
        } 
    
    message(paste(ifelse(multiple,"Positions","Position"), falsePos,ifelse(multiple,"are","is"), "not found in the",locus,alignType,"alignment.",sep=" "))
    
    return(FALSE)
    
  } else {TRUE}
  
}

################
##queryPositions
#'Extract Variant Information for Specific Positions in an Alignment
#'
#'Identifies the sequence variants at specific positions in specified alignments in HLAalignments. When 'table' is TRUE, data frames of counts and frequencies for each variant at each position are returned.
#'
#'@param alignType The type of alignment being searched. Allowed values are "prot", codon", "nuc" and "gen".  Only one 'alignType' value is allowed.
#'@param locus A locus supported by the IPD-IMGT/HLA Database.
#'@param positions A vector of numeric variant positions.
#'@param table A logical value specifying if a table of counts and frequencies for each variant at each value of 'positions' should be returned (table = TRUE). The default value is FALSE.
#'
#'@return When 'positions' specifies a single position, a character vector of variants is returned. When multiple positions are specified, a list of such character vectors is returned. When 'table' is TRUE, a list of data frames identifying the number (Count) and frequency (Frequency) of each variant is returned.
#'
#'@note This function requires that the HLAalignments object has been populated with alignments via the alignmentFull() function.
#'@note The '.' value in a vector of positions identifies absent sequence; e.g. following a nonsense mutation or in a region of sequence aligned to an insertion. The '*' value in a vector of positions identifies unknown sequence. 
#'
#'@examples
#'\dontrun{
#'queryPositions("prot","DRB1",c(5,10,86,129))
#'queryPositions("prot","DRB1",c(5,10,86,129),TRUE)
#'}
#'
#'@export
#'
queryPositions <- function(alignType,locus,positions,table = FALSE){
  
    if(!validatePositions(alignType,locus,positions)) {return(FALSE)}
  
        tables <- variants <- rep(list(NA),length(positions))
  
            for(i in 1:length(positions)) { 
                tables[[i]] <- variantTable(alignType,locus,positions[i])
                variants[[i]] <- tables[[i]]$Variant
            }
  
            names(tables) <- names(variants) <- paste(locus,positions,sep="_")
  
    ifelse(table,return(tables), return(variants)) 
}


################
##variantTable
#'Builds a Data Frame of Variant Information for Specific Positions in an Alignment
#'
#'Builds a data frame sequence variants at specific positions in specified alignments in HLAalignments, along with the counts and frequencies for each variant.
#'
#'@param alignType The type of alignment being searched. Allowed values are "prot", codon", "nuc" and "gen".  Only one 'alignType' value is allowed.
#'@param locus A locus supported by the IPD-IMGT/HLA Database.
#'@param position A numeric variant position.
#'
#'@return a data frame of three columns identifying the variants, and the number and frequency of each in the alignment. 
#'
#'@note For internal HLAtools use.
#'
#'@export
#'
variantTable <- function(alignType,locus,position) {

        primary <- HLAalignments[[alignType]][[locus]][,colnames(HLAalignments[[alignType]][[locus]]) == as.character(position)]
        ifelse(class(primary) == "data.frame", max <- nrow(primary), max <- length(primary))
        if(alignType == "codon")  { primary <- apply(primary[,colnames(primary)],1,paste,collapse="") }

            primary <- as.data.frame(table(primary))
            primary$value <- primary$Freq/max
            primary$primary <- as.character(primary$primary)
            colnames(primary) <- c("Variant","Count","Frequency")
            
            return(primary)
            
        }

#updateAlleleListHistory v1.2.0 2FEB2024

################
##updateAlleleListHistory
#'Build the AlleleList History R Object
#'
#'Consumes the ANHIG/IMGTHLA GitHub repository's allelelisthistory.txt file to create the AlleleListHistory object.
#'
#'@return A list object containing two data frames. alleleListHistory$Version identifies the date and version of alleleListHistory$AlleleListHistory. alleleListHistory$AlleleListHistory relates each IPD-IMGT/HLA Database Accession ID (HLA_ID) to the corresponding HLA allele name in each release version since 1.05.0.
#'
#'@importFrom utils read.table
#'
#'@export
#'
updateAlleleListHistory <- function() {
  url<-url("https://raw.githubusercontent.com/ANHIG/IMGTHLA/Latest/Allelelist_history.txt")
  alleleListHistoryHead <- readLines(url,n = 3)
  alleleListHistory <- read.table(url, skip = 6, header=TRUE, sep=",")
  AlleleListHistory <- list(ALversion <- data.frame(strsplit(alleleListHistoryHead[2]," ")[[1]][3],strsplit(alleleListHistoryHead[3]," ")[[1]][4]), ALtable <- alleleListHistory)
  names(AlleleListHistory) <- c("Version","AlleleListHistory")
  colnames(AlleleListHistory$Version) <- c("Date","Version")

  AlleleListHistory
}

## ExtractGenomicReferences v1.0.0 July 17, 2026 

# BuildGenomicReferences
#'Generate a Table of Genomic Reference Alleles and Characteristics of HLA Genes Defined by the IPD-IMGT/HLA Database 
#'
#'@description
#'buildGenomicReferences() extracts the table of genomic HLA reference alleles at 'www.ebi.ac.uk/ipd/imgt/hla/about/help/genomics/' and generates the IMGTGenomicReferences data object. 
#'
#'@return A list object of two elements -- GenomicReferences and Version. 
#'
#'@importFrom rvest read_html html_elements html_table
#'@importFrom stringr %>%
#'
#'@export
#'
#'@note This function requires internet access to function.
#'@note For internal HLAtools use.
#'
#'@references Barker et al. NAR (2025) 54:D1152–D1158 https://doi.org/10.1093/nar/gkaf1218
buildGenomicReferences <- function() {      
    
    url <- "https://www.ebi.ac.uk/ipd/imgt/hla/about/help/genomics/"
    
             page <- read_html(url)
    
          # Extract all HTML tables (currently #2 is the genome reference table)
            tables <- page %>%
             html_elements("table") %>%
                 html_table(fill = TRUE)
    
    genomicRefTab <- as.data.frame(tables[[2]])
    
    generationDate <- strsplit(as.character(Sys.time()),fixed = T," ")[[1]][1]
    
    IGR <- list(genomicRefTab,generationDate)
    names(IGR) <- c("GenomicReferences","version")
    
    IGR
}

### Alignment Full v3.4.0 20 July 2025

################
##alignmentFull
#'Build Sets of Protein, Codon, Coding Nucleotide and Genomic Nucleotide Alignments for Specified  Loci
#'
#'Applies buildAlignments() to build a set of alignments for loci supported in the ANHIG/IMGTHLA GitHub repository.
#'
#'@param loci A characer vector of the locus names for which alignments should be built. The default value ("all") generates alignments for all loci. 
#'@param alignType A character vector of alignment types. The allowed values are "all", "prot", "codon", "nuc", and "gen", which specify either all available alignments for a given locus or the respective protein, codon, nucleotide and genomic alignments, as determined by the HLAgazetteer.
#'@param version A character string describing the version of the ANHIG/IMGTHLA Github repository from which alignments are obtained. The default value, 'HLAgazetteer$version', generates alignments for the IPD-IMGT/HLA Database release under which the HLAgazetteer was built.
#'
#'@return A list object containing data frames of protein (prot), codon (codon), coding nucleotide (nuc), or genomic nucleotide (gen) alignments, for specified genes in the specified IPD-IMGT/HLA Database release, along with the pertinent reference database release.
#'
#'@note AlignmentFull() must be run to build the 'HLAalignments' object (HLAalignments <- alignmentFull()) before other functions that use alignments can be used, and therefore requires internet access.
#'@note Depending on local download speeds, building all available alignments for all loci can take several minutes.
#'@note Prior to IPD-IMGT/HLA Database release version 3.24.0, the HLA-DP and HLA-DQ sequence alignment files in the IPD-IMGT/HLA GitHub Repository did not all include a numerical suffix in the gene name (e.g., the protein sequence alignment file for the DQA1 gene was named 'DQA_prot.txt') because alignment files for the DPA2, DPB2, DQA2 and DQB2 genes had not been made available. Building DPA1, DPB1, DQA1, and DQB1 sequence alignments from releases prior to 3.24.0 require using a gene name that does not include the numerical suffix.
#'
#'@export
#'
alignmentFull <- function(loci = "all", alignType = "all", version = HLAgazetteer$version) {
  
  if(!"all" %in% loci) {
      loci <- multiLocusValidation(loci) }
  
  if(!"all" %in% alignType) {
  alignType <- checkAlignType(alignType) }
  
  NL1 <- NL2 <- NL3 <- NL4 <- NULL
  
  if(version != "Latest"){ #
    if(!validateVersion(version)){stop(paste(version," is not a valid IPD-IMGT/HLA Database release version."))}
        } else{ version <- getLatestVersion()}

  if(length(loci) == 1 && loci == "all" && alignType == "all") {
  
  #list of loci names
  NL4 <- NL1 <- as.list(HLAgazetteer$nuc) #cDNA and codon
  NL2 <- as.list(HLAgazetteer$gen) #gDNA
  NL3 <- as.list(HLAgazetteer$prot) #AA
  
  } else {
  
    if("all" %in% alignType)  { 
      NL1 <- as.list(loci[loci %in% as.list(HLAgazetteer$nuc)]) #nuc 
      NL2 <- as.list(loci[loci %in% as.list(HLAgazetteer$gen)]) #gen
      NL3 <- as.list(loci[loci %in% as.list(HLAgazetteer$prot)]) #prot
      NL4 <- as.list(loci[loci %in% as.list(HLAgazetteer$nuc)]) # codon
        } else {
            if("nuc" %in% alignType)  { NL1 <- as.list(loci[loci %in% as.list(HLAgazetteer$nuc)]) } #nuc 
            if("gen" %in% alignType)  { NL2 <- as.list(loci[loci %in% as.list(HLAgazetteer$gen)]) } #gen 
            if("prot" %in% alignType)  { NL3 <- as.list(loci[loci %in% as.list(HLAgazetteer$prot)]) } # prot
            if("codon" %in% alignType)  { NL4 <- as.list(loci[loci %in% as.list(HLAgazetteer$nuc)]) } # codon
          }
    
      }

  #creating nested lists for cDNA, gDNA and amino acid
  cList <- vector(mode='list', length=length(NL1))
  gList <- vector(mode='list', length=length(NL2))
  protList <- vector(mode='list', length=length(NL3))
  codonList <- vector(mode='list', length=length(NL4))

  #filling nested lists with alignments
#  if(NL1 != "") {
    if(length(NL1) > 0) { 
      for(i in 1:length(NL1)){
        cList[i] <- suppressWarnings(buildAlignments(as.character(NL1[i]), "cDNA", version = version)[[1]][2]) } 
    
        names(cList) <- NL1
      }
#    }
#  if(NL2 != "") {
    if(length(NL2) > 0) { 
      for(i in 1:length(NL2)){
        gList[i] <- suppressWarnings(buildAlignments(as.character(NL2[i]), "gDNA", version = version)[[1]][1]) }
    
        names(gList) <- NL2
      }
#    }
#  if(NL3 != "") {
    if(length(NL3) > 0) { 
      for(i in 1:length(NL3)){
        protList[i] <- suppressWarnings(buildAlignments(as.character(NL3[i]), "AA", version = version)[[1]][1]) }
    
        names(protList) <- NL3
      }
#    }
#  if(NL4 != "") {
    if(length(NL4) > 0) { 
      for(i in 1:length(NL4)){
        codonList[i] <- suppressWarnings(buildAlignments(as.character(NL4[i]), "cDNA", version = version)[[1]][1]) }
    
        names(codonList) <- NL4
      }
#    }

  #placing nested lists inside of larger list
  AllAlignment<-list(firstList <- protList,
                 fourthlist <- codonList,
                 secondList <- cList,
                 thirdlist <- gList,
                 versionString <- version)
#naming nested lists

  names(AllAlignment) <- c("prot","codon","nuc","gen","version")

  AllAlignment
}

##Fragment Feature Names v4.2.0 July 20, 2025

################
##FragmentFeatureNames (ffN)
#'Identify and Annotate Gene Features in Pseudogenes and Gene Fragments.
#'
#'@description
#'HLA pseudogenes and gene fragments many not share all of the gene features of their expressed homologs. The ffN() function generates a list object identifying the features present in a given gene fragment or pseudogene with a genomic alignment file. This includes annotations of the differences between the pseudogenes and gene fragments, and their expressed homologs.
#'Pseudogenes and gene fragments are identified in the IMGTHLAGeneTypes data object.
#'
#'Standard Features
#'* E - Exon, a peptide-encoding sequence
#'* I - Intron, an intervening sequence found between Exons
#'* U - UTR, an untranslated region of sequence preceding the first Exon or following the last Exon
#'
#'Additional Features used in these annotations, based on boundaries indicated in the sequence alignment, are:
#'* H - Hybrid, a sequence that includes at least one known feature sequence and one nucleotide sequence that does not correspond to a known feature
#'* J - Join, a sequence that includes two or more features that are separated by a feature boundary in the reference
#'* N - Novel, a novel sequence that does correspond to a known feature sequence
#'* S - Segment, a subset of a longer feature sequence
#'
#'Reference Genes
#'* HLA-C genomic sequence is used as the reference for class I pseudogenes and gene fragments.
#'* HLA-DPA1 and -DPB1 genomic sequences are used as the references for class II pseudogenes -DPA2 and -DPB2, respectively.
#'
#'@param version A character string identifying the pertinent IPD-IMGT/HLA Database release version (e.g., "3.55.0") under which the returned object is generated. This parameter does not impact the generation of the feature names or annotations, and is only included to provide IPD-IMGT/HLA Database release version context.
#'
#'@return A list object where each element is the name of a pseudogene or gene fragment. Each of these elements is a list of 'features' and 'annotation'. "Features" identifies the gene features for that gene, ordered from 5' to 3'. Because these genes are not-expressed, the three standard gene features (U, Untranslated Region (UTR); E, Exon; and I, Intron) may not always be present. In these cases, H, J, N and S features (defined below) are returned. Each feature identifier is followed by an identifying number (e.g. U.5 is the 5' UTR, and E.3 is Exon 3). "Annotation" provides the composition of the non-standard H, J, N and S features for each gene.
#'#'
#'@export
#'
#'@examples 
#'fragmentFeatureNames <- ffN("3.35.0")
#'
#'@note Features and their annotations have been identified manually. Feature annotations will not change unless a new pseudogene or gene fragment is added in a future release, in which case new annotations will be generated.
#'@note The H, J, N and S features are described for class I pseudogenes and gene fragment sequences, all of which are described relative to the HLA-C reference sequence. Feature length differences for DPA2 and DPB2, relative to the DPA1 and DPB1 references, are noted in annotations of standard feature abbreviations (E, I and U).
#'@note No annotations are included for the DRB2, DRB6, DRB7, and DRB9 genes, as genomic alignments for these genes are not available.
#'@note For internal HLAtools use.
#'
ffN <- function(version){ #Fragment Feature Names
  
  if(version != "Latest"){ #
    if(!validateVersion(version)){stop(paste(version," is not a valid IPD-IMGT/HLA Database release version."))}
  }else{ version <- getLatestVersion()}

  if(paste("X",gsub("\\.","",as.character(version)),sep="") %in% colnames(alleleListHistory$AlleleListHistory)) {

  featureNames <- rep(list(list()), length(HLAgazetteer$align[HLAgazetteer$align %in% unique(c(HLAgazetteer$pseudo,HLAgazetteer$frag))][!HLAgazetteer$align[HLAgazetteer$align %in% unique(c(HLAgazetteer$pseudo,HLAgazetteer$frag))] %in% HLAgazetteer$nogen]))
  names(featureNames) <-sort(HLAgazetteer$align[HLAgazetteer$align %in% unique(c(HLAgazetteer$pseudo,HLAgazetteer$frag))][!HLAgazetteer$align[HLAgazetteer$align %in% unique(c(HLAgazetteer$pseudo,HLAgazetteer$frag))] %in% HLAgazetteer$nogen])


  featureNames$DPA2$features <- c("U.5","E.1","I.1","E.2","I.2","E.3","I.3","E.4","U.3") #  8 boundaries -- **Unconfirmed
  featureNames$DPB2$features <- c("U.5","E.1","I.1","E.2","I.2","E.3","I.3","E.4","I.4","E.5","U.3") # 10 boundaries
  featureNames$H$features <- c("U.5","E.1","I.1","E.2","I.2","E.3","I.3","E.4","I.4","E.5","I.5","E.6","I.6","E.7","I.7","E.8","U.3") # 16 boundaries full length
  featureNames$J$features <- c("U.5","E.1","I.1","E.2","I.2","E.3","I.3","E.4","I.4","E.5","I.5","E.6","I.6","E.7","I.7","E.8","U.3") # 16 boundaries full length
  featureNames$K$features <- c("U.5","E.1","I.1","E.2","I.2","E.3","I.3","E.4","I.4","E.5","I.5","E.6","I.6","E.7","I.7","E.8","U.3") # 16 boundaries full length
  featureNames$L$features <- c("U.5","E.1","I.1","E.2","I.2","E.3","I.3","E.4","I.4","E.5","I.5","E.6","I.6","E.7","I.7","E.8","U.3") # 16 boundaries full length
  featureNames$N$features <- c("N.1","H.1","N.2") #  2 boundaries
  featureNames$P$features <- c("H.1","E.3","I.3","E.4","I.4","E.5","I.5","E.6","I.6","E.7","I.7","U.3") # 11 boundaries
  featureNames$R$features <- c("H.1","E.3","I.3","E.4","I.4","E.5","I.5","E.6","I.6","J.1","J.2") # 10 boundaries
  featureNames$S$features <- c("H.1","S.1","J.1","E.7","I.7","E.9","S.2") #  6 boundaries
  featureNames$T$features <- c("H.1","E.4","I.4","E.5","I.5","E.6","I.6","E.7","U.3") #  8 boundaries
  featureNames$U$features <- c("N.1","J.1","S.1") #  2 boundaries
  featureNames$V$features <- c("U.5","E.1","I.1","E.2","I.2","S.1","N.1") #6 boundaries
  featureNames$W$features <- c("H.1","E.3","I.3","E.4","I4","E.5","I.5","E.6","I.6","E.7","I.7","N.1","S.1") #12 boundaries
  featureNames$Y$features <- c("U.5","E.1","I.1","E.2","I.2","E.3","I.3","E.4","I.4","E.5","I.5","E.6","I.6","E.7","I.7","E.8","U.3") #16 boundaries full length

  featureNames$DPA2$annotation <- "All of the reference gene features are present. E.1 starts 28 nucleotides before the reference, and ends 88 nucleotides before the reference."
  featureNames$DPB2$annotation <- "All of the reference gene features are present. I.2 contains a very large insertion relative to the reference."
  featureNames$H$annotation <- "All of the reference gene features are present."
  featureNames$J$annotation <- "All of the reference gene features are present."
  featureNames$K$annotation <- "All of the reference gene features are present."
  featureNames$L$annotation <- "All of the reference gene features are present."
  featureNames$N$annotation <- "N.1 is 51 nucleotides of novel sequence that does not align to Exon 3. The last 40 nucleotides align poorly. The first 34 nucleotides of H.1 do not align to the Exon 4 reference; the last 135 nucleotides align to Exon 4. N.2 does not align to Exon 4 or Intron 4."
  featureNames$P$annotation <- "H.1 is 287 nucleotides of novel sequence followed by 168 nucleotides of the 5' end of Intron 2, which include a 15 nucleotide section of novel sequence. Exon 8 is absent."
  featureNames$R$annotation <- "H.1 is 287 nucleotides of novel sequence followed by 156 nucleotides of the 5'end of Intron 2, which includes a 5 nucleotide section of novel sequence. J.1 is all of Exon 7 and the first 12 nucleotides of Intron 7. J.2 is the remainder of Intron 7, and the 3' UTR. Exon 8 is absent."
  featureNames$S$annotation <- "H.1 is 37 nucleotides of novel sequence, followed by the last 185 nucleotides of Intron 5. S.1 is the first 27 nucleotides of Exon 6. J.1 is the last 4 nucleotides of Exon 6 (2 nucleotides in the reference have been deleted), followed by the last 100 nucleotides of Intron 6 (6 nucleotides in the reference have been deleted). E.9 in a 191 nucleotide long Exon in what is the 5' end of the 3' UTR in the reference. S.2 is the 3' end of the 3' UTR."
  featureNames$T$annotation <- "H.1 is 31 nucleotides of novel sequence, followed the last 93 nucleotides of Exon 3, and all of Intron 3."
  featureNames$U$annotation <- "N.1 is 54 nucleotides of novel sequence. J.1 includes 196 nucleoitdes that align to Exon 3, and the first 10 nucleotides of Intron 3. S.1 is 301 nucleotides of Intron 3. HLA-U*01:01:01:02, *01:02, 01:03 and 01:04 include a 41 nucleotide novel sequence insertion. U*01:04 includes an additional 145 nucleotides of novel sequence at the 3' end of S.1."
  featureNames$V$annotation <- "S.1 is the first 195 nucleotides of Exon 3. N.1 is 524 nucleotides of novel sequence that does not align to Exon or Intron 3."
  featureNames$W$annotation <- "H.1 is 327 nucleotides of novel sequence followed by 148 nucleotides of the 5' end of Intron 2, which include a 15 nucleotide section of novel sequence. N1 is a novel exon that follows the deletion of Exon 8. S1 is the remainder of the 3' UTR"
  featureNames$Y$annotation <- "U.5 is a single nucleotide, and U.3 is 171 nucleotides long in HLA-Y*02:01. There is no 5' or 3' UTR in HLA-Y*01:01 or *03:01."

  featureNames$version <- version

  featureNames  } else {message(paste(version, "is not a recognized release version.",sep=" "))}

}

### v1.0.1 7FEB2024
#This function was obtained from Reddit and written by Josh Bredeweg, user jbraids1421. 

################
##countSpaces
#'Count the Spaces in a Character String
#'
#'Counts the number of spaces in a character string.
#'
#'@param x A character string.
#'
#'@return A vector of numeric values identifying the number of instances of spaces (' ') in x, and the number of consecutive spaces in an instance.
#'
#'@author Josh Bredeweg, Reddit user jbraids1421
#'
#'@export
#'
#'@source https://www.reddit.com/r/rstats/comments/2th8ic/function_to_count_the_number_of_white_spaces/
#'
#'@examples
#'countSpaces("abc def")
#'countSpaces("abc def  ghi   jkl mno")
#'
countSpaces <- function(x){
  counter <- 0
  coll <- numeric()
  vec <- strsplit(x," ")[[1]]
  for(i in 1:length(vec)){
    if (vec[i]==""){
      counter <- counter+1
    }
    else{
      if (counter!=0) coll <- c(coll,counter)
      counter <- 1
    }
  }
  coll
}

#BuildAlignments v1.8.0 15JUN2026 - LT/SM

################
##BuildAlignments
#'Build Amino Acid, cDNA and gDNA Alignments
#'
#'Returns a list of data frames of amino acid, codon, nucleotide and genomic alignments, with version information.
#'
#'@param loci A character vector of HLA gene names (e.g., "DRB1", c("A","C")).
#'@param source A character vector of alignment types. The allowed values are "AA", "cDNA", and "gDNA". If 'source' is "cDNA", both codon and cDNA nucleotide alignments are generated. If source is 'AA' or 'gDNA', a single peptide or genomic nucleotide alignment is generated. Up to four alignments will be returned for a locus, as determined by its ability to be transcribed or translated.
#'@param version A character string describing the desired release version (branch) of the ANHIG/IMGTHLA Github repository (e.g. '3.53.0'). The default value ('Latest') returns alignments for the most recent release.
#'@param return_corr_table A logical. When return_corr_table = TRUE, correspondence tables are returned instead of the alignments. The default value is FALSE.
#'
#'@return A list object with a data frame of all allele names (and trimmed, two-field allele names) and their corresponding sequences (Amino Acid, codon, cDNA, or gDNA) for a specific locus, as well as version details for the returned information, when return_corr_table = FALSE. These alignments identify locations of feature boundaries in relation to amino acid, codon, cDNA, and gDNA sequences. When a three- or four-field allele name includes an expression variant suffix, that suffix is appended to the trimmed name. When return_corr_table = TRUE, correspondence tables, relating the named positions in an alignment to the ordinal numbers for each named position and the ordinal numbers of each insertion position, for the specified loci and source are returned, along with version details for the returned information.
#'
#'@importFrom stringr str_squish
#'@importFrom tibble add_column
#'@importFrom dplyr filter %>%
#'@importFrom utils head tail capture.output
#'
#'@note For internal HLAtools use.
#'
#'@export
#'

buildAlignments<-function(loci, source, version = "Latest", return_corr_table = FALSE){

  if(version != "Latest"){ #
    if(!validateVersion(version)){stop(paste(version," is not a valid IPD-IMGT/HLA Database release version."))}
  }else{ version <- getLatestVersion()}

  source <- checkSource(source)

  #checks if input locus is present in version 3.38.0 HLA loci
  #skip name checks for DRB1/3/4/5, as they are a part of the DRB alignment
  for(j in 1:length(loci)){
    if(loci[j]=="DRB1"|loci[j]=="DRB3"|loci[j]=="DRB4"|loci[j]=="DRB5") next
    for(x in 1:length(source)) {
      if(source[x] == "cDNA") {
        if(loci[j]%in% HLAgazetteer$nuc == FALSE) {
          return(warning(paste(loci[j], "is not currently supported for", source[x])))
        }
      }
      if(source[x] == "gDNA") {
        if(loci[j]%in% HLAgazetteer$gen == FALSE) {
          return(warning(paste(loci[j], "is not currently supported for", source[x])))
        }
      }
      if(source[x] == "AA") {
        if(loci[j] %in% HLAgazetteer$prot == FALSE) {
          return(warning(paste(loci[j], "is not currently supported for", source[x])))
        }
      }
    }
  }


  #creates empty variables for future for loops
  comb_list<-start<-end<-alignment<-list()

  #empty variables for correspondence table
  alignmentVersion<-pepsplit<-refexon<-aligned<-final_alignment<-AA_codon_alignments<-DNAalignments<-HLAalignments<-exonB<-inDels<-AA_atlas<-gDNA_atlas<-cDNA_atlas<-atlas<-corr_table<-cols<-downloaded_segments<-w<-alignment_positionsx3<-alignment_positions<-alignment_length<-DNA_start<-alignment_start<-space_diff<-prot_extractions<-refblock_number<-sapply(loci, function(x) NULL)

  #sub out periods in version, if there are any
  # version <- gsub(".", "", version, fixed=T) ## change to
  version <- repoVersion(version)

  #if version is not latest, turn into numeric object
  if(version != "Latest"){
    version <- as.numeric(version)
  }

  for(i in 1:length(loci)){


    for(j in 1:length(source)){
      HLAalignments<-sapply(loci, function(x) NULL)
      #expressions unique to each table source (AA or cDNA)
      if(source[j]=="AA"){
        suffix <- "_prot.txt"
        type <- "Prot"
        delete_lines <- c(1,2)
        divide <- 1
        sequence_name <- "AAsequence"
        } else if(source[j]=="cDNA"){
        suffix <- "_nuc.txt"
        type <- "cDNA"

        #change delete lines to first and second lines if locus does not have protein sequence ## LT
        if(loci[i] %in% HLAgazetteer$nuc[!HLAgazetteer$nuc %in% HLAgazetteer$prot]){
          delete_lines<-c(1,2)
                              #### Fix: DPA2, DPB2 and HLA-N cDNA alignments included an 'AA codon' row in several releases, through they are both pseudogenes
                              if(loci[i] == "DPA2" && version %in% c(3530,3520,3510,3500,3490,3480,3470,
                                                 3460,3450,3440,3430,3420,3410,3400,
                                                 3390,3380,3370,3360,3350,3340,3330,
                                                 3320,3310,3300,3290,3280,3270)) {
                                delete_lines <- c(1,2,3)} # 3.53.0 - 3.27.0

                              if(loci[i] == "DPB2" && version %in% c(3530,3520,3510,3500,3490,3480,3470,
                                                 3460,3450,3440,3430,3420,3410,3400,
                                                 3390,3380,3370,3360,3350,3340,3330,
                                                 3320,3310,3300,3290,3280,3270,3260,
                                                 3250,3240)) {
                                delete_lines <- c(1,2,3)} # 3.53.0 - 3.24.0

                              if(loci[i] %in% c("N","S","U","Y") && version %in% c(3350,3340,3330)) {
                                delete_lines <- c(1,2,3)} # 3.35.0 - 3.33.0

                              if(loci[i] =="Y" && version %in% c(3320,3310,3300,3320,3310,
                                                                 3300,3290,3280,3270,3260,
                                                                 3250,3240,3230,3220,3210,3200)) {
                               delete_lines <- c(1,2,3)} # 3.32.0 - 3.20.0

                              if(loci[i] %in% c("W","T") && version == 3270) {
                                delete_lines <- c(1,2,3)} # 3.35.0 - 3.33.0

                        } else{
                  delete_lines <- c(1,2,3)}
        divide <- 3
        sequence_name <- "cDNAsequence"
      } else if(source[j]=="gDNA"){
        suffix <- "_gen.txt"
        type <- "gDNA"
        delete_lines <- c(1,2)
        divide <- 1
        sequence_name <- "gDNAsequence"
      }

      if(version == 3131) {version <- 3130} ## Fix for downloading the alignments, as the repo URL includes "3130", but the files use "3131".

      #downloads relevant locus alignment file -- readLines allows for space preservation, which is important in
      #finding where the alignment sequence starts
      if(source[j] == "AA"|source[j] == "cDNA"){
        alignment[[loci[i]]] <- readLines(paste("https://raw.githubusercontent.com/ANHIG/IMGTHLA/", repoVersion(version), "/alignments/",paste(ifelse(loci[i]=="DRB1"|loci[i]=="DRB3"|loci[i]=="DRB4"|loci[i]=="DRB5","DRB",loci[i]),suffix,sep=""),sep=""),-1,ok=TRUE,skipNul = FALSE)
      }  else if(source[j] == "gDNA"){
        alignment[[loci[i]]] <- readLines(paste("https://raw.githubusercontent.com/ANHIG/IMGTHLA/", repoVersion(version), "/alignments/",paste(loci[i],suffix,sep=""),sep=""),-1,ok=TRUE,skipNul = FALSE)
      }

      #if version is equal to latest, or if version is numeric and > 3310, obtain
      #version number from line 3, and skip first 7 rows and last 3 rows
      if((version == "Latest") | (is.numeric(version) & version > 3310)){
        # alignmentVersion[[loci[i]]] <-alignment[[loci[i]]][3]
        alignmentVersion[[loci[i]]] <-substr(alignment[[loci[i]]][3],12,nchar(alignment[[loci[i]]][3]))

        #alters alignment file by cutting out non-pertinent information in beginning
        #and end of alignment file
        alignment[[loci[i]]] <- head(alignment[[loci[i]]],-3)
        alignment[[loci[i]]] <- tail(alignment[[loci[i]]],-7)
      }
          ### Fix for extraneous block of allele name rows in HLA-B cDNA alignment in 3.44.0 and 3.43.0.
      if(loci[i] == "B" && version == 3440 && source == "cDNA"){ alignment[[loci[i]]] <- alignment[[loci[i]]][-c(127540:135510)] } # 3.44.0
      if(loci[i] == "B" && version == 3430 && source == "cDNA"){ alignment[[loci[i]]] <- alignment[[loci[i]]][-c(124129:131886)] } # 3.43.0
         ### Fix for extraneous HLA-C allele name rows in version 3.2.0
      if(loci[i] == "C" && version == 320 && source == "cDNA"){ alignment[[loci[i]]] <- alignment[[loci[i]]][-c(14467:15430)] } # 3.43.0

         ### Fix for missing "AA codon" lines in HFE versions 3.27.0 to 3.22.0.
      if(loci[i] == "HFE" && version %in% c(3270,3260,3250,3240,3230,3220) && source == "cDNA") {

              firstPos <- -25
              alignment[[loci[i]]] <- addCodonLine(alignment[[loci[i]]],firstPos)

            }

         ### Fix for missing "AA codon" lines in DPA, DPB, TAP1 and TAP2 in version 0.00.0
      if(version == 300 && source == "cDNA" && loci[i] %in% c("DPA","DPB","TAP1","TAP2")) {

                  if(loci[i] == "DPA") { firstPos <- -31 }
                  if(loci[i] == "DPB") { firstPos <- -29 }
                  if(loci[i] %in% c("TAP1","TAP2")) { firstPos <- 1 }

                  alignment[[loci[i]]] <- addCodonLine(alignment[[loci[i]]],firstPos)
                }

      ### Fix for missing carriage-return between lines 2 and 3 in HLA-V versions 3.14.0, and converting "3.15.0" to "3.14.0"
       if(loci[i] == "V" && version == 3140) {
                  alignment[[loci[i]]] <- append(alignment[[loci[i]]],"Sequences Aligned: 2014 January 17",after=2)
                  alignment[[loci[i]]][2] <- "IMGT/HLA Release: 3.14.0"
                  }

      #if version is numeric and <= 3310, obtain version number from line 2, and
      #skip first 6 rows and last 2 rows
      if((is.numeric(version) & version <= 3310)){ ## this format is 'IPD-IMGT/HLA Release: 3.31.0'. I'd like it to be 'IPD-IMGT/HLA 3.31.0' to match the post 3.31.0 version structure ****
        #alignmentVersion[[loci[i]]] <-alignment[[loci[i]]][2]
        alignmentVersion[[loci[i]]] <- paste(substr(alignment[[loci[i]]][2],1,12),substr(alignment[[loci[i]]][2],gregexpr(pattern =':',alignment[[loci[i]]][2])[[1]][1]+2,nchar(alignment[[loci[i]]][2])),sep = " ")

        #alters alignment file by cutting out non-pertinent information in beginning
        #and end of alignment file
        alignment[[loci[i]]] <- head(alignment[[loci[i]]],-2)
        alignment[[loci[i]]] <- tail(alignment[[loci[i]]],-6)
      }

      #see countSpaces function at beginning of script (AA table only)
      #Counts difference between Prot to -30 and beginning of Prot to -30 + 1 due to zero number indexing to find where
      #the alignment sequence actually starts
      if(source[j]=="AA"){space_diff[[loci[i]]]<-(nchar(strsplit(alignment[[loci[i]]][3], " ")[[1]][2])+countSpaces(alignment[[loci[i]]][3])[2]+1)-countSpaces(alignment[[loci[i]]][2])[1]}

      #reduces repeated whitespace in alignment file and removes rows with empty values for proper
      #start and stop subsetting
      alignment[[loci[i]]] <-str_squish(alignment[[loci[i]]])
      alignment[[loci[i]]] <-alignment[[loci[i]]][-which(alignment[[loci[i]]] == "")]

      #determines positions of "cDNA" or "Prot" and the end of that reference block segment
      start[[loci[i]]] <-as.numeric(grep(type, alignment[[loci[i]]]))

      ## Fix for DOA in 3.01.0 - Only one sequence block
      if(length(start[[loci[i]]]) == 1) {
        end[[loci[i]]] <- length(alignment[[loci[i]]]) ## the only end is the end of the sequence block
          } else {
      end[[loci[i]]] <- as.numeric(c(start[[loci[i]]][2:length(start[[loci[i]]])]-1,length(alignment[[loci[i]]]))) }

      if(version == 3480 && source[j] == "gDNA" && loci[i] == "DRB1"){ ## Fix for DRB1*15:200:01:01N and DRB1*15:200:01:02N in 3.48.0 gDNA alignment
          alignment[[loci[i]]] <- gsub("DRB1*15:200:01:01N","DRB1*15:200:01:01N ",alignment[[loci[i]]], fixed=TRUE)
          alignment[[loci[i]]] <- gsub("DRB1*15:200:01:02N","DRB1*15:200:01:02N ",alignment[[loci[i]]], fixed=TRUE)
      }

      if(source[j]=="AA"){
        #extracts rows with "Prot" and reference sequence position information
        #extracts only relevant reference sequence positions
        #NOTE: the first row containing "Prot" contains two numbers -- -30 and 1 -- where only -30, is extracted,
        #as the actual sequence start will always be 1
        for (k in 1:length(start[[loci[i]]])){

          prot_extractions[[loci[i]]][k]<-strsplit(alignment[[loci[i]]][start[[loci[i]]][k]], " ")

          refblock_number[[loci[i]]][k]<-as.numeric(sapply(prot_extractions[[loci[i]]][k], "[", 2))

          #determines the alignment start by adding -30 to the difference between white spaces found above
          alignment_start[[loci[i]]]<-refblock_number[[loci[i]]][1]+space_diff[[loci[i]]]
        }
      }  else if(source[j]=="cDNA"){
        #these loci do not have protein sequences; set alignment start to 1
        if(loci[i] %in% HLAgazetteer$nuc[!HLAgazetteer$nuc %in% HLAgazetteer$prot]){
          alignment_start[[loci[i]]] <- 1
        } else{
        #determines the alignment start by finding the second vector in second list and removing "codon"
        alignment_start[[loci[i]]] <-as.numeric(sub("AA codon ", "", alignment[[loci[i]]][2]))
        }
        #print(alignment_start[[loci[i]]])
        DNA_start[[loci[i]]] <-as.numeric(sub("cDNA ", "", alignment[[loci[i]]][1]))
        #print(DNA_start[[loci[i]]])
      }  else if(source[j]=="gDNA"){
        DNA_start[[loci[i]]] <-as.numeric(sub("gDNA ", "", alignment[[loci[i]]][1]))
      }

      #closes all white space in the alignment file, except for the white space separating the allele and peptide sequence
      alignment[[loci[i]]] <-paste(substr(alignment[[loci[i]]],1,regexpr(" ",text = alignment[[loci[i]]],fixed = TRUE)), gsub(" ","",substr(alignment[[loci[i]]],regexpr(" ",text = alignment[[loci[i]]],fixed = TRUE),nchar(alignment[[loci[i]]]))),sep = "")

      #string splits at white spaces to yield allele and peptide sequences
      alignment[[loci[i]]]  <- strsplit(alignment[[loci[i]]]," ", fixed=T)

      #binds the previously split strings by row
      alignment[[loci[i]]] <- do.call(rbind,alignment[[loci[i]]])

      #if the seq column is equal to the allele column due to premature peptide termination,
      #insert a blank in place of the allele in the seq column (AA table only)
      if(source[j]=="AA"){alignment[[loci[i]]][which(alignment[[loci[i]]][,1]==alignment[[loci[i]]][,2]),2] <- ""}

      #renames columns to "alleles" and "seq"
      colnames(alignment[[loci[i]]])<-c(paste(loci[i], "alleles", sep="_"), "seq")

      #due to ANHIG formatting, cases where an allele contains newly reference peptide sequences will not
      #contain the same number of rows as previous reference peptide blocks.
      #as of version 3.64.0, this is limited to the last block.
      #this for loop is invoked to add "."for all other alleles for each character in the newly reference peptide
      #to preserve structural integrity
      #changes 10/9/19 to accommodate if there is more than one extraneous allele with an extended amino acid sequence

      # get last start and end positions to evaluate if extraneous alleles are present
      last_start <- start[[loci[[i]]]][length(start[[loci[[i]]]])]
      last_end <- end[[loci[[i]]]][length(end[[loci[[i]]]])]

      if(nrow(alignment[[loci[i]]][last_start:last_end,])!=nrow(alignment[[loci[i]]][start[[loci[i]]][1]:end[[loci[i]]][1],])){
        x<-as.data.frame(alignment[[loci[i]]][,1][start[[loci[i]]][1]:end[[loci[i]]][1]][-c(1,2)], stringsAsFactors = F)
        colnames(x)<-paste(loci[i], "alleles", sep="_")
        temp_vec<-alignment[[loci[i]]][last_start:last_end,][-c(1,2),]

        #if there is only one allele with an extended amino acid sequence, the conversion from a named
        #vector to a data frame is not properly
        #if the character vector is = 2, there is only one allele with an extended amino acid sequence, so use
        #as.list and then data.frame so data frame is created correctly
        if(length(temp_vec) == 2){
          temp_filter<-data.frame(as.list(temp_vec), stringsAsFactors = F)
        } else{
          temp_filter<-data.frame(temp_vec, stringsAsFactors = F)
        }

        #find the greatest number of peptides in extraneous alleles to determine
        #how many "." to add on
        max_nchar<-(temp_filter %>%
                      add_column(nchar = nchar(.$seq)) %>%
                      filter(nchar == max(nchar)))$nchar
        x<-cbind.data.frame(x, seq=as.character(paste(rep(".", max_nchar[1]), collapse = "")), stringsAsFactors=FALSE)
        y<-data.frame(tail(alignment[[loci[i]]], (nrow(alignment[[loci[i]]][last_start:last_end,][nrow(alignment[[loci[i]]][last_start:last_end,])!=nrow(alignment[[loci[i]]][start[[loci[i]]][1]:end[[loci[i]]][1],]),])-2)), stringsAsFactors = F)
        x$seq[match(y[,1], x[,1])]<-y$seq
        alignment[[loci[i]]]<-as.matrix(rbind(head(alignment[[loci[i]]], -(nrow(alignment[[loci[i]]][last_start:last_end,][nrow(alignment[[loci[i]]][last_start:last_end,])!=nrow(alignment[[loci[i]]][start[[loci[i]]][1]:end[[loci[i]]][1],]),])-2)), x))
        start[[loci[i]]]<-as.numeric(grep(type, alignment[[loci[i]]]))
        end[[loci[i]]] <- as.numeric(c(start[[loci[i]]][2:length(start[[loci[i]]])]-1,nrow(alignment[[loci[i]]])))
      }

      #if a locus has extra formatting, resulting in unequal rows, start and end will be updated to reflect subsetting
      #if a locus has no extra formatting, start and end will remain the same, as procured by earlier code
      if(length(start[[loci[i]]]) > 1) { # Fix #2a for DOA in 3.01.0 -- only need to bind together multiple sequence blocks
      for(m in 1:length(start[[loci[i]]])){
        HLAalignments[[loci[i]]]<-cbind(HLAalignments[[loci[i]]], alignment[[loci[i]]][start[[loci[i]]][m]:end[[loci[i]]][m],])
              }
          } else { HLAalignments[[loci[i]]] <- alignment[[loci[i]]]}

      #removes rows containing "cDNA", "AA codon", or "Prot"
      HLAalignments[[loci[i]]] <- HLAalignments[[loci[i]]][-delete_lines,]

      #designates columns to be combined as every other so allele names are not included
      #in pasting all the amino acid sequences together

      ### Fix #2b for DOA in 3.01.0 -- only need to bind together multiple sequence blocks
     if(length(start[[loci[i]]]) > 1) {
            cols<-seq(0, ncol(HLAalignments[[loci[i]]]), by=2)
            HLAalignments[[loci[i]]]<-cbind(HLAalignments[[loci[i]]][,1], apply(HLAalignments[[loci[i]]][,cols], 1 ,paste, collapse = ""))
            }

      #creates a new matrix with the number of columns equal to the number of characters in the reference sequence
      corr_table[[loci[i]]]<-matrix(, nrow = 3, ncol = as.numeric(nchar(HLAalignments[[loci[i]]][,2][1])))

      if(source[j]=="AA"|source[j]=="cDNA"){
        #if the first position enumeration is negative (i.e. has a leader peptide sequence), determines alignment length based on the total number of characters plus the alignment start (which is negative)
        if(grepl("-", alignment_start[[loci[i]]][[1]])==TRUE){
          alignment_length[[loci[i]]]<-(as.numeric(nchar(HLAalignments[[loci[i]]][,2][1]))/divide)+(alignment_start[[loci[i]]])
        } else{ #if there is no leader peptide (i.e sequence starts at 1), determines alignment length based on total number of characters
          alignment_length[[loci[i]]]<-as.numeric(nchar(HLAalignments[[loci[i]]][,2][1]))
        }

        #pastes alignment_start to alignment_length together in sequential order
        #captures output as "w"
        w[[loci[i]]] <- capture.output(cat(alignment_start[[loci[i]]]:alignment_length[[loci[i]]]))

        #splits string formed by cat for separate character variables
        alignment_positions[[loci[i]]]<-as.character(unlist(strsplit(w[[loci[i]]], " ")))

        #eliminates "0" if present in the alignment positions, as the alignment sequence from ANHIG does not contain 0
        if("0" %in% alignment_positions[[loci[i]]]==TRUE){
          alignment_positions[[loci[i]]]<-alignment_positions[[loci[i]]][-which(alignment_positions[[loci[i]]] == 0)]
        }
      }

      #triples alignment to account for codons (cDNA table only)
      if(source[j]=="cDNA"){
        for(n in 1:length(alignment_positions[[loci[i]]])){
          alignment_positionsx3[[loci[i]]]<-c(alignment_positionsx3[[loci[i]]],alignment_positions[[loci[i]]][n],alignment_positions[[loci[i]]][n],alignment_positions[[loci[i]]][n])
        }
        alignment_positions[[loci[i]]]<-alignment_positionsx3[[loci[i]]]
      }

      #string splits to extract locus in the allele name
      #assigns to new variable "aligned"
      aligned[[loci[i]]]<- as.matrix(do.call(rbind,strsplit(HLAalignments[[loci[i]]][,1],"[*]")))

      #adds a new column of pasted locus and trimmed two field alleles to aligned
      aligned[[loci[i]]]<- cbind(aligned[[loci[i]]], paste(aligned[[loci[i]]][,1], unlist(apply(aligned[[loci[i]]],MARGIN=c(1,2),FUN=alleleTrim,resolution=2,append=TRUE)[,2]), sep="*"))

      #binds aligned and HLAalignments -- renames columns
      HLAalignments[[loci[i]]] <- cbind(aligned[[loci[i]]], HLAalignments[[loci[i]]])

      colnames(HLAalignments[[loci[i]]]) <- c("locus", "full_allele", "trimmed_allele", "allele_name", sequence_name)

      if(source[j]=="AA"){
        #if locus is DRB3/4/5, use DRB line 1 as reference sequence
        if(loci[i]=="DRB3"|loci[i]=="DRB4"|loci[i]=="DRB5"){
          #sets refexon to a reference peptide for each HLA locus based on the reference sequences in HLAalignments
          refexon[[loci[i]]] <- rbind(HLAalignments[[loci[i]]][1,])[which(rbind(HLAalignments[[loci[i]]][1,])[,"locus"]=="DRB1"),sequence_name]
        } else{
                ## Fix error in version 3.23.0 with DPA1 cDNA alignments, where the filename is "DPA_prot.txt, and DPB1 with filenane "DPB_prot.txt".
                if(loci[i] %in% c("DPA","DPB","DQA","DQB") && source == "AA" && version %in% c(3230,3220,3210,3200,3190,3180,
                                                                                               3170,3160,3150,3140,3130,3120,
                                                                                               3110,3100,390,380,370,360,
                                                                                               350,340,330,320,310,300)) {

                 if(loci[i] == "DPA") {refexon[[loci[i]]] <- rbind(HLAalignments[[loci[i]]][1,])[which(rbind(HLAalignments[[loci[i]]][1,])[,"locus"]=="DPA1"),sequence_name]}
                 if(loci[i] == "DPB") {refexon[[loci[i]]] <- rbind(HLAalignments[[loci[i]]][1,])[which(rbind(HLAalignments[[loci[i]]][1,])[,"locus"]=="DPB1"),sequence_name]}
                 if(loci[i] == "DQA") {refexon[[loci[i]]] <- rbind(HLAalignments[[loci[i]]][1,])[which(rbind(HLAalignments[[loci[i]]][1,])[,"locus"]=="DQA1"),sequence_name]}
                 if(loci[i] == "DQB") {refexon[[loci[i]]] <- rbind(HLAalignments[[loci[i]]][1,])[which(rbind(HLAalignments[[loci[i]]][1,])[,"locus"]=="DQB1"),sequence_name]}

                  } else {
                      refexon[[loci[i]]] <- rbind(HLAalignments[[loci[i]]][1,])[which(rbind(HLAalignments[[loci[i]]][1,])[,"locus"]==loci[i]),sequence_name]
                          }
                      }
        #if input loci is DRB, use grep statement to match input loci to loci in HLAalignments
        if(loci[i]=="DRB"){
          refexon[[loci[i]]]<-rbind(HLAalignments[[loci[i]]][1,])[grepl(loci[i], rbind(HLAalignments[[loci[i]]][1,])[,"locus"]), sequence_name]}
          }

      #splits sequence column at every amino acid or nucleotide, resulting in a split list of each for each row
      pepsplit[[loci[i]]] <- sapply(HLAalignments[[loci[i]]][,sequence_name],strsplit,split="*")

      #fills in spaces with '.' for alleles with premature termination to make it the same number of characters
      #as the reference sequence (AA table only)
      if(source[j]=="AA"){pepsplit[[loci[i]]]<- lapply(pepsplit[[loci[i]]],function(x) c(x,rep(".",nchar(refexon[[loci[i]]])-length(x))))}

      #nullifies variable names
      names(pepsplit[[loci[i]]]) <- NULL

      #binds pep_split together by element in its previous list form by row ### need to skip this for DPA2 and DPB2??
      suppressWarnings(pepsplit[[loci[i]]]<- do.call(rbind,pepsplit[[loci[i]]]))

      #binds all columns together to form desired output, as described above
      HLAalignments[[loci[i]]] <- cbind.data.frame(HLAalignments[[loci[i]]][,1:4],pepsplit[[loci[i]]], stringsAsFactors=FALSE)

      if(source[j]=="AA"|source[j]=="cDNA"){
        #get reference sequence from DRB alignment
        if(loci[i]=="DRB3"|loci[i]=="DRB4"|loci[i]=="DRB5"){
          DRBref<-HLAalignments[[loci[i]]][1,]}

        #if the locus is DRB1/3/4/5, subset the specific locus from DRB alignment
        if(loci[i]=="DRB1"|loci[i]=="DRB3"|loci[i]=="DRB4"|loci[i]=="DRB5"){
          HLAalignments[[loci[i]]]<-HLAalignments[[loci[i]]][HLAalignments[[loci[i]]]$locus==loci[i],]}

        #add reference sequence to DRB3/4/5, reset row names to numerical order
        if(loci[i]=="DRB3"|loci[i]=="DRB4"|loci[i]=="DRB5"){
          HLAalignments[[loci[i]]]<- rbind(DRBref, HLAalignments[[loci[i]]])
          rownames(HLAalignments[[loci[i]]])<-NULL}
      }

      #inputs HLAalignments alignment sequence into the corr_table with "InDel" still present
      corr_table[[loci[i]]][1,]<-names(HLAalignments[[loci[i]]][5:ncol(HLAalignments[[loci[i]]])])

      #finds positions in HLAalignments that have ".", indicating an inDel
      inDels[[loci[i]]]<-colnames(HLAalignments[[loci[i]]][1, 5:ncol(HLAalignments[[loci[i]]])][HLAalignments[[loci[i]]][1, 5:ncol(HLAalignments[[loci[i]]])] %in% "."])

      #finds positions in HLAalignments that have "|", indicating a exon boundary
      exonB[[loci[i]]]<-colnames(HLAalignments[[loci[i]]][1, 5:ncol(HLAalignments[[loci[i]]])][HLAalignments[[loci[i]]][1, 5:ncol(HLAalignments[[loci[i]]])] %in% "|"])

      #inDel inclusion if there are inDels present
      if(length(inDels[[loci[i]]])!=0){
        for(o in 1:length(inDels[[loci[i]]])){
          corr_table[[loci[i]]][2,][inDels[[loci[i]]][[o]]==corr_table[[loci[i]]][1,]]<-paste("INDEL", o, sep="-")
          corr_table[[loci[i]]][3,][inDels[[loci[i]]][[o]]==corr_table[[loci[i]]][1,]]<-paste("INDEL", o, sep="-")
        }
      }

      #exonB inclusion if there are exon boundaries present
      if(length(exonB[[loci[i]]])!=0){
        for(p in 1:length(exonB[[loci[i]]])){
          corr_table[[loci[i]]][2,][exonB[[loci[i]]][[p]]==corr_table[[loci[i]]][1,]]<-paste("EXONB", p, sep="-")
          corr_table[[loci[i]]][3,][exonB[[loci[i]]][[p]]==corr_table[[loci[i]]][1,]]<-paste("EXONB", p, sep="-")
        }
      }

      if(source[j]=="AA"|source[j]=="cDNA"){
        #pastes alignment_positions into corr_table accounting for InDels and exon boundaries
        codon_count<-as.vector(1)
        for(q in 1:length(corr_table[[loci[i]]][2,])){
          if(is.na(corr_table[[loci[i]]][2,q])){
            corr_table[[loci[i]]][2,q]<-alignment_positions[[loci[i]]][codon_count]
            codon_count<-codon_count+1
          }
        }
      }

      # create number sequence for alignment (cDNA and gDNA tables only)
      if(source[j]=="cDNA"){ alignment_positions[[loci[i]]]<-as.character(unlist(strsplit(capture.output(cat(DNA_start[[loci[i]]]:length(corr_table[[loci[i]]][3,]))), " ")))
      } else if(source[j]=="gDNA"){ if(DNA_start[[loci[i]]] < 0) {alignment_positions[[loci[i]]]<-as.character(unlist(strsplit(capture.output(cat(c(DNA_start[[loci[i]]]:-1,1:length(corr_table[[loci[i]]][3,])))), " "))) ## most loci have negative start positions
      } else {alignment_positions[[loci[i]]]<-as.character(unlist(strsplit(capture.output(cat(c(DNA_start[[loci[i]]]:length(corr_table[[loci[i]]][3,])))), " ")))  #### HLA-P starts at position 1, but in theory some other locus could start at a different positive position
      }
      } # In 3.53.0, only the HLA-P alignment started with 1, so, using the original version above, the first four alignment_positions values are "1"   "0"   "-1"  "1"
      # pastes number sequence into corr_table accounting for InDels and exon boundaries (cDNA and gDNA tables only)

      #pastes number sequence into corr_table accounting for InDels and exon boundaries (cDNA and gDNA tables only)
      if(source[j]=="cDNA"|source[j]=="gDNA"){
        cDNAcount<-as.vector(1)
        for(r in 1:length(corr_table[[loci[i]]][3,])){
          if(is.na(corr_table[[loci[i]]][3,r])){
            corr_table[[loci[i]]][3,r]<-alignment_positions[[loci[i]]][cDNAcount]
            cDNAcount<-cDNAcount+1
          }
        }
      }

      #run if InDels or exon boundaries are present in alignment
      if(any(grepl("INDEL|EXONB", corr_table[[loci[i]]][2,]))){
        #finds which positions have InDels or exon boundaries
        v<-as.numeric(corr_table[[loci[i]]][1,][which(grepl("INDEL", corr_table[[loci[i]]][2,]))])
        w<-as.numeric(corr_table[[loci[i]]][1,][which(grepl("EXONB", corr_table[[loci[i]]][2,]))])

        #splits cumulative sequences
        vsplit<-split(v, cumsum(c(TRUE, diff(v) != 1L)))
        wsplit<-split(w, cumsum(c(TRUE, diff(w) != 1L)))


        if(any(grepl("INDEL", corr_table[[loci[i]]][2,]))){
          #adds increments of .1 to any InDels in row 2
          if(source[j]=="AA"|source[j]=="cDNA"){
            for(s in 1:length(vsplit)){
              if(grepl("-", corr_table[[loci[i]]][2,][vsplit[[s]][[1]]-1])){
                corr_table[[loci[i]]][2,][vsplit[[s]]]<-paste(corr_table[[loci[i]]][2,][vsplit[[s]][[1]]-1], gsub(0, "", seq(1, length(vsplit[[s]]))/10), sep="")
             # }
     #         else{corr_table[[loci[i]]][2,][vsplit[[s]]]<-as.numeric(corr_table[[loci[i]]][2,][vsplit[[s]][[1]]-1])+as.numeric(paste(0,".",seq(1, length(vsplit[[s]])),sep = ""))}
             }  else{corr_table[[loci[i]]][2,][vsplit[[s]]]<-paste(corr_table[[loci[i]]][2,][vsplit[[s]][[1]]-1], paste(".",seq(1,length(vsplit[[s]])),sep=""),sep="")}
            }
          }


          #adds increments of .1 to any InDels in row 3
          if(source[j]=="cDNA"|source[j]=="gDNA"){
            for(t in 1:length(vsplit)){
              corr_table[[loci[i]]][3,][vsplit[[t]]]<-paste(corr_table[[loci[i]]][3,][vsplit[[t]][[1]]-1],".",seq(1, length(vsplit[[t]])),sep = "")
            }
          }
        }


        if(any(grepl("EXONB", corr_table[[loci[i]]][2,]))){

          #changes EXONB to E.(# of last exon)-(# of next exon)
          if(source[j]=="cDNA"){
            for(u in 1:length(wsplit)){
              corr_table[[loci[i]]][2,][wsplit[[u]]]<-paste("E.",u,"-",u+1, sep = "")
              corr_table[[loci[i]]][3,][wsplit[[u]]]<-paste("E.",u,"-",u+1, sep = "")
            }
          }

          #changes EXONB to corresponding boundary e.g. UTR-E.1
          #also creates table (atlas) of gDNA boundaries and their locus
          if(source[j]=="gDNA"){
            if(loci[i] %in% names(fragmentFeatureNames)) { # simplified gDNA atlas builder for genes in the fragemtnFeatureNames object

              gDNA_atlas[[loci[i]]]<-matrix(, nrow = 1, ncol = length(wsplit))
              gDNA_atlas[[loci[i]]]<- as.numeric(corr_table[[i]][3,w+1])
              newNames <- rep(NA,length(fragmentFeatureNames[names(fragmentFeatureNames) %in% loci[i] == TRUE][[1]][[1]])-1)
              for(f in 1:(length(newNames))) {
                newNames[f] <- paste(fragmentFeatureNames[names(fragmentFeatureNames) %in% loci[i] == TRUE][[1]][[1]][f],"-",fragmentFeatureNames[names(fragmentFeatureNames) %in% loci[i] == TRUE][[1]][[1]][f+1], sep="")
              }
              gDNA_atlas[[loci[i]]] <- as.data.frame(t(gDNA_atlas[[loci[i]]]))
              colnames(gDNA_atlas[[loci[i]]]) <- newNames
              rownames(gDNA_atlas[[loci[i]]]) <- "gDNA"

            } else{
              intron<-as.vector(paste("I.",seq(1,((length(wsplit)-1)/2)-0.5),sep = ""))
              exon<-as.vector(paste("E.",seq(1,((length(wsplit)-1)/2)+0.5),sep = ""))
              a<-unlist(strsplit(paste0(exon," ",intron), " "))
              a<-c("UTR",head(a, length(a)-1),"UTR")
              b<-vector()
              for(y in 1:(length(a)-1)){
                b<-c(b,paste(a[[y]],"-",a[[y+1]],sep = ""))}
              gDNA_atlas[[loci[i]]]<-matrix(, nrow = 1, ncol = length(wsplit))
              #covers 3 loci
              if(length(b) != length(gDNA_atlas[[loci[i]]])) {
                gDNA_atlas[[loci[i]]]<-matrix(, nrow = 1, ncol = length(b))
              }
              colnames(gDNA_atlas[[loci[i]]])<-b
              rownames(gDNA_atlas[[loci[i]]])<-"gDNA"
              for(z in 1:length(wsplit)){
                gDNA_atlas[[loci[i]]][1,z]<-corr_table[[loci[i]]][3,][wsplit[[z]]+1]
                corr_table[[loci[i]]][3,][wsplit[[z]]]<-b[[z]]
              }
            }
          }

          #creates table (atlas) of cDNA boundaries and their locus
          if(source[j]=="cDNA"){
            cDNA_atlas[[loci[i]]]<-matrix(, nrow = 2, ncol = length(wsplit))
            colnames(cDNA_atlas[[loci[i]]])<-corr_table[[loci[i]]][2,][unlist(wsplit)]
            rownames(cDNA_atlas[[loci[i]]])<-c("cDNA","AA")
            for(z in 1:length(wsplit)){
              cDNA_atlas[[loci[i]]][1,z]<-corr_table[[loci[i]]][3,][wsplit[[z]]+1]
              cDNA_atlas[[loci[i]]][2,z]<-corr_table[[loci[i]]][2,][wsplit[[z]]+1]
            }
          }

          #creates table (atlas) of exon boundaries and their locus -- PROVISIONAL SJM
          if(source[j]=="AA"){
            AA_atlas[[loci[i]]]<-matrix(, nrow = 2, ncol = length(wsplit))
            colnames(AA_atlas[[loci[i]]])<-corr_table[[loci[i]]][2,][unlist(wsplit)]
            rownames(AA_atlas[[loci[i]]])<-c("cDNA","AA")
            for(z in 1:length(wsplit)){
              AA_atlas[[loci[i]]][1,z]<-corr_table[[loci[i]]][3,][wsplit[[z]]+1]
              AA_atlas[[loci[i]]][2,z]<-corr_table[[loci[i]]][2,][wsplit[[z]]+1]
            }
          }   ## End PROVISIONAL

        }

      }

      #Creates separate tables for AA/codon and DNA positions
      AA_codon_alignments[[loci[i]]]<-HLAalignments[[loci[i]]]
      DNAalignments[[loci[i]]]<-HLAalignments[[loci[i]]]

      if(source[j]=="cDNA"|source[j]=="AA"){
        #renames columns in AA_codon_alignments to codon names
        colnames(AA_codon_alignments[[loci[i]]]) <- c("locus","allele","trimmed_allele","allele_name", corr_table[[loci[i]]][2,])
        #distributes  reference sequence from row 1
        #into all other rows, if they contain a "-"
        #amino acids with changes will not be impacted
        for(w in 5:ncol(AA_codon_alignments[[loci[i]]])) {
          AA_codon_alignments[[loci[i]]][,w][which(AA_codon_alignments[[loci[i]]][,w]=="-")] <- AA_codon_alignments[[loci[i]]][,w][1]}
      }

      if(source[j]=="cDNA"|source[j]=="gDNA"){
        #renames columns in DNAalignments to cDNA names
        colnames(DNAalignments[[loci[i]]]) <- c("locus","allele","trimmed_allele","allele_name", corr_table[[loci[i]]][3,])
        #distributes  reference sequence from row 1
        #into all other rows, if they contain a "-"
        #amino acids with changes will not be impacted
        for(x in 5:ncol(DNAalignments[[loci[i]]])) {
          DNAalignments[[loci[i]]][,x][which(DNAalignments[[loci[i]]][,x]=="-")] <- DNAalignments[[loci[i]]][,x][1]}
      }

      if(source[j]=="AA"){
        final_alignment[[loci[i]]]<-c(final_alignment[[loci[i]]],AA = list(AA_codon_alignments[[loci[i]]]))
      }else if(source[j]=="cDNA"){
        comb_list[[loci[i]]]<-list(codon = AA_codon_alignments[[loci[i]]],cDNA = DNAalignments[[loci[i]]])
        final_alignment[[loci[i]]]<-c(final_alignment[[loci[i]]],comb_list[[loci[i]]])
      }else if(source[j]=="gDNA"){
        final_alignment[[loci[i]]]<-c(final_alignment[[loci[i]]],gDNA = list(DNAalignments[[loci[i]]]))
      }

      #Adds version number
      #     final_alignment[[loci[i]]]<-c(final_alignment[[loci[i]]],`ANHIG/IMGTHLA Alignments Version` = alignmentVersion[[loci[i]]])

      if(return_corr_table){

        object_to_return <- list(data.frame(corr_table[[loci[i]]]))
        #rename column names from 'X1,X2,X3...' to the first row in corr_table,
        #which contains amino acid position indices starting from 1
        colnames(object_to_return[[1]]) <- object_to_return[[1]][1,]

      } else {
        object_to_return <- final_alignment[[loci[i]]]
      }

      final_alignment[[loci[i]]]<-c(object_to_return,'Version' = alignmentVersion[[loci[i]]])

    }

  }
  # final_alignment<-c(final_alignment, atlas = list(atlas)) ## No need to add atlases as there is a separate atlas maker
  return(final_alignment)
}

#GLStoUNI (GL string to uniformat) v1.0.0 12APR2024

#library(stringr)

################
##GLStoUNI
#'Translate GL String to UNIFORMAT
#'
#'A wrapper function for GLtoUN, which translates strings from GL String format to UNIFORMAT format.
#'
#'@param GLstring A character string of HLA allele names and operators in the GL String format signifying their relation with one another.
#'@param prefix A character string of the pertinent gene-system prefix (default is "HLA-").
#'@param pre A logical that indicates whether returned allele names should contain 'prefix' (TRUE), or if 'prefix' should be excluded from returned names (FALSE).
#'
#'@return A version of 'GLstring' converted to UNIFORMAT format, or FALSE if 'GLstring' is invalid.
#'
#'@note This function does not function with the "?" operator, as the "?" operator has no cognate in UNIFORMAT.
#'
#'@export
#'
#'@examples
#'
#'GLStoUNI("HLA-A*02:01/HLA-A*02:02+HLA-A*03:01/HLA-A*03:02")
#'GLStoUNI("A+B/C~D^G|E^W+X/Y^Z+J")
#'
#'@references Nunes Tissue Antigens 2007;69(s1):203-205 https://doi.org/10.1111/j.1399-0039.2006.00808.x
#'@references Mack et al. HLA 2023;102(2):206-212 https://doi.org/10.1111/tan.15126
GLStoUNI <- function(GLstring, prefix = "HLA-", pre = FALSE) {
  #using validator to better protect code
  if(validateGLstring(GLstring,"1.0")) {
          #if all is well, call GLtoUNI
        GLtoUN(GLstring = GLstring, prefix = prefix, pre = pre) } else {
          
          return(FALSE)
        }
}

################
##multiGLStoUNI
#'Translate Multiple GL Strings to UNIFORMAT
#'
#'Translate a data frame or a vector of GL Strings to UNIFORMAT strings.
#'
#'@param GLstringArray A data frame or a vector containing GL String formatted data. If 'GLstringArray' is a data frame with more than one column, the first column should contain only identifiers. If 'GLstringArray' is a vector, it should contain only GL Strings.
#'@param prefix A string of the desired locus prefix (default is "HLA-").
#'@param pre A logical. If 'pre' is TRUE, all allele names will be prefixed with 'prefix'. If 'pre' is FALSE, no allele names will be prefixed.
#'
#'@return A version of 'GLStringArray' in which the GL String data has been converted to UNIFORMAT format. If a 'GLstringArray' was a data frame, a data frame is returned. If 'GLstringArray' was a vector, a vector is returned. 
#'
#'@note GL Strings that include the "?" operator will not be translated, as the "?" operator has no cognate in UNIFORMAT.
#'
#'@export
#'
#'@examples
#'multiGLStoUNI(GLstring.ex[[2]][1:5],version) ## converting a vector of GL Strings to UNIFORMAT
#'multiGLStoUNI(GLstring.ex[1:5,1:2])  ## converting a data.frame of GL Strings to UNIFORMAT
#'
#'@references Nunes Tissue Antigens 2007;69(s1):203-205 https://doi.org/10.1111/j.1399-0039.2006.00808.x
#'@references Mack et al. HLA 2023;102(2):206-212 https://doi.org/10.1111/tan.15126
multiGLStoUNI <- function(GLstringArray, prefix = "HLA-", pre = FALSE) {

    vec <- FALSE
    
      if(is.vector(GLstringArray)) {
        vec <- TRUE
        GLstringArray <-as.data.frame(GLstringArray)
      }
    
    names <- colnames(GLstringArray)

    for(x in ifelse(length(GLstringArray) == 1,1,2):length(GLstringArray)) {
      nn <- names[x]
      for(i in 1:length(GLstringArray[[1]])) {
        if(suppressMessages(validateGLstring(GLstringArray[[nn]][i],"1.0"))) {
          GLstringArray[[nn]][i] <- GLtoUN(GLstringArray[[nn]][i], prefix = prefix, pre = pre)
            } else {
                  message(paste("The GL String in row",i,"contains operators or characters that are not permitted in GL String version 1.0.",sep=" "))
                    }
               }
          }
  
  if(vec) { GLstringArray <- unname(unlist(GLstringArray)) }
    
  GLstringArray
}

################
##GLtoUN
#'Translate GL Strings to UNIFORMAT Strings
#'
#'A function that translates strings from GL String format to UNIFORMAT format.
#'
#'@param GLstring A character string of HLA allele names and operators in the GL String format signifying their relation with one another.
#'@param prefix A character string of the desired gene-system prefix (default is "HLA-").
#'@param pre A logical that indicates whether returned allele names should contain 'prefix' (TRUE), or if 'prefix' should be excluded from returned names (FALSE).
#'
#'@return A version of the input string converted to UNIFORMAT format.
#'
#'@note For internal use only.
#'@note This function does not function with the "?" operator, as the "?" operator has no cognate in UNIFORMAT.
#'
#'@importFrom stringr fixed str_replace_all
#'
#'@export
#'
#'@examples
#'GLtoUN("HLA-A*02:01/HLA-A*02:02+HLA-A*03:01/HLA-A*03:02")
#'GLtoUN("A+B/C~D^G|E^W+X/Y^Z+J")
#'
#'@references Nunes Tissue Antigens 2007;69(s1):203-205 https://doi.org/10.1111/j.1399-0039.2006.00808.x
#'@references Mack et al. HLA 2023;102(2):206-212 https://doi.org/10.1111/tan.15126
GLtoUN <- function(GLstring, prefix = "HLA-", pre = FALSE) {

  UNIret <- GLstring
  
  UNIret <- gsub(prefix, "", UNIret)

  #splitting into large sections, which will be split further
  firstSplit <- strsplit(UNIret, "\\^|[~]|[|]")[[1]]
  
  for (i in 1: length(firstSplit)) {

    #note: SS = second split, TS = third split
    secondSplit <- strsplit(firstSplit[i], "[+]")[[1]]
    thirdSplit <- strsplit(firstSplit[i], "[/]|[+]")[[1]]
    nameSS <-unique(strsplit(firstSplit[i], "[/]|[+]")[[1]])
    lengthSS <- length(secondSplit)
    lengthTS <- length(thirdSplit)
    
    #only invoked for specific scenarios involving cartesian products
    if(lengthSS>=2 && lengthTS >=3) {
      firstSSplit <- strsplit(secondSplit[1], "[/]")[[1]]
      secondSSplit <- strsplit(secondSplit[2], "[/]")[[1]]
      list1 <- vector("list", 1)[[1]]
      tempS <- ""
      #for (n in 1:lengthSS) {
      for (n in 1:length(firstSSplit)) {
        #for (m in 1:lengthTS) {
        for (m in 1:length(secondSSplit)) {
          tempSS <- ""
          tempSS <- paste0(paste0(firstSSplit[n], ","), secondSSplit[m])
          list1 <- append(list1, fixed(tempSS), (length(list1)))
          list1 <- unique(list1)
        }
      }
      fixOne <- paste0(list1, collapse = "|")
      UNIret <- str_replace_all(UNIret, fixed(firstSplit[i]), fixOne)
    }

  }

  #substituting the one to one translations
  UNIret <- gsub("[+]",",",UNIret)

  UNIret <- gsub("\\^"," ",UNIret)

  if (pre) {
    baseSplit <- unique(strsplit(UNIret, "[,]|[ ]|[\t]|[|]|[~]|\\^")[[1]])
    finalSplit <- unique(strsplit(UNIret, "[,]|[ ]|[\t]|[|]|[~]|\\^")[[1]])
    for (i in 1:length(finalSplit)) {
      finalSplit[i] <- paste0(prefix, finalSplit[i])
      UNIret <- str_replace_all(UNIret, fixed(baseSplit[i]), fixed(finalSplit[i]))
    }
  }

  UNIret
}

################
##validateGLstring
#'Validate a GL String
#'
#'Evaluates a GL String to identify unsupported characters for a specified GL String version.
#'
#'@param GLstring A character string of allele names and operators in the GL String format.
#'
#'@param version A character string identifying the GL String version for evaluation. Options are "1.0" and "1.1".
#'
#'@return A logical. TRUE is returned when all characters in a 'version' GL String are permitted. FALSE is returned when forbidden characters are present, or when an incorrect 'version' is specified.
#'
#'@export
#'
#'@examples
#'validateGLstring("HLA-A*02:01/HLA-A*02:02+HLA-A*03:01/HLA-A*03:02", version = "1.0")
#'
#'@references Mack et al. HLA 2023;102(2):206-212 https://doi.org/10.1111/tan.15126
validateGLstring <- function(GLstring, version) {
  #if there is anything not supposed to be there, badOperate will be false
  
  if(!version %in% c("1.0","1.1")) {return(FALSE)}
  
    if(version == "1.0") {
      badOperate <- all(strsplit(GLstring,split="")[[1]] %in% c("/","|","+","^","~",0:9,LETTERS, "w", "b", "l", "a", "n", "k", "g","*","-",":"))
        } 
  
    if(version == "1.1") {
      badOperate <- all(strsplit(GLstring,split="")[[1]] %in% c("/","|","+","^","~","?",0:9,LETTERS, "w", "b", "l", "a", "n", "k", "g","*","-",":"))
    }
    
     if (!badOperate) { message(paste(GLstring," contains characters or operators not permitted in GL String version ",version,".",sep="")) }

  badOperate
}

#### GIANT -- version 1.0.0 July 17, 2024

################
##GIANT
#'GL string code-Integrated Allele Name Translation
#'
#'Calls updateGL() to translate the HLA allele names in a vector or data frame between two IPD-IMGT/HLA Database release versions.
#'
#'@param data A data frame or vector of HLA allele name character string data. Allele names in vectors must include locus prefixes. Data frames must include column headers identifying a single locus for the allele names in that column. 
#'@param transFrom A dot-formatted character string identifying the IPD-IMGT/HLA Database version (e.g. 3.58.0) under which the HLA data were generated.
#'@param transTo A dot-formatted character string identifying the IPD-IMGT/HLA Database version (e.g. 2.25.0) to which which the HLA data are to be translated.
#'
#'@return A data frame or vector of HLA allele name data in the 'transTo' release. 
#'
#'@note GIANT() will ignore the first two columns of BIGDAWG-formatted data frames, which include non-HLA data. All columns in non-BIGDAWG-formatted data frames are expected to contain HLA allele names.
#'
#'@export
#'
#'@examples
#'GIANT(c("A*01:01:01:01","DOA*01:01:01:01"),"3.56.0","2.20.0")
#'updsHLAdata <- GIANT(sHLAdata,"3.56.0","2.25.0")
#'
GIANT <- function(data,transFrom,transTo) {

if(!checkVersion(transFrom)) {return(paste("Release version", transFrom,"is not loaded in the HLAtools package. Please provide a valid release version for 'transFrom'.",sep=" "))}
if(!checkVersion(transTo)) {return(paste("Release version", transTo,"is not loaded in the HLAtools package. Please provide a valid release version for 'transTo'.",sep=" "))}

### figure out data parameters 
        if(is.data.frame(data)) { ## determine type of data 
                  df.data <- TRUE # is a data frame 
                  is.vector <- FALSE
                  startCol <- 1 # assume all columns of a vector contain data 
                  prefix = ifelse(identical(grep("*",data[,3],fixed=TRUE),integer(0)),FALSE,TRUE) # determine if the alleles are prefixed
                  if(any(c(0,1) %in% data[,2])) {
                                bd.data <- TRUE
                                startCol <- 3 # ignore the first two columns of BD data
                                } 
                            } else {
                                  df.data <- FALSE
                                  is.vector <- TRUE} # is a vector
 ### for data.frames
          if(df.data){
                  
            fullframe <- data.frame(matrix(NA,nrow=nrow(data),ncol=ncol(data))) ## this dataframe contains full (suffixed) allele names 
            fullframe[,1:2] <- data[,1:2] ## if this is a BIGDAWG formatted dataset, otherwise it gets overwritten anyway when startCol = 1, below
            
                dots <-  grep(".",colnames(data),fixed=TRUE) ## strip off the .# suffixes if they persisted
                if(length(dots) != 0) {
                  colnames(data)[dots] <- substr(colnames(data)[dots],1,nchar(colnames(data)[dots])-2) } 
                
                  for(i in startCol:ncol(data)){
                    
                        if(!prefix) {   # Add prefixes if necessary
                          
                            fullframe[,i] <- paste(colnames(data)[i],data[,i],sep="*")
                                } else {
                           fullframe[,i] <- data[,i]
                                  }
                    
                            updVec <- GIANT(fullframe[,i],transFrom,transTo)
                            
                            if(!prefix) {
                              for(j in 1:length(updVec)) {
                                      ifelse(is.na(updVec[j]),fullframe[j,i] <- NA,fullframe[j,i] <- strsplit(updVec[j],"*",TRUE)[[1]][2])
                                }
                            } else {
                              fullframe[,i] <- updVec
                              }
                        } 

                  if(strsplit(transTo,".",fixed=TRUE)[[1]][1] %in% c(1,2)) { colnames(data)[colnames(data) == "C"] <- "Cw"}
                  if(strsplit(transTo,".",fixed=TRUE)[[1]][1] %in% c(1,2)) { colnames(data)[colnames(data) == "HLA-C"] <- "HLA-Cw"}
                  if(strsplit(transTo,".",fixed=TRUE)[[1]][1] %in% c(3)) { colnames(data)[colnames(data) == "Cw"] <- "C"}
                  if(strsplit(transTo,".",fixed=TRUE)[[1]][1] %in% c(3)) { colnames(data)[colnames(data) == "HLA-Cw"] <- "HLA-C"}
                colnames(fullframe) <- colnames(data)
            return(fullframe)
            
            # end of data frame section
        
 ### for vectors        
                } else { 
                        uniqueAlleles <- unique(data) 

          ignore <- grep("NA",uniqueAlleles,fixed=TRUE)

          uniqueAlleles <- uniqueAlleles[!1:length(uniqueAlleles) %in% ignore]

          #### translate the unique alleles

          GLScode <- paste("hla",transFrom,paste(uniqueAlleles,collapse="~"),sep="#")

          updAlleles <- gsub("HLA-","",strsplit(strsplit(updateGL(GLScode,transTo),"#",fixed = TRUE)[[1]][3],"~",fixed=TRUE)[[1]],fixed=TRUE)

                    if(length(updAlleles) != length(data)){
                      
                      retVec <- rep(NA,length(data))
                            
                      transTab <- rbind(uniqueAlleles,updAlleles)
                    
                          for(i in 1:length(data)) {
                        
                            if(any(transTab[1,] == data[i])) { 
                              retVec[i] <- transTab[2,transTab[1,] == data[i]]
                                  } else {
                              retVec[i] <- NA }
                          }
                      
                          return(retVec)
            
                            } else {
            
                        return(updAlleles) }
            
              } # end of vector section
          } # end

#alleleTrim v2.2.0 25/20/2025 -- Steven J. Mack

################
##alleleTrim
#'Trim All Versions of HLA Allele Names
#'
#'Trims an HLA allele name to a specified number of fields or number of digits, depending on the nomenclature epoch. Expression variant suffixes in full-length allele names can be appended to truncated allele names.
#'
#'@param allele A character string of an HLA allele name formatted as locus*allele_name, optionally including the "HLA-" prefix.
#'@param resolution A number identifying the number of fields to trim the allele name to.
#'@param version A number identifying the HLA nomenclature epoch under which the allele was named. Epoch 1 allele names are found in IPD-IMGT/HLA Database releases 1.0.0 to 1.16.0. Epoch 2 allele names are found in IPD-IMGT/HLA Database releases 2.0.0 to 2.28.0. Epoch 3 allele names are found in IPD_IMGT/HLA Database releases 3.0.0 and onward.
#'@param append A logical. When append = TRUE, the expression variant suffix of a full-length allele name is appended to a truncated allele name. The default value is FALSE.
#'
#'@return A character string of the trimmed allele name, shortened according to the input parameters.
#'
#'@note Expression variant suffixes will not be removed from full-length allele names that have fewer than four fields or eight digits.
#'
#'@export
#'
#'@examples
#'alleleTrim(allele = "A*03:01:01", resolution = 2)
#'alleleTrim(allele = "A*030101", resolution = 2,version = 2)
#'alleleTrim(allele = "A*0303N",resolution = 1,version = 1, append = TRUE)
#'alleleTrim("HLA-A*24020102L",resolution = 3,version = 2, append = TRUE)
#'
alleleTrim <- function(allele,resolution,version=3,append = FALSE){

  if(!resolution %in% c(1,2,3,4)) {
    message("Resolution must range from 1 and 4.")
    return(allele)
  }

  if(!version %in% c(1,2,3)) { 
    message("Version must range from 1 and 3.")
    return(allele)
    }
  
  if(version == 3){
  currRes <- lengths(regmatches(allele,gregexpr(":",allele)))+1

  if(currRes > resolution) {allele <- getField(allele,resolution,append)}

  }

  ## For versions 1 and 2, split out the allele prefix and suffix

  if(version %in% c(1,2)) {
    suffix <- strsplit(allele,"*",fixed=TRUE)[[1]][2]
    prefix <- strsplit(allele,"*",fixed=TRUE)[[1]][1]
    
      expVar <- ""
      if(append){
         if(substr(allele,nchar(allele),nchar(allele)) %in% c("N","L","S","C","A","Q")) {
                expVar <- substr(allele,nchar(allele),nchar(allele))
             }
       }
  }

  if(version == 2){
    currRes <- nchar(strsplit(allele,"*",fixed=TRUE)[[1]][2])/2
    if(round(currRes,0) > resolution) {allele <- paste(prefix,substr(suffix,1,2*resolution),sep="*")}
  }
  if(version == 1){
    suffixLen <- nchar(suffix)
     if(suffixLen == 4) {currRes <- 2}
     if(suffixLen == 5 || suffixLen == 6) {currRes <- 3} ## Here and in the next line, the second option is due to the presence of an expression variant (e.g. "N")
     if(suffixLen == 7 || suffixLen == 8) {currRes <- 4}
     if(suffixLen == 9) {
       return(alleleTrim(allele,resolution,version=2, append=append))
        } ## This is for the last version 1 release (1.16.0), which frustratingly used 2 digits for each field

     if(suffixLen < 8 && substr(suffix,nchar(suffix),nchar(suffix)) %in% LETTERS) {currRes <- currRes-1}

     if(currRes > resolution) {
       if(resolution %in% c(1,2) ) {allele <- paste(prefix,substr(suffix,1,resolution*2),sep="*")}
       if(resolution == 3) {allele <- paste(prefix,substr(suffix,1,5),sep="*")}

     }
  }
  
  if(version %in% c(1,2) && append && expVar != "") {
    
      if(!substr(allele,nchar(allele),nchar(allele)) %in% c("N","L","S","C","A","Q")) {
        allele <- paste(allele,expVar,sep="")
            }
      }
  
  allele
}

###############
##getField
#' Trim Colon-Delimited HLA Allele Names by Field
#' 
#'@description
#' Trims a properly formatted colon-delimited HLA allele name to a desired number of fields.
#' 
#' If an allele name with an expression-variant suffix is truncated, the suffix can be appended to the end of the truncated allele name. If a resolution value greater than the number of fields in the submitted field is specified, the original allele is returned.
#'
#' @param allele A character string representing an HLA allele.
#' @param res A numeric describing the resolution desired.
#' @param append A logical. When append = TRUE, the expression variant suffix of a full-length allele name is appended to a truncated allele name. The default value is FALSE.
#'
#' @note For internal HLAtools use.
#' 
#' @return A version of the 'allele' character string that has been trimmed to 'res' resolution.
#'
#' @export
#'
#'@examples
#'getField("HLA-A*01:01:01:01", 3)
#'getField("DRB1*11:01:01:12N", 2,TRUE)
#'
getField <- function(allele,res,append=FALSE) {
  Tmp <- unlist(strsplit(as.character(allele),":"))
  suffix <- ""
  
  if(!is.logical(append)) {stop(paste("The value of 'append' [",append,"] is not a logical.",sep=""))}
  
  if(length(Tmp) < res) { return(allele) } # added in case the requested resolution is larger than the allele's resolution
  
  if(grepl("[N,L,S,C,A,Q]",stringr::str_sub(allele, -1))) {suffix <- stringr::str_sub(allele, -1) }
  
  if (length(Tmp)<2) {
    return(allele)
  } else if(res==1) {
    return(paste(Tmp[1],ifelse(append,suffix,""),sep=""))
  } else if(res > 1) {
      if(stringr::str_sub(paste(Tmp[1:res],collapse=":"),-1) == suffix) {
           return(paste(Tmp[1:res],collapse=":")) }
    Out <- paste(paste(Tmp[1:res],collapse=":"),ifelse(append,suffix,""),sep="")
    return(Out)
  }
}

###############
##multiAlleleTrim
#' Trim Multiple HLA Allele Names
#' 
#'Trim a vector of allele names to a specified number of fields or digits.
#'
#'@param alleles A vector of HLA allele names from a single nomenclature epoch.
#'@param resolution A number identifying the number of fields to which the alleles should be trimmed.
#'@param version A number identifying the HLA nomenclature epoch under which the allele was named. Epoch 1 allele names are found in IPD-IMGT/HLA Database releases 1.0.0 to 1.16.0. Epoch 2 allele names are found in IPD-IMGT/HLA Database releases 2.0.0 to 2.28.0. Epoch 3 allele names are found in IPD_IMGT/HLA Database releases 3.0.0 and onward.
#'@param append A logical. When append = TRUE, the expression variant suffix of a full-length allele name is appended to a truncated allele name. The default value is FALSE.
#'
#'@return A vector of the trimmed allele name, shortened according to the input parameters.
#'
#'@note Expression variant suffixes will not be removed from full-length allele names that have fewer than four fields or eight digits.
#'
#'@export
#'
#'@examples
#'alleles <- c("A*02:01:01:02L","DRB1*08:07", "DQB1*04:02:01:16Q")
#'multiAlleleTrim(alleles,2)
#'multiAlleleTrim(alleles,2,append = TRUE)
#'
#'alleles <- c("A*01010102L","DRB1*1613N","HLA-Cw*0322Q")
#'multiAlleleTrim(alleles,1,2,TRUE)
#'multiAlleleTrim(alleles,2,2)
#'
multiAlleleTrim <- function(alleles,resolution,version=3, append = FALSE){
 
  intTrim <- Vectorize(HLAtools::alleleTrim)
                       
  unname(intTrim(alleles,resolution,version,append))
               
}

### Extract IPD-IMGT/HLA Database-Supported Gene Types v3.0.1 1 May 2025

####################
## BuildIMGTHLAGeneTypes
#'Describe IPD-IMGT/HLA Database Genes, Identifying Pseudogenes and Gene Fragments
#'
#'buildIMGTHLAGeneTypes() scrapes 'hla.alleles.org/pages/genes/genes_list' and generates a data frame identifying each gene. It requires internet access to function. As such, this function always returns data for the current IPD-IMGT/HLA Database release.
#'
#'@return A list object of two elements -- 'version' and 'GeneTypes'. The 'version' element identifies the date on which the source table was generated. The 'GeneTypes' element is a data frame of three columns, identifying each gene supported by the IPD-IMGT/HLA Database, along with its molecular characteristics and its status as either a pseudogene or a gene fragment.
#'
#'@importFrom rvest read_html html_table html_nodes
#'@importFrom stats setNames
#'@importFrom stringr %>%
#'
#'@export
#'
#'@note For internal HLAtools use.
#'
buildIMGTHLAGeneTypes <- function(){
  url <-  "https://hla.alleles.org/pages/genes/genes_list"
  
    # Setup and version extraction
    rawPage <- readLines(url,-1,warn = FALSE)
    updateLine <- which(rawPage[] == "        <p class=\"last-updated text-right content\">")+1 ## the line with the update information on it. 
    version <- strsplit(rawPage[updateLine],": ")[[1]][2]

          ## Extract the molecular characteristics table from the page HTML
          df <- url %>% 
          read_html() %>% 
          html_nodes("table") %>% 
          html_table(fill = T) %>%
          lapply(., function(x) setNames(x, c("Names", "Equivalent", "Alignments", 
                                        "Molecular Characteristics","Function","Links")))
  
          ## Capture columns 1 and 4, and ad a pseudogene/genefragment column
          baseTab <- as.data.frame(df[[1]][,c(1,4)])
          baseTab <- cbind(baseTab,rep("",nrow(baseTab)))
          colnames(baseTab)[3] <- "Pseudogene/Fragment"  
  
          ## Identify Pseudogenes and Gene Fragments
          pseudoRows <- grep("pseudogene",baseTab[,2],fixed = FALSE)
          fragmentRows <- grep("gene fragment",baseTab[,2],fixed = FALSE)

          baseTab[,3][pseudoRows] <- "Pseudogene"
          baseTab[,3][fragmentRows] <- "Fragment"
  
    geneTypes <- list(baseTab,version)
    names(geneTypes) <- c("GeneTypes","version")
    
    return(geneTypes)
}

## motifMatch v0.2 26May2024 

####################
##MotifMatch
#'Identify Alleles that Share a Sequence Motif
#'
#'Searches alignments for alleles that share a specific sequence motif. 
#'
#'@param motif A character string identifying a sequence variant motif in the following format: Locus*#$~#$~#$, where ## identifies a variant position, and $ identifies the sequence variant. Both nucleotide and peptide motifs can be provided, and any number of variants can be specified.
#'@param alignType A character string identifying the type of alignment being searched. Allowed values are "codon","gen", nuc" and "prot". Only one 'alignType' value is allowed.
#'@param full A logical value that specifies if full (TRUE) or truncated (FALSE) allele names should be returned.
#'
#'@return A character vector of allele names that contain the motif, or NA when no alleles contain the motif, NULL when no alignment is available for specified locus, and FALSE when the locus or motif is invalid.
#'
#'@note This function requires an HLAalignments object that has been populated with alignments via alignmentFull().
#'
#'@export
#'
motifMatch <- function(motif,alignType,full=TRUE){  
  
        aName <- 3 # return truncated names
            if(full == TRUE) {aName <- 4 } # return full names 
        
        if(!length(alignType) == 1) {stop("Please provide a single 'alignType' value.")}
        
        if(!validateMotif(motif,alignType)) {return(FALSE)}
  
        splitMotif <- strsplit(motif,"*",fixed = TRUE)[[1]]

        loc <- splitMotif[1]
 
        motifs <- strsplit(splitMotif[2],"~",fixed=TRUE)[[1]]
        numMotif <- length(motifs)
        splitMotifs <- strsplit(motifs,"",fixed=FALSE)
 
        posList <- as.list(rep(NA,numMotif)) ## 
        seqList <- as.list(rep(NA,numMotif)) ## may not need these
 
            for(i in 1:numMotif){
 
                  seqVar <- splitMotifs[[i]][splitMotifs[[i]] %in% c(letters,LETTERS) == TRUE]
                  seqPos <- paste(splitMotifs[[i]][!splitMotifs[[i]] %in% c(letters,LETTERS) == TRUE],collapse="")
   
                  posList[[i]] <- HLAalignments[[alignType]][[loc]][HLAalignments[[alignType]][[loc]][,colnames(HLAalignments[[alignType]][[loc]]) %in% seqPos] == seqVar,aName]
             }
  
            matchList <- Reduce(intersect,posList)
  
            if(identical(matchList,character(0))) { 
                  
                  message(paste("The",motif,"motif was not found in any",alignType,"alleles.",sep=" "))
                  return(NA)
            } else {
                matchList
        }
  
}

####################
### Validate Motif
#' Determine if a Motif is Properly Formatted
#'
#'Evaluates a motif to determine if the locus is valid and variants are valid. 
#'
#'@param motif A character string identifying a sequence variant motif in the following format: Locus*#$~#$~#$, where ## identifies a variant position, and $ identifies the sequence variant. Both nucleotide and peptide motifs can be provided, and any number of variants can be specified.
#'@param alignType A character string identifying the type of alignment being searched. Allowed values are "codon","gen", nuc" and "prot". Only one 'alignType' value is allowed.
#'
#'@return If the motif is valid, TRUE is returned. If the locus or body of the motif is invalid, FALSE is returned along with a brief message.
#'
#'@examples
#' validateMotif("A*-21M~2P","prot")
#' validateMotif("A*196G~301A~3046T","gen")
#'
#'@export
#'
validateMotif <- function(motif,alignType){

    alignType <- checkAlignType(alignType)
  
    splitMotif <- strsplit(motif,"*",fixed = TRUE)[[1]]
  
    if(!validateLocus(source = typeToSource(alignType),splitMotif[1])) {
      return(FALSE)
    }
  
    if(alignType == "prot") {
          if(any(strsplit(splitMotif[2],"")[[1]] %in% c(letters,":","*","B","J","O","U","Z","+","!","@","#","$","%","^","&","(",")","[","]","{","}",",","|","<",">","?"))) {
                message(paste(motif,"contains invalid characters.",sep = " "))
                return(FALSE)
                }
            } else { ## for "nuc" and "gen" alignments
              if(any(strsplit(splitMotif[2],"")[[1]] %in% c(letters,":","*","B","D","E","F","H","I","J","K","L","M","N","O","P","Q","R","S","U","V","W","X","Y","Z","+","!","@","#","$","%","^","&","(",")","[","]","{","}",",","|","<",">","?"))) {
                    message(paste(motif,"contains invalid characters.",sep = " "))
                    return(FALSE)
              }
      
      
        }
    
    return(TRUE)
} 

## BDtoPyPop v1.0.1 10 Jul 2024

################
##FormatHead
#'Format PyPop Data Frame Headers
#'
#'Format the header of a PyPop-formatted data frame.
#'
#'@param colHead A vector of column names. PyPop format requires that paired locus/gene names should end in '_1' and '_2', respectively.
#'
#'@return A vector in which the locus names are suffixed with '_1' and '_2'.
#'
#'@note This function assumes that the first two elements the 'colHead' vector are not locus/gene names.
#'
#'@references Lancaster et al. Front Immunol. 2024 Apr 2;15:1378512. https://pubmed.ncbi.nlm.nih.gov/38629078/
#'@references Pappas et al. Hum Immunol. 2016 Mar 77(3):283-287. https://pubmed.ncbi.nlm.nih.gov/26708359/
#'
#'@export
#'
#'@examples
#'formatHead(colHead = colnames(sHLAdata))
#'
formatHead <- function(colHead) {
  
  headLength <- length(colHead) 
  
  colHead[seq(3,headLength,by=2)] <- paste(colHead[seq(3,headLength,by=2)],"_1",sep="")
  colHead[seq(4,headLength,by=2)] <- paste(colHead[seq(4,headLength,by=2)],"_2",sep="")
  
  colHead
}

################
##convertAny
#' Convert Values Across an Entire Data Frame or Vector
#'
#' Converts all instances of a value in a data frame or vector to a specified value.
#'
#' @param dataset A data frame or vector of values.
#' @param change.from The value present in the data frame or vector that should be changed. The default value is NA.
#' @param change.to The value to which 'change.from' should be changed to. The default value is "****".
#'
#' @return A data frame or vector in which 'change.from' has been converted to 'change.to'.
#'
#' @import dplyr
#'
#' @export
#' 
#' @examples
#' dataset <- convertAny(sHLAdata)
#'
convertAny <- function(dataset,change.from=NA,change.to="****") {
  
  if(is.na(change.from)) {
    change.from <- is.na(dataset)
  } else {
    change.from <- (dataset == change.from)
  }
  dataset %>% replace(change.from,change.to)
}

################
##pypopHeaders
#'Convert BIGDAWG File Headers to PyPop Format 
#'
#'Convert the header of a BIGDAWG-formatted data frame into a Python for Population Genomics (PyPop) formatted header.
#'
#'@param colHead A vector of column names. BIGDAWG format requires that the first two header values are the sample identifier character string and the subject status (0 or 1). All other header values should be paired locus/gene names.
#'
#'@return A PyPop-formatted column header vector.
#'
#'@export
#'
#'@references Lancaster et al. Front Immunol. 2024 Apr 2;15:1378512. https://pubmed.ncbi.nlm.nih.gov/38629078/
#'@references Pappas et al. Hum Immunol. 2016 Mar 77(3):283-287. https://pubmed.ncbi.nlm.nih.gov/26708359/
#'
#'@note This function returns unique locus names for each column. For example, "A_1" and "A_2" headers will be returned for two "A" locus columns. PyPop configuration (.ini) files allow this format to be customized. 
#'@seealso [PyPop Configuration File](http://pypop.org/docs/guide-chapter-usage.html#a-minimal-configuration-file)
#'
#'@examples
#'pyHead <- pypopHeaders(colnames(sHLAdata))
#'
pypopHeaders <- function(colHead) {
  
  if(length(grep("[_][1]|[_][2]",colHead)) == 0) { ## no underscores
    
          if(length(grep("[.][1]|[.[2]",colHead)) == 0) { # no periods
           
            colHead <- formatHead(colHead)
      
            } else { ## there are *some* periods
              
              targets <- grep("[.][1]|[.[2]",colHead)
              if(length(targets) == length(colHead)-2) { # everything has a period
                
                colHead <- gsub(".","_",colHead,fixed=TRUE)
                
                    } else { ## guessing here all of the periods are ".1" on the second locus name
                
                      if(all(grep(".1",colHead,fixed = TRUE) == targets)) { ## confirm that guess
                      
                          colHead <- gsub(".1","",colHead,fixed=TRUE)
                          
                          colHead <- formatHead(colHead)
                      
                      } else { ## the mysterious fix -- strip everything 
                        
                        colhead <- formatHead(gsub("[.][1]|[.][2]","",colHead))
                        
                      }
                  }
             }
    
        }
  
  colHead
}

################
## BDtoPyPop
#'Convert BIGDAWG datasets to PyPop datasets
#'  
#'Converts a BIGDAWG-formatted data frame into a pair of PyPop-formatted case and control data frames.
#'
#'@param dataset A data frame containing a BIGDAWG-formatted case-control dataset
#'@param filename A character string identifying the desired path and name for the generated files or the elements of the returned list object.
#'@param save.file A logical that determines if a pair of files should be written (TRUE) or if a list object should be returned. The default value is TRUE.
#'@param save.path A character string identifying the path in which to write the pair of files when save.file is TRUE. The default value is tempdir().
#'
#'@return When save.file = TRUE, a pair of files named "'filename'.positive.pop" and "'filename'.negative.pop" are generated in the working directory. When save.file = FALSE, a list of two-elements, named "'filename'.positive" and "'filename'.negative", is returned. 
#'
#'@note Files generated by BDtoPyPop are intended for analysis using Pypop version 1.#.#. PyPop is not an R package, and must be installed separately. 
#'
#'@references Lancaster et al. Front Immunol. 2024 Apr 2;15:1378512. https://pubmed.ncbi.nlm.nih.gov/38629078/
#'@references Pappas et al. Hum Immunol. 2016 Mar 77(3):283-287. https://pubmed.ncbi.nlm.nih.gov/26708359/
#'
#'@export
#'
#'@examples
#' HLAdata.PP <- BDtoPyPop(sHLAdata,"BDHLA",FALSE)
#'
BDtoPyPop <- function(dataset, filename, save.file=TRUE,save.path = tempdir()) {
  
  colnames(dataset) <- pypopHeaders(colnames(dataset))
  dataset <- convertAny(dataset)
  
  posSet <- dataset[dataset[,2] %in% c(1,"1"),]
  negSet <- dataset[dataset[,2] %in% c(0,"0"),]
  
  if(save.file) {
    write.table(posSet,paste(save.path,paste(filename,"positive","pop",sep="."),sep=.Platform$file.sep),append = FALSE,quote = FALSE,sep ="\t",na = "****",row.names = FALSE,col.names = TRUE)
    write.table(negSet,paste(save.path,paste(filename,"negative","pop",sep="."),sep=.Platform$file.sep),append = FALSE,quote = FALSE,sep ="\t",na = "****",row.names = FALSE,col.names = TRUE)
  } else {
    
    retList <- list(posSet,negSet)
    names(retList) <- c(paste(filename,"positive",sep="."), paste(filename,"neagtive",sep="."))
    
    retList
    
  }
}

## v1.2.0 15 March 2024

####################
##checkgDNAStart
#'Identify gDNA Alignments in Which the First Feature Boundary is not Identified as Position +1.
#'
#'Checks the position of the first feature boundary (the 5' UTR - Exon 1 boundary in expressed genes) in each gDNA alignment, and identifies those alignments in which that position is not 1.
#'
#'This function reviews the gDNA atlases in the HLAatlas object and returns a data frame of the first feature boundary position for each locus. For expressed genes and some pseudogenes, this is the position of the start of Exon 1.
#'
#'@param verbose A logical indicating if loci with first feature boundary positions that are not 1 should be reported in the console (verbose = TRUE).
#'
#'@return A one-row data frame with one column for each locus with a gDNA alignment.
#'
#'@note This function requires that the HLAalignments object has been populated with alignments via the alignmentFull() function.
#'
#'@export
#'
#'@note For internal HLAtools use.
#'
checkgDNAstart <- function(verbose=FALSE){

  gLen <- length(HLAatlas$gen)
  findings <- matrix(rep(1,gLen),ncol=gLen)
  colnames(findings) <- names(HLAatlas$gen)

    tic <- 0
      for (i in 1:gLen){

        if(HLAatlas$gen[[i]][1,1] != 1) {

        if(verbose) {cat(paste("The protein start position for the ",names(HLAatlas$gen)[i]," locus is ",HLAatlas$gen[[i]][1,1],".","\n",sep=""))}
      findings[i] <- as.numeric(HLAatlas$gen[[i]][1,1])
    tic <- tic + 1
    }

        if(tic != 0 && verbose) {paste("All loci start with position 1.")}
        }
      as.data.frame(findings)
    }

#atlasMaker v1.8.0 17JUN2025

################
##atlasMaker
#'Identify the Gene-Feature Boundaries in HLA Region Genes
#'
#'Builds an 'atlas' of the gene-feature (exon, intron and UTR) boundaries for IPD-IMGT/HLA loci, using IPD-IMGT/HLA Database alignments.
#'
#'@param loci A character string identifying a specific HLA gene (ex. "DRB1"). More than one gene can be specified (e.g., c("DRB1","DQB1")).
#'@param source A character string identifying the type of alignment used to build the atlas. The allowed options are "AA", for protein alignments, "cDNA", for nucleotide alignments, and "gDNA", for genomic alignments. More than one alignment type can be specified (e.g., c("AA","gDNA")).
#'@param version A character string identifying the release version of the IPD-IMGT/HLA Github repository for which alignments should be generated.
#'
#'@return A list object of 'atlas' dataframes and a 'release version' character string for each locus and alignment type.
#'
#'@description
#'The 'AA' atlas identifies the amino acid (AA) positions encoded by codons that span each exon (E) boundary. The 'cDNA' atlas describes the 3' nucleotide position in the cDNA sequence that follows each exon boundary, as well as the codon that spans each exon boundary. The 'gDNA' atlas describes first 3' nucleotide position following each UTR (U), exon or intron (I) boundary. Each feature is followed by its rank (e.g U3 is the 3' UTR). Non-standard gene-features (H, J, N, and S) are detailed in the documentation for the ffN() function.
#'
#'@note For internal HLAtools use.
#'@note Nucleotide atlases for pseudogenes will include a 'codon' row populated with NA values.
#'@note Some HLA-DQB1\*05 and \*06 alleles include a 5th exon that is not present in the DQB1\*05:01:01:01 reference allele. In this case, the Exon 4 to Exon 5 boundary is defined as 'Ins' (insertion). For all other alleles there is no E.4-5 boundary.
#'@note Boundaries for non-standard hybrid (H), join (J), novel (N) and segment (S) features may be included in gene fragment and pseudogene atlases.
#'
#'@importFrom DescTools StrLeft IsOdd
#'@importFrom dplyr filter
#'@importFrom stringr str_replace str_squish
#'@importFrom tibble add_column
#'@importFrom utils capture.output head tail
#'
#'@export
#'
atlasMaker<-function(loci, source, version = "Latest"){

  if(version != "Latest"){ #
    if(!validateVersion(version)){stop(paste(version," is not a valid IPD-IMGT/HLA Database release version."))}
  }else{ version <- getLatestVersion()}

  source <- (checkSource(source))

  loci <- multiLocusValidation(loci, source) # Added as a check to make sure that HLAgazetteer#version is enforced; otherwise validateLocus could return FALSE

  if(validateLocus(loci=loci,source=source)) { ## primary 'check'.

    subSource<- expressed<-genestructure<-nucVersion<-pep_start<-nuc<-nuc_df<-extract_ref<-nuc_extract<-start<-end<-pipe_split<-column_names<-boundaries<-boundary_split<-AA_atlas<-sapply(loci, function(x) NULL)

    #creates empty variables for future for loops
    comb_list<-start<-end<-alignment<-list()

    #empty variables for correspondence table
    alignmentVersion<-pepsplit<-refexon<-aligned<-final_alignment<-AA_codon_alignments<-DNAalignments<-HLAalignments<-exonB<-inDels<-gDNA_atlas<-cDNA_atlas<-atlas<-corr_table<-cols<-downloaded_segments<-w<-alignment_positionsx3<-alignment_positions<-alignment_length<-DNA_start<-alignment_start<-space_diff<-prot_extractions<-refblock_number<-sapply(loci, function(x) NULL)

    #sub out periods in version, if there are any
    version <- gsub(".", "", version, fixed=T)

    #if version is not latest, turn into numeric object
    # if(version != "Latest"){

    version <- as.numeric(version)

    # }

    for(i in 1:length(loci)){ # main loop of i values

      expressed[[i]] <- !loci[i] %in% HLAgazetteer$nuc[HLAgazetteer$nuc %in% c(HLAgazetteer$pseudo,HLAgazetteer$frag)] # Exclude pseudogenes and gene fragments

      for(j in 1:length(source)){ #### Source Loop of j values

        subSource[[j]] <- source[[j]] ## identify the submitted source so that that the cDNA code can be used for both AA and cDNA
        if(subSource[[j]] == "AA") { source[[j]] <- "cDNA" ## Bypass all of the AA code for source = AA, in favor of the cDNA code
        } else {subSource[[j]] <- 0}

        if(source[j]=="AA"){ ## start of the AA routine
          # download nucleotide alignment from IPD-IMGT/HLA Github repository
          nuc[[loci[[i]]]]<-readLines(paste("https://raw.githubusercontent.com/ANHIG/IMGTHLA/", repoVersion(version),"/alignments/",paste(ifelse(loci[[i]] %in% c("DRB1", "DRB3", "DRB4", "DRB5"),"DRB",loci[[i]]),"_nuc.txt",sep=""),sep=""),-1,ok=TRUE,skipNul = FALSE)

          if((version == "Latest") | (is.numeric(version) & version > 3310)){ ## Account for the changes in file header structures

            nucVersion[[loci[[i]]]]<-substr(nuc[[loci[i]]][3],12,nchar(nuc[[loci[i]]][3])) } else {
              nucVersion[[loci[[i]]]] <- paste(substr(nuc[[loci[i]]][2],1,12),substr(nuc[[loci[i]]][2],gregexpr(pattern =':',nuc[[loci[i]]][2])[[1]][1]+2,nchar(nuc[[loci[i]]][2])),sep = " ")
            } # closes two lines above

          # reduces repeated whitespace between allele and nucleotide
          nuc[[loci[[i]]]]<-str_squish(nuc[[loci[[i]]]])

          # remove impertinent header/footer information
          nuc[[loci[[i]]]] <- head(nuc[[loci[[i]]]],-2)
          nuc[[loci[[i]]]] <- tail(nuc[[loci[[i]]]],-6)

          # remove all whitespace, except for the whitespace between the allele and nucleotide sequence
          nuc[[loci[[i]]]]<-paste(substr(nuc[[loci[i]]],1,regexpr(" ",text = nuc[[loci[i]]],fixed = TRUE)), gsub(" ","",substr(nuc[[loci[i]]],regexpr(" ",text = nuc[[loci[i]]],fixed = TRUE),nchar(nuc[[loci[i]]]))),sep = "")

          # split at white spaces to yield allele and nucleotide sequences
          nuc[[loci[i]]]  <- strsplit(nuc[[loci[i]]]," ", fixed=T)

          # bind the previously split strings by row
          nuc[[loci[i]]] <- do.call(rbind,nuc[[loci[i]]])

          #extracts beginning alignment enumeration
          pep_start[[loci[[i]]]]<-as.numeric(gsub("codon", "", nuc[[loci[[i]]]][[2,2]]))
          ## correction for nuc alignments that start at codon 1 -- currently TAP1 and TAP2

          colnames(nuc[[loci[[i]]]])<-c(paste(loci[[i]], "alleles", sep="_"), "pepseq")

          # find start and end of each "cDNA" block
          start[[loci[i]]]<-as.numeric(grep("cDNA", nuc[[loci[i]]]))
          end[[loci[i]]] <- as.numeric(c(start[[loci[i]]][2:length(start[[loci[i]]])]-1,nrow(nuc[[loci[i]]])))

          # Due to ANHIG formatting, cases where an allele contains newly reference nucleotide sequences will not
          # contain the same number of rows as previous reference peptide blocks
          # this for loop is invoked to add "."for all other alleles for each character in the new reference nucleotide sequence
          # to preserve structural integrity
          for(k in 1:length(start[[loci[i]]])){ ## pad out ... loop
            if(nrow(nuc[[loci[[i]]]][start[[loci[i]]][k]:end[[loci[i]]][k],])!=nrow(nuc[[loci[i]]][start[[loci[i]]][1]:end[[loci[i]]][1],])){
              x<-as.data.frame(nuc[[loci[i]]][,1][start[[loci[i]]][1]:end[[loci[i]]][1]][-c(1,2)], stringsAsFactors = F)
              colnames(x)<-paste(loci[[i]], "alleles", sep="_")
              x<-cbind.data.frame(x, pepseq=as.character(paste(rep(".", nchar(tail(nuc[[loci[i]]][,2], 1))), collapse = "")), stringsAsFactors=FALSE)
              y<-data.frame(tail(nuc[[loci[i]]], (nrow(nuc[[i]][start[[loci[i]]][k]:end[[loci[i]]][k],][nrow(nuc[[i]][start[[loci[i]]][k]:end[[loci[i]]][k],])!=nrow(nuc[[loci[i]]][start[[loci[i]]][1]:end[[loci[i]]][1],]),])-2)), stringsAsFactors = F)
              colnames(y)<-c(paste(loci[[i]], "alleles", sep="_"), "pepseq")
              x$pepseq[match(y[,1], x[,1])]<-y$pepseq
              nuc[[loci[i]]]<-as.matrix(rbind(head(nuc[[loci[i]]], -(nrow(nuc[[i]][start[[loci[i]]][k]:end[[loci[i]]][k],][nrow(nuc[[i]][start[[loci[i]]][k]:end[[loci[i]]][k],])!=nrow(nuc[[loci[i]]][start[[loci[i]]][1]:end[[loci[i]]][1],]),])-2)), x))
              start[[loci[i]]]<-as.numeric(grep("cDNA", nuc[[loci[i]]]))
              end[[loci[i]]] <- as.numeric(c(start[[loci[i]]][2:length(start[[loci[i]]])]-1,nrow(nuc[[loci[i]]])))}
          } # close ... loop

          # subset nucleotide alignment based on start and end blocks
          for(e in 1:length(start[[loci[[i]]]])){
            nuc_extract[[loci[i]]]<-cbind(nuc_extract[[loci[i]]], nuc[[loci[i]]][start[[loci[i]]][e]:end[[loci[i]]][e],])}

          # remove first two rows containing AA position and "Prot"
          nuc_extract[[loci[i]]] <- nuc_extract[[loci[i]]][-c(1,2,3),]

          # extract only the nucleotide reference sequence
          extract_ref[[loci[i]]]<-rbind(nuc_extract[[loci[[i]]]][1,])

          # paste every two columns together to get full nucleotide reference sequence without allele repetition
          cols<-seq(0, ncol(extract_ref[[loci[i]]]), by=2)
          extract_ref[[loci[i]]]<-cbind(extract_ref[[loci[i]]][,1], apply(extract_ref[[loci[i]]][,cols, drop=F], 1 ,paste, collapse = ""))

          # split reference sequences at pipes
          pipe_split[[loci[[i]]]]<-strsplit(extract_ref[[loci[[i]]]][,2], "|", fixed = T) ### Re: TAP1/TAP2 boundary error -- this pipe_spilt is correct

          for(z in 1:(length(pipe_split[[loci[[i]]]][[1]])-1)){ ## z-loop
            if(StrLeft(pipe_split[[loci[[i]]]][[1]][z], 2)!="..") {
              pipe_split[[loci[[i]]]][[1]][z]<-paste(pipe_split[[loci[[i]]]][[1]][z], StrLeft(pipe_split[[loci[[i]]]][[1]][z+1], 2), sep="")
            }

            if(StrLeft(pipe_split[[loci[[i]]]][[1]][z], 2)=="..") {
              pipe_split[[loci[[i]]]][[1]][z]<-pipe_split[[loci[[i]]]][[1]][z]
            }

            if(StrLeft(pipe_split[[loci[[i]]]][[1]][z+1], 2)=="..") {next}
            pipe_split[[loci[[i]]]][[1]][z+1]<-str_replace(pipe_split[[loci[[i]]]][[1]][z+1], substr(pipe_split[[loci[[i]]]][[1]][z+1],1, 2), "")
          } ## close z-loop
          # split every nucleotide within each boundary, and remove InDels
          for(k in 1:length(pipe_split[[loci[[i]]]][[1]])){ ## k-loop
            boundary_split[[loci[[i]]]][[k]]<-strsplit(pipe_split[[loci[[i]]]][[1]][[k]], "*")[[1]]
            if(boundary_split[[loci[[i]]]][[k]][[length(boundary_split[[loci[[i]]]][[k]])]]=="."){next}
            if(boundary_split[[loci[[i]]]][[k]][1]=="."){next}
            boundary_split[[loci[[i]]]][[k]]<-boundary_split[[loci[[i]]]][[k]][boundary_split[[loci[[i]]]][[k]]!="."]
          } ## close k-loop

          # form a blank dataframe with the number of rows equal to the number of exon boundaries
          AA_atlas[[loci[[i]]]]<-data.frame(matrix("", ncol=(length(boundary_split[[loci[[i]]]])-1)), stringsAsFactors = F)

          rownames(AA_atlas[[loci[[i]]]])<-"AA"

          # paste together individual nucleotides to form peptides; count number of peptides (codons) present between boundaries
          for(q in 1:(length(boundary_split[[loci[[i]]]])-1)){
            d <- seq.int(1L,length(boundary_split[[loci[[i]]]][[q]]),by = 3L)
            boundaries[[loci[[i]]]][[q]]<-length(paste0(boundary_split[[loci[[i]]]][[q]][d],boundary_split[[loci[[i]]]][[q]][d+1], boundary_split[[loci[[i]]]][[q]][d+2]))
          }

          # break out of for loop to add alignment start enumeration to actual start
          boundaries[[loci[[i]]]][[1]]<-boundaries[[loci[[i]]]][[1]]+pep_start[[loci[[i]]]] ### This seems to be the source of the discrepancy with TAP1 and TAP2 their pep_start values should be 0

          positions<-sapply(length(boundary_split[[loci[[i]]]])-1, function(x) NULL)

          tapLengths <- c(-1,-2,-2,-2,-2,-2,-2,-2,-3,-3) ## an ad hoc solution for the TAP1 and TAP2 genes. I'd prefer something programmatic, but I'm assuming that no more genes will be added
          # fill in AA_atlas information
          # boundaries obtained by adding up cumulative lengths
          for(q in 1:(length(boundary_split[[loci[[i]]]])-1)){
            column_names<-c(if(q==1){}else{column_names},column_names[[q]]<-paste("E", paste(seq(1, length(boundary_split[[loci[[i]]]]))[[q]], seq(1, length(boundary_split[[loci[[i]]]]))[[q+1]], sep="-"), sep="."))
            positions[[q]]<-as.numeric((ifelse(pep_start[[loci[[i]]]] == 1, tapLengths[q],0) + cumsum(boundaries[[loci[[i]]]])[[q]]))
            AA_atlas[[loci[[i]]]][1,q]<-positions[[q]]
          }
          colnames(AA_atlas[[loci[[i]]]])<-column_names

          atlas[[loci[i]]] <-c(atlas[[loci[i]]], AA = list(AA_atlas[[loci[i]]]),`Version` = nucVersion[[loci[[1]]]])
          ## end of the AA routine
        } else if(source[j]=="cDNA" | source[j]=="gDNA"){ ### cDNA/gDNA routine
          HLAalignments<-sapply(loci, function(x) NULL)
          # expressions unique to each table source (AA or cDNA)
          if(source[j]=="AA"){
            suffix <- "_prot.txt"
            type <- "Prot"
            delete_lines <- c(1,2)
            divide <- 1
            sequence_name <- "AAsequence"
          } else if(source[j]=="cDNA"){
            suffix <- "_nuc.txt"
            type <- "cDNA"
            delete_lines <- c(1,2,3) # cDNA alignments for unexpressed genes (e.g., HLA-H, J, K, L, N, S, T, U, V, W and Y) only have 2 lines
            if(!expressed[[i]]) {delete_lines <- c(1,2)} ## prior to release 3.54.0, DPA2 and DPB2 had 3 lines, but are pseudogenes

            ### Some cDNAs for pseudogenes still have 3 lines. These lines fix that.
            #### Fix: DPA2, DPB2 and HLA-N cDNA alignments included an 'AA codon' row in several releases, through they are both pseudogenes
            if(loci[i] == "DPA2" && version %in% c(3530,3520,3510,3500,3490,3480,3470,
                                                   3460,3450,3440,3430,3420,3410,3400,
                                                   3390,3380,3370,3360,3350,3340,3330,
                                                   3320,3310,3300,3290,3280,3270)) {
              delete_lines <- c(1,2,3)} # 3.53.0 - 3.27.0

            if(loci[i] == "DPB2" && version %in% c(3530,3520,3510,3500,3490,3480,3470,
                                                   3460,3450,3440,3430,3420,3410,3400,
                                                   3390,3380,3370,3360,3350,3340,3330,
                                                   3320,3310,3300,3290,3280,3270,3260,
                                                   3250,3240)) {
              delete_lines <- c(1,2,3)} # 3.53.0 - 3.24.0

            if(loci[i] %in% c("N","S","U","Y") && version %in% c(3350,3340,3330)) {
              delete_lines <- c(1,2,3)} # 3.35.0 - 3.33.0

            if(loci[i] =="Y" && version %in% c(3320,3310,3300,3320,3310,
                                               3300,3290,3280,3270,3260,
                                               3250,3240,3230,3220,3210,3200)) {
              delete_lines <- c(1,2,3)} # 3.32.0 - 3.20.0

            if(loci[i] %in% c("W","T") && version == 3270) {
              delete_lines <- c(1,2,3)} # 3.35.0 - 3.33.0

            divide <- 3
            sequence_name <- "cDNAsequence"
          } else { if(source[j]=="gDNA"){
            suffix <- "_gen.txt"
            type <- "gDNA"
            delete_lines <- c(1,2)
            divide <- 1
            sequence_name <- "gDNAsequence"
          } }

          if(version == 3131) {version <- 3130} ## Fix for downloading the alignments, as the repo URL includes "3130", but the files use "3131".

          # download relevant locus alignment file -- readLines allows for space preservation, which is important in
          # finding where the alignment sequence starts
          if(source[j] == "AA"|source[j] == "cDNA"){
            alignment[[loci[i]]] <- readLines(paste("https://raw.githubusercontent.com/ANHIG/IMGTHLA/", repoVersion(version), "/alignments/",paste(ifelse(loci[[i]]=="DRB1"|loci[[i]]=="DRB3"|loci[[i]]=="DRB4"|loci[[i]]=="DRB5","DRB",loci[[i]]),suffix,sep=""),sep=""),-1,ok=TRUE,skipNul = FALSE)
          } else { if(source[j] == "gDNA"){
            alignment[[loci[i]]] <- readLines(paste("https://raw.githubusercontent.com/ANHIG/IMGTHLA/", repoVersion(version), "/alignments/",paste(loci[[i]],suffix,sep=""),sep=""),-1,ok=TRUE,skipNul = FALSE)
          } }

          ### Fix for extraneous block of allele name rows in HLA-B cDNA alignment in 3.44.0 and 3.43.0.
          if(loci[i] == "B" && version == 3440 && source == "cDNA"){ alignment[[loci[i]]] <- alignment[[loci[i]]][-c(127540:135510)] } # 3.44.0
          if(loci[i] == "B" && version == 3430 && source == "cDNA"){ alignment[[loci[i]]] <- alignment[[loci[i]]][-c(124129:131886)] } # 3.43.0
          ### Fix for extraneous HLA-C allele name rows in version 3.2.0
          if(loci[i] == "C" && repoVersion(version) == 320 && source == "cDNA"){ alignment[[loci[i]]] <- alignment[[loci[i]]][-c(14467:15430)] } # 3.43.0

          ### Fix for missing "AA codon" lines in HFE versions 3.27.0 to 3.22.0.
          if(loci[i] == "HFE" && version %in% c(3270,3260,3250,3240,3230,3220) && source == "cDNA") {

            firstPos <- -25
            alignment[[loci[i]]] <- addCodonLine(alignment[[loci[i]]],firstPos)

          }

          ### Fix for missing "AA codon" lines in DPA, DPB, TAP1 and TAP2 in version 0.00.0
          if(repoVersion(version) == 300 && source == "cDNA" && loci[i] %in% c("DPA","DPB","TAP1","TAP2")) {

            if(loci[i] == "DPA") { firstPos <- -31 }
            if(loci[i] == "DPB") { firstPos <- -29 }
            if(loci[i] %in% c("TAP1","TAP2")) { firstPos <- 1 }

            alignment[[loci[i]]] <- addCodonLine(alignment[[loci[i]]],firstPos)
          }

          ### Fix for missing carriage-return between lines 2 and 3 in HLA-V versions 3.14.0, and converting "3.15.0" to "3.14.0"
          if(loci[i] == "V" && version == 3140) {
            alignment[[loci[i]]] <- append(alignment[[loci[i]]],"Sequences Aligned: 2014 January 17",after=2)
            alignment[[loci[i]]][2] <- "IMGT/HLA Release: 3.14.0"
          }


          # if version is equal to latest, or if version is numeric and > 3310, obtain
          # version number from line 3, and skip first 7 rows and last 3 rows
          if((version == "Latest") | (is.numeric(version) & version > 3310)){
            alignmentVersion[[loci[i]]] <-substr(alignment[[loci[i]]][3],12,nchar(alignment[[loci[i]]][3]))

            # alter alignment file by cutting out non-pertinent information in beginning
            # and end of alignment file
            alignment[[loci[i]]] <- head(alignment[[loci[i]]],-3)
            alignment[[loci[i]]] <- tail(alignment[[loci[i]]],-7)
          }

          # if version is numeric and <= 3310, obtain version number from line 2, and
          # skip first 6 rows and last 2 rows
          if((is.numeric(version) & version <= 3310)){
            alignmentVersion[[loci[i]]] <- paste(substr(alignment[[loci[i]]][2],1,12),substr(alignment[[loci[i]]][2],gregexpr(pattern =':',alignment[[loci[i]]][2])[[1]][1]+2,nchar(alignment[[loci[i]]][2])),sep = " ")

            # alters alignment file by cutting out non-pertinent information in beginning
            # and end of alignment file
            alignment[[loci[i]]] <- head(alignment[[loci[i]]],-2)
            alignment[[loci[i]]] <- tail(alignment[[loci[i]]],-6)
          }

          if(loci[[i]] %in%  c("DPA2","DPB2","Y")) { # In release version 3.53.0 and earlier these pseudogenes had 'AA codon' positions in their cDNA alignments, but in 3.54.0 those rows have been removed; similarly the HLA-Y pseudogene had an AA codon line prior to version 3.36.0
            alignment[[loci[i]]] <- alignment[[loci[[i]]]][!substr(alignment[[loci[[i]]]][],1,9) == " AA codon"]
          }

          if(version == 3480 && source[j] == "gDNA" && loci[i] == "DRB1"){ ## Fix for DRB1*15:200:01:01N and DRB1*15:200:01:02N in 3.48.0 gDNA alignment
            alignment[[loci[i]]] <- gsub("DRB1*15:200:01:01N","DRB1*15:200:01:01N ",alignment[[loci[i]]], fixed=TRUE)
            alignment[[loci[i]]] <- gsub("DRB1*15:200:01:02N","DRB1*15:200:01:02N ",alignment[[loci[i]]], fixed=TRUE)
          }


          # See countSpaces function (AA table only)
          # Counts difference between Prot to -30 and beginning of Prot to -30 + 1 due to zero number indexing to find where
          # the alignment sequence actually starts
          if(source[j]=="AA"){space_diff[[loci[i]]]<-(nchar(strsplit(alignment[[loci[i]]][3], " ")[[1]][2])+countSpaces(alignment[[loci[i]]][3])[2]+1)-countSpaces(alignment[[loci[i]]][2])[1]}

          # reduce repeated whitespace in alignment file and removes rows with empty values for proper
          # start and stop subsetting
          alignment[[loci[i]]] <-str_squish(alignment[[loci[i]]])
          alignment[[loci[i]]] <-alignment[[loci[i]]][-which(alignment[[loci[i]]] == "")]

          # determine positions of "cDNA" or "Prot" and the end of that reference block segment
          start[[loci[i]]] <-as.numeric(grep(type, alignment[[loci[i]]]))
          end[[loci[i]]] <- as.numeric(c(start[[loci[i]]][2:length(start[[loci[i]]])]-1,length(alignment[[loci[i]]])))

          if(source[j]=="AA"){
            # extract rows with "Prot" and reference sequence position information
            # extract only relevant reference sequence positions
            # NOTE: the first row containing "Prot" contains two numbers -- -30 and 1 -- where only -30, is extracted,
            # as the actual sequence start will always be 1
            for (k in 1:length(start[[loci[i]]])){

              prot_extractions[[loci[i]]][k]<-strsplit(alignment[[loci[i]]][start[[loci[i]]][k]], " ")

              refblock_number[[loci[i]]][k]<-as.numeric(sapply(prot_extractions[[loci[i]]][k], "[", 2))


              # determine the alignment start by adding -30 to the difference between white spaces found above
              alignment_start[[loci[i]]]<-refblock_number[[loci[i]]][1]+space_diff[[loci[i]]]
            }
          } else if(source[j]=="cDNA"){
            # determine the alignment start by finding the second vector in second list and removing "codon"
            if(expressed[[i]]) { # exclude loci with cDNA alignment but no AA alignment (e.g., not expressed)
              alignment_start[[loci[i]]] <-as.numeric(sub("AA codon ", "", alignment[[loci[i]]][2]))
            }
            if(is.null(alignment_start[[loci[i]]])) {
              alignment_start[[loci[i]]] <- 1
            }
            DNA_start[[loci[i]]] <-as.numeric(sub("cDNA ", "", alignment[[loci[i]]][1]))
          } else if(source[j]=="gDNA"){
            DNA_start[[loci[i]]] <-as.numeric(sub("gDNA ", "", alignment[[loci[i]]][1]))
          }

          # close all white space in the alignment file, except for the white space separating the allele and peptide sequence
          alignment[[loci[i]]] <-paste(substr(alignment[[loci[i]]],1,regexpr(" ",text = alignment[[loci[i]]],fixed = TRUE)), gsub(" ","",substr(alignment[[loci[i]]],regexpr(" ",text = alignment[[loci[i]]],fixed = TRUE),nchar(alignment[[loci[i]]]))),sep = "")

          # string split at white spaces to yield allele and peptide sequences
          alignment[[loci[i]]]  <- strsplit(alignment[[loci[i]]]," ", fixed=T)

          # bind the previously split strings by row
          alignment[[loci[i]]] <- do.call(rbind,alignment[[loci[i]]])

          # if the seq column is equal to the allele column due to premature peptide termination,
          # insert a blank in place of the allele in the seq column (AA table only)
          if(source[j]=="AA"){alignment[[loci[i]]][which(alignment[[loci[i]]][,1]==alignment[[loci[i]]][,2]),2] <- ""} ##### I think this can be eliminated, since it is not in the AA block

          #renames columns to "alleles" and "seq"
          colnames(alignment[[loci[i]]])<-c(paste(loci[[i]], "alleles", sep="_"), "seq")

          # Due to ANHIG formatting, cases where an allele contains newly reference peptide sequences will not
          # contain the same number of rows as previous reference peptide blocks
          # this for loop is invoked to add "." for all other alleles for each character in the newly reference peptide
          # to preserve structural integrity
          # changes (10/9/19) to accommodate if there is more than one extraneous allele with an extended amino acid sequence

          # get last start and end positions to evaluate if extraneous alleles are present
          last_start <- start[[loci[[i]]]][length(start[[loci[[i]]]])]
          last_end <- end[[loci[[i]]]][length(end[[loci[[i]]]])]

          if(nrow(alignment[[loci[i]]][last_start:last_end,])!=nrow(alignment[[loci[i]]][start[[loci[i]]][1]:end[[loci[i]]][1],])){
            x<-as.data.frame(alignment[[loci[i]]][,1][start[[loci[i]]][1]:end[[loci[i]]][1]][-c(1,2)], stringsAsFactors = F)
            colnames(x)<-paste(loci[i], "alleles", sep="_")
            temp_vec<-alignment[[loci[i]]][last_start:last_end,][-c(1,2),]

            #if there is only one allele with an extended amino acid sequence, the conversion from a named
            #vector to a data frame is not properly
            #if the character vector is = 2, there is only one allele with an extended amino acid sequence, so use
            #as.list and then data.frame so data frame is created correctly
            if(length(temp_vec) == 2){
              temp_filter<-data.frame(as.list(temp_vec), stringsAsFactors = F)
            } else{
              temp_filter<-data.frame(temp_vec, stringsAsFactors = F)
            }

            #find the greatest number of peptides in extraneous alleles to determine
            #how many "." to add on
            max_nchar<-(temp_filter %>%
                          add_column(nchar = nchar(.$seq)) %>%
                          filter(nchar == max(nchar)))$nchar
            x<-cbind.data.frame(x, seq=as.character(paste(rep(".", max_nchar[1]), collapse = "")), stringsAsFactors=FALSE)
            y<-data.frame(tail(alignment[[loci[i]]], (nrow(alignment[[loci[i]]][last_start:last_end,][nrow(alignment[[loci[i]]][last_start:last_end,])!=nrow(alignment[[loci[i]]][start[[loci[i]]][1]:end[[loci[i]]][1],]),])-2)), stringsAsFactors = F)
            x$seq[match(y[,1], x[,1])]<-y$seq
            alignment[[loci[i]]]<-as.matrix(rbind(head(alignment[[loci[i]]], -(nrow(alignment[[loci[i]]][last_start:last_end,][nrow(alignment[[loci[i]]][last_start:last_end,])!=nrow(alignment[[loci[i]]][start[[loci[i]]][1]:end[[loci[i]]][1],]),])-2)), x))
            start[[loci[i]]]<-as.numeric(grep(type, alignment[[loci[i]]]))
            end[[loci[i]]] <- as.numeric(c(start[[loci[i]]][2:length(start[[loci[i]]])]-1,nrow(alignment[[loci[i]]])))
          }


          # if(source[j]=="cDNA"|source[j]=="AA"&loci[i]=="TAP2"){ ## the logic in the first expression is odd, but it works
          #   for(l in 1:length(start[[loci[i]]])){
          #     if(nrow(alignment[[i]][start[[loci[i]]][l]:end[[loci[i]]][l],])!=nrow(alignment[[loci[i]]][start[[loci[i]]][1]:end[[loci[i]]][1],])){
          #       x<-as.data.frame(alignment[[loci[i]]][,1][start[[loci[i]]][1]:end[[loci[i]]][1]][-c(1,2)], stringsAsFactors = F)
          #       colnames(x)<-paste(loci[[i]], "alleles", sep="_")
          #       x<-cbind.data.frame(x, seq=as.character(paste(rep(".", nchar(tail(alignment[[loci[i]]][,2], 1))), collapse = "")), stringsAsFactors=FALSE)
          #       y<-data.frame(tail(alignment[[loci[i]]], (nrow(alignment[[i]][start[[loci[i]]][l]:end[[loci[i]]][l],][nrow(alignment[[i]][start[[loci[i]]][l]:end[[loci[i]]][l],])!=nrow(alignment[[loci[i]]][start[[loci[i]]][1]:end[[loci[i]]][1],]),])-2)), stringsAsFactors = F)
          #       x$seq[match(y[,1], x[,1])]<-y$seq
          #       alignment[[loci[i]]]<-as.matrix(rbind(head(alignment[[loci[i]]], -(nrow(alignment[[i]][start[[loci[i]]][l]:end[[loci[i]]][l],][nrow(alignment[[i]][start[[loci[i]]][l]:end[[loci[i]]][l],])!=nrow(alignment[[loci[i]]][start[[loci[i]]][1]:end[[loci[i]]][1],]),])-2)), x))
          #       start[[loci[i]]]<-as.numeric(grep(type, alignment[[loci[i]]]))
          #       end[[loci[i]]] <- as.numeric(c(start[[loci[i]]][2:length(start[[loci[i]]])]-1,nrow(alignment[[loci[i]]])))}
          #   }
          # }
          # # Like HLA-C, TAP2 has some alleles with extra nuc and prot 'tails' at the end of the alignment page.
          # if(source[j]=="AA"|source[j]=="gDNA"|source[j]!="AA"&loci[i]!="TAP2"){
          #   for(l in 1:length(start[[loci[i]]])){
          #     if(nrow(alignment[[i]][start[[loci[i]]][l]:end[[loci[i]]][l],])!=nrow(alignment[[loci[i]]][start[[loci[i]]][1]:end[[loci[i]]][1],])){ ####<<<< for HLA-C this is broken
          #       x<-as.data.frame(alignment[[loci[i]]][,1][start[[loci[i]]][1]:end[[loci[i]]][1]][-c(1,2)], stringsAsFactors = F)
          #       colnames(x)<-paste(loci[[i]], "alleles", sep="_")
          #       temp_vec<-alignment[[i]][start[[loci[i]]][l]:end[[loci[i]]][l],][-c(1,2),]
          #       cat(l,"\n")
          #       # If there is only one allele with an extended amino acid sequence, the conversion from a named
          #       # vector to a data frame is not properly done
          #       # If the character vector is = 2, there is only one allele with an extended amino acid sequence, so use
          #       # as.list and then data.frame so data frame is created correctly
          #       if(length(temp_vec) == 2){
          #         temp_filter<-data.frame(as.list(temp_vec), stringsAsFactors = F)
          #       } else{
          #         temp_filter<-data.frame(temp_vec, stringsAsFactors = F)
          #       }
          #
          #       # find the greatest number of peptides in extraneous alleles to determine
          #       # how many "." to add on
          #       max_nchar<-(temp_filter %>%
          #                     add_column(nchar = nchar(.$seq)) %>%
          #                     filter(nchar == max(nchar)))$nchar
          #       x<-cbind.data.frame(x, seq=as.character(paste(rep(".", max_nchar), collapse = "")), stringsAsFactors=FALSE)
          #       y<-data.frame(tail(alignment[[loci[i]]], (nrow(alignment[[i]][start[[loci[i]]][l]:end[[loci[i]]][l],][nrow(alignment[[i]][start[[loci[i]]][l]:end[[loci[i]]][l],])!=nrow(alignment[[loci[i]]][start[[loci[i]]][1]:end[[loci[i]]][1],]),])-2)), stringsAsFactors = F)
          #       x$seq[match(y[,1], x[,1])]<-y$seq
          #       alignment[[loci[i]]]<-as.matrix(rbind(head(alignment[[loci[i]]], -(nrow(alignment[[i]][start[[loci[i]]][l]:end[[loci[i]]][l],][nrow(alignment[[i]][start[[loci[i]]][l]:end[[loci[i]]][l],])!=nrow(alignment[[loci[i]]][start[[loci[i]]][1]:end[[loci[i]]][1],]),])-2)), x))
          #       start[[loci[i]]]<-as.numeric(grep("Prot", alignment[[loci[i]]]))
          #       end[[loci[i]]] <- as.numeric(c(start[[loci[i]]][2:length(start[[loci[i]]])]-1,nrow(alignment[[loci[i]]])))
          #     }
          #   }
          # } ## end of not-TAP2 block

          # if a locus has extra formatting, resulting in unqequal rows, start and end will be updated to reflect subsetting
          # if a locus has no extra formatting, start and end will remain the same, as procured by earlier code
          for(m in 1:length(start[[loci[i]]])){
            HLAalignments[[loci[i]]]<-cbind(HLAalignments[[loci[i]]], alignment[[loci[i]]][start[[loci[i]]][m]:end[[loci[i]]][m],])
          }

          # remove rows containing "cDNA", "AA codon", or "Prot"
          HLAalignments[[loci[i]]] <- HLAalignments[[loci[i]]][-delete_lines,] ## this was the pseudo/frag problem; With no AAcodon row, it takes one too many lines out. Now corrected.

          # designate columns to be combined as every other so allele names are not included
          # in pasting all the amino acid sequences together
          cols<-seq(0, ncol(HLAalignments[[loci[i]]]), by=2)
          HLAalignments[[loci[i]]]<-cbind(HLAalignments[[loci[i]]][,1], apply(HLAalignments[[loci[i]]][,cols], 1 ,paste, collapse = "")) # The full-length alignment, with 1 row per allele

          # create a new matrix with the number of columns equal to the number of characters in the reference sequence
          corr_table[[loci[i]]]<-matrix(, nrow = 3, ncol = as.numeric(nchar(HLAalignments[[loci[i]]][,2][1])))

          if(source[j]=="AA"|source[j]=="cDNA"){  # closes at line 384(381)
            # if the first position enumeration is negative (i.e. has a leader peptide sequence), determines alignment length based on the total number of characters plus the alignment start (which is negative)
            if(grepl("-", alignment_start[[loci[i]]][[1]])==TRUE){
              alignment_length[[loci[i]]]<-(as.numeric(nchar(HLAalignments[[loci[i]]][,2][1]))/divide)+(alignment_start[[loci[[i]]]])
            } else { # if there is no leader peptide (i.e sequence starts at 1), determines alignment length based on total number of characters
              alignment_length[[loci[i]]]<-as.numeric(nchar(HLAalignments[[loci[i]]][,2][1]))
            }
            if(expressed[[i]]){   # pertinent to cDNA alignments with expressed proteins (codons are nucleotide triplets)
              # paste alignment_start to alignment_length together in sequential order
              # capture output as "w"
              w[[loci[i]]] <- capture.output(cat(alignment_start[[loci[i]]]:alignment_length[[loci[i]]]))

              # split string formed by cat for separate character variables
              alignment_positions[[loci[i]]]<-as.character(unlist(strsplit(w[[loci[i]]], " ")))

              # eliminate "0" if present in the alignment positions, as the alignment sequence from ANHIG does not contain 0
              if("0" %in% alignment_positions[[loci[i]]]==TRUE){
                alignment_positions[[loci[i]]]<-alignment_positions[[loci[i]]][-which(alignment_positions[[loci[i]]] == 0)]
              }
            } # closing the if not a pseudogene or fragment (i.e., expressed)
          } # closing AA/cDNA

          if(expressed[[i]]) {
            # triple alignment to account for codons (cDNA table only) # Does not pertain to pseudogenes and fragments
            if(source[j]=="cDNA"){
              for(n in 1:length(alignment_positions[[loci[i]]])){
                alignment_positionsx3[[loci[i]]]<-c(alignment_positionsx3[[loci[i]]],alignment_positions[[loci[i]]][n],alignment_positions[[loci[i]]][n],alignment_positions[[loci[i]]][n])
              }
              alignment_positions[[loci[i]]]<-alignment_positionsx3[[loci[i]]]
            }

          } ## closing the expressed block

          # string splits to extract locus in the allele name
          # assign to new variable "aligned"
          aligned[[loci[i]]]<- as.matrix(do.call(rbind,strsplit(HLAalignments[[loci[i]]][,1],"[*]")))

          # add a new column of pasted locus and trimmed two field alleles to aligned
          aligned[[loci[i]]]<- cbind(aligned[[loci[i]]], paste(aligned[[loci[i]]][,1], apply(aligned[[loci[i]]],MARGIN=c(1,2),FUN=getField,res=2)[,2], sep="*"))

          # bind aligned and HLAalignments -- rename columns
          HLAalignments[[loci[i]]] <- cbind(aligned[[loci[i]]], HLAalignments[[loci[i]]])
          colnames(HLAalignments[[loci[i]]]) <- c("locus", "full_allele", "trimmed_allele", "allele_name", sequence_name)

          if(source[j]=="AA"){
            # if locus is DRB3/4/5, use DRB line 1 as reference sequence
            if(loci[[i]]=="DRB3"|loci[[i]]=="DRB4"|loci[[i]]=="DRB5"){
              # set refexon to a reference peptide for each HLA locus based on the reference sequences in HLAalignments
              refexon[[loci[i]]] <- rbind(HLAalignments[[loci[i]]][1,])[which(rbind(HLAalignments[[loci[i]]][1,])[,"locus"]=="DRB1"),sequence_name]
            } else{
              refexon[[loci[i]]] <- rbind(HLAalignments[[loci[i]]][1,])[which(rbind(HLAalignments[[loci[i]]][1,])[,"locus"]==loci[[i]]),sequence_name]
            }

            # if input loci is DRB, use grep statement to match input loci to loci in HLAalignments
            if(loci[[i]]=="DRB"){
              refexon[[loci[i]]]<-rbind(HLAalignments[[loci[i]]][1,])[grepl(loci[[i]], rbind(HLAalignments[[loci[i]]][1,])[,"locus"]), sequence_name]}
          } # close the AA specific block

          # split sequence column at every amino acid or nucleotide, resulting in a split list of each for each row
          pepsplit[[loci[i]]] <- sapply(HLAalignments[[loci[i]]][,sequence_name],strsplit,split="*")

          # fill in space with NA for alleles with premature termination to make it the same number of characters
          # as the reference sequence (AA table only)
          if(source[j]=="AA"){pepsplit[[loci[i]]]<- lapply(pepsplit[[loci[i]]],function(x) c(x,rep("NA",nchar(refexon[[loci[i]]])-length(x))))}

          # nullify variable names
          names(pepsplit[[loci[i]]]) <- NULL

          # bind pep_split together by element in its previous list form by row
          suppressWarnings(pepsplit[[loci[i]]]<- do.call(rbind,pepsplit[[loci[i]]])) ## this results in a warning when the appendix alignments are of different lengths

          # bind all columns together to form desired output, as described above
          HLAalignments[[loci[i]]] <- cbind.data.frame(HLAalignments[[loci[i]]][,1:4],pepsplit[[loci[i]]], stringsAsFactors=FALSE)

          if(source[j]=="AA"|source[j]=="cDNA"){
            # get reference sequence from DRB alignment
            if(loci[[i]] %in% c("DRB3","DRB4","DRB5")){
              DRBref<-HLAalignments[[loci[[i]]]][1,] #}

              #if the locus is DRB1/3/4/5, subset the specific locus from DRB alignment
              HLAalignments[[loci[i]]]<-HLAalignments[[loci[i]]][HLAalignments[[loci[[i]]]]$locus==loci[[i]],] #}

              # add reference sequence to DRB3/4/5, reset row names to numerical order
              HLAalignments[[loci[[i]]]]<- rbind(DRBref, HLAalignments[[loci[[i]]]])
              rownames(HLAalignments[[loci[[i]]]])<-NULL}
          } ## close DRB/DRB1 specific AA/cDNA block

          # input HLAalignments alignment sequence into the corr_table with "InDel" still present
          corr_table[[loci[[i]]]][1,]<-names(HLAalignments[[loci[[i]]]][5:ncol(HLAalignments[[loci[[i]]]])])

          # find positions in HLAalignments that have ".", indicating an inDel
          inDels[[loci[[i]]]]<-colnames(HLAalignments[[loci[[i]]]][1, 5:ncol(HLAalignments[[loci[[i]]]])][HLAalignments[[loci[[i]]]][1, 5:ncol(HLAalignments[[loci[[i]]]])] %in% "."])

          # find positions in HLAalignments that have "|", indicating a exon boundary
          exonB[[loci[[i]]]]<-colnames(HLAalignments[[loci[[i]]]][1, 5:ncol(HLAalignments[[loci[[i]]]])][HLAalignments[[loci[[i]]]][1, 5:ncol(HLAalignments[[loci[[i]]]])] %in% "|"])

          # inDel inclusion if there are inDels present
          if(length(inDels[[loci[[i]]]])!=0){
            for(o in 1:length(inDels[[loci[[i]]]])){
              corr_table[[loci[[i]]]][2,][inDels[[loci[[i]]]][[o]]==corr_table[[loci[[i]]]][1,]]<-paste("INDEL", o, sep="-")
              corr_table[[loci[[i]]]][3,][inDels[[loci[[i]]]][[o]]==corr_table[[loci[[i]]]][1,]]<-paste("INDEL", o, sep="-")
            }
          } # close indel block

          # exonB inclusion if there are exon boundaries present
          if(length(exonB[[loci[[i]]]])!=0){
            for(p in 1:length(exonB[[loci[[i]]]])){
              corr_table[[loci[[i]]]][2,][exonB[[loci[[i]]]][[p]]==corr_table[[loci[[i]]]][1,]]<-paste("EXONB", p, sep="-")
              corr_table[[loci[[i]]]][3,][exonB[[loci[[i]]]][[p]]==corr_table[[loci[[i]]]][1,]]<-paste("EXONB", p, sep="-")
            }
          }

          if(expressed[[i]]) {
            if(source[j]=="AA"|source[j]=="cDNA"){
              # paste alignment_positions into corr_table accounting for InDels and exon boundaries
              codon_count<-as.vector(1)
              for(q in 1:length(corr_table[[loci[[i]]]][2,])){
                if(is.na(corr_table[[loci[[i]]]][2,q])){
                  corr_table[[loci[[i]]]][2,q]<-alignment_positions[[loci[i]]][codon_count]
                  codon_count<-codon_count+1
                }
              }
            }
          }

          # create number sequence for alignment (cDNA and gDNA tables only)
          if(source[j]=="cDNA"){ alignment_positions[[loci[i]]]<-as.character(unlist(strsplit(capture.output(cat(DNA_start[[loci[i]]]:length(corr_table[[loci[[i]]]][3,]))), " ")))
          } else if(source[j]=="gDNA"){ if(DNA_start[[loci[i]]] < 0) {alignment_positions[[loci[i]]]<-as.character(unlist(strsplit(capture.output(cat(c(DNA_start[[loci[i]]]:-1,1:length(corr_table[[loci[[i]]]][3,])))), " "))) ## most loci have negative start positions
          } else {alignment_positions[[loci[i]]]<-as.character(unlist(strsplit(capture.output(cat(c(DNA_start[[loci[i]]]:length(corr_table[[loci[[i]]]][3,])))), " ")))  #### HLA-P starts at position 1, but in theory some other locus could start at a different positive position
          }
          } # In 3.53.0, only the HLA-P alignment started with 1, so, using the original version above, the first four alignment_positions values are "1"   "0"   "-1"  "1"
          # pastes number sequence into corr_table accounting for InDels and exon boundaries (cDNA and gDNA tables only)
          if(source[j]=="cDNA"|source[j]=="gDNA"){ # cDNA/gDNA correspondence table block
            cDNAcount<-as.vector(1)
            for(r in 1:length(corr_table[[loci[[i]]]][3,])){
              if(is.na(corr_table[[loci[[i]]]][3,r])){
                corr_table[[loci[[i]]]][3,r]<-alignment_positions[[loci[i]]][cDNAcount]
                cDNAcount<-cDNAcount+1
              }
            }
          } # Close cDNA/gDNA block

          # run if InDels or exon boundaries are present in alignment
          if(any(grepl("INDEL|EXONB", corr_table[[loci[[i]]]][2,]))){ ## if there are no boundaries or indels, this never gets run so N and U cDNA don't get their corr_tables made
            # find which positions have InDels or exon boundaries
            v<-as.numeric(corr_table[[loci[[i]]]][1,][which(grepl("INDEL", corr_table[[loci[[i]]]][2,]))])
            w<-as.numeric(corr_table[[loci[[i]]]][1,][which(grepl("EXONB", corr_table[[loci[[i]]]][2,]))])

            # split cumulative sequences
            vsplit<-split(v, cumsum(c(TRUE, diff(v) != 1L))) # absolute corr_table coordinates (row 1)
            wsplit<-split(w, cumsum(c(TRUE, diff(w) != 1L))) # absolute corr_table coordinates (row 1)

            if(any(grepl("INDEL", corr_table[[loci[[i]]]][2,]))){ # indel block
              # add increments of .1 to any InDels in row 2
              if(source[j]=="AA"|source[j]=="cDNA"){ # AA/cDNA indel block
                for(s in 1:length(vsplit)){
                  if(grepl("-", corr_table[[loci[[i]]]][2,][vsplit[[s]][[1]]-1])) {
                    corr_table[[loci[[i]]]][2,][vsplit[[s]]]<-paste(corr_table[[loci[[i]]]][2,][vsplit[[s]][[1]]-1], gsub(0, "", seq(1, length(vsplit[[s]]))/10), sep="")
                  } else {
                    corr_table[[loci[[i]]]][2,][vsplit[[s]]]<-as.numeric(corr_table[[loci[[i]]]][2,][vsplit[[s]][[1]]-1])+as.numeric(paste(0,".",seq(1, length(vsplit[[s]])),sep = ""))
                  }
                }
              } # close AA/cDNA indel block

              # add increments of .1 to any InDels in row 3
              if(source[j]=="cDNA"|source[j]=="gDNA"){ # cDNA/gDNA block
                for(t in 1:length(vsplit)){
                  corr_table[[loci[[i]]]][3,][vsplit[[t]]]<-paste(corr_table[[loci[[i]]]][3,][vsplit[[t]][[1]]-1],".",seq(1, length(vsplit[[t]])),sep = "")
                }
              } # close cDNA/gDNAA block
            } # close indel block

            if(any(grepl("EXONB", corr_table[[loci[[i]]]][2,]))){
              # change EXONB to E.(# of last exon)-(# of next exon)
              if(source[j]=="cDNA"){  # make Exon -> E.
                for(u in 1:length(wsplit)){
                  corr_table[[loci[[i]]]][2,][wsplit[[u]]]<-paste("E.",u,"-",u+1, sep = "")
                  corr_table[[loci[[i]]]][3,][wsplit[[u]]]<-paste("E.",u,"-",u+1, sep = "")
                }
              } # close Exon -> E.

              boundaryIndel <- w[which((w+1) %in% v)] # Identify the position of the ExonBoundary immediately after which an indel occurs
              if(length(boundaryIndel) != 0) {        # These are special procedures for HLA-C and -DQB1 that should be generalizable
                insertion <- FALSE
                toReplace <- toTarget <- NA
                for(ff in 1:length(boundaryIndel)) { # boundary Indel Loop
                  if(length(boundaryIndel[ff]) != 0) { # then we have an indel immediately following a boundary
                    vsplitSub <- NA
                    for(o in 1:length(vsplit)) {
                      if((boundaryIndel[ff]+1) %in% vsplit[[o]]) {vsplitSub <- o}
                      if(!is.na(vsplitSub)) {
                        toReplace <- vsplit[vsplitSub][[as.character(o)]][length(vsplit[vsplitSub][[as.character(o)]])]+1
                        if(length(grep("E",corr_table[[loci[[i]]]][2,toReplace])) != 0) {insertion <- TRUE}
                        toTarget <- vsplit[vsplitSub][[as.character(o)]][1]
                        vsplitSub <- NA
                      }
                    }
                    #             } # close length(boundaryIndel) loop --- THIS NEEDS TO CLOSE AFTER THE BELOW
                    if(insertion) { # Currently pertains only to DQB1 as of 3.54.0
                      corr_table[[loci[[i]]]][2,toTarget] <- corr_table[[loci[[i]]]][3,toTarget] <- "Ins"
                    } else {
                      corr_table[[loci[[i]]]][2,toTarget] <- corr_table[[loci[[i]]]][2,toReplace]
                      corr_table[[loci[[i]]]][3,toTarget] <- corr_table[[loci[[i]]]][3,toReplace]
                    }
                  } # close length(boundaryIndel[ff]) loop
                } # close boundaryIndel loop
              } # close boundaryIndel > 0 check

              # change EXONB to corresponding boundary e.g. UTR-E.1
              # create atlas of gDNA boundaries and their locus
              if(source[j]=="gDNA"){ # start of the gDNA construction
                if(loci[[i]] %in% names(fragmentFeatureNames)) { # simplified gDNA atlas builder for genes in the fragemtnFeatureNames object

                  gDNA_atlas[[loci[[i]]]]<-matrix(, nrow = 1, ncol = length(wsplit))
                  gDNA_atlas[[loci[[i]]]]<- as.numeric(corr_table[[i]][3,w+1])
                  newNames <- rep(NA,length(fragmentFeatureNames[names(fragmentFeatureNames) %in% loci[[i]] == TRUE][[1]][[1]])-1)
                  for(f in 1:(length(newNames))) {
                    newNames[f] <- paste(fragmentFeatureNames[names(fragmentFeatureNames) %in% loci[[i]] == TRUE][[1]][[1]][f],"-",fragmentFeatureNames[names(fragmentFeatureNames) %in% loci[[i]] == TRUE][[1]][[1]][f+1], sep="")
                  }
                  gDNA_atlas[[loci[[i]]]] <- as.data.frame(t(gDNA_atlas[[loci[[i]]]]))
                  colnames(gDNA_atlas[[loci[[i]]]]) <- newNames
                  rownames(gDNA_atlas[[loci[[i]]]]) <- "gDNA"

                } else { intron<-as.vector(paste("I.",seq(1,((length(wsplit)-1)/2)-0.5),sep = "")) # } # start of the 'expressed gene' notation section
                exon<-as.vector(paste("E.",seq(1,((length(wsplit)-1)/2)+0.5),sep = ""))
                a<-unlist(strsplit(paste0(exon," ",intron), " "))
                a<-c("U.5",head(a, length(a)-1),"U.3")
                b<-vector()
                for(y in 1:(length(a)-1)){
                  b<-c(b,paste(a[[y]],"-",a[[y+1]],sep = ""))}
                gDNA_atlas[[loci[[i]]]]<-matrix(, nrow = 1, ncol = length(wsplit))
                # cover 3 loci
                if(length(b) != length(gDNA_atlas[[loci[[i]]]])) {
                  gDNA_atlas[[loci[[i]]]]<-matrix(, nrow = 1, ncol = length(b))
                }
                colnames(gDNA_atlas[[loci[[i]]]])<-b
                rownames(gDNA_atlas[[loci[[i]]]])<-"gDNA"
                for(z in 1:length(wsplit)){
                  gDNA_atlas[[loci[[i]]]][1,z]<-corr_table[[loci[[i]]]][3,][wsplit[[z]]+1]
                  corr_table[[loci[[i]]]][3,][wsplit[[z]]]<-b[[z]]
                }
                } # End of expressed gene construction
              } # End of the gDNA construction routine

              # create atlas of cDNA boundaries and their locus
              if(source[j]=="cDNA"){

                # if there are no boundaries, make a 1 column atlas
                if(length(exonB[[loci[[i]]]]) == 0) { # if exonB[[loci[[i]]]] has a length of 0, then it has no boundaries, and this alignment represents one feature (HLA-N and -U only as of 3.53.0).

                  cDNA_atlas[[loci[[i]]]]<-matrix(, nrow = 1, ncol = 1)
                  colnames(cDNA_atlas[[loci[i]]]) <- "No.Exon.Boundaries"
                  rownames(cDNA_atlas[[loci[[i]]]])<-c("cDNA")
                  cDNA_atlas[[loci[[i]]]][1,1]<-corr_table[[loci[[i]]]][2,1]

                } else {

                  cDNA_atlas[[loci[[i]]]]<-matrix(, nrow = 2, ncol = length(wsplit))
                  colnames(cDNA_atlas[[loci[[i]]]])<-corr_table[[loci[[i]]]][2,][unlist(wsplit)]
                  rownames(cDNA_atlas[[loci[[i]]]])<-c("cDNA","codon")
                  for(z in 1:length(wsplit)){
                    cDNA_atlas[[loci[[i]]]][1,z]<-corr_table[[loci[[i]]]][3,][wsplit[[z]]+1]
                    cDNA_atlas[[loci[[i]]]][2,z]<-corr_table[[loci[[i]]]][2,][wsplit[[z]]+1]
                  }
                } # closes if length exon B = 0's else
              } # closes if(source[j] == "cDNA")
            } # closes the if(any(grepl("EXONB" block
          } # closes the if any INDEL|EXONB block

          # Create a cDNA Atlas for N and U (and any other cDNA that doesn't have boundaries)
          if((source[j] == "cDNA") & (length(exonB[[loci[[i]]]]) == 0)) {   # if exonB[[loci[[i]]]] has a length of 0, then it has no boundaries, and this alignment represents one feature (HLA-N and -U only as of 3.53.0).

            cDNA_atlas[[loci[[i]]]]<-matrix(, nrow = 1, ncol = 1)
            colnames(cDNA_atlas[[loci[i]]]) <- "No.Exon.Boundaries"
            rownames(cDNA_atlas[[loci[[i]]]])<-c("cDNA")
            cDNA_atlas[[loci[[i]]]][1,1]<-corr_table[[loci[[i]]]][2,1]

          }
          # pull the AA atlas from the cDNA atlas
          if(subSource[[j]] == "AA") { #return things back to the way they should be after jumping the AA atlas routine
            source[[j]] <- "AA" ## restore the proper source
            AA_atlas[[loci[[i]]]] <- as.data.frame(cDNA_atlas[[loci[[i]]]])[2,]
            rownames(AA_atlas[[loci[i]]]) <- c("AA")
            atlas[[loci[i]]] <-c(atlas[[loci[i]]], AA = list(AA_atlas[[loci[i]]]),`Version` = alignmentVersion[[loci[[1]]]])
            subSource[[j]] <- 0 ## clear the submitted source
          }

          if(source[j]=="cDNA"){ # With no boundaries, the Atlas is empty
            atlas[[loci[i]]] <-c(atlas[[loci[i]]], cDNA = list(cDNA_atlas[[loci[i]]]),`Version` = alignmentVersion[[loci[[i]]]])
          }else if(source[j]=="gDNA"){
            atlas[[loci[i]]] <-c(atlas[[loci[i]]], gDNA = list(gDNA_atlas[[loci[i]]]),`Version` = alignmentVersion[[loci[[i]]]])
          } # close cDNA & gDNA creation code

        }

        if(source[j]=="cDNA" && loci[i] %in% c("R","S","T","V","W")) { ## post-hoc update of cDNA feature names for fragments and pseudogenes

          repLen <- length(colnames(atlas[[loci[i]]]$cDNA))
          featNames <- fragmentFeatureNames[[names(fragmentFeatureNames)[names(fragmentFeatureNames) %in% loci[i]]]]$feature
          featRep <- featNames[which(1:length(featNames)%%2==0)]
          newPairs <- rep(NA,length(featRep)-1)

          for(g in 1:length(newPairs)){
            newPairs[g] <- paste(featRep[g],featRep[g+1],sep="-")
          }

          colnames(atlas[[loci[i]]]$cDNA) <- newPairs
        }

      } # end of the loop if j source values
    } # end of main loop of i length values

    for(k in 1:length(loci)){
      if("AA" %in% names(atlas[[loci[k]]])) { atlas[[loci[k]]]$AA <- as.data.frame(atlas[[loci[k]]]$AA)   }
      if("cDNA" %in% names(atlas[[loci[k]]])) { atlas[[loci[k]]]$cDNA <- as.data.frame(atlas[[loci[k]]]$cDNA)   }
      if("gDNA" %in% names(atlas[[loci[k]]])) { atlas[[loci[k]]]$gDNA <- as.data.frame(atlas[[loci[k]]]$gDNA)   }
    }

    return(atlas)
  } else { return("The 'locus' and/or 'source' parameters were invalid.") } ## end of the LocusValidator check
}

##Accessory Functions v5.1.1 30APR2025

################
##posSort
#'Numerical Sort of Alignment Positions that Contain Indels
#'
#'@description
#'Sorts sequence alignment positions that contain indels in numerical order; e.g., three indels following position X are are identified as X.1, X.2 and X.3.
#'
#'Positions are validated against a specified locus for a specified alignment type. Positions that do not exist for that locus-alignment combination are not returned.
#'
#'@param posVec A character vector of variant positions.
#'@param alignType A character string describing the type of alignment the positions are found in. The values 'prot', 'codon', 'nuc' and 'gen' are valid options. Only one 'alignType' value is allowed.
#'@param locus A character string describing a locus in the HLAalignments object for the specified alignType.
#'
#'@return A character string of the correctly sorted sequence.
#'
#'@note Indel positions must be text-formatted (e.g. "607.12"). C.f., posSort(c(2,4,3,1,5), "nuc","DRB1") vs posSort(c("607.23","607.10","607.3","607.4"),"nuc","DRB1").
#'
#'@export
#'
posSort <- function(posVec,alignType, locus){
  
    alignType <- checkAlignType(alignType)  
    
    if(length(alignType)!=1) {stop(paste("Please specify only one 'alignType'."))}
    
    if(!locus %in% names(HLAalignments[[alignType]])) {stop(paste(locus,"is not included among the",alignType,"alignments.\n",sep=" "))}
  
  tab <- as.data.frame(cbind(match(posVec,names(HLAalignments[[alignType]][[locus]])),posVec))
  tab$V1 <- as.numeric(tab$V1)
  tab <- tab[order(tab$V1),]
  tab$posVec[!is.na(tab$V1)]
}

################
##numFields
#'Identify the Number of Fields in a Colon-Delimited Allele Name
#'
#'Returns the number of fields in a colon-delimited HLA allele name. A value of 1 is returned for digit-delimited HLA allele names.
#'
#'@param allele A character string of a colon-delimited HLA allele name.
#'
#'@return A numeric value describing the number of fields in the allele name.
#'
#'@export
#'
#'@examples
#'numFields("HLA-A*01:01")
#'numFields("DRB1*04:03:01")
#'numFields("13:02:01:01")
numFields <- function(allele) {
  length(strsplit(allele,":",fixed=TRUE)[[1]])
}

#################
##validateAllele
#'Validate Allele-Name Format and Presence in HLAalignments
#'
#'Returns TRUE if an allele name is found in HLAalignments in either the 'allele_name' column of full-length allele names or the 'trimmed_allele' column of two-field allele names in the pertinent genomic alignment. Returns FALSE if the allele name is not properly formed, or if the allele name is not found in HLAalignments.
#'
#'@param allele A character string of the colon-delimited HLA allele name.
#'
#'@return A logical identifying if the allele name is present in the alignments (TRUE) or, if it is not in the alignments or is not valid (FALSE).
#'
#'@note Messages will be returned to the console if the allele name is malformed, or the locus is invalid; e.g., 'validateAllele("C\*01:01:01:01")' or 'validateAllele("A\*01:01:01:01:01")'.
#'@note The locus being evaluated must be included in HLAalignments.
#'
#'@export
#'
validateAllele <- function(allele) {
  alleleParts <- strsplit(allele,"*",fixed=TRUE)[[1]]
  if(length(alleleParts) != 2) {
    message(paste("No asterisk ('*') is present in ",allele,".",sep=""))
    return(FALSE)
    }
  if(length(strsplit(alleleParts[2],":",fixed=TRUE)[[1]]) == 1) {
    message(paste("No colon (':') is present in ",allele,".",sep=""))
    return(FALSE)
  }
  locus <- alleleParts[1]
  if(!suppressMessages(validateLocus(alleleParts[1],"gDNA")) && !suppressMessages(validateLocus(alleleParts[1],"cDNA")) && !suppressMessages(validateLocus(alleleParts[1],"prot"))) {
    message(paste(locus, "is not a valid locus.", sep=" "))
    return(FALSE)
  } 
  if((allele %in% HLAalignments$prot[[locus]]$trimmed_allele) || (allele %in% HLAalignments$nuc[[locus]]$trimmed_allele) || (allele %in% HLAalignments$gen[[locus]]$trimmed_allele) || (allele %in% HLAalignments$prot[[locus]]$allele_name) || (allele %in% HLAalignments$nuc[[locus]]$allele_name) || (allele %in% HLAalignments$gen[[locus]]$allele_name)) {
    return(TRUE) } else { message(paste(allele, "is not found in version", HLAalignments$version,"alignments.",sep=" "))
    return(FALSE)
   }
}


#################
##verifyAllele
#'Determine if an Allele Name Ever Existed, and (if so) in Which IPD-IMGT/HLA Database Releases
#'
#'Returns TRUE if an allele name is present in AlleleListHistory or FALSE it is absent, c(TRUE,version), where 'version' can be the most recent, original, or all IPD-IMGT/HLA Database releases in which that name appears, and c(TRUE,version,acc), where 'acc' is the IPD-IMGT/HLA accession number for the allele.
#'
#'@param allele A character string of an HLA allele name. Colon-delimited and field-delimited names are both accepted.
#'@param version A logical that indicates if the most recent nomenclature release version in which that name was valid should be returned. The default value is FALSE.
#'@param all A logical that indicates if all of the nomenclature release versions in which that name was valid should be returned. The default value is FALSE.
#'@param first A logical that indicates if only the original nomenclature release version in which that name was valid should be returned. The default value is FALSE.
#'@param acc A logical that indicates if the IPD-IMGT/HLA accession number for an allele name should be returned. The default value is FALSE.  
#'
#'@return A logical identifying if the allele name is found in AlleleListHistory (TRUE) or not (FALSE), c(TRUE,version) if version = TRUE, and c(TRUE,acc) or c(TRUE,version,acc) if acc = TRUE.
#'
#'@export
#'
#'@examples
#'verifyAllele("A*01:01:01:01")
#'verifyAllele("A*01:01:01:01",TRUE)
#'verifyAllele("A*010101",TRUE,TRUE)
#'verifyAllele("A*0101",TRUE)
#'verifyAllele("A*02:01:01:01",FALSE,FALSE,FALSE,TRUE)
#'verifyAllele("A*02:01:01:01",TRUE,TRUE,FALSE,TRUE)
#'
verifyAllele <- function(allele, version=FALSE,all=FALSE,first=FALSE,acc=FALSE){
  
  if(first && !all) {all = TRUE}
  
  resArray <- which(alleleListHistory$AlleleListHistory == allele,arr.ind = TRUE)
  if(length(resArray)==0) {return(FALSE)}
  
  accession <- alleleListHistory$AlleleListHistory[as.data.frame(resArray)[1,1],1]
  
  if(!version) {ifelse(acc,return(c(TRUE,accession)),return(TRUE))
      } else {
        if(!all) {
          rawVersion <- colnames(alleleListHistory$AlleleListHistory)[resArray[1,2]]
          {ifelse(acc,return(c(TRUE,expandVersion(substr(rawVersion,start=2,stop = nchar(rawVersion))),accession)),return(c(TRUE,expandVersion(substr(rawVersion,start=2,stop = nchar(rawVersion))))))}
        } else {
            allVersion <- unlist(lapply(unlist(lapply(colnames(alleleListHistory$AlleleListHistory)[resArray[,2]],FUN = substring,2,5)),FUN=expandVersion))
            if(!first) {
              {ifelse(acc,return(c(TRUE,allVersion,accession)),return(c(TRUE, allVersion)))}
            } else {
              {ifelse(acc,return(c(TRUE,allVersion[length(allVersion)],accession)),return(c(TRUE,allVersion[length(allVersion)])))}
            }
          }
        } 
      }

#################
##parseAlignmentHead
#'Guides for Parsing the Header Blocks of Alignment Files
#'
#'Returns a vector describing the location of key information in the header blocks of alignment files.
#'
#'@param version A character string of a validated IPD-IMGT/HLA Database release version (e.g., '3.25.0' or '3250').
#'
#'@return Either FALSE if the version is not valid, or two-value numerical vector describing (1) the header line in which alignment version data appears and (2) the length of the character string in that line preceding version data.
#'
#'@export
#'
#'@examples
#'parseAlignmentHead("3.25.0")
#'
parseAlignmentHead <- function(version){
  
  if(validateVersion(version) ) { version <- squashVersion(version)
      } else {
          return(FALSE) }
  
    field1 <- as.numeric(substr(version,1,1))
        if(field1 != 3) { 
                  message(paste("Alignments are not available for version",field1,"releases.",sep=" "))
                     return(FALSE) 
              }
    field2 <- as.numeric(substr(version,2,3))
    
    if(field2 >= 32) {return(c(3,24))} # on line three, skip the first 24 characters 
    
    if(field2 %in% c(25,27:31)) {return(c(2,22))} # on line 2, skip the first 22 characters
    
    if(field2 %in% c(0:24,26)) {return(c(2,18))} # on line 2, skip the first 18 characters
    
    return(FALSE) ## In case something else makes it through.
    
#    field3 <- substr(version,4,4) ## not needed.
  
}

##################
##queryRelease
#'Search Allele Names Across Release Versions 
#'
#'Searches specific release versions in the AlleleListHistory object for user-defined allele variants. 
#'
#'@param rel An IPD-IMGT/HLA Database release version, represented as either a character (e.g., "3.56.0") or a numeric (e.g., 3560) value.
#'
#'@param variant A character string. The value of 'variant' can be any part of a locus or allele name (e.g., "DR", "02:01", "DRB1*08:07"). The default ("") specifies all alleles in 'rel'.
#'
#'@param all A logical. When 'all' = TRUE, a vector of all instances of 'variant' in 'rel' is returned.  When 'all' = FALSE, the number of instances of 'var' in 'rel' is returned. 
#
#'@return A character vector of all matches to 'variant' in 'rel' or the number of all matches to 'variant' in 'rel'.
#'
#'@export
#'
#'@examples
#' # Identify the number of DRB9 alleles in releases 3.30.0 and 3.31.0.
#'queryRelease("3.30.0","DRB9",FALSE) 
#'queryRelease("3.31.0","DRB9",FALSE)
#'
#' # Identify the total number of alleles in release 3.56.0.
#' queryRelease(3560)
#'
queryRelease <- function(rel, variant="", all= FALSE){
  
  if(validateVersion(rel)) {rel <- paste("X",squashVersion(rel),sep="")} else {return(paste("Release version",ifelse(is.numeric(rel),expandVersion(rel),rel),"is not in the local copy of 'AlleleListHisory'.",sep=" "))}
  
      alleles <- alleleListHistory$AlleleListHistory[,rel][!is.na(alleleListHistory$AlleleListHistory[,rel][])]
  
      if(variant == "") {
          
        matchAllele <- alleles 
          
            } else { 
            
            matchAllele <- grep(variant,alleles,fixed=TRUE)
            
              }
  
      numAllele <- length(matchAllele)
  
          if(all) { 
            
                  if(variant == "") {
                          return(matchAllele) 
                    } else {
                   return(alleles[matchAllele])
                    }
    
                 } else { return(numAllele)
    
             }
  }

##################
##multiQueryRelease
#'Search Multiple Elements of Allele Names Across Release Versions 
#'
#'Searches specific release versions in the AlleleListHistory object for a set of user-defined allele-name elements. 
#'
#'@param rel An IPD-IMGT/HLA Database release version, represented as either a character (e.g., "3.58.0") or a numeric (e.g., 3580) value.
#'
#'@param variants A vector of character strings. The values in 'variants' can be any part of a locus or allele name (e.g., "DR", "02:01", "DRB1*08:07"). The default ("") specifies all alleles in 'rel'.
#'
#'@param all A logical. When 'all' = TRUE, a vector of all alleles in 'rel' that share all elements of 'variants' is returned.  When 'all' = FALSE, the number of alleles that share all elements of 'variants' in 'rel' is returned. 
#
#'@return A character vector of all alleles that share all 'variants' in 'rel' or the number of alleles that share all 'variants' in 'rel'.
#'
#'@export
#'
#'@examples
#' # Identify DRB1 allele names that include a 'Q' character in release 3.58.0.
#' multiQueryRelease("3.58.0",c("DRB1","Q"),TRUE) 
#' # Identify the number of null (N) DRB1 alleles.
#' multiQueryRelease("3.58.0",c("DRB1","N"),FALSE)
#'
multiQueryRelease <- function(rel,variants = "", all = TRUE) {
  
  if(length(variants) == 1) {return(queryRelease(rel,variants,all))}
  
  result <- NA
  
  for(i in 1:length(variants)) {
    if(i == 1) { result <- queryRelease(rel,variants[i],TRUE) 
    } else {
      result <- result[grep(variants[i],result)]
      }
   }
  ifelse(all,return(result),return(length(result))) 
}


##################
##checkAlignType
#'Ensure that AlignType Values are Valid
#'
#'Evaluates 'alignType' values to ensure that only "prot", "nuc", "codon" and "gen" are passed to downstream functions. If any other values are entered, a message describing excluded values is generated. If no valid 'alignType' values are are present, an error is generated, and any calling function is ended. 
#'
#'@param alignType A vector of character values specifying sequence alignment types to be used for a function.
#'
#'@return A character vector that includes only allowed 'alignType' values.
#'
#'@export
#'
#'@examples
#'checkAlignType(c("nuc","prot","gDNA")) 
#'
checkAlignType <- function(alignType) {
  
  if(!all(alignType %in% c("codon","nuc","prot","gen"))) {
    
    fixedType <- alignType[alignType %in% c("codon","nuc","prot","gen")]
    
    if(length(fixedType) == 0 ){stop(paste("None of the values",paste(alignType,collapse=", "),"are valid 'alignType' values."))}
    
    message(paste("The value '",alignType[!alignType %in% fixedType],"' was removed from 'alignType', as it is not a valid 'alignType' value.\n",sep=""))
    
    return(fixedType)
    }
  
  alignType
}

##################
##checkSource
#'Ensure that Source Values are Valid
#'
#'Evaluates 'source' values to ensure that only "AA", "cDNA", and "gDNA" are passed to downstream functions. If any other values are entered, a message describing excluded values is generated. 
#'
#'@param source A vector of character values specifying sequence alignment file sources to be used for a function.
#'
#'@return A character vector that includes only allowed 'source' values.
#'
#'@export
#'
#'@examples
#'checkSource(c("AA","cDNA","codon")) 
#'
checkSource <- function(source) {
  
  if(!all(source %in% c("AA","cDNA","gDNA"))) {
    
    fixedSource <- source[source %in% c("AA","cDNA","gDNA")]
    
    if(length(fixedSource) == 0 ){stop(paste("None of the values",paste(source,collapse=", "),"are valid 'source' values."))}
    
    message(paste("The value '",source[!source %in% fixedSource],"' was removed from 'source', as it is not a valid 'source' value.\n",sep=""))
    
    return(fixedSource)
  }
  
  source
}

##################
##TypeToSource
#'Convert AlignType Values to Source Values
#'
#'Converts between 'alignType' values, identifying four types of sequence alignments and 'source' values, identifying three kinds of alignment files.
#'
#'@param alignVector A vector of character values specifying kinds of four sequence alignment types ("prot","nuc","codon" and "gen"), or three source alignment files ("AA","cDNA", and "gDNA").
#'
#'@param toSource A logical. If 'toSource' is true, 'alignType' values are converted to 'source' values. If 'toSource' is false, 'source' values are concerted to 'alignType' values
#'
#'@return A character string of the converted 'align' or 'source' value.  
#'
#'@export
#'
#'@examples
#'typeToSource(c("nuc","prot","gen"),TRUE) 
#'
typeToSource <- function(alignVector,toSource=TRUE){
  
  if(toSource) {
    
    alignVector <- checkAlignType(alignVector)
    
    alignVector[alignVector == "prot"] <- "AA"
    alignVector[alignVector == "nuc"] <- "cDNA"
    alignVector[alignVector == "codon"] <- "cDNA"
    alignVector[alignVector == "gen"] <- "gDNA"
    
    unique(alignVector)
  } else {
    
      alignVector <- checkSource(alignVector)
    
      alignVector[alignVector == "AA"] <- "prot"
      alignVector[alignVector == "cDNA"] <- "nuc"
      alignVector[alignVector == "gDNA"] <- "gen"
    
      if("nuc" %in% alignVector) {alignVector <- append(alignVector,"codon",length(alignVector))} ## cDNA is both nuc and codon
    
      unique(alignVector)
      
      }  
}

#################
##AddCodonLine
#'Add an 'AA codon' Line to Alignments When Missing.
#'
#'Modifies cDNA alignment objects that are missing "AA codon" lines to include these lines in the correct location with the correct codon position information.
#'
#'@param cDNAalign A matrix of cDNA alignment lines, generated from an alignment file that is missing "AA codon" lines.
#'@param firstPos A numeric value identifying the position number of the transcript's N-terminal codon, based on a complete cDNA alignment in another release.
#'@param afterLine A numeric value identifying the number of the line below the first "cDNA" line in the alignment. The default value is 8.
#'@param codons A numeric value identifying the number of codons in each line of the nucleotide alignment. The default value is 25.
#'
#'@return A complete cDNA alignment data frame that includes "AA codon" rows.
#'
#'@export
#'
#'@note For internal HLAtools use.
#'
addCodonLine <- function(cDNAalign,firstPos,afterLine = 8, codons = 25){
  
  # Assigns the AA codon positions using the first position in the codon row of another release, and adds the missing lines.
  afterRow <- rev(which(cDNAalign == cDNAalign[afterLine])-1)
  numRow <- length(afterRow)
  startPos <- rep(firstPos,numRow) ## this needs to be set based on a correct alignment 
  for(k in 2:length(startPos)) {startPos[k] <- startPos[k-1]+codons}
  if(any(startPos < 1)) {startPos[startPos > 0] = startPos[startPos > 0 ]+1 }
  startPos <- rev(startPos)
  
  for(f in 1:length(startPos)) {
    cDNAalign <- append(cDNAalign,paste(" AA codon          ",startPos[f],sep=""),after=afterRow[f])
  }      
  
  cDNAalign
}

#buildGazetteer v7.0.0 27MAY2026

##############
##buildGazetteer
#'Define Categories of Genes Supported by the IPD-IMGT/HLA Database
#'
#'@description
#'Consumes information in the ANHIG/IMGTHLA GitHub repository and at hla.alleles.org/genes to define specific categories of genes supported by the IPD-IMGT/HLA Database, which are represented as nineteen elements of the HLAgazetteer object.
#'
#'Elements: 
#'*  All genes with alignments ($align)
#'*  Genes that that do and do not have amino acid ($prot/$noprot), nucleotide ($nuc/$nonuc), and genomic ($gen/$nogen) alignments 
#'*  HLA genes ($hla), pseudogenes ($pseudo), gene fragments ($frag)
#'*  Genes that are expressed ($expressed) or not expressed ($notexpressed)
#'*  Genes in the Class I region ($classireg), Class I HLA genes ($classIhla), Genes in the Class II region ($classiireg), and Class II HLA genes ($classiihla)
#'*  Classical HLA genes ($classical) and non-classical expressed HLA genes ($nonclassical)
#'*  HLA region genes present in specific conserved ancestral HLA region haplotype blocks ($alpha, $kappa, $beta, $delta and $epsilon)
#'*  All genes presented in map order ($map)
#'
#'The twentieth element ($version) identifies the IPD-IMGT/HLA Database version used to build the HLAgazetteer.
#'
#'@param version A string identifying the desired IPD-IMGT/HLA Database release version to which the gazetteer should be updated. The default value is most recent IPD-IMGT/HLA Database release version.
#'
#'@note The *$prot* and *$nuc* vectors include a 'DRB' "gene". While 'DRB' is not a gene name, the DRB_prot.txt file includes combined alignments for the DRB1, DRB3, DRB4, and DRB5 genes, and the DRB_nuc.txt file includes combined alignments for the DRB1, DRB2, DRB3, DRB4, DRB5, DRB6, DRB7, DRB8, and DRB9 genes. 'DRB' is included in these vectors for the purpose of validation when these combined alignments are desired.
#'
#'@return A list object of vectors organizing the genes in the IPD-IMGT/HLA Database into specific categories.
#'
#'@note For interal HLAtools use.
#'@note These elements are constructed using data compiled from hla.alleles.org/genes/, github.com/ANHIG/IMGTHLA/tree/Latest/alignments, DOI:10.2741/a317, DOI:10.1111/tan.15180, DOI: 10.1038/s41439-022-00226-5, and Human Genome Assembly GRCh38.p14 reference assembly NC_000006.12.
#'@note Additional information about these genes can be found in HLAtools::IMGTHLAGeneTypes.
#'@note The $map element does not distinguish the order of the DRB3, DRB4 and DRB5 genes (DRB3/4/5) or the DR6 and DR7 (DR6/7) genes, as the individual elements of these gene sets are found on different DRB haplotypes. 
#'@note This function requires internet access to function.
#'
#'@importFrom stringr fixed
#'@importFrom utils read.table
#'
#'@source Andersson G. Evolution of the human HLA-DR region. Front. Biosci. 1998 Jul 27:3:d739-45.
#'@source Alexandrov et al. HLA-OLI: A new MHC class I pseudogene and HLA-Y are located on a 60 kb indel in the human MHC between HLA-W and HLA-J. HLA 2023 102(5):599-606.
#'@source Dawkins et al. Genomics of the major histocompatibility complex: haplotypes, duplication, retroviruses and disease. Immunological Reviews 1999 167:275-304.
#'@source Kulski et al. Human leukocyte antigen super-locus: nexus of genomic supergenes, SNPs, indels, transcripts, and haplotypes. Human Genome Variation 2022 9:1–15.
#'
#'@export
#'
buildGazetteer <- function(version = getLatestVersion()) {
  
  if(!validateVersion(version)) {stop(paste(version,"is not a valid IPD-IMGT/HLA Database release.",sep=" "))}
  
  URL <- paste("https://github.com/ANHIG/IMGTHLA/tree/",repoVersion(version),"/alignments",sep="")
  
  # Fix for version 3.13.0/3.13.1, where "3130" is used in the URL, but "3.13.1" is used in the AlleleListHistory file and alignments
  if(version == "3.13.1") { URL <- "https://github.com/ANHIG/IMGTHLA/tree/3130/alignments" }
  
  narrow <- getAlignmentNames(URL)
  
  nucList <- vector("list", 1)
  genList <- vector("list", 1)
  protList <- vector("list", 1)

 for(i in 1:length(narrow)) {
        if (grepl("Class", narrow[i], fixed = TRUE) == FALSE) {
           if (grepl("nuc", narrow[i], fixed = TRUE)) {
       #   if(!narrow[i] == "DRB_nuc.txt") { #Exclude combined "DRB" alignments
          pure <- strsplit(narrow[i], "_")
          nucList[[1]] <- append(nucList[[1]], fixed(pure[[1]][1]), after = length(nucList[[1]]))
      #    }
        }
        if (grepl("gen", narrow[i], fixed = TRUE)) {
          pure <- strsplit(narrow[i], "_")
          genList[[1]] <- append(genList[[1]], fixed(pure[[1]][1]), after = length(genList[[1]]))
         }
        if (grepl("prot", narrow[i], fixed = TRUE)) {
      #    if(!narrow[i] == "DRB_prot.txt") { #Exclude combined "DRB" alignments
          pure <- strsplit(narrow[i], "_")
          protList[[1]] <- append(protList[[1]], fixed(pure[[1]][1]), after = length(protList[[1]]))
     #     }
        }
      }
  }
  locList <- list(zerothList <- sort(unique(c(unlist(genList),unlist(nucList),unlist(protList),c("DRB2", "DRB6", "DRB7", "DRB8", "DRB9")))), #all aligned; these last five are in the DRB_nuc.txt file. Currently, only one DRB2 and one DRB8 sequence is known.
                      firstList <- unlist(genList),  ## prot
                      secondList <- unlist(nucList), ## nuc
                      thirdlist <- unlist(protList)) ## gen

  names(locList) <- c("align","gen","nuc","prot")
  
  locList
  
  locList$align <- locList$align[!locList$align %in% "DRB"] # remove "DRB" for the list of genes with alignments

  pseudo <- IMGTHLAGeneTypes$GeneTypes$Names[IMGTHLAGeneTypes$GeneTypes$`Pseudogene/Fragment` == "Pseudogene"]
  pseudo <- sub("HLA-","",pseudo,fixed=TRUE)

  frag <- IMGTHLAGeneTypes$GeneTypes$Names[IMGTHLAGeneTypes$GeneTypes$`Pseudogene/Fragment` == "Fragment"]
  frag <- sub("HLA-","",frag,fixed=TRUE)
  
  hla <-IMGTHLAGeneTypes$GeneTypes$Names[grep("HLA-",IMGTHLAGeneTypes$GeneTypes$Names,fixed=TRUE)]
  hla <- sort(sub("HLA-","",hla,fixed=TRUE))

  locList$nogen <- sort(c(pseudo[is.element(pseudo,locList$gen) == FALSE],frag[is.element(frag,locList$gen) == FALSE]))
  locList$nonuc <- sort(c(pseudo[is.element(pseudo,locList$nuc) == FALSE],frag[is.element(frag,locList$nuc) == FALSE]))
  locList$noprot <- unique(sort(c(locList$gen[!locList$gen %in% locList$prot],locList$nonuc)))

  locList$pseudo <- pseudo
  locList$frag <- frag
  
  locList$hla <- hla
  
  expressed <- IMGTHLAGeneTypes$GeneTypes$Names[IMGTHLAGeneTypes$GeneTypes$`Pseudogene/Fragment`==""]
  expressed <- expressed[!expressed %in% "HLA-DQB3"]
  locList$expressed <- sub("HLA-","",expressed,fixed=TRUE)
  locList$notexpressed <- locList$hla[!locList$hla %in% locList$expressed]
  
  classireg <- sort(c(IMGTHLAGeneTypes$GeneTypes$Names[grep("Class I",IMGTHLAGeneTypes$GeneTypes$`Molecular Characteristics`,fixed=TRUE)],IMGTHLAGeneTypes$GeneTypes$Names[grep("class I",IMGTHLAGeneTypes$GeneTypes$`Molecular Characteristics`,fixed=TRUE)]))
  classiireg <- IMGTHLAGeneTypes$GeneTypes$Names[!IMGTHLAGeneTypes$GeneTypes$Names %in% classireg]
  classihla <- sub("HLA-","",classireg[grep("HLA-",classireg,fixed=TRUE)],fixed = TRUE)
  classiihla <- sub("HLA-","",classiireg[grep("HLA-",classiireg,fixed=TRUE)],fixed = TRUE)
  classireg <- sort(sub("HLA-","",classireg[classireg != c("HLA-Z")],fixed=TRUE)) # HLA-Z is a class I gene fragment in the class II region
  classiireg <- sort(c(sub("HLA-","",classiireg,fixed=TRUE),"Z"))
  
  locList$classireg <- classireg
  locList$classihla <- classihla
  locList$classiireg <- classiireg
  locList$classiihla <- classiihla
  
  locList$classical <- sort(c("A","B","C","DRA","DRB1","DRB3","DRB4","DRB5","DQA1","DQB1","DPA1","DPB1"))
  locList$nonclassical <- sort(c("F","G","E","DQA2","DQB2","DOB","DMB","DMA","DOA","DPA2","DPB2"))
  
  locList$alpha <- sort(c("F","MICE","V","P","G","H","T","K","U","A","W","MICD","Y","R","J"))
  locList$kappa <- sort(c("L","N","MICC","E"))
  locList$beta <- sort(c("C","B","S","MICA","X","MICB"))
 # locList$gamma <- c("") Class III region, not covered
  locList$delta <- sort(c("DRA","DRB9","DRB3/4/5","DRB8","DRB6/7","DRB2","DRB1","DQA1","DQB1","DQB3","DQA2","DQB2","DOB"))
  locList$epsilon <- sort(c("Z","DMB","DMA","DOA","DPA1","DPB1","DPA2","DPB2","DPA3"))
  
  locList$map <- c("HFE","F","MICE","V","P","G","H","T","K","U","A","W","MICD","Y","R","J","L","N","MICC","E","C","B","S","MICA","X","MICB","DRA","DRB9","DRB3/4/5","DRB8","DRB6/7","DRB2","DRB1","DQA1","DQB1","DQB3","DQA2","DQB2","DOB","TAP2","PSMB8","TAP1","PSMB9","Z","DMB","DMA","DOA","DPA1","DPB1","DPA2","DPB2","DPA3")

  locList$version <- version
    
  locList
}

##############
##getAlignmentNames
#'Retrieve Alignment Filenames for HLA Genes
#'
#'Retrieves the filenames of the protein, nucleotide and genomic alignments available in a specific branch of the IMGTHLA GitHub Repository
#'
#'@param URL A chararacter string containing a Uniform Resource Locator (URL) identifying of the desired IPD-IMGT/HLA Database release version from which the alignment filenames should be retrieved.
#'
#'@return A character vector of all of the filenames. 
#'
#'@importFrom stringr str_detect
#'
#'@note For internal HLAtools use.
#'
#'@export
#'
getAlignmentNames <- function(URL){
  on.exit(closeAllConnections())
  rough <- readLines(URL)
  rough <- strsplit(rough, split = "[,]|[:]|[\\{]|[\\}]")
  
  roughList <- c()
  k <- 0
  
  for(i in 1:length(rough)) {
    if(length(rough[[i]])>0) {
      for(j in 1:length(rough[[i]])) {
        if(nchar(rough[[i]][j])!= 0){
          if(any(!is.na(str_detect(rough[[i]][j],c("alignments"))))) {
            if(str_detect(rough[[i]][j],c("alignments")) && str_detect(rough[[i]][j],c(".txt"))) {
              roughList <- append(roughList,rough[[i]][j],length(roughList))
              k <- k+1
              roughList[k] <- substr(roughList[k],13,nchar(roughList[k])-1)
            }
          }        
        }
      }
    }
  }
  
  roughList <- roughList[!is.na(roughList)]
  roughList <- unique(roughList)
  roughList[!str_detect(roughList[],"Link--primary")]
}

## Locus Validation Functions v1.2.0 15JUL2026

################
##validateLocus
#'Determine if a Locus Name is in the HLAgazetteer
#'
#'Checks a vector of HLA locus names against the HLAgazetteer to determine if the locus name is valid for a specific type of alignment.
#'
#'@param loci A character vector of HLA gene names (ex. "DRB1", c("DRB1","DQB1")).
#'@param source A character vector of alignment source types. "AA", "cDNA", and "gDNA" are allowed types.
#'
#'@return A logical value. TRUE indicates that all of the names and source types are valid. FALSE indicates that at least one locus name or alignment source type is invalid.
#'
#'@export
#'
#'@note The results of this check should only be considered valid for the version of the HLAgazetteer included in the HLAtools package.
#'
#'@examples
#'validateLocus(loci = "DRB1", source = "AA")
#'validateLocus(loci = c("V"), source = c("cDNA","gDNA"))
#'validateLocus(loci = c("E","F","G"), source = "gDNA")
validateLocus<-function(loci, source){
      valid <- TRUE

      source <- checkSource(source)

      for(j in 1:length(loci)){
          if(loci[j]=="DRB1"|loci[j]=="DRB3"|loci[j]=="DRB4"|loci[j]=="DRB5") next
      for(x in 1:length(source)) {
          if(source[x] == "cDNA") {
              if(loci[j]%in% HLAgazetteer$nuc == FALSE) {
                  message(paste("The", loci[j], "locus is not present in", source[x],"alignments.",sep=" "))
                  valid <- FALSE
              }
            }
       if(source[x] == "gDNA") {
          if(loci[j]%in% HLAgazetteer$gen == FALSE) {
            message(paste("The", loci[j], "locus is not present in", source[x],"alignments.",sep=" "))
            valid <- FALSE
            }
          }
        if(source[x] == "AA") {
            if(loci[j]%in% HLAgazetteer$prot == FALSE) {
              message(paste("The", loci[j], "locus is not present in", source[x],"alignments.",sep=" "))
              valid <- FALSE
            }
          }
        }
      }
    valid
}

################
##multiLocusValidation
#'Apply validateLocus() to Multiple Loci
#'
#'Applies validateLocus() to a vector of locus names, validates them against HLAgazetteer$gen (by default), and returns a vector of validated locus names. 
#'
#'@param loci A character vector of locus names found in the current HLAgazetteer. 
#'
#'@param source A character vector describing the type of alignment 'loci' should be validated against. Allowed options are 'AA' (protein), 'cDNA' (nucleotide) and 'gDNA' (genomic). Only a single value should be provided. The default value is 'gen'.
#'
#'@param verbose A logical value. If 'verbose' = TRUE, messages describing invalid 'loci' or 'source' values are generated. If 'verbose' = FALSE, no messages are generated.
#'
#'@return A character vector of locus names that are present in HLAgazetteer$gen. If none of the locus names are present in HLAgazetteer$gen, an Error is generated.
#'
#'@export
#'
#'@note The results of this check should only be considered valid for the IPD-IMGT/HLA Database release version of the current HLAgazetteer.
#'
#'@examples
#'multiLocusValidation(loci = c("DRB1","DPB1","DQB8"))
#'multiLocusValidation(loci = c("A","B","C","D","Q"))
#'
multiLocusValidation <- function(loci, source = "gDNA", verbose = TRUE) {
    loci <- unique(loci)
  
    lociTest <- rep(FALSE,length(loci))
    
    funCaller <- as.character(sys.call(sys.parent())) ## checking to see if alignmentFull() called mLV(), because its source is always "gDNA", but user may specify "DRB" as a locus.

    source <- checkSource(source)
    if(length(source) != 1) {stop(if(verbose){paste("Please provide only a single 'source' value.\n")})}
    
    subSource <- source ### record the submitted source value for subverting 'source=gDNA' for cases when 'loci=DRB' when called from alignmentFull() and 'loci != all'.
    
      for(i in 1:length(loci)) {
            if(!loci[i] == "all") { 

                                    if(funCaller[1] == "alignmentFull" && loci[i] == "DRB") { 
                                            source <- "cDNA"} ### change source to "cDNA" for "DRB" only
              
                if(suppressMessages(validateLocus(loci[i],source))) { 
                   
                      lociTest[i] <- TRUE 
                      } 
                                    if(subSource != source ) { source <- subSource} ### return source to its original submitted version
                }
              }
        if(any(lociTest == FALSE)) {if(verbose) {
            message(paste("The",loci[lociTest == FALSE],"locus is invalid in version", HLAgazetteer$version ,"and has been removed.\n",sep=" "))
              }
         }
    
      loci <- loci[lociTest == TRUE]
          if(length(loci) == 0) { stop(if(verbose) {paste("None of the provided 'loci' values are valid.")} ) }
    loci
}

##data v3.0.0 16JULY2026

##alleleListHistory
#'Allele Names Across All Release Versions
#'
#'A large list object of two elements. The first element contains version information. The second element contains all releases of allele name lists in the IPD-IMGT/HLA Database.
#'This object is built by the updateAlleleListHistory() function. 
#'
#' @docType data
#' @name alleleListHistory
#' @usage data(alleleListHistory)
#' @format A large list of 2 elements that identify the table of allele names for each IPD-IMGT/HLA reference Database release and the source release
#' \itemize{
#'    \item(Version: the Date and IPD-IMGT HLA Release version under which the alleleListHistory object was made)
#'    \item(AlleleListHistory: a dataframe version of the alleleListHistory.txt file)
#' }
#' @source https://raw.githubusercontent.com/ANHIG/IMGTHLA/Latest/Allelelist_history.txt
#' 
#' @note The Allele List History file in IPD-IMGT/HLA Database release version 3.60.0 is too large to be included in the HLAtools package. The alleleListHistory object for release version 3.59.0 is included in the package. The alleleListHistory object in release versions 3.60.0 and later can be loaded via the updateAll() function.
#' 
"alleleListHistory"

##fragmentFeatureNames
#'Gene Features of HLA Pseudogenes and Gene Fragments
#'
#'A list object of 15 elements. Each of the 15 elements corresponds to one of the 15 HLA pseudogenes and gene fragments, and contains two items. The first item identifies the gene features for that locus. The second item contains an annotation detailing information regarding non-standard gene structures.
#'This object is built by the ffN() function.
#'
#' @docType data
#' @name fragmentFeatureNames
#' @usage data(fragmentFeatureNames)
#' @format A list of 15 elements that identify the gene features for each locus wih a genomic alignment and annotate non-standard features
#' @source https://hla.alleles.org/genes/index.html
#' @source https://github.com/ANHIG/IMGTHLA/tree/Latest/alignments
"fragmentFeatureNames"

##GLSC.ex
#'Example Data Frame of Genotype List String Code Data
#'
#'A two-column data frame (modified from pould::hla.hap.demo) including GL String data in GL String Code format.
#'Column one identifies each subject's status, while column two identifies each subject's HLA genotype in GL String Code format.
#'This data is provided for use in examples and demonstrations.
#'
#' @docType data
#' @name GLSC.ex
#' @usage data(GLSC.ex)
#' @format A dataframe with 419 rows and 2 columns.
#' @source pould::hla.hap.demo
#' @references Mack et al. HLA 2023 Oct;102(4):501-507 https://doi.org/10.1111/tan.15145
"GLSC.ex"

##GLstring.ex
#'Example Data Frame of Genotype List String Data.
#'
#'A two-column example data frame (from pould::hla.hap.demo) including GL String Data.
#'Column one identifies each subject's status, while column two identifies each subject's HLA genotype in GL String format.
#'This data is provided for use in examples and demonstrations.
#'
#' @docType data
#' @name GLstring.ex
#' @usage data(GLstring.ex)
#' @format A dataframe with 419 rows and 2 columns.
#' @source pould::hla.hap.demo
#' @references Mack et al. HLA 2023;102(2):206-212 https://doi.org/10.1111/tan.15126
"GLstring.ex"

##HLAgazetteer
#'Functional and Organizational Categories of Genes Supported by the IPD-IMGT/HLA Database
#'
#'A list object of nineteen vectors that identify genes in the HLA region that share specific commonalities. 
#'This object is built by the buildGazetteer() function.
#'
#' @docType data
#' @name HLAgazetteer
#' @usage data(HLAgazetteer)
#' @format A large list of 24 vectors that define specific categories of genes supported by the IPD-IMGT/HLA Database
#' \itemize{
#'.   \item(align: all genes with alignments in the IPD/IMGT-HLA GitHub Repository)
#'    \item(gen: genes with genomic alignments in the IPD/IMGT-HLA GitHub Repository)
#'    \item(nuc: genes with nucleotide alignments in the IPD/IMGT-HLA GitHub Repository)
#'    \item(prot: genes with protein alignments in the IPD/IMGT-HLA GitHub Repository)
#'    \item(nogen: genes with no genomic alignments)
#'    \item(nonuc: genes with no nucleotide alignments)
#'    \item(noprot: genes with no protein alignments)
#'    \item(pseudo: pseudogenes)
#'    \item(frag: gene fragments)
#'    \item(hla: HLA genes)
#'    \item(expressed: genes that are expressed)
#'    \item(notexpressed: genes that are not expressed)
#'    \item(classireg: genes found in the HLA class I region)
#'    \item(classihla: class I HLA genes)
#'    \item(classiireg: genes found in the HLA class II region)
#'    \item(classiihla: class II HLA genes)
#'    \item(classical: classical HLA genes)
#'    \item(nonclassical: non-classical HLA genes)
#'    \item(alpha: genes found in the alpha ancestral haplotype block)
#'    \item(kappa: genes found in the kappa ancestral haplotype block)
#'    \item(beta: genes found in the beta ancestral haplotype block)
#'    \item(delta: genes found in the delta ancestral haplotype block)
#'    \item(epsilon: genes found in the epsilon ancestral haplotype block)
#'    \item(map: all genes organized by 5' to 3' map order on the genomic reference + strand)
#'    \item(version: IPD-IMGT/HLA Database version used to build the HLAgazetteer)
#' }
#' @source https://hla.alleles.org/genes
#' @source https://github.com/ANHIG/IMGTHLA/tree/Latest/alignments
#' @source Andersson Front. Biosci. 1998, 3(4), 739–745. https://doi.org/10.2741/a317
#' @source Alexandrov et al. HLA 2023, Vol.102(5), p.599-606. https://doi.org/10.1111/tan.15180
#' @source Dawkins et al. Immunol Rev. 1999, Vol. 167, p.275-304. https://doi.org/10.1111/j.1600-065x.1999.tb01399.x 
#' @source Kulski et al. Hum Genome Var. 2022 Dec 21;9(1):49. https://doi.org/10.1038/s41439-022-00226-5 
#' @source https://www.ncbi.nlm.nih.gov/nucleotide/NC_000009.12
"HLAgazetteer"

##IMGTHLAGeneTypes
#'Molecular characteristics of the Genes Curated by the IPD-IMGT/HLA Database
#'
#'A data frame of three columns identifying each gene supported by the IPD-IMGT/HLA Database, its molecular characteristics, and its status as a gene fragment or pseudogene.
#'This object is built by the BuildIMGTHLAGeneTypes() function.
#'
#' @docType data
#' @name IMGTHLAGeneTypes
#' @usage data(IMGTHLAGeneTypes)
#' @format A list object of two elements:
#' \itemize{
#'    \item(GeneTypes: a data frame of three columns)
#'    \item(version: a character string identifying the date that the source file was written)
#' }
#' @source @source https://hla.alleles.org/pages/genes/genes_list
"IMGTHLAGeneTypes"

##UNIFORMAT.example
#'Example Data Frame of UNIFORMAT Data.
#'
#'A two-column example data frame including UNIFORMAT data.
#'Column one contains sample identifiers, while column two identifies each subject's HLA genotype in UNIFORMAT format.
#'To be used for example runs and demonstrations.
#'
#' @docType data
#' @name UNIFORMAT.example
#' @usage data(UNIFORMAT.example)
#' @format A dataframe with 20 rows and 2 columns.
#' @source https://hla-net.eu/wp/wp-content/uploads/example-three-loci.unif_.txt
#' @references Nunes Tissue Antigens 2007;69(s1):203-205 https://doi.org/10.1111/j.1399-0039.2006.00808.x
"UNIFORMAT.example"

##HLAatlas
#'Boundary Positions of Exons, Introns and UTRs in Amino Acid, cDNA and gDNA Alignments 
#'
#'A list object of two wlements. Sub-lists of R dataframes (atlases) for each locus with a protein (prot), cDNA (nuc), and gDNA (gen) alignment. Each atlas identifies the position of the exon (E), intron (I) or UTR (U) gene-feature boundaries in an alignment. Boundaries for non-standard hybrid (H), join (J), novel (N) and segment (S) features may be included in gene fragment and pseudogene atlases.
#'This object is built by the atlasFull() function.
#'
#' @docType data
#' @name HLAatlas
#' @usage data(HLAatlas)
#' @format A list of 4 elements that include the gene features boundaries in each ANHIG/IMGTHLA sequence alignment. 
#' \itemize{
#'    \item(prot: peptide-alignment atlases)
#'    \item(nuc: cDNA- and codon-alignment atlases)
#'    \item(gen: gDNA-alignment atlases)
#'    \item(version: The IPD-IMGT/HLA Database release version under which these data were generated)
#'    }
#' @source https://github.com/ANHIG/IMGTHLA/tree/Latest/alignments
"HLAatlas"

##sHLAdata
#'Synthetic HLA Data for use with Package Examples
#'
#'A BIGDAWG-formatted data frame of 18 columns and 47 rows containing synthetic HLA-A, -C, -B, -DRB1, -DQA1, -DQB1, -DPA1, and -DPB1 genotype data for 24 control subjects and 23 case subjects. Allele name data were recorded under IPD-IMGT/HLA Database release version 3.56.0. These are synthetic data generated for the imaginary UchiTelle population, and do not represent biologically-derived HLA genotypes. 
#'
#' @docType data
#' @name sHLAdata
#' @usage data(sHLAdata)
#' @format A data frame of 18 columns and 47 rows.
#' @source These synthetic data were generated as part of the 13th International HLA Workshop to demonstrate PyPop functions. See <http://pypop.org/docs/guide-chapter-usage.html#data-minimal-noheader> for additional details.
"sHLAdata"

##IMGTGenomicReferences
#'Genomic Reference Alleles for each Gene in the IPD-IMGT/HLA Database
#'
#'A data.frame characterizing a genomic reference allele for each IPD-IMGT/HLA Database gene, and a vector identifying the date that the table was generated.
#'This object is built by the buildGenomicReferences() function.
#'
#' @docType data
#' @name IMGTGenomicReferences
#' @usage data(IMGTGenomicReferences)
#' @format A list of two elements.
#' \itemize{
#'    \item(GenomicReferenceGenes: A table of 44 rows, one for each gene, and 6 columns, identifying each gene, its Ensembl version, GRChg38 location, genomic coordinate of that gene's start codon, GRChg38 reference allele, and IPD-IMGT/HLA reference allele)
#'    \item(Version: The date on which this object was generated)
#'    }
#' @source https://www.ebi.ac.uk/ipd/imgt/hla/about/help/genomics/  
"IMGTGenomicReferences"

#declaring global variables
utils::globalVariables(".")
utils::globalVariables("HLAalignments")
utils::globalVariables("alleleListHistory")
utils::globalVariables("HLAgazetteer")
utils::globalVariables("fragmentFeatureNames")
utils::globalVariables("IMGTHLAGeneTypes")
utils::globalVariables("HLAatlas")
utils::globalVariables("IMGTGenomicReferences")

#### BDstrat -- multi-allele stratification in BIGDAWG formatted datasets  
####            v2.1.1 SJM July 10, 2024 

####################
##BDstrat
#'Stratify BIGDAWG Datasets by Specific Alleles
#'
#'Divides a BIGDAWG-formatted dataset into two subsets (strata) that do and do not include specified alleles.
#'
#'@param dataset A BIGDAWG-formatted data frame or a path to a BIGDAWG-formatted, tab-delimited text file.
#'@param alleles A vector of allele names in the locus-asterisk-allele_name format (e.g., "A*01:01:01:01").
#'@param warnBelow An integer that defines a low number of subjects in a stratum, generating a warning message. The default value is 21.
#'
#'@return A list-object of two BIGDAWG-formatted data frames titled dataset$`<alleles>-positive` and dataset$`<alleles>-negative`. The positive list element includes all subjects with the specified alleles, and the negative list element includes all subjects without those specified alleles.
#'
#'@examples
#' HLA_data.multi.strat <- BDstrat(sHLAdata,c("DRB1*16:02:01:01","DRB1*04:07:01:01","A*25:01:01:01"))
#' HLA_data.single.strat <- BDstrat(sHLAdata,"DRB1*16:02:01:01")
#'
#'@export
#'
#'@references \href{https://CRAN.R-project.org/package=BIGDAWG/vignettes/BIGDAWG.html#input-data}{BIGDAWG Data Format}
#' 
BDstrat <- function(dataset,alleles,warnBelow = 21){  
  
   ## loading the dataset if it is a path
  if(!is.data.frame(dataset)) { 
    dataset <- read.table(dataset,header = TRUE,sep="\t",check.names = TRUE,stringsAsFactors = FALSE)
    # Since all loci are duplicated, check.names = TRUE generates "locus","Locus.1" name pairs if they are not unique
  } 
  
  ## checking for locus suffixes in the dataset's column header
  suffix <- NA
  suffVec <- grepl("_",colnames(dataset)[3:length(colnames(dataset))],fixed=TRUE)
  
  #no mixture of suffixes allowed
  if(length(unique(suffVec)) == 2) { stop("Each locus name must either be sufffixed with '_1' and '_2, or have no suffix.") }
  
  if(unique(suffVec) == FALSE) {suffix = FALSE} else {suffix = TRUE} ## set suffixed or not 
  
  # and make.names(x,unique=TRUE) does the same for data frames
  if(!suffix) { colnames(dataset) <- make.names(colnames(dataset),unique = TRUE) }
   
  datafalse <- as.data.frame(matrix(data=FALSE,nrow=nrow(dataset),ncol = ncol(dataset)))
  colnames(datafalse) <- colnames(dataset)
  allcols <- c()
  
  for(s in 1:length(alleles)) { # main loop for all of the stratified alleles
    
        allele <- c(strsplit(alleles[s],"*",fixed = TRUE)[[1]])
                                                                                                                  
        if(suffix) {
          currLoc <- (c(paste(allele[1],"1",sep="_"),paste(allele[1],"2",sep="_"))) } else {
            currLoc <- make.names(c(allele[1],allele[1]),unique=TRUE) }
                                                                                                                  
        allcols <- append(allcols,currLoc) ## the targeted column names
                                                                 
        for(t in 1:nrow(datafalse)) {               
          
          if(!is.na(dataset[t,colnames(dataset) %in% currLoc[1]])) {
                if(dataset[t,colnames(dataset) %in% currLoc[1]] == allele[2]) { datafalse[t,colnames(dataset) %in% currLoc[1]] <- TRUE}
                      } #end NA check
          if(!is.na(dataset[t,colnames(dataset) %in% currLoc[2]])) {
                if(dataset[t,colnames(dataset) %in% currLoc[2]] == allele[2]) { datafalse[t,colnames(dataset) %in% currLoc[2]] <- TRUE}
                      } #end NA check
                } # end of t loop
            
          }  # end of s loop
  
  datafalse <- datafalse[,colnames(datafalse) %in% unique(allcols)]
  
  sumVec <- rep(0,nrow(datafalse))
  
      for(i in 1:nrow(datafalse)){
          sumVec[i] <- sum(datafalse[i,])
        }

  posStrat <- dataset[sumVec[] > 0,]
  negStrat <- dataset[sumVec[] == 0,]
 
  # Add them as elements of the stratPair list, named with the selected or excluded alleles
  stratPair <- list()
  stratPair[[paste(paste(alleles,collapse="+"),"-positive",sep="")]] <- posStrat
  stratPair[[paste(paste(alleles,collapse="+"),"-negative",sep="")]] <- negStrat
  
  for(i in 1:2){
    # Strip suffixes from column names if they have any
    colnames(stratPair[[i]]) <- gsub(".1","",colnames(stratPair[[i]]),fixed=TRUE)
    
    # Silently eliminate empty rows
    stratPair[[i]][is.na(stratPair[[i]])] <- ""
    stratPair[[i]] <- stratPair[[i]][!substr(rownames(stratPair[[i]]),1,2) == "NA",]
    
    # Return Phenotype to integer
    stratPair[[i]][,2] <- as.integer(stratPair[[i]][,2])
    
    # Make blanks into NAs
    stratPair[[i]][stratPair[[i]][,] == ""] <- NA
    
    ## alert if number of rows in a stratum is < 20 (an arbitrary cutoff, to be adjusted)
    if(nrow(stratPair[[i]])<warnBelow){warning(paste("The",names(stratPair[i]),"stratum contains",nrow(stratPair[[i]]),ifelse(nrow(stratPair[[i]])==1,"subject.","subjects."),sep=" "))}
  }
  #return object
  stratPair
  
}

##atlasFull v1.3.0 - 20 July 2025

################
##atlasFull
#'Generate a Complete set of Protein, Coding nucleotide and Genomic Nucleotide Atlases
#'
#'Applies atlasMaker() to build a set of 'atlases' identifying the gene-feature (exon, intron, and UTR) boundaries in the protein, coding nucleotide, and genomic nucleotide alignments of all HLA loci.
#'
#'@param version A character string identifying the version of the ANHIG/IMGTHLA Github repository used to generate atlases. The default value, "Latest", generates an atlas for the most recent IPD-IMGT/HLA Database release.
#'
#'@return A list object containing a list of data frames of protein (prot), coding nucleotide (nuc), and genomic nucleotide (gen) atlases for all genes in the IPD-IMGT/HLA Database release, along with a character vector identifying the pertinent IPD-IMGT/HLA Database release.
#'
#'@note Data informing the atlases were downloaded from the ANHIG/IMGTHLA Github repository.
#'
#'@note
#'For internal HLAtools use.
#'
#'@export
#'
atlasFull <- function(version = "Latest") {

  if(version != "Latest"){ #
    if(!validateVersion(version)){stop(paste(version," is not a valid IPD-IMGT/HLA Database release version."))}
  }else{ version <- getLatestVersion()}

   #list of loci names
  NL1 <- as.list(HLAgazetteer$nuc) #cDNA
  NL2 <- as.list(HLAgazetteer$gen) #gDNA
  NL3 <- as.list(HLAgazetteer$prot) #AA
  #creating nested lists for cDNA, gDNA and amino acid
  cList <- vector(mode='list', length=length(NL1))
  gList <- vector(mode='list', length=length(NL2))
  protList <- vector(mode='list', length=length(NL3))
  #filling nested lists with alignments
  for(i in 1:length(NL1)){
    cList[i] <- suppressWarnings(atlasMaker(as.character(NL1[i]), "cDNA", version = version)[[1]][1])
  }
  for(i in 1:length(NL2)){
    gList[i] <- suppressWarnings(atlasMaker(as.character(NL2[i]), "gDNA", version = version)[[1]][1])
  }
  for(i in 1:length(NL3)){
    protList[i] <- suppressWarnings(atlasMaker(as.character(NL3[i]), "AA", version = version)[[1]][1])
  }

  #naming inside of nested lists
  names(cList) <- NL1
  names(gList) <- NL2
  names(protList) <- NL3

  #placing nested lists inside of larger list
  HLAatlas<-list(firstList <- protList,
                      secondList <- cList,
                      thirdlist <- gList,
                      versionString <- version)
                    # versionString <- ifelse(version == "Latest",getLatestVersion(),version))
  #naming nested lists
  names(HLAatlas) <- c("prot","nuc","gen","version")

  HLAatlas

}

## HLA RRisk Calculation -- Steven J Mack July 10, 2024 v 1.4.1

#######################
#' Calculate Relative Risk for Individual Alleles and Genotypes in BIGDAWG-formatted Non-Case-Control Datasets
#' 
#' Returns a list object containing relative risk, confidence interval and p-value data for the individual alleles and individual genotypes at each locus in a BIGDAWG-formatted non-case-control genotype data frame or file. 
#' 
#' @param dataset A character string describing the name of a non-case-control genotype dataset using the BIGDAWG format. Here, "non-case-control" means that while two subject categories are required, the categories should not be patients and controls; instead, the categories may be, e.g., for a dataset of patients, either of two disease states, where one disease state is coded as 0 and the other is coded as 1 in the second column of the dataset. Either a tab-delimited file or a data frame can be specified. 
#' @param return A logical identifying if the list object should be returned (return=TRUE), or if pairs of tab-delimited text files of results (one for alleles and one for genotypes) should be written to the working directory for each locus.
#' @param save.path A character string identifying the path in which to write the pair of files when return is FALSE. The default value is tempdir().
#' 
#' @keywords relative risk genotype allele
#' 
#' @importFrom fmsb riskratio
#' 
#' @importFrom utils read.table capture.output write.table
#' 
#' @return A list object of two lists ("alleles" and "genotypes"), each of which contains a list of nine-column data frames containing results for each unique allele or genotype (in rows) at each locus. Column headers in each dataframe are, *Locus*, *Variant*, *Status_1*, *Status_0*, *RelativeRisk*, *CI.low*, *CI.high*, *p.value*, and *Significant*.
#' 
#' @export
#' 
#' @examples
#' rr <- relRisk(sHLAdata)
#' 
#' @references \href{https://CRAN.R-project.org/package=BIGDAWG/vignettes/BIGDAWG.html#input-data}{BIGDAWG Data Format}
#' 
relRisk <- function(dataset,return=TRUE,save.path = tempdir()){ ## if return == TRUE, a list object is returned; if return==FALSE, two tab-delimited text files are written for each locus; one for alleles and one for genotypes

      if(!is.data.frame(dataset)){
      # load in HLA BIGDAWG formatted data
        HLAgenoData <- read.table(dataset,header=TRUE,sep="\t",quote = "",stringsAsFactors = FALSE,colClasses = "character",check.names = FALSE, na.strings = c("****","NA","","-","na","Na"))
      } else {
              HLAgenoData <- dataset }
  
        colnames(HLAgenoData) <- sub("\\.\\d|\\_\\d","",colnames(HLAgenoData))
  
        loci <- unique(colnames(HLAgenoData)[3:ncol(HLAgenoData)])

        ## sorting genotypes in ascending order
        for(j in 1:nrow(HLAgenoData)) {
              for(i in 1:length(loci)){
                  if(!NA %in% HLAgenoData[j,colnames(HLAgenoData) %in% loci[i]]) { #ignore NAs
              HLAgenoData[j,colnames(HLAgenoData) %in% loci[i]] <- sort(unlist(HLAgenoData[j,colnames(HLAgenoData) %in% loci[i]]),decreasing = FALSE)
              }
        }
    
    }
        cases <- HLAgenoData[HLAgenoData[,2]=="1",] # status_1
        controls <- HLAgenoData[HLAgenoData[,2]=="0",] # status_0

        ## variant level RRs 
        RRtab <- masterTab <- list("alleles"=list(),
                           "genotypes"=list())
        for(k in 1:2){ ## looping through alleles RRtab[[1]], and genotypes RRtab[[2]]

            for(i in 1:length(loci)){ # looping through loci

            if(k == 2){
            ## Genotype evaluations
            masterTab[[k]][[i]] <- merge(as.data.frame(table(apply(cases[,colnames(cases) %in% loci[i]],1,paste,collapse="+")),stringsAsFactors = FALSE),as.data.frame(table(apply(controls[,colnames(controls) %in% loci[i]],1,paste,collapse="+")),stringsAsFactors = FALSE),by.x="Var1",by.y="Var1",all=TRUE)
                } else {
                    ## Allele evaluations
                      masterTab[[k]][[i]] <- merge(as.data.frame(table(unlist(cases[,colnames(cases) %in% loci[i]])),stringsAsFactors = FALSE),as.data.frame(table(unlist(controls[,colnames(controls) %in% loci[i]])),stringsAsFactors = FALSE),by.x="Var1",by.y="Var1",all=TRUE)
                    }

                colnames(masterTab[[k]][[i]]) <- c(loci[i],"Cases","Controls")
                masterTab[[k]][[i]][is.na(masterTab[[k]][[i]])] <- 0
                masterTab[[k]][[i]] <- masterTab[[k]][[i]][!masterTab[[k]][[i]][,1] %in% "NA+NA",] # Remove any NA+NA genotypes
                masterTab[[k]][[i]] <- masterTab[[k]][[i]][substr(masterTab[[k]][[i]][,1],1,3) != "NA+",] # Remove any NA+X genotypes (should't be any ideally)

                totCases <- sum(masterTab[[k]][[i]]$Cases)
                totControls <- sum(masterTab[[k]][[i]]$Controls)

# Matrix for Relative Risk via fmsb:riskratio()
#  Allele | # cases | # controls
# --------+---------+-----------
#    A    |    a    |     b
#  not-A  |    c    |     d
#
# riskratio() parameters
#    X = a
#    Y = c = (totCases - a)
#   m1 = a+b 
#   m2 = c+d = (totCases - a) + (totControls - b) 

            RRtab[[k]][[i]] <- data.frame(Locus=character(),
                 Variant=character(),
                 Status_1=character(),
                 Status_0=character(),
                 RelativeRisk=numeric(), 
                 CI.low=numeric(), 
                 CI.high=numeric(),
                 p.value=numeric(),
                 Significant=character(),
                 stringsAsFactors=FALSE)

            for(j in 1:nrow(masterTab[[k]][[i]])){ # looping through variants
                  ## This is sort of kludgy -- in some instances when these values go directly into riskratio, the p-value is not calculated, and a warning:
                  ## "NAs produced by integer overflow" is produced, but by converting these to characters and then back into integers, that goes away.
                  x <- as.character(masterTab[[k]][[i]][j,2])
                  y <- as.character(totCases-masterTab[[k]][[i]][j,2])
                  M1 <- as.character(masterTab[[k]][[i]][j,2]+masterTab[[k]][[i]][j,3])
                  M2 <- as.character((totCases-masterTab[[k]][[i]][j,2])+(totControls-masterTab[[k]][[i]][j,3]))
  
              silenceRR <- capture.output(RR <- riskratio(as.numeric(x),as.numeric(y),as.numeric(M1),as.numeric(M2)))
              RRtab[[k]][[i]][j,1:9] <- c(loci[i],masterTab[[k]][[i]][j,1],masterTab[[k]][[i]][j,2],masterTab[[k]][[i]][j,3],RR$estimate,RR$conf.int[1],RR$conf.int[2],RR$p.value,ifelse(RR$p.value<0.05,"*",""))
                    } # end of variant loop j  

              if(!return){write.table(RRtab[[k]][[i]],file=paste(save.path,paste(loci[i],ifelse(k == 2,"genotype","allele"),"relative_risk.txt",sep="."),sep=.Platform$file.sep),append = FALSE,quote = FALSE,sep = "\t",row.names = FALSE,col.names = TRUE)
                    message(loci[i],ifelse(k == 2,".genotype",".allele"),".relative_risk.txt file written.")
                      }
                } # end of Loci loop i 
            } # end of allele/genotype loop k
        if(return){return(RRtab)}
      } # end of Function

#UNItoGLS (uniformat to GL string) v1.0.0 7FEB2024

#library(stringr)

################
##UNItoGLS
#'Translate UNIFORMAT to GL String
#'
#'A wrapper function for UNtoGL() which translates strings from UNIFORMAT format to GL String format.
#'
#'@param uniformat A character string of HLA allele names and operators in the UNIFORMAT format signifying their relation with one another.
#'@param prefix A character string of the desired gene-system prefix (default is "HLA-").
#'@param pre A logical that indicates whether returned allele names should contain 'prefix' (TRUE), or if 'prefix' should be excluded from returned names (FALSE).
#'
#'@return  An version of 'uniformat' converted to GL String format, or FALSE if 'uniformat' is invalid.
#'
#'@note This function does not return the "?" operator, as the "?" operator has no cognate in UNIFORMAT.
#'
#'@export
#'
#'@examples
#'UNItoGLS("A*02:01,A*03:01|A*02:01,A*03:02|A*02:02,A*03:01|A*02:02,A*03:02")
#'UNItoGLS("A,B|A,C~D G|E W,X|W,Y Z,J J,Z")
#'
#'@references Nunes Tissue Antigens 2007;69(s1):203-205 https://doi.org/10.1111/j.1399-0039.2006.00808.x
#'@references Mack et al. HLA 2023;102(2):206-212 https://doi.org/10.1111/tan.15126
UNItoGLS <- function(uniformat, prefix = "HLA-", pre = TRUE) {

  if(validateUniformat(uniformat)){
    
          UNtoGL(uniformat = uniformat, prefix = prefix, pre = pre) } else { 
              FALSE}
}

################
##multiUNItoGLS
#'Translate Multiple UNIFORMAT Strings to GL Strings
#'
#'Translate a data frame or vector of UNIFORMAT strings to GL Strings.
#'
#'@param UniformatArray A data frame or vector of UNIFORMAT formatted strings. If 'UniformatArray' is a data frame with more than one column, the first column should contain only identifiers. If 'UniformatgArray' is a vector, it should contain only UNIFORMAT strings.
#'@param prefix A character string of the desired locus prefix (default is "HLA-").
#'@param pre A logical. If 'pre' is TRUE, all allele names will be prefixed with 'prefix'. If 'pre' is FALSE, no allele names will be prefixed.
#'
#'@return A version of 'UniformatArray' in which the UNIFORMAT data have been converted to GL String format. If a 'UniformatArray' was a data frame, a data frame is returned. If 'UniformatArray' was a vector, a vector is returned. 
#'
#'@note This function does not return the GL String "?" operator, as the "?" operator has no cognate in UNIFORMAT.
#'
#'@export
#'
#'@examples
#'multiUNItoGLS(unname(as.vector(UNIFORMAT.example[2]))[[1]])
#'multiUNItoGLS(UNIFORMAT.example[1:2])
#'
#'@references Nunes Tissue Antigens 2007;69(s1):203-205 https://doi.org/10.1111/j.1399-0039.2006.00808.x
#'@references Mack et al. HLA 2023;102(2):206-212 https://doi.org/10.1111/tan.15126
multiUNItoGLS <- function(UniformatArray, prefix = "HLA-", pre = TRUE) {
 
  vec <- FALSE
  if (is.vector(UniformatArray)) {
      vec <- TRUE
      UniformatArray <-as.data.frame(UniformatArray)
    }
  
  names <- colnames(UniformatArray)

    for(x in ifelse(length(UniformatArray) == 1,1,2):length(UniformatArray)) {
 
      nn <- names[x]
      for (i in 1:length(UniformatArray[[x]])) {
 
                if (suppressMessages(validateUniformat(UniformatArray[[nn]][i]))) {
          UniformatArray[[nn]][i] <- UNtoGL(UniformatArray[[nn]][i], prefix = prefix, pre = pre)
        } else {

          message( paste( "The UNIFORMAT string in row ",i,"contains operators or characters that are not permitted in UNIFORMAT."))
        }
      }
    }

  if(vec) { UniformatArray <- unname(unlist(UniformatArray)) }
  
  UniformatArray
}

################
##UNtoGL
#'Translate UNIFORMAT Strings to GL Strings
#'
#'A function that translates strings from UNIFORMAT format to GL String format.
#'
#'@param uniformat A character string of HLA allele names and operators in the UNIFORMAT format signifying their relation with one another.
#'@param prefix A character string of the desired prefix (default is "HLA-").
#'@param pre A logical that indicates whether user would like all allele names to contain the prefix of their choice (TRUE), or if the prefix should not be appended to allele names (FALSE).
#'
#'@return An altered version of the input character string, converted to GL String format.
#'
#'@note For internal use only.
#'@note This function does not return the "?" operator, as the "?" operator has no cognate in UNIFORMAT.
#'
#'@importFrom stringr fixed str_replace_all
#'
#'@export
#'
#'@examples
#'UNtoGL("A*02:01,A*03:01|A*02:01,A*03:02|A*02:02,A*03:01|A*02:02,A*03:02")
#'UNtoGL("A,B|A,C|D,B|D,C|E,B|E,C", pre = FALSE)
#'
#'@references Nunes Tissue Antigens 2007;69(s1):203-205 https://doi.org/10.1111/j.1399-0039.2006.00808.x
#'@references Mack et al. HLA 2023;102(2):206-212 https://doi.org/10.1111/tan.15126
UNtoGL <- function(uniformat, prefix = "HLA-", pre = TRUE) {
 
  GLret <- uniformat
  
  preTemp <- prefix

  GLret <- gsub(prefix, "", GLret)

  #splitting across some operators, allowing more difficult translations to be carried out on segments
  firstSplit <- strsplit(GLret, "[ ]|[~]|[\t]")[[1]]
  for (i in 1: length(firstSplit)) {
    #divide alleles amongst these lists
    list1 <- vector("list", 1)
    list2 <- vector("list", 1)
    #more refined splits and other information, to be used when creating "/" denoted lists of ambiguous alleles
    #note: SS = second split, TS = third split
    secondSplit <- strsplit(firstSplit[i], "[|]")[[1]]
    nameSS <-unique(strsplit(firstSplit[i], "[|]|[,]")[[1]])
    #print(nameSS)
    lengthSS <- length(secondSplit)
    lengthTS <- length(strsplit(firstSplit[i], "[|]|[,]")[[1]])
    #only invoked for specific scenarios involving cartesian products
    if(lengthSS>=2 && lengthTS>2) {
      fixOne <- fixTwo <- ""
      #populating lists with first pair
      initialPopulate <- strsplit(secondSplit[1], "[,]")[[1]]
      list1 <- append(list1[[1]], initialPopulate[1])
      list2 <- append(list2[[1]], initialPopulate[2])
      #removing initial populaters names from nameSS
      nameSS <- nameSS[nameSS != initialPopulate[1]]
      nameSS <- nameSS[nameSS != initialPopulate[2]]
      #secondSplit <- secondSplit[secondSplit != secondSplit[1]]
      #print(secondSplit)
      #print(length(nameSS))
      for (n in 1:length(nameSS)) {
        #sorting to either list, if two items are in same section (e.g. |a,c|) they will be placed on different lists
        FirstList <- TRUE
        #specific scenario involving the "blank" allele
        isBlank <- FALSE
        for (w in 1:lengthSS){
          SmallSub <- strsplit(secondSplit[w], "[,]")[[1]]
          #if there are no names, this means there are only repeats, so set isBlank to true and break
          if (length(nameSS) == 0) {
            isBlank <- TRUE
            break()
          }
          if (nameSS[n] %in% SmallSub && list1[1] %in% SmallSub){
            FirstList = FALSE
            break()
          }
          if (nameSS[n] %in% SmallSub && list2[1] %in% SmallSub){
            FirstList = TRUE
            break()
          }
        }
        #handling if there is a blank allele possibility
        if (isBlank) {
          if ("blank" %in% list1) {
            list1 <- append(list1, list2, (length(list1)))
            list1 <- unique(list1)
          }
          if ("blank" %in% list2) {
            list2 <- append(list2, list1, (length(list1)))
            list2 <- unique(list2)
          }
          if (length(secondSplit) == 2) {
            list1 <- list1[list1 %in% NA == FALSE]
            list2 <- list2[list2 %in% NA == FALSE]
          }
        }
        if (FirstList) {
          #list1[[length(list1)+1]] = nameSS[n]
          list1 <- append(list1, nameSS[n], (length(list1)))
        }
        #if (!FirstList) {
        else {
          #list2[[length(list2)+1]] = nameSS[n]
          list2 <- append(list2, nameSS[n], (length(list2)))
        }
      }
      #creating single string with "/" denoted lists of ambiguous alleles
      fixOne <- paste0(list1, collapse="/")
      fixOne <- paste0(fixOne, "+")
      fixTwo <- paste0(list2, collapse="/")
      fixOne <- paste0(fixOne, fixTwo)
      #print(fixOne)
      GLret <- str_replace_all(GLret, fixed(firstSplit[i]), fixOne)
    }

  }
  #substituting the one to one translations
  GLret <- gsub(",","+",GLret)
  GLret <- gsub(" ","^",GLret)
  GLret <- gsub("\t","^",GLret)

  if (pre) {
    baseSplit <- unique(strsplit(GLret, "[.]|[/]|[+]|[|]|[~]|\\^")[[1]])
    finalSplit <- unique(strsplit(GLret, "[.]|[/]|[+]|[|]|[~]|\\^")[[1]])
    #print(finalSplit)
    for (i in 1:length(finalSplit)) {
      #print(finalSplit[i])
      finalSplit[i] <- paste0(prefix, fixed(finalSplit[i]))
      GLret <- str_replace_all(GLret, fixed(baseSplit[i]), fixed(finalSplit[i]))
    }
  }

  GLret
}

################
##validateUniformat
#'Validate a UNIFORMAT String
#'
#'Evaluates a UNIFORMAT string to identify unsupported characters.
#'
#'@param uniformat A character string of allele names and operators in the UNIFORMAT format.
#'
#'@return A logical. TRUE is returned when all characters in 'uniformat' are permitted. FALSE is returned when forbidden characters are present.
#'
#'@export
#'
#'@examples
#'validateUniformat("A*02:01,A*03:01|A*02:01,A*03:02|A*02:02,A*03:01|A*02:02,A*03:02")
#'
#'@references Nunes Tissue Antigens 2007;69(s1):203-205 https://doi.org/10.1111/j.1399-0039.2006.00808.x
validateUniformat <- function(uniformat) {
 
  badOperate <- all(strsplit(uniformat,split="")[[1]] %in% c(" ","|",",","\t","~",0:9,LETTERS,"b", "l", "a", "n", "k", "g","*","-",":"))
  
  #populating return sentence with what is wrong
  if (!badOperate) {  message(paste(uniformat, "contains operators or characters not permitted in UNIFORMAT.",sep=" "))  }

  badOperate

}

### Functions for Validating IPD-IMGT/HLA Release Version Values v2.4.2 30 Mar 2026

################
##CheckVersion
#'Check IPD-IMGT/HLA Release Version Allele Names
#'
#'Determines if allele name data for a given IPD-IMGT/HLA Release Version is present in the local HLAtools package.
#'
#'@param version A character string identifying the release version (branch) of the ANHIG/IMGTHLA Github repository (e.g. '3.53.0'). The value 'Latest' refers to the most recent release.
#'
#'@return A logical. TRUE indicates that data for 'version' is available to the local HLAtools package. FALSE means that data for 'version' is not available.
#'
#'@export
#'
#'@examples
#'checkVersion(version = "3.25.0")
#'checkVersion(version = "Latest")
#'
#'@note
#'For internal HLAtools use.
checkVersion <- function(version){
  if(version == "Latest") { version <- getLatestVersion()}
  
  
  paste("X",gsub(".","",version,fixed=TRUE),sep="") %in% colnames(alleleListHistory$AlleleListHistory)
  
}

################
##ValidateVersion
#'Validate an IPD-IMGT/HLA Release Version
#'
#'Determines if an IPD-IMGT/HLA Release Version is referenced in the AlleleListHistory file.
#'
#'@param version A character string identifying the release version (branch) of the ANHIG/IMGTHLA Github repository (e.g. '3.53.0'). The value 'Latest' refers to the most recent release.
#'
#'@return A logical. TRUE indicates that data for 'version' is available in the ANHIG/IMGTHLA GitHub repository. FALSE means that data for 'version' is not available.
#'
#'@export
#'
#'@examples
#'validateVersion(version = "3.31.1")
#'validateVersion(version = "3.13.0")
#'
#'@note
#'For internal HLAtools use.
validateVersion <- function(version) {
  if(!checkVersion(version)) {
    
    releases <- strsplit(readLines(url("https://raw.githubusercontent.com/ANHIG/IMGTHLA/Latest/Allelelist_history.txt"),n=7,ok=TRUE,skipNul = FALSE)[7],",")[[1]]
    releases <- releases[2:length(releases)]
    
    on.exit(close(url("https://raw.githubusercontent.com/ANHIG/IMGTHLA/Latest/Allelelist_history.txt")))
    
    squashVersion(version) %in% releases
    
  } else {TRUE}
  
}

################
##GetLatestVersion
#'Identify the Latest IPD-IMGT/HLA Database Release
#'
#'Identifies the most recent version of the IPD-IMGT/HLA Database available in the ANHIG/IMGTHLA Github repository. 
#'@return A dot-delimited character string identifying the latest release version (branch) of the ANHIG/IMGTHLA Github repository.
#'If the most recent release version is not a 'primary' release version (i.e., following the X.YY.O naming format), the integer 
#'in the third field of the release version is replaced with a '0'. 
#'
#'
#'@export
#'
#'@examples
#'getLatestVersion()
#'
#'@note
#'For internal HLAtools use.
getLatestVersion <- function() {
  
  vers <- strsplit(readLines(url("https://raw.githubusercontent.com/ANHIG/IMGTHLA/Latest/release_version.txt"),n=3,ok=TRUE,skipNul = FALSE)[3]," ")[[1]][4]
  
  # if the incremental version value in the third field of a release is not 0, replace the incremental version with 0
  if(!strsplit(vers, '.',fixed = TRUE)[[1]][[3]] == 0) {
    vers <- paste(strsplit(vers,'.',fixed=TRUE)[[1]][1],strsplit(vers,'.',fixed=TRUE)[[1]][2],"0",sep=".")
  }
  
  on.exit(close(url("https://raw.githubusercontent.com/ANHIG/IMGTHLA/Latest/release_version.txt")))
  vers
}

################
##SquashVersion
#'Reduce a Release Version to Numerals
#'
#'Removes the '.' delimiters from an IPD-IMGT/HLA Database release version name.
#'
#'@param ver A character vector-formatted IPD-IMGT/HLA Database release version name (e.g., '3.54.0').
#'@param num A Logial. A numeric value is returned when num = TRUE. A character vector is returned when num = FALSE. The default value is FALSE.
#'
#'@return A 4-digit value, as either a character string or a number.
#'
#'@export
#'
#'@examples
#'squashVersion("3.45.0",TRUE)
#'
#'@note
#'For internal HLAtools use.
squashVersion <- function(ver,num = FALSE){
  
  ver <- gsub(".","",ver,fixed=TRUE)
  if(num){as.numeric(ver)} else {ver}
}

################
##ExpandVersion
#'Add 'Dot' Delimiters to a Numeric Release Version
#'
#'Adds the '.' delimiters to a 'squashed' IPD-IMGT/HLA Database Release Version name
#'
#'@param ver A four character/numeral IPD-IMGT/HLA Database Release Version name (e.g., '3540' or 3450).
#'
#'@return A character vector of the expanded IPD-IMGT/HLA Database Release Version name.
#'
#'@export
#'
#'@examples
#'expandVersion(3450)
#'
#'@note
#'For internal HLAtools use.
#'@note
#'This function assumes single-character-length first and third fields of IPD-IMGT/HLA Database Release Version names.
#'
expandVersion <- function(ver){
  
  if(is.numeric(ver)) {ver <- as.character(ver)}
  ver <- strsplit(ver,split="")[[1]]
  paste(ver[1],paste(ver[2:(length(ver)-1)],collapse = ""),ver[length(ver)],sep=".")
}

##################
##RepoVersion
#'Convert an AlleleListHistory Release Version to the GitHub Repository Version
#'
#'Removes periods from IMGT/HLA Database release versions and converts release versions '3.00.0'-'3.09.0' to '300'-'390' to facilitate working with these GitHub repo branches.
#'
#'@param version A character string of a validated IPD-IMGT/HLA Database release version.
#'
#'@return A character string of the release version appropriate for use in an IMGT/HLA GitHub repository URL.
#' 
#'@examples
#'repoVersion("3.05.0")
#'
#'@note
#'For internal HLAtools use.
#'
#'@export
#'
repoVersion <- function(version){
  
  version <- squashVersion(version)
  
    verVec <- strsplit(version,"")[[1]]
    if(verVec[2] == "0" && length(verVec) == 4) {
      return(paste(c(verVec[1],verVec[3:4]),collapse=""))
      }  
    
    version
}

## UpdateAll v02.4.0 19 July 2026

################
##updateAll
#'Update All Package Data Objects Derived from IPD-IMGT/HLA Database Resources
#'
#'@description
#'Applies updateAlleleListHistory(), atlasFull(), buildGazetteer(), extractGeneTypes(), ffN() and buildGenomicReferences() to update the alleleListHistory, HLAatlas, HLAgazetteer, IMGTHLAGeneTypes, fragmentFeatureNames and IMGTGenomicReferences data objects.
#'
#'A new alleleListHistory data object should be generated with each IPD-IMGT/HLA Database release. The other data objects will likely only change when new genes are added to the IPD-IMGT/HLA Database.
#'
#'@param updateType A character vector of the names of data objects to be updated. By default, updateAll() builds all six data objects (updateType="all"). Alternatively specific data objects can be updated; e.g., updateType="alleleListHistory" or updateType=c("alleleListHistory","fragmentFeatureNames").
#'@param version A numeric value or character string identifying the version of the ANHIG/IMGTHLA Github repository to build these objects from. By default, updateAll() calls the getLatestVersion() function to identify the most recent IPD-IMGT/HLA Database release.
#'
#'@return No value is returned. The desired versions of the specified data objects are built into the environment that called updateAll().
#'
#'@note Generating a new HLAatlas can take 5 minutes or more to complete.
#'
#'@export
updateAll <-function(updateType="all",version = getLatestVersion()){

    HLTpath <- path.package("HLAtools")
    HLTDpath <- paste(HLTpath,"/data",sep="")
    
  #check and see if a version of an object exists, build new local objects, then load them in turn into the environment

  #IMGTHLAGeneTypes
    if("all" %in% updateType || "IMGTHLAGeneTypes" %in% updateType) {
      rawPage <- readLines("https://hla.alleles.org/genes/index.html",-1,warn = FALSE)
       IHGTversion <- strsplit(rawPage[grep("Last updated",rawPage,fixed=TRUE)][1],": ",fixed = TRUE)[[1]][2]
      
      if(IHGTversion == IMGTHLAGeneTypes$version){
        message(paste("IMGTHLAGeneTypes for version",IHGTversion,"is already loaded.",sep=" "))
      } else { 
        
        if(file.exists(paste(HLTDpath,paste(IHGTversion,"IMGTHLAGeneTypes.rda",sep="."),sep="/"))) {
        load(paste(HLTDpath,paste(IHGTversion,"IMGTHLAGeneTypes.rda",sep="."),sep="/"),envir = parent.frame())
        message(paste("IMGTHLAGeneTypes for version",IHGTversion,"has been loaded.",sep=" "))
      } else { 
        
      IMGTHLAGeneTypes <- buildIMGTHLAGeneTypes() # has version, no version argument
        IHGTversion <- IMGTHLAGeneTypes$version
      save(IMGTHLAGeneTypes,file=paste(HLTDpath,paste(IHGTversion,"IMGTHLAGeneTypes.rda",sep="."),sep="/"))
      load(paste(HLTDpath,paste(IHGTversion,"IMGTHLAGeneTypes.rda",sep="."),sep="/"),envir = parent.frame())
      message(paste("IMGTHLAGeneTypes for version",IHGTversion,"has been built and loaded.",sep=" "))
          }
       }
    }
      
    if("all" %in% updateType || "HLAgazetteer" %in% updateType) {
      
      if(version == HLAgazetteer$version){
        message(paste("HLAgazetteer for version",version,"is already loaded.",sep=" "))
      } else {
        if(file.exists(paste(HLTDpath,paste(version,"HLAgazetteer.rda",sep="."),sep="/"))) {
          load(paste(HLTDpath,paste(version,"HLAgazetteer.rda",sep="."),sep="/"),envir = parent.frame())
          message(paste("HLAgazetteer for version",version,"has been loaded.",sep=" "))
        } else { 
        
      HLAgazetteer <- buildGazetteer(version) ## has version, takes version
        HGversion <- HLAgazetteer$version
      save(HLAgazetteer,file=paste(HLTDpath,paste(HGversion,"HLAgazetteer.rda",sep="."),sep="/"))
      load(paste(HLTDpath,paste(HGversion,"HLAgazetteer.rda",sep="."),sep="/"),envir = parent.frame())
      message(paste("HLAgazetteer for version",version,"has been built and loaded.",sep=" "))
        }
       }
    }
    
    if("all" %in% updateType || "alleleListHistory" %in% updateType) {
      if(version == alleleListHistory$Version$Version){
        message(paste("alleleListHistory for version",version,"is already loaded.",sep=" "))
      } else {
        if(file.exists(paste(HLTDpath,paste(version,"alleleListHistory.rda",sep="."),sep="/"))) {
          load(paste(HLTDpath,paste(version,"alleleListHistory.rda",sep="."),sep="/"),envir = parent.frame())
          message(paste("alleleListHistory for version",version,"has been loaded",sep=" "))
        } else { 
          
      alleleListHistory <- updateAlleleListHistory() ## has version, takes no versiomn argument
        ALHversion <- alleleListHistory$Version$Version
      save(alleleListHistory,file=paste(HLTDpath,paste(ALHversion,"alleleListHistory.rda",sep="."),sep="/"))
      load(paste(HLTDpath,paste(ALHversion,"alleleListHistory.rda",sep="."),sep="/"),envir = parent.frame())
      message(paste("alleleListHistory for version",version,"has been built and loaded",sep=" "))
        }
      }
    } 

    if("all" %in% updateType || "fragmentFeatureNames" %in% updateType) {
      if(version == fragmentFeatureNames$version){
        message(paste("fragmentFeatureNames for version",version,"is already loaded.",sep=" "))
      } else {      
        if(file.exists(paste(HLTDpath,paste(version,"fragmentfeatureNames.rda",sep="."),sep="/"))) {
          load(paste(HLTDpath,paste(version,"fragmentFeatureNames.rda",sep="."),sep="/"),envir = parent.frame())
          message(paste("fragmentFeatureNames for version",version,"has been loaded.",sep=" "))
        } else {  
        
      fragmentFeatureNames <- ffN(version) ## has version, takes version
        FFNversion <- fragmentFeatureNames$version
      save(fragmentFeatureNames,file=paste(HLTDpath,paste(FFNversion,"fragmentFeatureNames.rda",sep="."),sep="/"))
      load(paste(HLTDpath,paste(FFNversion,"fragmentFeatureNames.rda",sep="."),sep="/"),envir = parent.frame())
      message(paste("fragmentFeatureNames for version",version,"has been built and loaded.",sep=" "))
        }
      }
    }  
     
    if("all" %in% updateType || "HLAatlas" %in% updateType) {
      if(version == HLAatlas$version){
        message(paste("HLAatlas for version",version,"is already loaded.",sep=" "))
      } else {      
        if(file.exists(paste(HLTDpath,paste(version,"HLAatlas.rda",sep="."),sep="/"))) {
          load(paste(HLTDpath,paste(version,"HLAatlas.rda",sep="."),sep="/"),envir = parent.frame())
          message(paste("HLAatlas for version",version,"has been loaded.",sep=" "))
        } else {  
          
      HLAatlas <- atlasFull(version) ## has version, takes version
        HAversion <- HLAatlas$version
      save(HLAatlas,file=paste(HLTDpath,paste(HAversion,"HLAatlas.rda",sep="."),sep="/"))
      load(paste(HLTDpath,paste(HAversion,"HLAatlas.rda",sep="."),sep="/"),envir = parent.frame())
      message(paste("HLAatlas for version",version,"has been built and loaded.",sep=" "))
        }
      }
    }
  ##
    
    if("all" %in% updateType || "IMGTGenomicReferences" %in% updateType) {
      if(version == IMGTGenomicReferences$version){
        message(paste("IMGTGenomicReferences for version",version,"is already loaded.",sep=" "))
      } else {      
        if(file.exists(paste(HLTDpath,paste(version,"IMGTGenomicReferences.rda",sep="."),sep="/"))) {
          load(paste(HLTDpath,paste(version,"IMGTGenomicReferences.rda",sep="."),sep="/"),envir = parent.frame())
          message(paste("IMGTGenomicReferences for version",version,"has been loaded.",sep=" "))
        } else {  
          
          IMGTGenomicReferences <- buildGenomicReferences() ## has version, version is built in
            IGRversion <- IMGTGenomicReferences$version
          save(IMGTGenomicReferences,file=paste(HLTDpath,paste(IGRversion,"IMGTGenomicReferences.rda",sep="."),sep="/"))
          load(paste(HLTDpath,paste(IGRversion,"IMGTGenomicReferences.rda",sep="."),sep="/"),envir = parent.frame())
          message(paste("IMGTGenomicReferences for version",version,"has been built and loaded.",sep=" "))
        }
      }
    }
    
    
    
    
}

#GLupdater v2.0.0 27FEB2025 R. Nickens & S.J. Mack

###############
##multiUpdateGL
#'Update a Vector of GL String Code Data to a Desired IPD-IMGT/HLA Database Version
#'
#'Updates a column from a data frame in GL String Code format to a desired reference database version.
#'
#'@param GLSCs A vector of GL String Codes. 
#'@param to A character string identifying the IPD-IMGT/HLA Database release version to translate 'GLSCs' to. Values can range from version 1.05.0 to the loaded version of the alleleListHistory object.
#'@param expand A logical that indicates if the lowest-numbered truncated allele names that match truncated allele names in 'GLSCs' should be returned (expand = FALSE), or if a slash-delimited string of all matching full-length allele names should be returned (expand = TRUE). The default value is FALSE.
#'@param verbose A logical that indicates if messages regarding the update process should be sent to the console (TRUE) or not (FALSE). The default value is FALSE.
#'
#'@return A version GLSCs (in a data frame) updated to the desired version. NA values are returned in place of alleles that are not present in 'to'.
#'
#'@export
#'
#'@examples
#'multiUpdateGLs <- multiUpdateGL(GLSC.ex$GL.String.Code[1:50],"2.25.0")
#'multiUpdateGLs <- multiUpdateGL(GLSC.ex$GL.String.Code[1:50],"3.58.0")
#'
#'@references Mack et al. HLA 2023;102(2):206-212 https://doi.org/10.1111/tan.15126
#'@references Mack et al. HLA 2023;102(4):501-507 https://doi.org/10.1111/tan.15145
#'
multiUpdateGL <- function(GLSCs,to,expand=FALSE,verbose=FALSE){
  
  if(!checkVersion(to)) {stop(paste("Release version",to,"is not loaded in the HLAtools package. Please provide a loaded release version for 'to', or update the alleleListHistory object."))}
  if(!is.logical(expand)) {stop(paste(expand,"is not a valid value for 'expand'. Please provide a logical value."))}
  if(!is.logical(verbose)) {stop(paste(verbose,"is not a valid value for 'verbose'. Please provide a logical value."))}
  
  outVec <-rep(NA,length(GLSCs))
  
  for(i in 1:length(GLSCs)){
    
    outVec[i] <- updateGL(GLSCs[i],to,expand,verbose)
    
  }
  
  outVec
}        

###############
##multiTranslateGLstring
#'Translate a Vector of GL Strings to a Desired IPD-IMGT/HLA Database Version
#'
#'Translates a column from a data frame in GL String Code format to a desired reference database version.
#'
#'@param GLstrings A vector of GL Strings encoded in the same IPD-IMGT/HLA Database release version. 
#'@param from A character string identifying the IPD-IMGT/HLA Database release version of 'GLstrings'.
#'@param to A character string identifying the IPD-IMGT/HLA Database release version to translate 'GLstrings' to. Values can range from version 1.05.0 to the loaded version of the alleleListHistory object.
#'@param expand A logical that indicates if the lowest-numbered truncated allele names that match truncated allele names in 'GLstrings' should be returned (expand = FALSE), or if a slash-delimited string of all matching full-length allele names should be returned (expand = TRUE). The default value is FALSE.
#'@param verbose A logical that indicates if messages regarding the trnslation process should be sent to the console (TRUE) or not (FALSE). The default value is FALSE.
#'
#'@return A version GLSCs (in a data frame) updated to the desired version. NA values are returned in place of alleles that are not present in 'to'.
#'
#'@export
#'
#'@examples
#'multiGLs <- multiTranslateGLstring(GLstring.ex$Gl.String[1:50],"3.01.0","2.15.0")
#'multiGLs <- multiTranslateGLstring(GLstring.ex$Gl.String[1:50],"3.30.0","3.58.0")
#'
#'@references Mack et al. HLA 2023;102(2):206-212 https://doi.org/10.1111/tan.15126
#'
multiTranslateGLstring <- function(GLstrings,from,to,expand=FALSE,verbose=FALSE){
  
  if(!checkVersion(to)) {stop(paste("Release version",to,"is not loaded in the HLAtools package. Please provide a loaded release version for 'to', or update the alleleListHistory object."))}
  if(!checkVersion(from)) {stop(paste("Release version",from,"is not loaded in the HLAtools package. Please provide a loaded release version for 'from', or update the alleleListHistory object."))}
  if(!is.logical(expand)) {stop(paste(expand,"is not a valid value for 'expand'. Please provide a logical value.\n"))}
  if(!is.logical(verbose)) {stop(paste(verbose,"is not a valid value for 'verbose'. Please provide a logical value.\n"))}
  
  outVec <-rep(NA,length(GLstrings))
  
  for(i in 1:length(GLstrings)){
    
    outVec[i] <- translateGLstring(GLstrings[i],from,to,expand,verbose)
    
  }
  
  outVec
}        

###############
##updateGL
#'Update a GL String Code to a Specified IPD-IMGT/HLA Database Version
#'
#'Updates the elements of a GLString Code across IPD-IMGT/HLA Database release versions. Truncated allele names can be expanded to the list of all allele names that contain the truncated name.
#'
#'@param GLSC A GL String Code containing HLA allele names. 
#'@param to A character string identifying the IPD-IMGT/HLA Database release version to translate 'GLSC' to. Values can range from version 1.05.0 to the loaded version of the alleleListHistory object.
#'@param expand A logical that indicates if the lowest-numbered truncated allele names that match truncated allele names in 'GLSC' should be returned (expand = FALSE), or if a slash-delimited string of all matching full-length allele names should be returned (expand = TRUE). The default value is FALSE.
#'@param verbose A logical that indicates if messages regarding the update process should be sent to the console (TRUE) or not (FALSE). The default value is FALSE.
#'
#'@return A GL String Code updated the 'to' release version.
#'
#'@export
#'
#'@examples
#'updateGL(GLSC.ex$GL.String.Code[1],"2.15.0",FALSE,FALSE)
#'updateGL(GLSC.ex$GL.String.Code[1],"2.15.0",TRUE,FALSE)
#'
#'@references Mack et al. HLA 2023;102(2):206-212 https://doi.org/10.1111/tan.15126
#'@references Mack et al. HLA 2023;102(4):501-507 https://doi.org/10.1111/tan.15145
#'
updateGL <- function(GLSC,to,expand = FALSE,verbose = FALSE) {
  
  if(!checkVersion(to)) {stop(paste("Release version",to,"is not loaded in the HLAtools package. Please provide a loaded release version for 'to', or update the alleleListHistory object."))}
  if(!is.logical(expand)) {stop(paste(expand,"is not a valid value for 'expand'. Please provide a logical value."))}
  if(!is.logical(verbose)) {stop(paste(verbose,"is not a valid value for 'verbose'. Please provide a logical value."))}
  
  parts <- strsplit(GLSC,"#",TRUE)[[1]]
  fromVers <- parts[2] # -- version to translate from
  if(!checkVersion(fromVers)) {stop(paste("Release version",fromVers,"is not loaded in the HLAtools package. Please update the alleleListHistory object."))}
  
  inString <- parts[3] # -- GL string to translate to 
  outString <- rep(NA,length(inString)) ## -- the translated string 
  
  outString <- translateGLstring(inString,fromVers,to,expand,verbose)
  
  paste("hla",to,outString,sep="#")
  
}

###############
##GLupdate
#'Update a GL String Code to a Specified IPD-IMGT/HLA Database Version
#'
#'Updates the elements of a GLString Code across IPD-IMGT/HLA Database release versions. Truncated allele names can be expanded to the list of all allele names that contain the truncated name.
#'
#'@param GLSC A GL String Code containing HLA allele names. 
#'@param to A character string identifying the IPD-IMGT/HLA Database release version to translate 'GLSC' to. Values can range from version 1.05.0 to the loaded version of the alleleListHistory object.
#'@param expand A logical that indicates if the lowest-numbered truncated allele names that match truncated allele names in 'GLSC' should be returned (expand = FALSE), or if a slash-delimited string of all matching full-length allele names should be returned (expand = TRUE). The default value is FALSE.
#'@param verbose A logical that indicates if messages regarding the update process should be sent to the console (TRUE) or not (FALSE). The default value is FALSE.
#'
#'@export
#'
#'@return A GL String Code updated to the 'to' release version.
#'
#'@note This function is maintained for compatibility with previous releases, and will be removed in a future release. HLAtools users are advised to use updateGL().
#'
#'@examples
#'GLupdate(GLSC.ex$GL.String.Code[1],"2.15.0",FALSE,FALSE)
#'GLupdate(GLSC.ex$GL.String.Code[1],"2.15.0",TRUE,FALSE)
#'
#'@references Mack et al. HLA 2023;102(2):206-212 https://doi.org/10.1111/tan.15126
#'@references Mack et al. HLA 2023;102(4):501-507 https://doi.org/10.1111/tan.15145
#'
GLupdate <- function(GLSC,to,expand = FALSE,verbose = FALSE) {
  
  updateGL(GLSC,to,expand,verbose)
  
}
###############
##translateGLstring
#'Translate a GL String across IPD-IMGT/HLA Database Release Versions
#'
#'Translates a single Genotype List (GL) string across IPD-IMGT/HLA Database release versions. Truncated allele names can be expanded to the list of all allele names that contain the truncated name.  
#'
#'@param GLstring A GL String of HLA allele names. 
#'@param from A character string identifying the IPD-IMGT/HLA Database release version of 'GLstring'.
#'@param to A character string identifying the IPD-IMGT/HLA Database release version to translate 'GLstring' to. Values can range from version 1.05.0 to the loaded version of the alleleListHistory object.
#'@param expand A logical that indicates if the lowest-numbered truncated allele names that match truncated allele names in the GL String should be returned (expand = FALSE), or if a slash-delimited string of all matching full-length allele names should be returned (expand = TRUE). The default value is FALSE.
#'@param verbose A logical that indicates if messages regarding allele translations should be sent to the console (TRUE) or not (FALSE). The default value is FALSE.
#'
#'@export
#'
#'@return A GL String of allele names in the 'to' release version. If the GL String is determined to be invalid, FALSE is returned. 
#'
#'@examples
#'translateGLstring(GLstring.ex$Gl.String[1],"3.01.0","2.15.0")
#'translateGLstring(GLstring.ex$Gl.String[1],"3.01.0","2.15.0",TRUE)
#'translateGLstring(GLstring.ex$Gl.String[1],"3.01.0","2.15.0",FALSE,TRUE)
#'translateGLstring(GLstring.ex$Gl.String[1],"3.01.0","2.15.0",TRUE,TRUE)
#'
#'@references Mack et al. HLA 2023;102(2):206-212 https://doi.org/10.1111/tan.15126
#'
translateGLstring <- function(GLstring,from,to,expand=FALSE,verbose=FALSE) {      
  
  if(!checkVersion(to)) {stop(paste("Release version",to,"is not loaded in the HLAtools package. Please provide a loaded release version for 'to', or update the alleleListHistory object."))}
  if(!checkVersion(from)) {stop(paste("Release version",from,"is not loaded in the HLAtools package. Please provide a loaded release version for 'from', or update the alleleListHistory object."))}
  if(!is.logical(expand)) {stop(paste(expand,"is not a valid value for 'expand'. Please provide a logical value.\n"))}
  if(!is.logical(verbose)) {stop(paste(verbose,"is not a valid value for 'verbose'. Please provide a logical value.\n"))}
  
      if(verbose) {
        if(!validateGLstring(GLstring,"1.1")) {return(FALSE)}
        } else {
          if(suppressMessages(validateGLstring(GLstring,"1.1")) == FALSE) {return(FALSE)}
        }
  
        #set up all of the parts
          bits <- strsplit(GLstring,"[+,/,|,?,^,~]")[[1]] ## the alleles to be translated, in order of appearance 
          
            if(length(bits) == 1) {
              return(translateAllele(bits,from,to,expand,verbose))
                  }
          
          pieces <- rep(NA,length(bits)-1) ## the operators that delimit the alleles, in order.
          nuts <- ifelse(grep("HLA-",GLstring),TRUE,FALSE) ## determine if the returned vector elements should be prefixed with "HLA-".
          bolts <- 0 ## spacers for determining the nature and order of delimiters
          screws <- rep(NA,length(bits)) ## the new, translated allele/strings, corresponding to each bit
  
          for(i in 1:length(pieces)) { # determine the nature and order of delimiters
              pieces[i] <- substr(GLstring,bolts+nchar(bits[i])+1,bolts+nchar(bits[i])+1)
              bolts <- bolts+nchar(bits[i])+1
            }
           
            for(i in 1:length(bits)) { ## looping through each allele in the string
              screws[i] <- translateAllele(bits[i],from,to,expand,verbose)
                } 
          
            paste(paste(c(rbind(screws[1:length(screws)-1],pieces)),collapse=""),screws[length(screws)],sep="")
                
}

################
##translateAllele
#'Translate HLA Allele Names across IPD-IMGT/HLA Database Release Versions
#'
#'Translates a single HLA allele name across IPD-IMGT/HLA Database release versions.  Truncated allele names in one version can be expanded to a list of all allele names that match the provided allele name in a chosen version.
#'
#'@param allele A full or truncated HLA allele name. 
#'@param from A character string identifying the IPD-IMGT/HLA Database release version of 'allele'.
#'@param to A character string identifying the IPD-IMGT/HLA Database release version to translate 'allele' to. Values can range from version 1.05.0 to the loaded version of the the alleleListHistory object. 
#'@param expand A logical that indicates if the lowest-numbered truncated allele name that matches a truncated allele name should be returned (expand = FALSE), or if a slash-delimited string of all matching full-length allele names should be returned (expand = TRUE). The default value is FALSE.
#'@param verbose A logical that indicates if messages regarding allele translations should be sent to the console (TRUE) or not (FALSE). The default value is FALSE.
#'
#'@return A character string of either a single matching allele or all matching alleles in the 'to' release version. If an allele does not have a cognate match in 'to', NA is returned for that allele. 
#'
#'@export
#'         
#'@examples
#'translateAllele("A*01:01","3.01.0","2.20.0",FALSE)
#'translateAllele("A*0101","2.20.0","3.01.0",TRUE)
#'translateAllele("B*57:01","3.12.0","3.12.0",TRUE)
#'
translateAllele <- function(allele,from,to,expand=FALSE,verbose = FALSE) {
  
  if(!checkVersion(to)) {stop(paste("Release version",to,"is not loaded in the HLAtools package. Please provide a loaded release version for 'to', or update the alleleListHistory object."))}
  if(!checkVersion(from)) {stop(paste("Release version",from,"is not loaded in the HLAtools package. Please provide a loaded release version for 'from', or update the alleleListHistory object."))}
  if(!is.logical(expand)) {stop(paste(expand,"is not a valid value for 'expand'. Please provide a logical value.\n"))}
  if(!is.logical(verbose)) {stop(paste(verbose,"is not a valid value for 'verbose'. Please provide a logical value.\n"))}
  
  nuts <- ifelse(substr(allele,1,4) == "HLA-", TRUE,FALSE) ## should the returned vector elements be prefixed with "HLA-".
  
  origField <- origDigit <- NA ## determine if 'allele' is field or digit delimited, and how many fields
  if(length(grep(":",allele)) == 0) {
    origDigit <- nchar(strsplit(allele,"*",fixed = TRUE)[[1]][2])  #vs 1 & 2 allele names
  } else { origField <- length(strsplit(allele,":",fixed=TRUE)[[1]]) # v3 allele names
  }
  
  #---- single allele translation
  loc <- strsplit(stringr::str_sub(allele,-(nchar(allele)-ifelse(nuts,4,0))),"*",fixed = TRUE)[[1]][1]
  check <- queryRelease(from,stringr::str_sub(allele,-(nchar(allele)-ifelse(nuts,4,0))),TRUE) # the set of all allele names matching the current allele in the current release
  check <- sort(check[substr(check,1,nchar(loc)) == loc]) ## exclude alleles from other loci, sort in numerical order
  ## exclude partial-match-loci. e.g, "DMA*01:01" when the allele is "A*01:01"
  
  if(length(check) == 0) {
    if(verbose) {message(paste("Allele",stringr::str_sub(allele,-(nchar(allele)-ifelse(nuts,4,0))),"is not found in release",paste(from,".",sep=""),sep=" "))}
    currFrom <- NA
  } else {
    if(length(check) > 1) {
      if(verbose) {message(paste(length(check), "alleles matching",paste(ifelse(nuts,"HLA-",""),stringr::str_sub(allele,-(nchar(allele)-ifelse(nuts,4,0))),sep=""),"are found in release",paste(from,".",sep=""),"Translation will be performed using the", paste(ifelse(nuts,"HLA-",""),check[1],sep=""),"allele.",sep=" "))}
      currFrom <- check[1]
    } else { ## there is only one allele
      currFrom <- check 
    }
  }
  
  if(is.na(currFrom)) {  # no currFrom -- return NA
    return(NA)
  } else { # there is a currFrom
    
    ## determine which row the current allele is in
    currRow <- match(currFrom,alleleListHistory$AlleleListHistory[,colnames(alleleListHistory$AlleleListHistory) == GLV2(from)])
    
    updAllele <- alleleListHistory$AlleleListHistory[currRow,GLV2(to)] ## only need the exact match 
    
    if(is.na(updAllele)) {
      if(verbose){paste("The",ifelse(nuts,paste("HLA",currFrom,sep="-"),currFrom),"allele has no cognate in release",paste(to,".",sep=""),sep=" ")}
      return(NA)
    } ## allele is not found in 'to'.
    
    updDigit <- updField <- NA ## determine if the updated allele is field or digit delimited
    
    if(length(grep(":",updAllele)) == 0) {
      updDigit <- nchar(strsplit(updAllele,"*",fixed = TRUE)[[1]][2])  #vs 1 & 2 allele names
    } else { 
      updField <- length(strsplit(updAllele,":",fixed=TRUE)[[1]]) # v3 allele names
    }
    
    if(!is.na(origField)) { ## original allele was colon-field delimited
      if(!is.na(updField)) { ## updated allele was colon-field delimited
        updAllele <- alleleTrim(allele = updAllele,resolution = origField,version = as.numeric(strsplit(to,split = ".",fixed = TRUE)[[1]][1]))
      } else { ## updated allele was digit-delimited
        updAllele <- alleleTrim(allele = updAllele,resolution = origField,version = as.numeric(strsplit(to,split = ".",fixed = TRUE)[[1]][1]))
      }
    } else { ## original allele was digit delimited
      if(!is.na(updField)) { ## updated allele was field-delimited
        updAllele <- alleleTrim(allele = updAllele,resolution = ceiling(origDigit/2),version = as.numeric(strsplit(to,split = ".",fixed = TRUE)[[1]][1]))
      } else { ## updated field was digit-delimited
        updAllele <- alleleTrim(allele = updAllele,resolution = ceiling(origDigit/2),version = as.numeric(strsplit(to,split = ".",fixed = TRUE)[[1]][1]))
      }
    }
    
    ## this single allele is what will be returned if expand = FALSE
    if(!expand) {return(ifelse(nuts,paste("HLA-",updAllele,sep=""),updAllele))}  
    
    trans <- updAllele  
    
    if(expand) { ## multiple alleles
      
      check <- queryRelease(to,trans,TRUE) # the set of all allele names matching the current allele in the current release
      loc <- strsplit(stringr::str_sub(check[1],-(nchar(check[1])-ifelse(nuts,4,0))),"*",fixed = TRUE)[[1]][1] ## need to do it again for 'C'vs'Cw'
      updAlleles <- sort(check[substr(check,1,nchar(loc)) == loc]) ## exclude alleles from other loci, sort in numerical order
      if(verbose) {message(paste(length(updAlleles), "version", to, "alleles match",paste(allele,".",sep=""),sep=" "))}
      
      if(nuts){ updAlleles <- paste("HLA-",updAlleles,sep="") }
      
      trans <- paste(updAlleles,collapse="/")
      
    } ## end of multiple allele expansion
  } ## end of 'is there a currFrom?'
  
  trans # return the translated allele
}

################
##GLV
#'Retrieve Version from Input GL String
#'
#'Extracts the version data from an input GL String Code, or provides appropriate options if version input is not present in the IPD-IMGT/HLA Database.
#'
#'@param GLString A character string of HLA allele names and operators in GL String Code format, signifying their relation with one another and the pertinent IPD-IMGT/HLA Database release version.
#'
#'@return Returns A character string of the version of the release as it appears in the pertinent alleleListHistory$AlleleListHistory-column-header.
#'
#'@export
#'
#'@note For internal use only.
#'
#'@examples
#'GLV("hla#3.25.0#HLA-B15:35")
#'
#'@references Mack et al. HLA 2023;102(2):206-212 https://doi.org/10.1111/tan.15126
#'@references Mack et al. HLA 2023;102(4):501-507 https://doi.org/10.1111/tan.15145
#'
GLV <- function(GLString) {
  ve1 <-""
  ve2 <- ""
  version <- strsplit(strsplit(GLString,"#")[[1]][2], "[.]")[[1]]
  long <- strsplit(gsub("[.]", "", strsplit(GLString,"#")[[1]][2]), "")
  for(i in 1:length(version)) {
    ve1 <- paste0(ve1,version[[i]])
  }
  ve1 <- paste0("X",ve1)
  if (length(long[[1]]) == 3) {
    ve2 <- strsplit(ve1, "")[[1]]
    ve2 <- append(ve2, "0", 2)
    ve2 <- paste0(ve2, collapse = "")
    ve1 <- ve2
  }
  if (!ve1 %in% names(alleleListHistory$AlleleListHistory)) {
    ve1 <- redec(ve1)
    GLVhelper(ve1)
  }
  else {
    ve1
  }
}

################
##GLV2
#'Format GL String Code Version Number
#'
#'Returns a compressed IPD-IMGT/HLA Database release version or a vector of dot-delimited release version options if the specified version is not present in the alleleListHistory object.
#'
#'@param Version A character string of the desired IPD-IMGT/HLA Database version, going back to version 1.05.0.
#'
#'@return Returns either a single, compressed character string-formatted release version, or a character vector of potential release versions.
#'
#'@export
#'
#'@note For internal use only.
#'
#'@examples
#'GLV2("3.34.0")
#'GLV2("3.0.0")
#'
#'@references Mack et al. HLA 2023;102(2):206-212 https://doi.org/10.1111/tan.15126
#'@references Mack et al. HLA 2023;102(4):501-507 https://doi.org/10.1111/tan.15145
#
GLV2 <- function(Version) {
  ve1 <- ve2 <- ""
  version <- strsplit(Version, "[.]")[[1]]
  long <- strsplit(gsub("[.]", "", Version), "")
  for(i in 1:length(version)) {
    ve1 <- paste0(ve1,version[[i]])
  }
  ve1 <- paste0("X",ve1)
  if (length(long[[1]]) == 3) {
    ve2 <- strsplit(ve1, "")[[1]]
    ve2 <- append(ve2, "0", 2)
    ve2 <- paste0(ve2, collapse = "")
    ve1 <- ve2
  }
  
  if (!ve1 %in% names(alleleListHistory$AlleleListHistory)) {
    GLVhelper(Version)
  } else {
    ve1
  }
}

################
##GLVhelper
#'Locate Matches for an Incomplete IPD-IMGT/HLA Database Version
#'
#'Uses the provided version information to locate and return possible matches for an incompletely defined IPD-IMGT/HLA Database release version.
#'
#'@param Version A character string of the desired version to which the alleles should be updated, going back to version 1.05.0.
#'
#'@return A list of character strings of possible matches to a given incomplete input.
#'
#'@export
#'
#'@note For internal use only.
#'
#'@examples
#'GLVhelper("2.25")
#'GLVhelper("3.9.0")
#'
#'
#'@references Mack et al. HLA 2023;102(2):206-212 https://doi.org/10.1111/tan.15126
#'@references Mack et al. HLA 2023;102(4):501-507 https://doi.org/10.1111/tan.15145
#'
GLVhelper <- function(Version) {
  ve5 <-ve4 <-ve3 <-ve2 <- ""
  vte1 <- vte3 <- vector("list", 1)
  fin <- FALSE
  #making list of only the numbers
  version <- strsplit(gsub("[.]","",Version), "")[[1]]
  #check all names second letter, if its same as input, format with decimals and return options to choose from
  if (length(version) == 1) {
    for (i in 2:(length(names(alleleListHistory$AlleleListHistory)))) {
      if(strsplit(names(alleleListHistory$AlleleListHistory)[i], "")[[1]][2] == version) {
        vte1[[1]] <- append(vte1[[1]], names(alleleListHistory$AlleleListHistory)[i])
      }
    }
    if (!is.null(vte1[[1]])) {
      message("Returning all version ", version, " releases. Please re-enter one of these options.")
      for (i in 1:length(vte1[[1]])) {
        vte1[[1]][i] <- redec(vte1[[1]][i])
      }
      return(vte1[[1]])
      fin <- TRUE
    }  else {
      message(version, " is not a valid major release version.")
      fin <- TRUE
    }
  }
  #add 1 zero to the end, test for multiple other options
  if (length(version) == 2) {
    message("Version is truncated. Searching for matches by appending '0'.")
 
    ve2 <- redec(paste0("X",paste(append(version, "0"),collapse="")))

     return(GLVhelper(ve2))
     fin <- TRUE
  }
  if (length(version) == 3) {
    tte3 <- vector("list", 1)

    vv3 <- paste(version,collapse="")
    for(i in 1:length(version)) {
      xt <- append(version, "0", i)
      for (n in 1:length(xt)) {
        ve3 <- paste0(ve3,xt[n])
      }
      vte3[[1]][i] <- paste0("X",ve3)
      ve3 <- ""
    }
    for (i in 1:length(vte3[[1]])) {
      if(vte3[[1]][i] %in% names(alleleListHistory$AlleleListHistory)) {
        #message about what it is returning (its returning earliest version of input)
        tte3[[1]] <- append(tte3[[1]], vte3[[1]][i])
      }
    } 
    if(!is.null(tte3[[1]])) { # added 7/7
    for (i in 2:(length(names(alleleListHistory$AlleleListHistory)))) {
      yy <-names(alleleListHistory$AlleleListHistory)[i]
      yy <- substr(yy, start = 2, stop = 4)
      if(yy == vv3) {
        tte3[[1]] <- append(tte3[[1]], names(alleleListHistory$AlleleListHistory)[i])
          }
        }
    } else { ## when TTE is null, then none of the options exist
      
      message("No direct matches were found for release version '", Version, "', please try one of these options:")
      return(as.vector(sapply(names(alleleListHistory$AlleleListHistory)[substr(names(alleleListHistory$AlleleListHistory)[1:length(alleleListHistory$AlleleListHistory)],2,2) == version[1]],redec)))
         }
    
    if (!is.null(tte3[[1]])) {

      tte3[[1]] <- unique(tte3[[1]])
      for (i in 1:length(tte3[[1]])) {
        tte3[[1]][i] <- redec(tte3[[1]][i])
      }
      message("No direct matches were found for release version '", vv3, "', please try one of these options:")
      return(tte3[[1]])

      fin <- TRUE
    } else {    
      return(GLVhelper(version[1]))
      fin <- TRUE
    }
  }
  
  if(!fin) {
  #remove last element from list, and see if that works
  if (length(version) > 3) {
    
    xnames <- as.vector(sapply(names(alleleListHistory$AlleleListHistory)[-1],redec,simplify = TRUE))
    
    if(Version %in% xnames) {
      
      return(Version)
      
    } else {
      message("No direct matches were found for release version '", Version, "', please try one of these options:")

         return(GLVhelper(version[1]))
              }
          }
      }
  }

################
##redec
#'Reintroduce Version Decimals
#'
#'Correctly put decimals back into an AlleleListHistory column-formatted version name.
#'
#'@param Cname A character string describing a version name in AlleleListHistory column format.
#'
#'@return A character string describing a version name with decimals in appropriate places according to the IPD-IMGT/HLA Database.
#'
#'@export
#'
#'@note For internal use only.
#'
#'@examples
#'redec("X3090")
#'
redec <- function(Cname) {
  Rname <- ""
  Cname <- gsub('^.', '', Cname)
  Dname <- strsplit(Cname, "")[[1]]
  if(length(Dname)==4) {
    Dname <- append(append(Dname, ".", 1),".", 4)
      }
  if(length(Dname)==3) {
    Dname <- append(append(Dname, ".", 1),".", 3)
      }
  if(length(Dname)==2) {
    Dname <- append(Dname, ".", 1)
      }
  for(i in 1:length(Dname)) {
    Rname <- paste0(Rname,Dname[i])
      }
  Rname
}

################
##GLvalidate
#'Validate a GL String Code
#'
#'A lightweight validator that inspects a GL String Code for correct structure. If the namespace field contains a value other than value of 'namespace', the namespace is changed to 'hla'. The version and GL String fields are not evaluated.
#'
#'@param GLString A character string describing a string formatted to GL String Code specifications.
#'@param namespace A character vector of allowed namespace strings. The default value is 'c("hla","kir")'.
#'
#'@return A character string describing either a well-formatted GL String Code or the value FALSE.
#'
#'@note For internal use only.
#'
#'@export
#'
#'@examples
#'GLvalidate("ha#3.25.0#hla-B15:35") 
#'
#'@references Mack et al. HLA 2023;102(2):206-212 https://doi.org/10.1111/tan.15126
#'@references Mack et al. HLA 2023;102(4):501-507 https://doi.org/10.1111/tan.15145
#'
GLvalidate <- function(GLString,namespace = c("hla","kir")) {
  GLcopy <- GLString
  GLret <- ""
  if (length(strsplit(GLString, "#")[[1]]) > 3) {
    message("The input GL String code is formatted incorrectly, there are too many '#' delimited sections.")
    return(FALSE)
      }
      if (length(strsplit(GLString, "#")[[1]]) < 3) {
        message("The input GL String code is formatted incorrectly, there are too few '#' delimited sections.")
        return(FALSE)
      }
  if (length(strsplit(GLcopy, "#")[[1]]) == 3) {
    whole <- strsplit(GLcopy, '#')[[1]]
    first <- strsplit(GLcopy, '#')[[1]][1]
    se <- GLV(GLString)
    third <- strsplit(GLcopy, "#")[[1]][3]
        if (!first %in% namespace) {
          whole[1] <- "hla"
        }
    third <- gsub("hla-", "HLA-", third)
        for (i in 1:2) {
          GLret <- paste0(GLret, paste0(whole[i], "#"))
        }
    GLret <- paste0(GLret, third)

      return(GLret)
      }
  }

---
title: "HLAtools"
author: 
- Ryan Nickens -- rynickens@gmail.com
- Livia Tran -- livia.tran@ucsf.edu
- Leamon Crooms IV -- lcroomsiv@gmail.com
- Derek Pappas, PhD -- djpappas75@gmail.com
- Vinh Luu -- vinhluu864@berkeley.edu
- Josh Bredeweg
- Steven J. Mack, PhD -- steven.mack@ucsf.edu
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{HLAtools}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

```{r setup}
library(HLAtools)
```

-   Package Version: 1.6.3

# Overview

The *HLAtools* package contains a set of computable resources and a suite of tools intended to facilitate the effective application and analysis of the named genes in the HLA region.

HLAtools includes:

-   Reference datasets that describe the genes supported by the IPD-IMGT/HLA Database, organize them into useful categories, and identify their gene feature boundaries.
-   The ability to build local amino acid, codon, coding nucleotide, and genomic nucleotide sequence alignments for HLA genes of interest.
-   Advanced tools of the comparison and dissection of allele names and allele sequences, and the construction of customized multi-gene sequence alignments.
-   Functions that translate allele names, vectors and data frames of allele names, and Genotype List (GL) String Code messages across IPD-IMGT/HLA database release versions, and inter-convert between GL String and UNIFORMAT formats.
-   Data-analysis tools that extend the utility of Bridging Immunogenomic Data-Analysis Workflow Gaps (BIGDAWG) formatted datasets by performing multiallele and multilocus stratification, calculating Relative Risk measures, and converting BIGDAWG datasets to PyPop datasets.

------------------------------------------------------------------------

# 1. Reference Datasets

## Atlas of Gene Feature Boundaries

The *HLAatlas* data object is a list object of sub-lists of R data frames identifying the location of boundary positions between gene-features (exons, introns and untranslated regions [UTRs]) identified in the protein, nucleotide and genomic alignments for each gene supported in the [ANHIG/IMGTHLA GitHub Repository](https://github.com/ANHIG/IMGTHLA/tree/Latest/alignments). Example protein (*prot*), nucleotide (*nuc*) and genomic (*gen*) atlases for HLA-A are shown below. A new *HLAatlas* can be built after each IPD-IMGT/HLA Database release using the *updateAll()* function, although the atlases are not expected to change unless new genes are added to the IPD-IMGT/HLA Database.

Building the *HLAatlas* object requires internet access, and it can take several minutes to build a full complement of atlases. For example, building all of the atlases for IPD-IMGT/HLA Database release version 3.56.0 took 4.2 minutes on a 3.3 GHz 12-Core Intel Xenon W 2019 Mac Pro with 288 GB of 2933 MHZ DDR4 RAM, 4.6 minutes on a 2.4 GHz 8-Core Intel Core i9 2019 MacBook Pro with 64 GB of 2667 MHZ DDR4 RAM, and 5.24 minutes on a 2.4 GHz Intel Core i9-10885H Dell Precision 5500 with 32 GB of RAM.

*HLAatlas* objects can be built for all HLA region genes with available genomic, nucleotide and amino acid alignments in release versions 3.00.0 to 3.59.0. Barring changes in the structure of the source data for this object in future IPD-IMGT/HLA Database releases, it should be possible to build *HLAatlas* objects for future releases.

### Protein Sequence Atlases

The column headers for *prot* atlases identify the peptide residues encoded by codons that follow or contain an exon (E) boundary. The HLA-A *prot* atlas below illustrates that the codon encoding amino acid 1 included the Exon 1: Exon 2 boundary (E.1-2).

```         
HLAatlas$prot$A
   E.1-2 E.2-3 E.3-4 E.4-5 E.5-6 E.6-7 E.7-8
AA     1    91   183   275   314   325   341
```

### Nucleotide Sequence Atlases

The column headers for *nuc* atlases identify the cDNA and codon positions preceded by an exon (E) boundary. The HLA-A *nuc* atlas below illustrates that the Exon 1:Exon 2 boundary (E.1-2) is between nucleotide positions 73 and 74 and within codon 1.

```         
HLAatlas$nuc$A
      E.1-2 E.2-3 E.3-4 E.4-5 E.5-6 E.6-7 E.7-8
cDNA     74   344   620   896  1013  1046  1094
codon     1    91   183   275   314   325   341
```

### Genomic Sequence Atlases

The column headers for *gen* atlases identify the genomic nucleotide positions preceded by a UTR (U), exon (E) or intron (I) boundary. The HLA-A *gen* atlas below illustrates that the boundary between the 5' UTR (U.5) and Exon 1 (E.1) is between genomic positions -1 and 1, and that the boundary between Intron 3 and Exon 4 is between positions 1569 and 1570 (presented as a table for readability).

|   | U.5-E.1 | E.1-I.1 | I.1-E.2 | E.2-I.2 | I.2-E.3 | E.3-I.3 | I.3-E.4 | E.4-I.4 | I.4-E.5 | E.5-I.5 | I.5-E.6 | E.6-I.6 | I.6-E.7 | E.7-I.7 | I.7-E.8 | E.8-U.3 |
|:--:|:--:|:--:|:--:|:--:|:--:|:--:|:--:|:--:|:--:|:--:|:--:|:--:|:--:|:--:|:--:|:--:|
| gDNA | 1 | 74 | 204 | 474 | 715 | 991 | 1570 | 1846 | 1948 | 2065 | 2507 | 2540 | 2682 | 2730 | 2899 | 2904 |

## Historical Catalogue of HLA Allele Names

The *alleleListHistory* data object is a data frame that identifies all HLA, MIC and TAP allele names and their accession identifiers (HLA_IDs) for all IPD-IMGT/HLA Database release versions (e.g., "X3350" represents version 3.35.0) going back to version 1.05.0 (January of 2000), and includes allele names that have been changed or deleted. A new *alleleListHistory* can be built after each IPD-IMGT/HLA Database release using the *updateAll()* function.

```         
alleleListHistory$AlleleListHistory[1:5,1:5]
    HLA_ID          X3560          X3550          X3540          X3530
1 HLA00001  A*01:01:01:01  A*01:01:01:01  A*01:01:01:01  A*01:01:01:01
2 HLA00002  A*01:02:01:01  A*01:02:01:01  A*01:02:01:01  A*01:02:01:01
3 HLA00003  A*01:03:01:01  A*01:03:01:01  A*01:03:01:01  A*01:03:01:01
4 HLA00004 A*01:04:01:01N A*01:04:01:01N A*01:04:01:01N A*01:04:01:01N
5 HLA00005  A*02:01:01:01  A*02:01:01:01  A*02:01:01:01  A*02:01:01:01

alleleListHistory$AlleleListHistory[c(65,100:105,2094),c(1,23,58:59,81,96,101,102)]
       HLA_ID         X3350      X3000     X2280  X2090  X1110  X1060  X1050
65   HLA00069       A*24:19    A*24:19    A*2419 A*2419 A*2419   <NA>   <NA>
100  HLA00107       A*33:04    A*33:04    A*3304 A*3304 A*3304 A*3304 A*3304
101  HLA00108 A*34:01:01:01 A*34:01:01  A*340101 A*3401 A*3401 A*3401 A*3401
102  HLA00109 A*34:02:01:01    A*34:02    A*3402 A*3402 A*3402 A*3402 A*3402
103  HLA00110       A*36:01    A*36:01    A*3601 A*3601 A*3601 A*3601 A*3601
104  HLA00111       A*43:01    A*43:01    A*4301 A*4301 A*4301 A*4301 A*4301
105  HLA00112 A*66:01:01:01    A*66:01    A*6601 A*6601 A*6601 A*6601 A*6601
2094 HLA02186     MICB*021N  MICB*021N MICB*021N   <NA>   <NA>   <NA>   <NA>
```

The *alleleListHistory* data object is expected to change with each IPD-IMGT/HLA Database release. Third-party functions that reference the *alleleListHistory* data object should rely on an internal version of this data object built using the *HLAtools::updateAlleleListHistory()* function.

## Functional and Organizational Categories of Genes in the HLA Region

### Molecular Characteristics of HLA Region Genes

The *IMGTHLAGeneTypes* data object describes the [named genes within the HLA region](https://hla.alleles.org/pages/genes/genes_list/) curated by the IPD-IMGT/HLA Database. *IMGTHLAGeneTypes* distinguishes pseudogenes and gene fragments from expressed genes, and summarizes each gene's molecular characteristics.

The information in this object can be found online at <https://hla.alleles.org/pages/genes/genes_list/>. A new *IMGTHLAGeneTypes* can be built using the *updateAll()* function, although the source data is not expected to change unless new genes are added to the IPD-IMGT/HLA Database.

### Gazetteer of HLA Region Genes

The *HLAgazetteer* data object is a list object that organizes the HLA region genes supported by the IPD-IMGT/HLA Database in nineteen vectors describing the availability of alignments, gene functionality, group identity and map order. A new *HLAgazetteer* can be built after each IPD-IMGT/HLA Database release using the *updateAll()* function, although the *HLAgazetteer* is not expected to change unless new genes are added to the IPD-IMGT/HLA Database.

```         
names(HLAgazetteer)
 [1] "align" "gen" "nuc" "prot" "nogen" "nonuc" "noprot" "pseudo" "frag" "hla" "expressed" "notexpressed" "classireg" "classihla" "classiireg" "classiihla" "classical" "nonclassical" "map" "version"     
```

For example, the *\$align* vector includes all of the genes for which **sequence alignments** are available:

```         
HLAgazetteer$align
 [1] "A" "B" "C" "DMA" "DMB" "DOA" "DOB" "DPA1" "DPA2" "DPB1" "DPB2" "DQA1" "DQA2" "DQB1" "DQB2" "DRA"  "DRB1" "DRB2" "DRB3" "DRB4" "DRB5" "DRB6" "DRB7" "DRB8" "DRB9" "E" "F" "G" "H" "HFE" "J" "K" "L" "MICA" "MICB" "N" "P" "S" "T" "TAP1" "TAP2" "U" "V" "W" "Y"     
```

The *\$prot* vector includes all of the genes with **protein** alignments:

```         
HLAgazetteer$prot
 [1] "A" "B" "C" "DMA" "DMB" "DOA" "DOB" "DPA1" "DPB1" "DQA1" "DQA2" "DQB1" "DQB2" "DRA" "DRB1" "DRB3" "DRB4" "DRB5" "DRB" "E" "F" "G" "HFE" "MICA" "MICB" "TAP1" "TAP2"
```

Note that the *\$prot* and *\$nuc* vectors include a 'DRB' "gene". While 'DRB' is not a gene name, the DRB_prot.txt file includes combined alignments for the DRB1, DRB3, DRB4, and DRB5 genes, and the DRB_nuc.txt file includes combined alignments for the DRB1, DRB2, DRB3, DRB4, DRB5, DRB6, DRB7, DRB8, and DRB9 genes. 'DRB' is included in these vectors for the purpose of validation when these combined alignments are desired.

## Annotation of Pseudogene and Gene Fragment Features

The *fragmentFeatureNames* object identifies and annotates the non-standard gene features found in pseudogenes and gene fragments, based on the positions of feature boundaries ("\|") in the pertinent genomic alignment sequence. Where the features of functional genes are limited to Introns (I), Exons (E) and Untranslated Regions (U), the non-standard features in pseudogenes and gene fragments are described as:

-   Joins (J), when a feature includes sequence from two or more sequences that are separated by a boundary in the reference
-   Segments (S), when the sequence of a feature is a subset of a longer feature sequence
-   Novel (N), when the sequence of a feature does not correspond to a known sequence
-   Hybrid (H), when a feature includes a sequence to belongs to a known feature as well as nucleotide sequence that does not correspond to a feature

For each pseudogene or gene fragment, the *fragmentFeatureNames* element contains a \$features element identifying the gene features in the 5' to 3' direction, and an \$annotation element that provides some detail about the feature.

These non-standard gene features are included in the feature boundaries of the gene fragments and pseudogenes in the *HLAatlas*.

The current annotations should not change across IPD-IMGT/HLA Database releases. When a new pseudogene or gene-fragment is added to the IPD-IMGT/HLA Database, a new annotation will be added as part of a package update.

```         
fragmentFeatureNames$DPA2
$features
[1] "U.5" "E.1" "I.1" "E.2" "I.2" "E.3" "I.3" "E.4" "U.3"
$annotation
[1] "All of the reference gene features are present. E.1 starts 28 nucleotides before the reference, and ends 88 nucleotides before the reference."

fragmentFeatureNames$L
$features
 [1] "U.5" "E.1" "I.1" "E.2" "I.2" "E.3" "I.3" "E.4" "I.4" "E.5" "I.5" "E.6" "I.6" "E.7" "I.7" "E.8" "U.3"
$annotation
[1] "All of the reference gene features are present."

fragmentFeatureNames$P
$features
 [1] "J.1" "E.3" "I.3" "E.4" "I.4" "E.5" "I.5" "E.6" "I.6" "E.7" "I.7" "U.3"
$annotation
[1] "J.1 is ~350 nucleotides of novel sequence followed by ~120 nucleotides from the 5' end of Intron 2. Exon 8 is absent."

fragmentFeatureNames$S
$features
[1] "H.1" "S.1" "J.1" "E.7" "I.7" "E.9" "S.2"
$annotation
[1] "H.1 is 37 nucleotides of novel sequence, followed by the last 185 nucleotides of Intron 5. S.1 is the first 27 nucleotides of Exon 6. J.1 is the last 4 nucleotides of Exon 6 (2 nucleotides in the reference have been deleted), followed by the last 100 nucleotides of Intron 6 (6 nucleotides in the reference have been deleted). E.9 is a 191 nucleotide-long Exon in what is the 5' end of the 3' UTR in the reference. S.2 is the 3' end of the 3' UTR."
```

## Sequence Alignments

The *HLAalignments* data object is a list object of sub-lists of data frames of peptide (*prot*), codon (*codon*), coding nucleotide (*nuc*) and genomic nucleotide (*gen*) alignments for the HLA and HLA-region genes supported in the [ANHIG/IMGTHLA GitHub Repository](https://github.com/ANHIG/IMGTHLA/tree/Latest/alignments), along with a *version* character string that identifies the IPD-IMGT/HLA Database release under which the *HLAalignments* object was built.

Given the size of all the combined alignments, *HLAalignments* **is not bundled with the HLAtools package**. All alignments, or desired subsets of each type of alignment for a given gene, can be built using the *alignmentFull()* function, as described in Section 2 *Working with Sequence Alignments*, below. *HLAalignments* objects can be built for all HLA region genes with available genomic, nucleotide and amino acid alignments in IPD-IMGT/HLA release versions 3.00.0 to 3.59.0. Barring changes in the structure of the source data for this object in future IPD-IMGT/HLA Database releases, it should be possible to build *HLAalignments* objects for future releases.

These protein, codon, nucleotide and genomic DNA sequence alignments are described in detail below.

The first four columns of each alignment identify the locus, allele, two-field allele name (trimmed_allele), and full allele name (allele_name) for each sequence, while the remaining columns identify the sequence at each position, as illustrated below.

### Protein (AA) Aligments

The column header for the *prot* alignments identifies individual peptide positions, starting from the first peptide of the leader sequence (a negative position) or the first peptide of the native protein when there is no leader sequence. For example, as shown below, the HLA-A *prot* alignment begins at residue -24.

```         
HLAalignments$prot$A[1:5,1:15]
  locus       allele trimmed_allele    allele_name -24 -23 -22 -21 -20 -19 -18 -17 -16 -15 -14
1     A  01:01:01:01        A*01:01  A*01:01:01:01   M   A   V   M   A   P   R   T   L   L   L
2     A 01:01:01:02N        A*01:01N A*01:01:01:02N   M   A   V   M   A   P   R   T   L   L   L
3     A  01:01:01:03        A*01:01  A*01:01:01:03   M   A   V   M   A   P   R   T   L   L   L
4     A  01:01:01:04        A*01:01  A*01:01:01:04   M   A   V   M   A   P   R   T   L   L   L
5     A  01:01:01:05        A*01:01  A*01:01:01:05   M   A   V   M   A   P   R   T   L   L   L
```

In contrast, the TAP *prot* alignment begins at residue 1, as shown below.

```         
HLAalignments$prot$TAP1[1:5,1:15]
  locus      allele trimmed_allele      allele_name 1 2 3 4 5 6 7 8 9 10 11
1  TAP1 01:01:01:01     TAP1*01:01 TAP1*01:01:01:01 M A S S R C P A P  R  G
2  TAP1 01:01:01:02     TAP1*01:01 TAP1*01:01:01:02 M A S S R C P A P  R  G
3  TAP1 01:01:01:03     TAP1*01:01 TAP1*01:01:01:03 M A S S R C P A P  R  G
4  TAP1 01:01:01:04     TAP1*01:01 TAP1*01:01:01:04 M A S S R C P A P  R  G
5  TAP1 01:01:01:05     TAP1*01:01 TAP1*01:01:01:05 M A S S R C P A P  R  G
```

### Codon and Individual Nucleotide (cDNA) Sequence Alignments

Codon-triplet (*codon*) and individual nucleotide (*nuc*) alignments are included as separate data frames.

The column heads for the *codon* alignments identify the individual nucleotide positions within each codon, starting with the first codon. As shown below, the HLA-A *codon* alignment begins at the first nucleotide in codon position -24, which is identified as *-24*; the second nucleotide in codon -24 is identified as *-24.1* and the third nucleotide in codon -24 is identified as *-24.2*.

```         
HLAalignments$codon$A[1:5,1:12]
  locus       allele trimmed_allele    allele_name -24 -24.1 -24.2 -23 -23.1 -23.2 -22 -22.1
1     A  01:01:01:01        A*01:01  A*01:01:01:01   A     T     G   G     C     C   G     T
2     A 01:01:01:02N        A*01:01N A*01:01:01:02N   A     T     G   G     C     C   G     T
3     A  01:01:01:03        A*01:01  A*01:01:01:03   A     T     G   G     C     C   G     T
4     A  01:01:01:04        A*01:01  A*01:01:01:04   A     T     G   G     C     C   G     T
5     A  01:01:01:05        A*01:01  A*01:01:01:05   A     T     G   G     C     C   G     T
```

The column heads for the *nuc* aligmnents identify each individual nucleotide position, starting from the first transcribed nucleotide (1), as shown below.

```         
HLAalignments$nuc$A[1:5,1:21]
  locus       allele trimmed_allele    allele_name 1 2 3 4 5 6 7 8 9 10 11 12 13 14 15 16 17
1     A  01:01:01:01        A*01:01  A*01:01:01:01 A T G G C C G T C  A  T  G  G  C  G  C  C
2     A 01:01:01:02N        A*01:01N A*01:01:01:02N A T G G C C G T C  A  T  G  G  C  G  C  C
3     A  01:01:01:03        A*01:01  A*01:01:01:03 A T G G C C G T C  A  T  G  G  C  G  C  C
4     A  01:01:01:04        A*01:01  A*01:01:01:04 A T G G C C G T C  A  T  G  G  C  G  C  C
5     A  01:01:01:05        A*01:01  A*01:01:01:05 A T G G C C G T C  A  T  G  G  C  G  C  C
```

### Genomic (gDNA) Sequence Alignments

The column heads for the *gen* alignments identify the individual nucleotide positions in the 'full gene' sequence, which usually starts in the 5' untranslated region (UTR) sequence for expressed genes.

```         
HLAalignments$gen$A[1:5,1:13]
  locus       allele trimmed_allele    allele_name -300 -299 -298 -297 -296 -295 -294 -293 -292
1     A  01:01:01:01        A*01:01  A*01:01:01:01    C    A    G    G    A    G    C    A    G
2     A 01:01:01:02N        A*01:01N A*01:01:01:02N    *    *    *    *    *    *    *    *    *
3     A  01:01:01:03        A*01:01  A*01:01:01:03    C    A    G    G    A    G    C    A    G
4     A  01:01:01:04        A*01:01  A*01:01:01:04    *    *    *    *    *    *    *    *    *
5     A  01:01:01:05        A*01:01  A*01:01:01:05    *    *    *    *    *    *    *    *    *
```

### Insertion-Deletion (Indel) Variant Representation

Indel positions in these alignments are numbered sequentially, following the first position 5'/N-terminal of the indel, with decimal values appended to the 5' reference position in the indel positions, starting from ".1", as illustrated below.

```         
HLAalignments$nuc$DQB1[c(1,1138:1140),c(1:4,57:63)]
     locus      allele trimmed_allele      allele_name 53 54 54.1 54.2 54.3 54.4 55
1     DQB1 05:01:01:01     DQB1*05:01 DQB1*05:01:01:01  T  C    .    .    .    .  A
1138  DQB1      06:421    DQB1*06:421      DQB1*06:421  *  *    .    .    .    .  *
1139  DQB1     06:422N   DQB1*06:422N     DQB1*06:422N  T  C    T    G    T    C  A
1140  DQB1     06:423N   DQB1*06:423N     DQB1*06:423N  *  *    .    .    .    .  *

HLAalignments$gen$DQB1[c(1,378:380),c(1:4,589:594)]
    locus      allele trimmed_allele      allele_name 52 52.1 52.2 52.3 52.4 53
1    DQB1 05:01:01:01     DQB1*05:01 DQB1*05:01:01:01  G    .    .    .    .  T
378  DQB1 06:41:01:03     DQB1*06:41 DQB1*06:41:01:03  G    .    .    .    .  T
379  DQB1     06:422N   DQB1*06:422N     DQB1*06:422N  G    T    C    T    G  T
380  DQB1      06:424    DQB1*06:424      DQB1*06:424  G    .    .    .    .  T
```

Indels in *codon* alignment positions are similarly numbered from the codon in the N-terminal direction. However, due to a peculiarity of the R environment, which generates unique column-header names, the last two positions of the N-terminal-ward codon **before** the deletion will be numbered with the decimal values following those used to identify the insertion positions, as shown below. This only occurs when indel positions are included in a selected range. In this example, the *allele* and *trimmed_allele* columns have been omitted for spacing.

```         
HLAalignments$codon$DQB1[c(1,1138:1140),c(1,4,56:65)]
     locus      allele_name -15 -15.5 -15.6 -15.1 -15.2 -15.3 -15.4 -14 -14.1 -14.2
1     DQB1 DQB1*05:01:01:01   G     T     C     .     .     .     .   A     C     C
1138  DQB1      DQB1*06:421   *     *     *     .     .     .     .   *     *     *
1139  DQB1     DQB1*06:422N   G     T     C     T     G     T     C   A     C     C
1140  DQB1     DQB1*06:423N   *     *     *     .     .     .     .   *     *     *
```

------------------------------------------------------------------------

# 2. Working with Sequence Alignments

As noted above, the **HLAtools** package includes several functions for working with sequence alignments. All of these functions require that alignments first be built, as the complete set of protein, codon, nucleotide and genomic sequence alignments is too large to include in the **HLAtools** package. [**The functions that perform operations in sequence alignments expect the alignments to be in a variable named** ***HLAalignments***.]{style="color:red"} For example the HLA-A protein alignment is identified as *HLAalignments\$prot\$A*. If alignments are built into a different data object, they will not be accessible to the functions described below.

## Building Alignments

*alignmentFull()* is a wrapper function that applies *buildAlignments()* to populate the *HLAalignments* object. When *alignmentFull()* is run, the resulting alignments will be built using the IPD-IMGT/HLA Database release under which the *HLAgazetteer* object was built ('HLAgazetteer\$version'). When alignments for a different database release are desired, *updateAll()* should be used to update the *HLAgazetteer* to the desired release version before applying *alignmentFull()*.

**Requirements**

The *alignmentFull()* function requires internet access, and can take significant time to run to build a full complement of alignments.

For example, to build all of the alignments in IPD-IMGT/HLA Database release version 3.56.0, *HLAalignments \<- alignmentFull()* completed in 6.2 minutes on a 3.3 GHz 12-Core Intel Xenon W 2019 Mac Pro with 288 GB of 2933 MHZ DDR4 RAM, 6.5 minutes on a 2.4 GHz 8-Core Intel Core i9 2019 MacBook Pro with 64 GB of 2667 MHZ DDR4 RAM, and 9.9 minutes on a 2.4 GHz Intel Core i9-10885H Dell Precision 5500 with 32 GB of RAM.

**Parameters**

-   *loci*: A vector of the locus names for which alignments should be built. The default value ("all") specifies all loci. Allowed loci are listed in *HLAgazetteer\$align*.
-   *alignType*: A vector of alignment types for which alignments should be built. Allowed values are "prot", "codon", "nuc" and "gen", which specify respective protein, codon, nucleotide and genomic alignments, as well as the default value ("all"), which specifies all four alignment types. Loci with alignments for a specific *alignType* are defined in *HLAgazetteer\$prot*, *HLAgazetteer\$nuc* and *HLAgazetteer\$gen*.
-   *version*: A character string identifying the IPD-IMGT/HLA Database release version for which alignments should be built. The default value ("Latest") generates alignments for the most recent IPD-IMGT/HLA Database release. Other releases are identified by their IPD-IMGT/HLA Database version; e.g., "3.55.0" identifies IPD-IMGT/HLA Database version 3.55.0.

**Return** A list object containing data frames of protein (*prot*), codon (*codon*), coding nucleotide (*nuc*), and genomic nucleotide (*gen*) alignments for specified genes in the specified IPD-IMGT/HLA Database release, and a character string identifying the pertinent reference database version is returned.

Generate a full set of all alignments for all supported genes in the current release.

```         
HLAalignments <- alignmentFull()
```

Generate all alignments for four genes in the current release.

```         
HLAalignments <- alignmentFull(loci = c("C","DQB1","DPA1","DRB5")) 
```

Generate a protein alignment for one gene in release version 3.54.0.

```         
updateAll(updateType = "HLAgazetteer",version = "3.54.0")
HLAalignments <- alignmentFull("DRB1","prot","3.54.0") 
```

## Working with Alignments in Past Releases

As indicated above, *alignmentFull()* can be applied for any IPD-IMGT/HLA Database release version. The *updateAll()* function must be applied to build the *HLAgazetteer* for the desired release version **before** applying *alignmentFull()*.

```         
> updateAll()
IMGTHLAGeneTypes for version 12-10-2022 is already loaded.
HLAgazetteer for version 3.56.0 has been built and loaded.
alleleListHistory for version 3.56.0 has been built and loaded
fragmentFeatureNames for version 3.56.0 has been built and loaded.
HLAatlas for version 3.56.0 has been built and loaded.

HLAalignments <- alignmentFull()
```

### Caveats for Alignments from Previous Releases

Prior to IPD-IMGT/HLA Database release version 3.24.0, the names of the sequence alignment files for the HLA-DP and HLA-DQ genes in the IPD-IMGT/HLA GitHub Repository did not include a numerical suffix in the gene name (e.g., the protein sequence alignment file for the DQA1 gene was 'DQA_prot.txt') because sequence alignment files for the DPA2, DPB2, DQA2 and DQB2 genes had not been made available.

Similarly, prior to release version 3.51.0, the 'DRB_prot.txt' and 'DRB_nuc.txt' alignment files were the source of alignments for DRB1, DRB3, DRB4 and DRB5. In release versions 3.0.0 to 3.23.0, genomic DRB1 alignments are in 'DRB_gen.txt' files, and these alignments are in 'DRB1_gen.txt' files in subsequent releases. Building DPA1, DPB1, DQA1, DQB1, DRB1, DRB3, DRB4 and DRB5 sequence alignments from these earlier releases may require specifying a gene name that does not include the numerical suffix. When the *HLAgazetteer* has been updated to the desired release version, the available gene names for that release can be found in the *HLAgazetteer*.

In all version 3.\*.\* releases (3.0.0 to 3.59.0), nucleotide alignments for the DRB2, DRB6, DRB7 and DRB9 genes are included in the 'DRB_nuc.txt' file.

As noted above, when building alignments for past releases, the *HLAgazetteer* should be updated to reflect the gene names included in that release, and consulted to determine which gene names should be provided to alignmentFull().

```         
updateAll(version = "3.20.0")
IMGTHLAGeneTypes for version 12-10-2022 is already loaded.
HLAgazetteer for version 3.20.0 has been built and loaded.
alleleListHistory for version 3.20.0 has been built and loaded
fragmentFeatureNames for version 3.20.0 has been built and loaded.
HLAatlas for version 3.20.0 has been built and loaded.

HLAgazetteer$version
[1] "3.20.0"
HLAgazetteer$align
 [1] "A" "B" "C" "DMA" "DMB" "DOA" "DOB" "DPA" "DPB" "DQA" "DQB" "DRA" "DRB3" "DRB4" "DRB" "E" "F" "G" "H" "J" "K" "L" "MICA" "MICB" "P" "TAP1" "TAP2" "V" "Y"

HLAalignments <- alignmentFull(c("C","DQB"),alignType = c("nuc","gen"),version = "3.20.0")
```

------------------------------------------------------------------------

# 3. Trim, Search and Query Functions

## Trimming HLA Allele Names by Digits or Fields

The *multiAlleleTrim()* function shortens HLA allele names in a vector to a specified number of fields or digits, depending on the pertinent nomenclature epoch. Epoch 1 allele names are found in IPD-IMGT/HLA Database releases 1.0.0 to 1.16.0, epoch 2 allele names are found in releases 2.0.0 to 2.28.0, and epoch 3 allele names are found in releases 3.0.0 and onward. Expression variant suffixes in full-length allele names can be appended to truncated allele names, but will not be removed from full-length allele names that have fewer than four fields or eight digits.

The vector must contain only allele names, and all allele names in the vector must belong to the same nomenclature epoch.

-   *alleles*: A vector of HLA allele names from a single nomenclature epoch.
-   *resolution*: A number identifying the number of fields to which the alleles should be trimmed.
-   *version*: A number identifying the HLA nomenclature epoch under which the allele was named. The default value is 3.
-   *append*: A logical. When append = TRUE, the expression variant suffix of a full-length allele name is appended to a truncated allele name. The default value is FALSE.

**Return** A vector of trimmed allele names, shortened according to the input parameters, is returned.

```         
multiAlleleTrim(alleles = "A*03:01:01", resolution = 2)
[1] "A*03:01"

multiAlleleTrim(alleles = "A*0303N",resolution = 1,version = 1, append = TRUE)
[1] "A*03N

multiAlleleTrim("HLA-A*24020102L",resolution = 3,version = 2, append = TRUE)
[1] "HLA-A*240201L"

alleles <- c("A*02:01:01:02L","DRB1*08:07", "DQB1*04:02:01:16Q")
multiAlleleTrim(alleles,2)
[1] "A*02:01"    "DRB1*08:07" "DQB1*04:02"

multiAlleleTrim(alleles,2,append = TRUE)
[1] "A*02:01L"    "DRB1*08:07"  "DQB1*04:02Q"
```

## Identifying Differences Between Alleles at a Locus

The *compareSequences()* function identifies sequence differences between two alleles at a locus for a specific type of alignment.

**Parameters**

-   *alignType*: the alignment type of either "codon", "gen", nuc" or "prot".
-   *alleles*: a vector of the two alleles to be compared, e.g., "c("DPA1\*01:03:38:01","DPA1\*01:03:38:02")". The 'HLA-' prefix is excluded for HLA alleles.

**Return** If there are no differences between the alleles for the specified 'alignType', a message is returned. When there are differences, a data frame identifying the positions and sequences variants that distinguish the alleles is returned.

```         
compareSequences("prot",c("DPA1*01:03:38:01","DPA1*01:03:38:02"))
[1] "There are no differences between DPA1*01:03:38:01 and DPA1*01:03:38:02 in the protein alignment."

compareSequences("nuc",c("DPA1*01:03:38:01","DPA1*01:03:38:02"))
[1] "There are no differences between DPA1*01:03:38:01 and DPA1*01:03:38:02 in the nucleotide alignment."

compareSequences("codon",c("DPA1*01:03:38:01","DPA1*01:03:38:02"))
[1] "There are no differences between DPA1*01:03:38:01 and DPA1*01:03:38:02 in the codon alignment."

compareSequences("gen",c("DPA1*01:03:38:01","DPA1*01:03:38:02"))
       allele_name 1544 1723 3318 4149
1 DPA1*01:03:38:01    G    G    C    G
2 DPA1*01:03:38:02    A    A    G    C
```

## Searching Allele Names Across IPD-IMGT/HLA Database Release Versions

The *multiQueryRelease()* function searches the AlleleListHistory object for user-defined allele name variants in a specific IPD-IMGT/HLA release.

**Parameters**

-   *rel*: An IPD-IMGT/HLA Database release version, represented as either a character (e.g. "3.56.0") or a numeric (e.g., 3560) value.
-   *variants*: A vector of character strings representing any part of a locus or allele name (e.g., "DR", "02:01", "DRB1\*08:07"). The default ("") specifies all alleles in 'rel'.
-   *all*: A logical. When 'all' = TRUE, a vector of all instances of 'variants' in 'rel' is returned. When 'all' = FALSE, the number of instances of 'variants' in 'rel' is returned.

```         
multiQueryRelease("3.30.0","DRB9",FALSE) 
[1] 1
multiQueryRelease("3.30.0","DRB9",TRUE) 
[1] "DRB9*01:01"
multiQueryRelease("3.31.0","DRB9",FALSE) 
[1] 6
multiQueryRelease("1.05.0","304",TRUE) 
[1] "A*0304"    "A*3304"    "B*1304"    "Cw*03041"  "Cw*03042"  "DQB1*0304" "DRB1*0304" "DRB1*1304" "B*5304"    "A*2304"   
multiQueryRelease("3.58.0",c("DPB","2N"),TRUE)
[1] "DPB1*786:01:02N" "DPB1*401:01:02N"
```

The *verifyAllele()* function searches the AlleleListHistory object to determine if a specific allele name ever existed, and optionally returns the IPD-IMGT/HLA Database release versions in which that allele name existed and the allele's accession number.

```  
verifyAllele("A*02011",TRUE)
[1] "TRUE"   "1.15.0"
> verifyAllele("A*02011",TRUE,TRUE)
[1] "TRUE"   "1.15.0" "1.14.0" "1.13.0" "1.12.0" "1.11.0" "1.10.0" "1.09.0" "1.08.0" "1.07.0" "1.06.0" "1.05.0"
> verifyAllele("A*02011",TRUE,TRUE,TRUE,TRUE)
[1] "TRUE"     "1.05.0"   "HLA00005"
> verifyAllele("A*02011",TRUE,TRUE,FALSE,TRUE)
[1] "TRUE"     "1.15.0"   "1.14.0"   "1.13.0"   "1.12.0"   "1.11.0"   "1.10.0"   "1.09.0"   "1.08.0"   "1.07.0"   "1.06.0"   "1.05.0"   "HLA00005"
> verifyAllele("A*0201101",TRUE,TRUE,FALSE,TRUE)
[1] FALSE
verifyAllele("A*020101",TRUE,FALSE,TRUE)
[1] "TRUE"   "1.16.0"
verifyAllele("A*02:01:01:01",TRUE,FALSE,TRUE,TRUE)
[1] "TRUE"     "3.00.0"   "HLA00005"
verifyAllele("A*02:01:01:01",TRUE,FALSE,FALSE,TRUE)
[1] "TRUE"     "3.60.0"   "HLA00005"

```

## Searching Alignments at Specific Positions

The *alignmentSearch()* function returns the variants for specific positions in a single allele for a specific type of alignment.

**Parameters**

-   *alignType*: The alignment type of either "codon", "gen", nuc" or "prot".
-   *allelename*: A full-length or two-field allele name, including the locus name, but excluding the "-HLA" prefix for HLA alleles.
-   *positions*: A vector of sequence positions (e.g., c(1,2,3,4,5), c(1:5), c(1,2,3,15:20,30)). If positions do not exist in an alignment, a message is returned.
-   *prefix*: A logical that indicates if the position numbers should be included in the result (TRUE) or excluded (FALSE).
-   *sep*: The value used to separate the sequences for each position. If sep="", a sequence block is returned.

```         
alignmentSearch("nuc","DPA1*01:06:03",107)
[1] "107A"
alignmentSearch("nuc","DPA1*01:06:03",c(3,30,130,230))
[1] "3*~30*~130G~230A"
alignmentSearch("nuc","DPA1*01:06:03",c(120,125,126:128,130))
[1] "120T~125C~126G~127T~128T~130G"
alignmentSearch("nuc","DPA1*01:06:03",c(120,125,126:128,130),prefix=FALSE)
[1] "T~C~G~T~T~G"
alignmentSearch("nuc","DPA1*01:06:03",c(120,125,126:128,130),prefix=FALSE,sep="")
[1] "TCGTTG"
alignmentSearch("nuc","DPA1*01:06:03",c(120,125,126:128,130),prefix=FALSE,sep=":")
[1] "T:C:G:T:T:G"
```

When sequences include indel-positions, all of the values in the vector of sequence positions must be characters.

```         
alignmentSearch("nuc","DPA1*01:103",c(181:182),prefix=FALSE,sep="")
[1] "GA"
alignmentSearch("nuc","DPA1*01:103",c("181","181.1","181.2","182"),prefix=FALSE,sep="")
[1] "G..A"
```

## Identifying Sequence Variants at Specific Positions and their Frequencies Across all Alleles

The *queryPositions()* function identifies all of the variants at a single position in a specified locus for a specific type of alignment.

**Parameters**

-   *alignType*: The alignment type of either "codon", "gen", nuc" or "prot". Only a single alignType can be specified.
-   *locus*: A locus in HLAalignments.
-   *positions*: A vector of sequence positions (e.g., 86, c(1,2,3,4,5), c(1:5), c(1,2,3,15:20,30)). If positions do not exist in a specified alignment, a message is returned.
-   *table*: A logical value indicating if a data frame describing the counts and frequencies for each variant should be returned (TRUE) or if a vector of all variants should be returned (FALSE). The default value is FALSE.

``` 
queryPositions("prot","DRB1",86)
 [1] "." "A" "C" "D" "G" "L" "M" "Q" "R" "S" "V" "W" "X"
 
queryPositions("prot","DRB1",86,TRUE)
$DRB1_86
   Variant Count    Frequency
1        .    71 0.0185620915
2        A     6 0.0015686275
3        C    10 0.0026143791
4        D     5 0.0013071895
5        G  1887 0.4933333333
6        L     4 0.0010457516
7        M     4 0.0010457516
8        Q     1 0.0002614379
9        R     1 0.0002614379
10       S     2 0.0005228758
11       V  1825 0.4771241830
12       W     5 0.0013071895
13       X     4 0.0010457516
``` 

## Searching Alignments for Sequence Motifs

The *motifMatch()* function returns the names of alleles that share specific nucleotide or peptide motifs.

**Parameters**

-   *motif*: A nucleotide or peptide sequence variant motif structured as Locus\*#\$\~#\$\~#\$, where \# identifies a variant position, and \$ identifies the sequence variant. Any number of variants can be specified.
-   *alignType*: The alignment type of either "codon", "gen", nuc" or "prot". Only a single alignType can be specified.
-   *full*: A logical value indicating if full (TRUE) or truncated (FALSE) allele names should be returned.

```         
motifMatch("A*-21M~2P","prot")
[1] "A*02:774"  "A*11:284"  "A*11:417N" "A*68:216N"

motifMatch("A*196G~301A~3046T","gen",FALSE)
[1] "A*01:09"

motifMatch("A*196G~301A~3046T","gen",TRUE)
[1] "A*01:09:01:01" "A*01:09:01:02"
```

## Building Custom Alignments

The *customAlign()* function returns a customized peptide, codon, coding nucleotide or genomic nucleotide alignment for specified alleles at specified positions.

**Parameters**

-   *alignType* The alignment type of either "codon", "gen", nuc" or "prot".
-   *alleles* A vector of allele names, which can incude alleles at multiple loci. The 'HLA-' prefix is excluded for HLA alleles.
-   *positions* Either a single vector positions, against which all loci will be aligned, or a list of vectors of positions, for which there is one vector for each allele in 'alleles'. Each allele in 'alleles' will be aligned on the basis of the corresponding vector.

```         
customAlign("codon",c("DPB1*01:01:01:01","DQA1*01:01:01:01","DQB1*05:01:01:01"),c(1:4,6:9))
            Allele   1   2   3   4   6   7   8   9
1 DPB1*01:01:01:01 AGG GCC ACT CCA AAT TAC GTG TAC
2 DQA1*01:01:01:01 GAA GAC ATT GTG GAC CAC GTT GCC
3 DQB1*05:01:01:01 AGA GAC TCT CCC GAT TTC GTG TAC
```

When the length of vectors in a 'positions' list varies, it is recommended to coordinate the values in 'alleles' and 'positions' so that the longest vector is generated first.

```         
customAlign("codon",c("DPB1*01:01:01:01","DQA1*01:01:01:01","DQB1*05:01:01:01"),list(19:35,1:4,6:9))
              DPB1  19  20  21  22  23  24  25  26  27  28  29  30  31  32  33  34  35
1 DPB1*01:01:01:01 AAT GGG ACA CAG CGC TTC CTG GAG AGA TAC ATC TAC AAC CGG GAG GAG TAC
2             DQA1   1   2   3   4                                                    
3 DQA1*01:01:01:01 GAA GAC ATT GTG                                                    
4             DQB1   6   7   8   9                                                    
5 DQB1*05:01:01:01 GAT TTC GTG TAC
```

------------------------------------------------------------------------

# 4. Data Translation and Conversion Tools

## Working with Genotype List String (GL String), GL String Code (GLSC), and UNIFORMAT Data

[Genotype List (GL) String](https://onlinelibrary.wiley.com/doi/10.1111/tan.15126) is a grammar for describing the relationships between allele names reported in a genotype, using a set of six case-defined “AND” and “OR” delimiters ('?', '\^', '\|', '+', '\~', '\~', and '/').

[UNIFORMAT](https://hla-net.eu/tools/overview/) is a parallel grammar that uses an alternate set of operators (",", "\|", " ", and "\~") to describe the relationships between allele names in a genotype. Additional details of UNIFORMAT can be found at [hla-net.eu](https://hla-net.eu/tools/uniformate/).

[GL String Code](https://onlinelibrary.wiley.com/doi/10.1111/tan.15145) (GLSC) is a format that encapsulates a GL String with information describing the associated gene family namespace and the nomenclature version (or date) under which the GL String was generated. Additional information about GL String and GL String Code can be found at [glstring.org](https://glstring.org).

An example GLSC for an HLA GL String generated under IPD-IMGT/HLA Database release version 3.01.0 is presented below.

```         
hla#3.1.0#HLA-A*02:01~HLA-C*07:02~HLA-B*07:02~HLA-DRB1*15:01~HLA-DQB1*06:02~HLA-DPB1*04:02+HLA-A*01:01~HLA-C*06:02~HLA-B*57:01~HLA-DRB1*07:01~HLA-DQB1*03:03~HLA-DPB1*04:01"
```

## Translate GL String Codes Across IPD-IMGT/HLA Database Release Versions

### updateGL

The *updateGL()* function translates the HLA allele names in a GL String Code across IPD-IMGT/HLA Database release versions, going back to version 1.05.0. Allele names can be matched across release versions via the allele Accession ID, and via allele name fields shared across alleles in a given release, as illustrated below.

**Parameters**

-   *GLStringCode* A string of HLA allele names and operators in GL String Code format, signifying their relation with one another and the associated IMGT/HLA Database release version.
-   *Version* A string identifying of the desired IPD-IMGT/HLA Database release version to which the alleles should be updated.
-   *expand* A logical that indicates whether user would like to return all allele names that match the allele name in 'GLStringCode' (TRUE), or if only the direct HLA accession ID match should be returned (FALSE).
-   *verbose* A logical that indicates if messages regarding the update process should be sent to the console (TRUE) or not (FALSE). The default value is FALSE.

In the examples below, a GLSC containing a single DPB1 allele name, as recorded in release version 1.05.0, is translated to version 3.52.0. In the first example, 'expand = FALSE', and the HLA accession ID of the submitted allele is used to perform the translation to a single allele. In the second example, 'expand = TRUE' and all alleles that match the submitted allele in release 3.52.0 are returned.

```         
updateGL("hla#1.05.0#HLA-DPA1*0106", "3.52.0", expand = FALSE)
[1] "hla#3.52.0#HLA-DPA1*01:06:01"

updateGL("hla#1.05.0#HLA-DPA1*0106", "3.52.0", expand = TRUE)
[1] "hla#3.52.0#HLA-DPA1*01:06:01/HLA-DPA1*01:06:02/HLA-DPA1*01:06:03"
```

### multiUpdateGL

The *multiUpdateGL()* function translates columns of GL String Code data across IPD-IMGT/HLA Database release versions in data frames. If a single column of a data frame is provided, it must contain GL string codes. If multiple columns are entered, the first column should contain identifiers, which will not be updated. Translations can be performed for all IPD-IMGT/HLA Database release versions in the loaded *alleleListHistory* data object, going back to IPD-IMGT/HLA Database version 1.05.0.

**Parameters**

-   *GLstringArray* An array of HLA allele names and operators in GL String code format identifying their relation with one another and the pertinent IPD-IMGT/HLA Database release version.
-   *Version* A text string identifying the desired version to which for the alleles should be updated.
-   *expand* A logical specifying whether to include only the direct HLA ID match (FALSE), or all possible allele matches (TRUE). The default value is FALSE.
-   *verbose* A logical that indicates if messages regarding the update process should be sent to the console (TRUE) or not (FALSE). The default value is FALSE.

```         
GLstringArray
multiUpdateGL(GLSC.ex[[2]][1], Version = "3.53.0")
1 hla#3.53.0#HLA-A*02:01:01:01~HLA-C*07:02:01:01~HLA-B*07:02:01:01~HLA-DRB1*15:01:01:01~HLA-DQB1*06:02:01:01~HLA-DPB1*04:02:01:01+HLA-A*01:01:01:01~HLA-C*06:02:01:01~HLA-B*57:01:01:01~HLA-DRB1*07:01:01:01~HLA-DQB1*03:03:02:01~HLA-DPB1*04:01:01:01
```

## Translate vectors and data frames of HLA allele name data across IPD-IMGT/HLA Database Release Versions

The GIANT() function applies translateGLstring() to translate HLA allele name data across IPD-IMGT/HLA Database release versions, back to version 1.05.0, and functions on both vectors and data frames of HLA allele name data, including BIGDAWG-formatted HLA datasets. Data frames that are not BIGDAWG-formatted must be entirely composed of columns of HLA allele name data.

**Parameters**

- *data* A data frame or vector of HLA allele name character string data. Allele names in vectors must include locus prefixes. Data frames must include column headers identifying a single locus for the allele names in that column.
- *transFrom* A dot-formatted character string identifying the IPD-IMGT/HLA Database version (e.g. 3.59.0) under which the HLA data were generated.
- *transTo*	A dot-formatted character string identifying the IPD-IMGT/HLA Database version (e.g. 2.25.0) to which which the HLA data are to be translated.

*Translation of vectors of HLA allele names*

```         
GIANT(c("A*01:01:01:01","DQA1*01:01:01:01"),"3.56.0","2.20.0")
[1] "A*01010101"  "DQA1*010101"
GIANT(c("A*01:01:01:01","DQA1*01:01:01:01"),"3.56.0","3.00.0")
[1] "A*01:01:01:01" "DQA1*01:01:01"
GIANT(c("A*01:01:01:01","DQA1*01:01:01:01"),"3.56.0","1.09.0")
[1] "A*01011"   "DQA1*0101"
```

*Translation of BIGDAWG-formatted HLA data*

```         
sHLAdata[1:5,c(1:8,15:18)]
   Subject Status           A         A.1           C         C.1           B         B.1
1 UT900-23      0        <NA>        <NA> 01:02:01:01    02:10:06 13:01:01:01    18:01:02
2 UT900-24      0 01:01:01:01 02:01:01:01 03:07:01:01       06:05 14:01:01:01 39:02:01:01
3 UT900-25      0       02:10    03:01:02       07:12 01:02:01:01       15:20 13:01:01:01
4 UT900-26      0 01:01:01:01       02:18 08:04:01:01    12:02:01 35:09:01:01 40:05:01:01
5 UT910-01      0 25:01:01:01 02:01:01:01 15:07:01:01 03:07:01:01    51:01:03 14:01:01:01

         DPA1      DPA1.1        DPB1      DPB1.1
1 02:01:01:01 02:01:01:01      103:01 14:01:01:01
2 01:03:01:01    02:01:03 14:01:01:01 14:01:01:01
3 01:03:01:01 01:03:01:01      103:01       58:01
4 01:03:01:01 01:03:01:01 14:01:01:01      103:01
5 01:03:01:01 02:01:01:01 14:01:01:01       58:01

GIANT(sHLAdata,"3.56.0","2.10.0")[1:5,c(1:8,15:18)]
   Subject Status        A      A.1      C    C.1      B    B.1   DPA1 DPA1.1 DPB1 DPB1.1
1 UT900-23      0     <NA>     <NA> 010201 020205   1301 180102 020101 020101 0403   1401
2 UT900-24      0 01010101 02010101   0307   0605   1401 390201 010301 020103 1401   1401
3 UT900-25      0     0210   030102   0712 010201   1520   1301 010301 010301 0403   5801
4 UT900-26      0 01010101     0218   0804 120201 350901   4005 010301 010301 1401   0403
5 UT910-01      0   250101 02010101   1507   0307 510103   1401 010301 020101 1401   5801

GIANT(sHLAdata,"3.56.0","1.15.0")[1:5,c(1:8,15:18)]
DPB1*103:01 has been removed from the Allele List or has had its name changed.
DPB1*103:01 has been removed from the Allele List or has had its name changed.
   Subject Status     A   A.1    C   C.1     B   B.1  DPA1 DPA1.1 DPB1 DPB1.1
1 UT900-23      0  <NA>  <NA> 0102 02025  1301 18012 02011  02011 <NA>   1401
2 UT900-24      0 01011 02011 0307  0605  1401 39021 01031  02013 1401   1401
3 UT900-25      0  0210 03012 0712  0102  1520  1301 01031  01031 <NA>   5801
4 UT900-26      0 01011  0218 0804 12021 35091  4005 01031  01031 1401   <NA>
5 UT910-01      0  2501 02011 1507  0307 51013  1401 01031  02011 1401   5801
```

In these examples, the DPB1\*103:01 allele was originally named DPB1\*04:03, but this allele was not identified until release version 2.0.0.

## Convert GL String-formatted data to UNIFORMAT-formatted data

### validateGLstring

The *validateGLstring()* function evaluates version 1.0 and 1.1 GL Strings for incorrect characters. TRUE is returned when all characters in the GL String are valid for the specified version. FALSE is returned when the GL String contains invalid characters for the version.

**Parameters**

-   *GLstring* A character string of GL String-formatted HLA allele names and GL String 1.0 or 1.1 operators.
-   *version* A character string of either '1.0' or '1.1' identifying the GL String version.

```         
GLstring <- "HLA-A*02:01/HLA-A*02:02?HLA-A*03:01/HLA-A*03:02"
validateGLstring(GLstring,"1.0")
HLA-A*02:01/HLA-A*02:02?HLA-A*03:01/HLA-A*03:02 contains characters or operators not permitted in GL String version 1.0.
[1] FALSE
validateGLstring(GLstring,"1.1")
[1] TRUE
```

### GLStoUNI

The *GLStoUNI()* function converts genotype data from GL String format to UNIFORMAT format. Because the [GL String '?' delimiter](https://glstring.org/syntax-1.1.html) has no UNIFORMAT cognate, this function does not support the '?' delimiter or the [GL String v1.1](https://pubmed.ncbi.nlm.nih.gov/37286192) format.

**Parameters**

-   *GLstring* A character string of GL String-formatted HLA allele names and GL String 1.0 operators.
-   *prefix* A string containing a gene-system prefix (default is "HLA-").
-   *pre* A logical that indicates whether 'prefix' should be appended to the allele names (TRUE) or not (FALSE). The default value is FALSE.

```         
GLStoUNI("HLA-A*02:01/HLA-A*02:02+HLA-A*03:01/HLA-A*03:02")
[1] "A*02:01,A*03:01|A*02:01,A*03:02|A*02:02,A*03:01|A*02:02,A*03:02"
```

### multiGLStoUNI

The *multiGLStoUNI()* function converts GL string formatted data in data frames or vectors to uniformat format. If a vector is provided, all elements must be GL Strings. If a data frame is provided, the first column is assumed to contain identifiers and will not be converted; the remaining columns should contain GL strings.

**Parameters**

-   *GLstringArray* A data frame or a vector of GL String 1.0 formatted data. If 'GLstringArray' is a data frame with more than one column, the first column should contain only identifiers. If 'GLstringArray' is a vector, it should contain only GL Strings.
-   *prefix* A string containing a gene-system prefix (default is "HLA-").
-   *pre* A logical that indicates whether 'prefix' should be appended to the allele names (TRUE) or not (FALSE). The default value is FALSE.

**Return** If 'GLstringArray' is a data frame, a data frame of UNIFORMAT values is returned. If 'GLstringArray' is a vector, a vector of UNIFORMAT values is returned.

*Example GL String data frame*

```         
GLstring.ex[1:3,]
Relation    Gl.String
Subject    1 HLA-A*02:01~HLA-C*07:02~HLA-B*07:02~HLA-DRB1*15:01~HLA-DQB1*06:02~HLA-DPB1*04:02+HLA-A*01:01~HLA-C*06:02~HLA-B*57:01~HLA-DRB1*07:01~HLA-DQB1*03:03~HLA-DPB1*04:01
Subject    2 HLA-A*03:01~HLA-C*07:01~HLA-B*49:01~HLA-DRB1*04:05~HLA-DQB1*03:02~HLA-DPB1*02:01+HLA-A*01:01~HLA-C*07:01~HLA-B*08:01~HLA-DRB1*13:01~HLA-DQB1*06:03~HLA-DPB1*04:01
Subject    3 HLA-A*11:01~HLA-C*04:01~HLA-B*15:01~HLA-DRB1*04:01~HLA-DQB1*03:02~HLA-DPB1*04:02+HLA-A*03:01~HLA-C*08:02~HLA
```

*Example GL String vectors*

```         
GLstring.ex[2][1:3,]
[1] "HLA-A*02:01~HLA-C*07:02~HLA-B*07:02~HLA-DRB1*15:01~HLA-DQB1*06:02~HLA-DPB1*04:02+HLA-A*01:01~HLA-C*06:02~HLA-B*57:01~HLA-DRB1*07:01~HLA-DQB1*03:03~HLA-DPB1*04:01"
[2] "HLA-A*03:01~HLA-C*07:01~HLA-B*49:01~HLA-DRB1*04:05~HLA-DQB1*03:02~HLA-DPB1*02:01+HLA-A*01:01~HLA-C*07:01~HLA-B*08:01~HLA-DRB1*13:01~HLA-DQB1*06:03~HLA-DPB1*04:01"
[3] "HLA-A*11:01~HLA-C*04:01~HLA-B*15:01~HLA-DRB1*04:01~HLA-DQB1*03:02~HLA-DPB1*04:02+HLA-A*03:01~HLA-C*08:02~HLA-B*14:02~HLA-DRB1*13:02~HLA-DQB1*06:09~HLA-DPB1*04:01"
```

*Convert GL String to UNIFORMAT*

```         
multiGLStoUNI(GLstring.ex[1:5,1:2])
  Relation                                                                                                         Gl.String
1  Subject A*02:01~C*07:02~B*07:02~DRB1*15:01~DQB1*06:02~DPB1*04:02,A*01:01~C*06:02~B*57:01~DRB1*07:01
~DQB1*03:03~DPB1*04:01
2  Subject A*03:01~C*07:01~B*49:01~DRB1*04:05~DQB1*03:02~DPB1*02:01,A*01:01~C*07:01~B*08:01~DRB1*13:01
~DQB1*06:03~DPB1*04:01
3  Subject A*11:01~C*04:01~B*15:01~DRB1*04:01~DQB1*03:02~DPB1*04:02,A*03:01~C*08:02~B*14:02~DRB1*13:02
~DQB1*06:09~DPB1*04:01
4  Subject A*68:01~C*15:02~B*40:06~DRB1*16:02~DQB1*05:02~DPB1*10:01,A*68:01~C*06:02~B*45:01~DRB1*04:05
~DQB1*03:02~DPB1*11:01
5  Subject A*02:01~C*05:01~B*44:02~DRB1*15:01~DQB1*06:02~DPB1*02:01,A*30:01~C*06:02~B*13:02~DRB1*07:01
~DQB1*02:02~DPB1*17:01

multiGLStoUNI(GLstring.ex[[2]][1:5])
[1] "A*02:01~C*07:02~B*07:02~DRB1*15:01~DQB1*06:02~DPB1*04:02,A*01:01~C*06:02~B*57:01~
DRB1*07:01~DQB1*03:03~DPB1*04:01"
[2] "A*03:01~C*07:01~B*49:01~DRB1*04:05~DQB1*03:02~DPB1*02:01,A*01:01~C*07:01~B*08:01~
DRB1*13:01~DQB1*06:03~DPB1*04:01"
[3] "A*11:01~C*04:01~B*15:01~DRB1*04:01~DQB1*03:02~DPB1*04:02,A*03:01~C*08:02~B*14:02~
DRB1*13:02~DQB1*06:09~DPB1*04:01"
[4] "A*68:01~C*15:02~B*40:06~DRB1*16:02~DQB1*05:02~DPB1*10:01,A*68:01~C*06:02~B*45:01~
DRB1*04:05~DQB1*03:02~DPB1*11:01"
[5] "A*02:01~C*05:01~B*44:02~DRB1*15:01~DQB1*06:02~DPB1*02:01,A*30:01~C*06:02~B*13:02~
DRB1*07:01~DQB1*02:02~DPB1*17:01"
```

## Convert UNIFORMAT-formatted data to GL String-formatted data

### validateUniformat

The *validateUniformat()* function evaluates UNIFORMAT data for incorrect characters. TRUE is returned when all characters in the UNIFORMAT data are valid for the specified version. FALSE is returned when the UNIFORMAT data contain invalid characters.

**Parameters**

-   *uniformat* A character string of allele names and operators in the UNIFORMAT format.

```         
validateUniformat("A*02:01,A*03:01|A*02:01,A*03:02|A*02:02,A*03:01|A*02:02,A*03:02")
[1] TRUE

validateUniformat("A*02:01+A*03:01|A*02:01,A*03:02|A*02:02,A*03:01|A*02:02,A*03:02")
A*02:01+A*03:01|A*02:01,A*03:02|A*02:02,A*03:01|A*02:02,A*03:02 contains operators or characters not permitted in UNIFORMAT.
[1] FALSE
```

### UNItoGLS

The *UNItoGLS()* function converts a single UNIFORMAT string to GL String format.

**Parameters**

-   *uniformat* A string of HLA allele names and operators in the UNIFORMAT format signifying their relation with one another.
-   *prefix* A character string of the desired gene-system prefix (default is "HLA-").
-   *pre* A logical that indicates whether returned allele names should contain 'prefix' (TRUE), or if 'prefix' should be excluded from returned names (FALSE).

```         
UNItoGLS("A*02:01,A*03:01|A*02:01,A*03:02|A*02:02,A*03:01|A*02:02,A*03:02")
[1] "HLA-A*02:01/HLA-A*02:02+HLA-A*03:01/HLA-A*03:02"
```

### multiUNItoGLS

The *multiUNItoGLS()* function translates columns of arrays from uniformat format to GL string format. If one column is entered, it must be in UNIFORMAT format. If multiple columns are entered, the first column (furthest left) should contain an identification factor of some kind, as the first column will not be translated, and the remaining columns should contain UNIFORMAT strings. Examples of input formats are presented below.

*Example UNIFORMAT Data Frame*

```         
> UNIFORMAT.example[1:3,]
   sample.id  genotype
1 hhrv_id195  blank,A*02:01|A*02:01,A*02:01 blank,B*44:03|B*44:03,B*44:03 blank,DRB1*07:01|DRB1*07:01,DRB1*07:01
2 hhrv_id454  A*02:01,A*23:01|A*02:01,A*23:17 B*27:05,B*44:03|B*27:13,B*44:03 DRB1*07:01,DRB1*11:04
3 hhrv_id642  A*66:01,A*31:08|A*31:08,A*66:08 B*27:05,B*07:02|B*27:05,B*07:61|B*27:13,B*07:02|B*27:13,B*07:61 DRB1*01:01,DRB1*16:02
```

*Example UNIFORMAT Vector*

```         
> UNIFORMAT.example[2][1:3,]
[1] "blank,A*02:01|A*02:01,A*02:01 blank,B*44:03|B*44:03,B*44:03 blank,DRB1*07:01|DRB1*07:01,DRB1*07:01"
[2] "A*02:01,A*23:01|A*02:01,A*23:17 B*27:05,B*44:03|B*27:13,B*44:03 DRB1*07:01,DRB1*11:04"
[3] "A*66:01,A*31:08|A*31:08,A*66:08 B*27:05,B*07:02|B*27:05,B*07:61|B*27:13,B*07:02|B*27:13,B*07:61 DRB1*01:01,DRB1*16:02"
```

**Parameters**

-   *UniformatArray* A data frame or vector of UNIFORMAT formatted strings.
-   *prefix* A character string of the desired gene-system prefix (default is "HLA-").
-   *pre* A logical that indicates whether returned allele names should contain 'prefix' (TRUE), or if 'prefix' should be excluded from returned names (FALSE).

**Return** If 'uniformat'is a data frame, a data frame of GL String values is returned. If 'uniformat' is a vector, a vector of GL String values is returned.

*Convert UNIFORMAT to GL String*

```         
multiUNItoGLS(UNIFORMAT.example[1:3,])
   sample.id                                                                                                          genotype
1 hhrv_id195       HLA-blank/HLA-A*02:01+HLA-A*02:01^HLA-blank/HLA-B*44:03+HLA-B*44:03^HLA-blank/HLA-DRB1*07:01+HLA-DRB1*07:01
2 hhrv_id454             HLA-A*02:01+HLA-A*23:01/HLA-A*23:17^HLA-B*27:05/HLA-B*27:13+HLA-B*44:03^HLA-DRB1*07:01+HLA-DRB1*11:04
3 hhrv_id642 HLA-A*66:01/HLA-A*66:08+HLA-A*31:08^HLA-B*27:05/HLA-B*27:13+HLA-B*07:02/HLA-B*07:61^HLA-DRB1*01:01+HLA-DRB1*16:02

multiUNItoGLS(UNIFORMAT.example[2][1:3,])
[1] "HLA-blank/HLA-A*02:01+HLA-A*02:01^HLA-blank/HLA-B*44:03+HLA-B*44:03^HLA-blank/HLA-DRB1*07:01+HLA-DRB1*07:01"      
[2] "HLA-A*02:01+HLA-A*23:01/HLA-A*23:17^HLA-B*27:05/HLA-B*27:13+HLA-B*44:03^HLA-DRB1*07:01+HLA-DRB1*11:04"            
[3] "HLA-A*66:01/HLA-A*66:08+HLA-A*31:08^HLA-B*27:05/HLA-B*27:13+HLA-B*07:02/HLA-B*07:61^HLA-DRB1*01:01+HLA-DRB1*16:02"
```

*Convert GL String to UNIFORMAT*

```         
multiGLStoUNI(GLstring.ex[1:5,1:2])
  Relation                                                                                                         Gl.String
1  Subject A*02:01~C*07:02~B*07:02~DRB1*15:01~DQB1*06:02~DPB1*04:02,A*01:01~C*06:02~B*57:01~DRB1*07:01
~DQB1*03:03~DPB1*04:01
2  Subject A*03:01~C*07:01~B*49:01~DRB1*04:05~DQB1*03:02~DPB1*02:01,A*01:01~C*07:01~B*08:01~DRB1*13:01
~DQB1*06:03~DPB1*04:01
3  Subject A*11:01~C*04:01~B*15:01~DRB1*04:01~DQB1*03:02~DPB1*04:02,A*03:01~C*08:02~B*14:02~DRB1*13:02
~DQB1*06:09~DPB1*04:01
4  Subject A*68:01~C*15:02~B*40:06~DRB1*16:02~DQB1*05:02~DPB1*10:01,A*68:01~C*06:02~B*45:01~DRB1*04:05
~DQB1*03:02~DPB1*11:01
5  Subject A*02:01~C*05:01~B*44:02~DRB1*15:01~DQB1*06:02~DPB1*02:01,A*30:01~C*06:02~B*13:02~DRB1*07:01
~DQB1*02:02~DPB1*17:01

multiGLStoUNI(GLstring.ex[[2]][1:5])
[1] "A*02:01~C*07:02~B*07:02~DRB1*15:01~DQB1*06:02~DPB1*04:02,A*01:01~C*06:02~B*57:01~
DRB1*07:01~DQB1*03:03~DPB1*04:01"
[2] "A*03:01~C*07:01~B*49:01~DRB1*04:05~DQB1*03:02~DPB1*02:01,A*01:01~C*07:01~B*08:01~
DRB1*13:01~DQB1*06:03~DPB1*04:01"
[3] "A*11:01~C*04:01~B*15:01~DRB1*04:01~DQB1*03:02~DPB1*04:02,A*03:01~C*08:02~B*14:02~
DRB1*13:02~DQB1*06:09~DPB1*04:01"
[4] "A*68:01~C*15:02~B*40:06~DRB1*16:02~DQB1*05:02~DPB1*10:01,A*68:01~C*06:02~B*45:01~
DRB1*04:05~DQB1*03:02~DPB1*11:01"
[5] "A*02:01~C*05:01~B*44:02~DRB1*15:01~DQB1*06:02~DPB1*02:01,A*30:01~C*06:02~B*13:02~
DRB1*07:01~DQB1*02:02~DPB1*17:01"
```

------------------------------------------------------------------------

# 5. Data Analysis Tools

The functions described in this section either perform data analyses or generate specific data sets suitable for additional analyses. All consume BIGDAWG-formatted datasets.

### Example Data

To facilitate testing and experimentation with these functions, the 'sHLAdata' data object is included in the package. This BIGDAWG-formatted HLA genotype data object represents completely synthetic IPD-IMGT/HLA release 3.56.0 HLA-A, -C, -B, -DRB1, -DQA1, -DQB1, -DPA1, and -DPB1 genotype data for 24 control subjects and 23 case subjects, and does not represent any true human population.

## Calculating Relative Risk with BIGDAWG-formatted Datasets

### relRisk

The relRisk() function calculates relative risk (RR) for individual alleles and genotypes in [BIGDAWG-formatted datasets](https://CRAN.R-project.org/package=BIGDAWG/vignettes/BIGDAWG.html#input-data). RR analysis is intended for *non-case-control* datasets. While two subject categories (0 and 1) are required in the **Subject** column of the BIGDAWG-formatted datafile, the categories should *not* be patients and controls. Instead, the categories may be, e.g., two groups of patients with the same disease, where one group has more severe symptoms (coded as 1), and other has mild symptoms (coded as 0). In this case, relRisk() identifies the risk of sefverity associated with a given allele or genotype.

relRisk() returns a list object of two lists ("alleles" and "genotypes"), each of which contains relative risk, confidence interval and p-value data for the individual alleles and individual genotypes at each locus in a BIGDAWG-formatted non-case-control genotype data frame or tab-delimited text file.

**Parameters**

-   *dataset* The name of a non-case-control genotype dataset using the BIGDAWG format.
-   *return* A logical identifying if the list object should be returned (return=TRUE), or if pairs of tab-delimited text files of results (one for alleles and one for genotypes) should be written to the working directory for each locus.
-   *save.path* A character string identifying the path in which to write the pair of files when return is FALSE. The default value is tempdir().

```         
rr <- relRisk(sHLAdata[,1:4])
```

Column headers in each returned data frame are, Locus, Variant, Status_1, Status_0, RelativeRisk, CI.low, CI.high, p.value, and Significant.

## Division of BIGDAWG-formatted Datasets for Stratification Analyses

### BDstrat: Multilocus Allele Stratification for BIGDAWG Datasets

The BDstrat() function divides a [BIGDAWG-formatted](https://CRAN.R-project.org/package=BIGDAWG/vignettes/BIGDAWG.html#input-data)] dataset into two strata, one of which contains all subjects with any of a set of specified alleles, while the other contains all subjects without any of the specified alleles. If the positive-stratum and negative-stratum datasets have sufficient numbers of case and control (or exposed and unexposed) subjects, each stratum can be analyzed via the BIGDAWG package or relRisk(). Sample size is an important consideration when determining how many alleles on which to stratify.

**Parameters**

-   *dataset*: A BIGDAWG-formatted data frame or a path to a BIGDAWG-formatted, tab-delimited text file.
-   *alleles*: A vector of allele names in the <locus>\*<allele name> format (e.g., "A\*01:01:01:01" for HLA alleles or "KIR2DL1\*001" for KIR alleles).
-   *warnBelow* : An integer that defines a low number of subjects in a stratum, generating a warning message. The default value is 21.

**Return** A list-object of two BIGDAWG-formatted data frames titled 'dataset'\$'alleles'-positive' and 'dataset'\$'alleles'-negative\`. The positive list element includes all subjects with the specified alleles, and the negative list element includes all subjects without those specified alleles.

```         
Stratify on a single allele in the HLA_data dataset bundled with BIGDAWG
HLA_data.single.strat <- BDstrat(sHLAdata,"DRB1*08:02:01:01") 

Stratify on three alleles at two loci in the HLA_data dataset bundled with BIGDAWG
HLA_data.multi.strat <- BDstrat(sHLAdata,c("DRB1*04:02:01","DRB1*04:07:01:01","A*32:04"))

Run BIGDAWG() on both multi-allele strata
for(i in 1:2) {BIGDAWG(HLA_data.multi.strat[[i]],HLA = TRUE,Run.Tests = "L")}
```

## Conversion of BIGDAWG-formatted Datasets to *Python for Population Genetics* (PyPop)-formatted Datasets

#### BDtoPyPop : Enabling Population Genetic Analyses of Case and Control Datasets

-   BDtoPyPop() converts a [BIGDAWG-formatted](https://CRAN.R-project.org/package=BIGDAWG/vignettes/BIGDAWG.html#input-data) case-control dataset into two [PyPop version 1.\*.\*](http://pypop.org/docs/guide-chapter-usage.html) datasets -- one for all 'case' subjects and one for all 'control' subjects. PyPop datasets can be exported as a pair of ".pop" files for analysis, or can be generated as a list object containing two data frames for additional manipulation before analysis.

**Parameters**

-   *dataset* A BIGDAWG-formatted data frame containing a case-control dataset.
-   *filename* A character string identifying either the desired path and name for the generated files, or the elements of the returned list object.
-   *save.file* A logical indicating if the pair of files should be written (TRUE) or if a list object should be returned. The default value is TRUE.
-   *save.path* A character string identifying the path in which to write the pair of files when return is FALSE. The default value is tempdir().

**Return** When save.file = TRUE a pair of PyPop-formatted text files named "'filename'.positive.pop" and "'filename'.negative.pop" are generated in the specified directory. The "'filename'.positive.pop" file includes all subjects with a value of 1 in the second column of the BIGDAWG data file, and the "'filename'.negative.pop" file includes all subjects with a value of 0 in the second column of the BIGDAWG data file.

When save.file = FALSE, a list-object of two PyPop-formatted data frames titled "'filename'.positive" and "'filename'-negative" is returned. The positive list element includes all subjects a value of 1 in the second column of the BIGDAWG data file, and the negative list element includes all subjects with a value of 0 in the second column of the BIGDAWG data file.

**Note** PyPop is not an R package. In order to analyze the files generated by this function with PyPop version 1.\*.\*, [the pypop-genomics package must be installed](http://pypop.org/docs/guide-chapter-install.html) on your system, and a [PyPop configuration file](http://pypop.org/docs/guide-chapter-usage.html#the-configuration-file), detailing the analyses to be performed, is required. A PyPop configuration file is not generated by BDtoPyPop().

```         
Generating a list object of two PyPop datasets from the HLA_data dataset bundled with BIGDAWG.
HLAdata.PP <- BDtoPyPop(sHLAdata,"BDHLA",FALSE)
 
Generating a pair PyPop data files from the HLA_data dataset bundled with BIGDAWG.
BDtoPyPop(sHLAdata,"BDHLA",TRUE)
```

------------------------------------------------------------------------

::: {align="center"}
END OF VIGNETTE
:::

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/VersionValidation.R
\name{squashVersion}
\alias{squashVersion}
\title{Reduce a Release Version to Numerals}
\usage{
squashVersion(ver, num = FALSE)
}
\arguments{
\item{ver}{A character vector-formatted IPD-IMGT/HLA Database release version name (e.g., '3.54.0').}

\item{num}{A Logial. A numeric value is returned when num = TRUE. A character vector is returned when num = FALSE. The default value is FALSE.}
}
\value{
A 4-digit value, as either a character string or a number.
}
\description{
Removes the '.' delimiters from an IPD-IMGT/HLA Database release version name.
}
\note{
For internal HLAtools use.
}
\examples{
squashVersion("3.45.0",TRUE)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/BDtoPyPop.R
\name{BDtoPyPop}
\alias{BDtoPyPop}
\title{Convert BIGDAWG datasets to PyPop datasets

Converts a BIGDAWG-formatted data frame into a pair of PyPop-formatted case and control data frames.}
\usage{
BDtoPyPop(dataset, filename, save.file = TRUE, save.path = tempdir())
}
\arguments{
\item{dataset}{A data frame containing a BIGDAWG-formatted case-control dataset}

\item{filename}{A character string identifying the desired path and name for the generated files or the elements of the returned list object.}

\item{save.file}{A logical that determines if a pair of files should be written (TRUE) or if a list object should be returned. The default value is TRUE.}

\item{save.path}{A character string identifying the path in which to write the pair of files when save.file is TRUE. The default value is tempdir().}
}
\value{
When save.file = TRUE, a pair of files named "'filename'.positive.pop" and "'filename'.negative.pop" are generated in the working directory. When save.file = FALSE, a list of two-elements, named "'filename'.positive" and "'filename'.negative", is returned.
}
\description{
Convert BIGDAWG datasets to PyPop datasets

Converts a BIGDAWG-formatted data frame into a pair of PyPop-formatted case and control data frames.
}
\note{
Files generated by BDtoPyPop are intended for analysis using Pypop version 1.#.#. PyPop is not an R package, and must be installed separately.
}
\examples{
HLAdata.PP <- BDtoPyPop(sHLAdata,"BDHLA",FALSE)

}
\references{
Lancaster et al. Front Immunol. 2024 Apr 2;15:1378512. https://pubmed.ncbi.nlm.nih.gov/38629078/

Pappas et al. Hum Immunol. 2016 Mar 77(3):283-287. https://pubmed.ncbi.nlm.nih.gov/26708359/
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{GLstring.ex}
\alias{GLstring.ex}
\title{Example Data Frame of Genotype List String Data.}
\format{
A dataframe with 419 rows and 2 columns.
}
\source{
pould::hla.hap.demo
}
\usage{
data(GLstring.ex)
}
\description{
A two-column example data frame (from pould::hla.hap.demo) including GL String Data.
Column one identifies each subject's status, while column two identifies each subject's HLA genotype in GL String format.
This data is provided for use in examples and demonstrations.
}
\references{
Mack et al. HLA 2023;102(2):206-212 https://doi.org/10.1111/tan.15126
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/BuildAlleleListHistory.R
\name{updateAlleleListHistory}
\alias{updateAlleleListHistory}
\title{Build the AlleleList History R Object}
\usage{
updateAlleleListHistory()
}
\value{
A list object containing two data frames. alleleListHistory$Version identifies the date and version of alleleListHistory$AlleleListHistory. alleleListHistory$AlleleListHistory relates each IPD-IMGT/HLA Database Accession ID (HLA_ID) to the corresponding HLA allele name in each release version since 1.05.0.
}
\description{
Consumes the ANHIG/IMGTHLA GitHub repository's allelelisthistory.txt file to create the AlleleListHistory object.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/GLStoUNI.R
\name{multiGLStoUNI}
\alias{multiGLStoUNI}
\title{Translate Multiple GL Strings to UNIFORMAT}
\usage{
multiGLStoUNI(GLstringArray, prefix = "HLA-", pre = FALSE)
}
\arguments{
\item{GLstringArray}{A data frame or a vector containing GL String formatted data. If 'GLstringArray' is a data frame with more than one column, the first column should contain only identifiers. If 'GLstringArray' is a vector, it should contain only GL Strings.}

\item{prefix}{A string of the desired locus prefix (default is "HLA-").}

\item{pre}{A logical. If 'pre' is TRUE, all allele names will be prefixed with 'prefix'. If 'pre' is FALSE, no allele names will be prefixed.}
}
\value{
A version of 'GLStringArray' in which the GL String data has been converted to UNIFORMAT format. If a 'GLstringArray' was a data frame, a data frame is returned. If 'GLstringArray' was a vector, a vector is returned.
}
\description{
Translate a data frame or a vector of GL Strings to UNIFORMAT strings.
}
\note{
GL Strings that include the "?" operator will not be translated, as the "?" operator has no cognate in UNIFORMAT.
}
\examples{
multiGLStoUNI(GLstring.ex[[2]][1:5],version) ## converting a vector of GL Strings to UNIFORMAT
multiGLStoUNI(GLstring.ex[1:5,1:2])  ## converting a data.frame of GL Strings to UNIFORMAT

}
\references{
Nunes Tissue Antigens 2007;69(s1):203-205 https://doi.org/10.1111/j.1399-0039.2006.00808.x

Mack et al. HLA 2023;102(2):206-212 https://doi.org/10.1111/tan.15126
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/UNItoGLS.R
\name{validateUniformat}
\alias{validateUniformat}
\title{Validate a UNIFORMAT String}
\usage{
validateUniformat(uniformat)
}
\arguments{
\item{uniformat}{A character string of allele names and operators in the UNIFORMAT format.}
}
\value{
A logical. TRUE is returned when all characters in 'uniformat' are permitted. FALSE is returned when forbidden characters are present.
}
\description{
Evaluates a UNIFORMAT string to identify unsupported characters.
}
\examples{
validateUniformat("A*02:01,A*03:01|A*02:01,A*03:02|A*02:02,A*03:01|A*02:02,A*03:02")

}
\references{
Nunes Tissue Antigens 2007;69(s1):203-205 https://doi.org/10.1111/j.1399-0039.2006.00808.x
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/GLupdater.R
\name{GLvalidate}
\alias{GLvalidate}
\title{Validate a GL String Code}
\usage{
GLvalidate(GLString, namespace = c("hla", "kir"))
}
\arguments{
\item{GLString}{A character string describing a string formatted to GL String Code specifications.}

\item{namespace}{A character vector of allowed namespace strings. The default value is 'c("hla","kir")'.}
}
\value{
A character string describing either a well-formatted GL String Code or the value FALSE.
}
\description{
A lightweight validator that inspects a GL String Code for correct structure. If the namespace field contains a value other than value of 'namespace', the namespace is changed to 'hla'. The version and GL String fields are not evaluated.
}
\note{
For internal use only.
}
\examples{
GLvalidate("ha#3.25.0#hla-B15:35") 

}
\references{
Mack et al. HLA 2023;102(2):206-212 https://doi.org/10.1111/tan.15126

Mack et al. HLA 2023;102(4):501-507 https://doi.org/10.1111/tan.15145
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/VersionValidation.R
\name{getLatestVersion}
\alias{getLatestVersion}
\title{Identify the Latest IPD-IMGT/HLA Database Release}
\usage{
getLatestVersion()
}
\value{
A dot-delimited character string identifying the latest release version (branch) of the ANHIG/IMGTHLA Github repository.
If the most recent release version is not a 'primary' release version (i.e., following the X.YY.O naming format), the integer
in the third field of the release version is replaced with a '0'.
}
\description{
Identifies the most recent version of the IPD-IMGT/HLA Database available in the ANHIG/IMGTHLA Github repository.
}
\note{
For internal HLAtools use.
}
\examples{
getLatestVersion()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/SearchandBuildAlignments.R
\name{customAlign}
\alias{customAlign}
\title{Generate a Customized Peptide, Codon or Nucleotide Sequence Alignment.}
\usage{
customAlign(alignType, alleles, positions)
}
\arguments{
\item{alignType}{The type of alignment being searched. Allowed values are "prot", codon", "nuc" and "gen". Only one 'alignType' value is allowed.}

\item{alleles}{A vector of un-prefixed HLA allele names.}

\item{positions}{Either a vector of variant positions, against which all loci will be aligned, or a list of vectors of nucleotide positions, exactly one vector for each allele, against which each corresponding allele will be aligned.}
}
\value{
A data frame of allele names and the corresponding nucleotide sequences for each desired nucleotide position. FALSE is returned when 'positions' includes values that do not exist in the specified alignment, and when a specified locus is not available in the HLAalignments object.
}
\description{
Generates a peptide, codon, coding or genomic nucleotide alignment table for user-specified HLA alleles at user-specified positions.
}
\note{
This function requires that the HLAalignments object has been populated with alignments via the alignmentFull() function.
}
\examples{
\dontrun{
customAlign("codon",c("DRB1*01:01","DQB1*02:01","DPB1*01:01"),c(1,2,3,7,8,9,13,14,15))
customAlign("codon",c("DPB1*01:01:01:01","DQA1*01:01:01:01","DQB1*05:01:01:01"),list(19:35,1:4,6:9))
customAlign("prot",c("DRB1*08:07","DPB1*14:01"),list(c(12,27,30,39,57,84),c(15,16,39,60,67,81)))
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/extractGeneTypes.R
\name{buildIMGTHLAGeneTypes}
\alias{buildIMGTHLAGeneTypes}
\title{Describe IPD-IMGT/HLA Database Genes, Identifying Pseudogenes and Gene Fragments}
\usage{
buildIMGTHLAGeneTypes()
}
\value{
A list object of two elements -- 'version' and 'GeneTypes'. The 'version' element identifies the date on which the source table was generated. The 'GeneTypes' element is a data frame of three columns, identifying each gene supported by the IPD-IMGT/HLA Database, along with its molecular characteristics and its status as either a pseudogene or a gene fragment.
}
\description{
buildIMGTHLAGeneTypes() scrapes 'hla.alleles.org/pages/genes/genes_list' and generates a data frame identifying each gene. It requires internet access to function. As such, this function always returns data for the current IPD-IMGT/HLA Database release.
}
\note{
For internal HLAtools use.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/compareSequences.R
\name{compareSequences}
\alias{compareSequences}
\title{Identify Sequence Differences Between Two Alleles}
\usage{
compareSequences(alignType, alleles)
}
\arguments{
\item{alignType}{A character string identifying the type of alignment being searched. Allowed values are "codon","gen", nuc" and "prot". Only one 'alignType' value is permitted.}

\item{alleles}{A character vector containing two full-length names for alleles at the two different loci.}
}
\value{
A two-row data frame identifying the positions and sequences at which the two alleles differ. E.g., compareLoci(alignType = "gen", alleles = c("B\*07:02:01:01","C\*02:02:02:01")). Positions are nor shared between the two loci, or for which the sequence of either allele is unknown are ignored.
}
\description{
Compares the sequences of two alleles, and identifies the differences between them at shared polymorphic sequence positions.
Alleles at different loci can be compared.
}
\examples{
\dontrun{
compareSequences("gen",c("DPA1*01:03:38:01","DPA1*01:03:38:02"))
compareSequences("prot",c("DRB1*01:01:01:01","DRB3*01:01:02:01"))
compareSequences("prot", c("B*57:01:01:01", "A*24:02:01:01"))
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/VersionValidation.R
\name{checkVersion}
\alias{checkVersion}
\title{Check IPD-IMGT/HLA Release Version Allele Names}
\usage{
checkVersion(version)
}
\arguments{
\item{version}{A character string identifying the release version (branch) of the ANHIG/IMGTHLA Github repository (e.g. '3.53.0'). The value 'Latest' refers to the most recent release.}
}
\value{
A logical. TRUE indicates that data for 'version' is available to the local HLAtools package. FALSE means that data for 'version' is not available.
}
\description{
Determines if allele name data for a given IPD-IMGT/HLA Release Version is present in the local HLAtools package.
}
\note{
For internal HLAtools use.
}
\examples{
checkVersion(version = "3.25.0")
checkVersion(version = "Latest")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/UNItoGLS.R
\name{UNItoGLS}
\alias{UNItoGLS}
\title{Translate UNIFORMAT to GL String}
\usage{
UNItoGLS(uniformat, prefix = "HLA-", pre = TRUE)
}
\arguments{
\item{uniformat}{A character string of HLA allele names and operators in the UNIFORMAT format signifying their relation with one another.}

\item{prefix}{A character string of the desired gene-system prefix (default is "HLA-").}

\item{pre}{A logical that indicates whether returned allele names should contain 'prefix' (TRUE), or if 'prefix' should be excluded from returned names (FALSE).}
}
\value{
An version of 'uniformat' converted to GL String format, or FALSE if 'uniformat' is invalid.
}
\description{
A wrapper function for UNtoGL() which translates strings from UNIFORMAT format to GL String format.
}
\note{
This function does not return the "?" operator, as the "?" operator has no cognate in UNIFORMAT.
}
\examples{
UNItoGLS("A*02:01,A*03:01|A*02:01,A*03:02|A*02:02,A*03:01|A*02:02,A*03:02")
UNItoGLS("A,B|A,C~D G|E W,X|W,Y Z,J J,Z")

}
\references{
Nunes Tissue Antigens 2007;69(s1):203-205 https://doi.org/10.1111/j.1399-0039.2006.00808.x

Mack et al. HLA 2023;102(2):206-212 https://doi.org/10.1111/tan.15126
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{fragmentFeatureNames}
\alias{fragmentFeatureNames}
\title{Gene Features of HLA Pseudogenes and Gene Fragments}
\format{
A list of 15 elements that identify the gene features for each locus wih a genomic alignment and annotate non-standard features
}
\source{
https://hla.alleles.org/genes/index.html

https://github.com/ANHIG/IMGTHLA/tree/Latest/alignments
}
\usage{
data(fragmentFeatureNames)
}
\description{
A list object of 15 elements. Each of the 15 elements corresponds to one of the 15 HLA pseudogenes and gene fragments, and contains two items. The first item identifies the gene features for that locus. The second item contains an annotation detailing information regarding non-standard gene structures.
This object is built by the ffN() function.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/GLupdater.R
\name{updateGL}
\alias{updateGL}
\title{Update a GL String Code to a Specified IPD-IMGT/HLA Database Version}
\usage{
updateGL(GLSC, to, expand = FALSE, verbose = FALSE)
}
\arguments{
\item{GLSC}{A GL String Code containing HLA allele names.}

\item{to}{A character string identifying the IPD-IMGT/HLA Database release version to translate 'GLSC' to. Values can range from version 1.05.0 to the loaded version of the alleleListHistory object.}

\item{expand}{A logical that indicates if the lowest-numbered truncated allele names that match truncated allele names in 'GLSC' should be returned (expand = FALSE), or if a slash-delimited string of all matching full-length allele names should be returned (expand = TRUE). The default value is FALSE.}

\item{verbose}{A logical that indicates if messages regarding the update process should be sent to the console (TRUE) or not (FALSE). The default value is FALSE.}
}
\value{
A GL String Code updated the 'to' release version.
}
\description{
Updates the elements of a GLString Code across IPD-IMGT/HLA Database release versions. Truncated allele names can be expanded to the list of all allele names that contain the truncated name.
}
\examples{
updateGL(GLSC.ex$GL.String.Code[1],"2.15.0",FALSE,FALSE)
updateGL(GLSC.ex$GL.String.Code[1],"2.15.0",TRUE,FALSE)

}
\references{
Mack et al. HLA 2023;102(2):206-212 https://doi.org/10.1111/tan.15126

Mack et al. HLA 2023;102(4):501-507 https://doi.org/10.1111/tan.15145
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/UNItoGLS.R
\name{UNtoGL}
\alias{UNtoGL}
\title{Translate UNIFORMAT Strings to GL Strings}
\usage{
UNtoGL(uniformat, prefix = "HLA-", pre = TRUE)
}
\arguments{
\item{uniformat}{A character string of HLA allele names and operators in the UNIFORMAT format signifying their relation with one another.}

\item{prefix}{A character string of the desired prefix (default is "HLA-").}

\item{pre}{A logical that indicates whether user would like all allele names to contain the prefix of their choice (TRUE), or if the prefix should not be appended to allele names (FALSE).}
}
\value{
An altered version of the input character string, converted to GL String format.
}
\description{
A function that translates strings from UNIFORMAT format to GL String format.
}
\note{
For internal use only.

This function does not return the "?" operator, as the "?" operator has no cognate in UNIFORMAT.
}
\examples{
UNtoGL("A*02:01,A*03:01|A*02:01,A*03:02|A*02:02,A*03:01|A*02:02,A*03:02")
UNtoGL("A,B|A,C|D,B|D,C|E,B|E,C", pre = FALSE)

}
\references{
Nunes Tissue Antigens 2007;69(s1):203-205 https://doi.org/10.1111/j.1399-0039.2006.00808.x

Mack et al. HLA 2023;102(2):206-212 https://doi.org/10.1111/tan.15126
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/accessoryFunctions.R
\name{multiQueryRelease}
\alias{multiQueryRelease}
\title{Search Multiple Elements of Allele Names Across Release Versions}
\usage{
multiQueryRelease(rel, variants = "", all = TRUE)
}
\arguments{
\item{rel}{An IPD-IMGT/HLA Database release version, represented as either a character (e.g., "3.58.0") or a numeric (e.g., 3580) value.}

\item{variants}{A vector of character strings. The values in 'variants' can be any part of a locus or allele name (e.g., "DR", "02:01", "DRB1*08:07"). The default ("") specifies all alleles in 'rel'.}

\item{all}{A logical. When 'all' = TRUE, a vector of all alleles in 'rel' that share all elements of 'variants' is returned.  When 'all' = FALSE, the number of alleles that share all elements of 'variants' in 'rel' is returned.}
}
\value{
A character vector of all alleles that share all 'variants' in 'rel' or the number of alleles that share all 'variants' in 'rel'.
}
\description{
Searches specific release versions in the AlleleListHistory object for a set of user-defined allele-name elements.
}
\examples{
# Identify DRB1 allele names that include a 'Q' character in release 3.58.0.
multiQueryRelease("3.58.0",c("DRB1","Q"),TRUE) 
# Identify the number of null (N) DRB1 alleles.
multiQueryRelease("3.58.0",c("DRB1","N"),FALSE)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/SearchandBuildAlignments.R
\name{uniSearch}
\alias{uniSearch}
\title{Search Sequences at a Single Position for an Allele}
\usage{
uniSearch(alignType, locus, allele, position, prefix = TRUE, trimmed = FALSE)
}
\arguments{
\item{alignType}{The type of alignment being searched. Allowed values are "prot", "codon", "nuc" and "gen". Only one 'alignType' value is allowed.}

\item{locus}{A specific HLA locus.}

\item{allele}{The name of the allele being searched.}

\item{position}{The specified position.}

\item{prefix}{A logical that indicates if the position number(s) should be included in the result.}

\item{trimmed}{A logical identifying whether 'allele' is a two-field, 'trimmed' allele name (trimmed=TRUE), or a full-length name (trimmed=FALSE). The default value is FALSE.}
}
\value{
The peptide residue, codon or nucleotide sequence at specified position for specified allele as available in the ANHIG/IMGTHLA Github Repository.
}
\description{
Generates a character string of the peptide, codon or nucleotide sequence at the specified position for the specified HLA allele.
}
\note{
For internal HLAtools use.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/accessoryFunctions.R
\name{addCodonLine}
\alias{addCodonLine}
\title{Add an 'AA codon' Line to Alignments When Missing.}
\usage{
addCodonLine(cDNAalign, firstPos, afterLine = 8, codons = 25)
}
\arguments{
\item{cDNAalign}{A matrix of cDNA alignment lines, generated from an alignment file that is missing "AA codon" lines.}

\item{firstPos}{A numeric value identifying the position number of the transcript's N-terminal codon, based on a complete cDNA alignment in another release.}

\item{afterLine}{A numeric value identifying the number of the line below the first "cDNA" line in the alignment. The default value is 8.}

\item{codons}{A numeric value identifying the number of codons in each line of the nucleotide alignment. The default value is 25.}
}
\value{
A complete cDNA alignment data frame that includes "AA codon" rows.
}
\description{
Modifies cDNA alignment objects that are missing "AA codon" lines to include these lines in the correct location with the correct codon position information.
}
\note{
For internal HLAtools use.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/GLupdater.R
\name{translateGLstring}
\alias{translateGLstring}
\title{Translate a GL String across IPD-IMGT/HLA Database Release Versions}
\usage{
translateGLstring(GLstring, from, to, expand = FALSE, verbose = FALSE)
}
\arguments{
\item{GLstring}{A GL String of HLA allele names.}

\item{from}{A character string identifying the IPD-IMGT/HLA Database release version of 'GLstring'.}

\item{to}{A character string identifying the IPD-IMGT/HLA Database release version to translate 'GLstring' to. Values can range from version 1.05.0 to the loaded version of the alleleListHistory object.}

\item{expand}{A logical that indicates if the lowest-numbered truncated allele names that match truncated allele names in the GL String should be returned (expand = FALSE), or if a slash-delimited string of all matching full-length allele names should be returned (expand = TRUE). The default value is FALSE.}

\item{verbose}{A logical that indicates if messages regarding allele translations should be sent to the console (TRUE) or not (FALSE). The default value is FALSE.}
}
\value{
A GL String of allele names in the 'to' release version. If the GL String is determined to be invalid, FALSE is returned.
}
\description{
Translates a single Genotype List (GL) string across IPD-IMGT/HLA Database release versions. Truncated allele names can be expanded to the list of all allele names that contain the truncated name.
}
\examples{
translateGLstring(GLstring.ex$Gl.String[1],"3.01.0","2.15.0")
translateGLstring(GLstring.ex$Gl.String[1],"3.01.0","2.15.0",TRUE)
translateGLstring(GLstring.ex$Gl.String[1],"3.01.0","2.15.0",FALSE,TRUE)
translateGLstring(GLstring.ex$Gl.String[1],"3.01.0","2.15.0",TRUE,TRUE)

}
\references{
Mack et al. HLA 2023;102(2):206-212 https://doi.org/10.1111/tan.15126
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/GLupdater.R
\name{multiTranslateGLstring}
\alias{multiTranslateGLstring}
\title{Translate a Vector of GL Strings to a Desired IPD-IMGT/HLA Database Version}
\usage{
multiTranslateGLstring(GLstrings, from, to, expand = FALSE, verbose = FALSE)
}
\arguments{
\item{GLstrings}{A vector of GL Strings encoded in the same IPD-IMGT/HLA Database release version.}

\item{from}{A character string identifying the IPD-IMGT/HLA Database release version of 'GLstrings'.}

\item{to}{A character string identifying the IPD-IMGT/HLA Database release version to translate 'GLstrings' to. Values can range from version 1.05.0 to the loaded version of the alleleListHistory object.}

\item{expand}{A logical that indicates if the lowest-numbered truncated allele names that match truncated allele names in 'GLstrings' should be returned (expand = FALSE), or if a slash-delimited string of all matching full-length allele names should be returned (expand = TRUE). The default value is FALSE.}

\item{verbose}{A logical that indicates if messages regarding the trnslation process should be sent to the console (TRUE) or not (FALSE). The default value is FALSE.}
}
\value{
A version GLSCs (in a data frame) updated to the desired version. NA values are returned in place of alleles that are not present in 'to'.
}
\description{
Translates a column from a data frame in GL String Code format to a desired reference database version.
}
\examples{
multiGLs <- multiTranslateGLstring(GLstring.ex$Gl.String[1:50],"3.01.0","2.15.0")
multiGLs <- multiTranslateGLstring(GLstring.ex$Gl.String[1:50],"3.30.0","3.58.0")

}
\references{
Mack et al. HLA 2023;102(2):206-212 https://doi.org/10.1111/tan.15126
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{UNIFORMAT.example}
\alias{UNIFORMAT.example}
\title{Example Data Frame of UNIFORMAT Data.}
\format{
A dataframe with 20 rows and 2 columns.
}
\source{
https://hla-net.eu/wp/wp-content/uploads/example-three-loci.unif_.txt
}
\usage{
data(UNIFORMAT.example)
}
\description{
A two-column example data frame including UNIFORMAT data.
Column one contains sample identifiers, while column two identifies each subject's HLA genotype in UNIFORMAT format.
To be used for example runs and demonstrations.
}
\references{
Nunes Tissue Antigens 2007;69(s1):203-205 https://doi.org/10.1111/j.1399-0039.2006.00808.x
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/GLupdater.R
\name{GLVhelper}
\alias{GLVhelper}
\title{Locate Matches for an Incomplete IPD-IMGT/HLA Database Version}
\usage{
GLVhelper(Version)
}
\arguments{
\item{Version}{A character string of the desired version to which the alleles should be updated, going back to version 1.05.0.}
}
\value{
A list of character strings of possible matches to a given incomplete input.
}
\description{
Uses the provided version information to locate and return possible matches for an incompletely defined IPD-IMGT/HLA Database release version.
}
\note{
For internal use only.
}
\examples{
GLVhelper("2.25")
GLVhelper("3.9.0")


}
\references{
Mack et al. HLA 2023;102(2):206-212 https://doi.org/10.1111/tan.15126

Mack et al. HLA 2023;102(4):501-507 https://doi.org/10.1111/tan.15145
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/atlasMaker.R
\name{atlasMaker}
\alias{atlasMaker}
\title{Identify the Gene-Feature Boundaries in HLA Region Genes}
\usage{
atlasMaker(loci, source, version = "Latest")
}
\arguments{
\item{loci}{A character string identifying a specific HLA gene (ex. "DRB1"). More than one gene can be specified (e.g., c("DRB1","DQB1")).}

\item{source}{A character string identifying the type of alignment used to build the atlas. The allowed options are "AA", for protein alignments, "cDNA", for nucleotide alignments, and "gDNA", for genomic alignments. More than one alignment type can be specified (e.g., c("AA","gDNA")).}

\item{version}{A character string identifying the release version of the IPD-IMGT/HLA Github repository for which alignments should be generated.}
}
\value{
A list object of 'atlas' dataframes and a 'release version' character string for each locus and alignment type.
}
\description{
The 'AA' atlas identifies the amino acid (AA) positions encoded by codons that span each exon (E) boundary. The 'cDNA' atlas describes the 3' nucleotide position in the cDNA sequence that follows each exon boundary, as well as the codon that spans each exon boundary. The 'gDNA' atlas describes first 3' nucleotide position following each UTR (U), exon or intron (I) boundary. Each feature is followed by its rank (e.g U3 is the 3' UTR). Non-standard gene-features (H, J, N, and S) are detailed in the documentation for the ffN() function.
}
\details{
Builds an 'atlas' of the gene-feature (exon, intron and UTR) boundaries for IPD-IMGT/HLA loci, using IPD-IMGT/HLA Database alignments.
}
\note{
For internal HLAtools use.

Nucleotide atlases for pseudogenes will include a 'codon' row populated with NA values.

Some HLA-DQB1\*05 and \*06 alleles include a 5th exon that is not present in the DQB1\*05:01:01:01 reference allele. In this case, the Exon 4 to Exon 5 boundary is defined as 'Ins' (insertion). For all other alleles there is no E.4-5 boundary.

Boundaries for non-standard hybrid (H), join (J), novel (N) and segment (S) features may be included in gene fragment and pseudogene atlases.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/BDstrat.R
\name{BDstrat}
\alias{BDstrat}
\title{Stratify BIGDAWG Datasets by Specific Alleles}
\usage{
BDstrat(dataset, alleles, warnBelow = 21)
}
\arguments{
\item{dataset}{A BIGDAWG-formatted data frame or a path to a BIGDAWG-formatted, tab-delimited text file.}

\item{alleles}{A vector of allele names in the locus-asterisk-allele_name format (e.g., "A*01:01:01:01").}

\item{warnBelow}{An integer that defines a low number of subjects in a stratum, generating a warning message. The default value is 21.}
}
\value{
A list-object of two BIGDAWG-formatted data frames titled dataset$\verb{<alleles>-positive} and dataset$\verb{<alleles>-negative}. The positive list element includes all subjects with the specified alleles, and the negative list element includes all subjects without those specified alleles.
}
\description{
Divides a BIGDAWG-formatted dataset into two subsets (strata) that do and do not include specified alleles.
}
\examples{
HLA_data.multi.strat <- BDstrat(sHLAdata,c("DRB1*16:02:01:01","DRB1*04:07:01:01","A*25:01:01:01"))
HLA_data.single.strat <- BDstrat(sHLAdata,"DRB1*16:02:01:01")

}
\references{
\href{https://CRAN.R-project.org/package=BIGDAWG/vignettes/BIGDAWG.html#input-data}{BIGDAWG Data Format}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/accessoryFunctions.R
\name{validateAllele}
\alias{validateAllele}
\title{Validate Allele-Name Format and Presence in HLAalignments}
\usage{
validateAllele(allele)
}
\arguments{
\item{allele}{A character string of the colon-delimited HLA allele name.}
}
\value{
A logical identifying if the allele name is present in the alignments (TRUE) or, if it is not in the alignments or is not valid (FALSE).
}
\description{
Returns TRUE if an allele name is found in HLAalignments in either the 'allele_name' column of full-length allele names or the 'trimmed_allele' column of two-field allele names in the pertinent genomic alignment. Returns FALSE if the allele name is not properly formed, or if the allele name is not found in HLAalignments.
}
\note{
Messages will be returned to the console if the allele name is malformed, or the locus is invalid; e.g., 'validateAllele("C\*01:01:01:01")' or 'validateAllele("A\*01:01:01:01:01")'.

The locus being evaluated must be included in HLAalignments.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/buildGazetteer.R
\name{getAlignmentNames}
\alias{getAlignmentNames}
\title{Retrieve Alignment Filenames for HLA Genes}
\usage{
getAlignmentNames(URL)
}
\arguments{
\item{URL}{A chararacter string containing a Uniform Resource Locator (URL) identifying of the desired IPD-IMGT/HLA Database release version from which the alignment filenames should be retrieved.}
}
\value{
A character vector of all of the filenames.
}
\description{
Retrieves the filenames of the protein, nucleotide and genomic alignments available in a specific branch of the IMGTHLA GitHub Repository
}
\note{
For internal HLAtools use.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{IMGTGenomicReferences}
\alias{IMGTGenomicReferences}
\title{Genomic Reference Alleles for each Gene in the IPD-IMGT/HLA Database}
\format{
A list of two elements.
\itemize{
\item(GenomicReferenceGenes: A table of 44 rows, one for each gene, and 6 columns, identifying each gene, its Ensembl version, GRChg38 location, genomic coordinate of that gene's start codon, GRChg38 reference allele, and IPD-IMGT/HLA reference allele)
\item(Version: The date on which this object was generated)
}
}
\source{
https://www.ebi.ac.uk/ipd/imgt/hla/about/help/genomics/
}
\usage{
data(IMGTGenomicReferences)
}
\description{
A data.frame characterizing a genomic reference allele for each IPD-IMGT/HLA Database gene, and a vector identifying the date that the table was generated.
This object is built by the buildGenomicReferences() function.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/locusValidator.R
\name{validateLocus}
\alias{validateLocus}
\title{Determine if a Locus Name is in the HLAgazetteer}
\usage{
validateLocus(loci, source)
}
\arguments{
\item{loci}{A character vector of HLA gene names (ex. "DRB1", c("DRB1","DQB1")).}

\item{source}{A character vector of alignment source types. "AA", "cDNA", and "gDNA" are allowed types.}
}
\value{
A logical value. TRUE indicates that all of the names and source types are valid. FALSE indicates that at least one locus name or alignment source type is invalid.
}
\description{
Checks a vector of HLA locus names against the HLAgazetteer to determine if the locus name is valid for a specific type of alignment.
}
\note{
The results of this check should only be considered valid for the version of the HLAgazetteer included in the HLAtools package.
}
\examples{
validateLocus(loci = "DRB1", source = "AA")
validateLocus(loci = c("V"), source = c("cDNA","gDNA"))
validateLocus(loci = c("E","F","G"), source = "gDNA")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/accessoryFunctions.R
\name{posSort}
\alias{posSort}
\title{Numerical Sort of Alignment Positions that Contain Indels}
\usage{
posSort(posVec, alignType, locus)
}
\arguments{
\item{posVec}{A character vector of variant positions.}

\item{alignType}{A character string describing the type of alignment the positions are found in. The values 'prot', 'codon', 'nuc' and 'gen' are valid options. Only one 'alignType' value is allowed.}

\item{locus}{A character string describing a locus in the HLAalignments object for the specified alignType.}
}
\value{
A character string of the correctly sorted sequence.
}
\description{
Sorts sequence alignment positions that contain indels in numerical order; e.g., three indels following position X are are identified as X.1, X.2 and X.3.

Positions are validated against a specified locus for a specified alignment type. Positions that do not exist for that locus-alignment combination are not returned.
}
\note{
Indel positions must be text-formatted (e.g. "607.12"). C.f., posSort(c(2,4,3,1,5), "nuc","DRB1") vs posSort(c("607.23","607.10","607.3","607.4"),"nuc","DRB1").
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{sHLAdata}
\alias{sHLAdata}
\title{Synthetic HLA Data for use with Package Examples}
\format{
A data frame of 18 columns and 47 rows.
}
\source{
These synthetic data were generated as part of the 13th International HLA Workshop to demonstrate PyPop functions. See \url{http://pypop.org/docs/guide-chapter-usage.html#data-minimal-noheader} for additional details.
}
\usage{
data(sHLAdata)
}
\description{
A BIGDAWG-formatted data frame of 18 columns and 47 rows containing synthetic HLA-A, -C, -B, -DRB1, -DQA1, -DQB1, -DPA1, and -DPB1 genotype data for 24 control subjects and 23 case subjects. Allele name data were recorded under IPD-IMGT/HLA Database release version 3.56.0. These are synthetic data generated for the imaginary UchiTelle population, and do not represent biologically-derived HLA genotypes.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/accessoryFunctions.R
\name{queryRelease}
\alias{queryRelease}
\title{Search Allele Names Across Release Versions}
\usage{
queryRelease(rel, variant = "", all = FALSE)
}
\arguments{
\item{rel}{An IPD-IMGT/HLA Database release version, represented as either a character (e.g., "3.56.0") or a numeric (e.g., 3560) value.}

\item{variant}{A character string. The value of 'variant' can be any part of a locus or allele name (e.g., "DR", "02:01", "DRB1*08:07"). The default ("") specifies all alleles in 'rel'.}

\item{all}{A logical. When 'all' = TRUE, a vector of all instances of 'variant' in 'rel' is returned.  When 'all' = FALSE, the number of instances of 'var' in 'rel' is returned.}
}
\value{
A character vector of all matches to 'variant' in 'rel' or the number of all matches to 'variant' in 'rel'.
}
\description{
Searches specific release versions in the AlleleListHistory object for user-defined allele variants.
}
\examples{
# Identify the number of DRB9 alleles in releases 3.30.0 and 3.31.0.
queryRelease("3.30.0","DRB9",FALSE) 
queryRelease("3.31.0","DRB9",FALSE)

# Identify the total number of alleles in release 3.56.0.
queryRelease(3560)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/SearchandBuildAlignments.R
\name{alignmentSearch}
\alias{alignmentSearch}
\title{Search Alignments for Specific Positions in a Specific Allele}
\usage{
alignmentSearch(alignType, allelename, positions, prefix = TRUE, sep = "~")
}
\arguments{
\item{alignType}{The type of alignment being searched. Allowed values are "codon","gen", nuc" and "prot". Only one 'alignType' value is allowed.}

\item{allelename}{A full or 2-field HLA allele name, excluding the "HLA-" prefix.}

\item{positions}{A vector of sequence positions (e.g., c(-17,1,22,130)); in nucleotide and genomic alignments, indel positions are named using decimals. So the first indel between positions 26 and 27 is named 26.1, and the second indel between 26 and 27 is 26.2, etc.}

\item{prefix}{A logical that indicates if the position number(s) should be included in the result.}

\item{sep}{The value that will be used to separate the sequences for each position. No value can be specified (sep=""). If prefix=FALSE and sep="", a sequence block is returned.}
}
\value{
A character string containing the corresponding peptide, codon or nucleotide sequence for each position in 'positions' for 'allelename'. A sequence is not returned if 'allelename' is not not found in the pertinent alignment. A position will be empty if 'allelename' does not have a value at the specified position.
}
\description{
Searches ANHIG/IMGT-HLA alignments and returns protein, codons or nucleotide sequences for a submitted allele name and position(s).
}
\note{
This function requires that the HLAalignments object has been populated with alignments via the alignmentFull() function.

Indel positions must be text-formatted (e.g. "607.12").
}
\examples{
\dontrun{
alignmentSearch("gen","DRB1*15:07:01",11:22)
alignmentSearch("nuc","DRB1*15:07:01",11:22)
alignmentSearch("prot","DRB1*15:07:01",11:22)
alignmentSearch("codon","DRB1*15:07:01",11:22)
alignmentSearch("nuc","DRB1*11:250N",c(605,"607.1","607.12",608))
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/alignmentFull.R
\name{alignmentFull}
\alias{alignmentFull}
\title{Build Sets of Protein, Codon, Coding Nucleotide and Genomic Nucleotide Alignments for Specified  Loci}
\usage{
alignmentFull(loci = "all", alignType = "all", version = HLAgazetteer$version)
}
\arguments{
\item{loci}{A characer vector of the locus names for which alignments should be built. The default value ("all") generates alignments for all loci.}

\item{alignType}{A character vector of alignment types. The allowed values are "all", "prot", "codon", "nuc", and "gen", which specify either all available alignments for a given locus or the respective protein, codon, nucleotide and genomic alignments, as determined by the HLAgazetteer.}

\item{version}{A character string describing the version of the ANHIG/IMGTHLA Github repository from which alignments are obtained. The default value, 'HLAgazetteer$version', generates alignments for the IPD-IMGT/HLA Database release under which the HLAgazetteer was built.}
}
\value{
A list object containing data frames of protein (prot), codon (codon), coding nucleotide (nuc), or genomic nucleotide (gen) alignments, for specified genes in the specified IPD-IMGT/HLA Database release, along with the pertinent reference database release.
}
\description{
Applies buildAlignments() to build a set of alignments for loci supported in the ANHIG/IMGTHLA GitHub repository.
}
\note{
AlignmentFull() must be run to build the 'HLAalignments' object (HLAalignments <- alignmentFull()) before other functions that use alignments can be used, and therefore requires internet access.

Depending on local download speeds, building all available alignments for all loci can take several minutes.

Prior to IPD-IMGT/HLA Database release version 3.24.0, the HLA-DP and HLA-DQ sequence alignment files in the IPD-IMGT/HLA GitHub Repository did not all include a numerical suffix in the gene name (e.g., the protein sequence alignment file for the DQA1 gene was named 'DQA_prot.txt') because alignment files for the DPA2, DPB2, DQA2 and DQB2 genes had not been made available. Building DPA1, DPB1, DQA1, and DQB1 sequence alignments from releases prior to 3.24.0 require using a gene name that does not include the numerical suffix.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/motifMatch.R
\name{validateMotif}
\alias{validateMotif}
\title{Determine if a Motif is Properly Formatted}
\usage{
validateMotif(motif, alignType)
}
\arguments{
\item{motif}{A character string identifying a sequence variant motif in the following format: Locus*#$~#$~#$, where ## identifies a variant position, and $ identifies the sequence variant. Both nucleotide and peptide motifs can be provided, and any number of variants can be specified.}

\item{alignType}{A character string identifying the type of alignment being searched. Allowed values are "codon","gen", nuc" and "prot". Only one 'alignType' value is allowed.}
}
\value{
If the motif is valid, TRUE is returned. If the locus or body of the motif is invalid, FALSE is returned along with a brief message.
}
\description{
Evaluates a motif to determine if the locus is valid and variants are valid.
}
\examples{
validateMotif("A*-21M~2P","prot")
validateMotif("A*196G~301A~3046T","gen")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/SearchandBuildAlignments.R
\name{multiAlign}
\alias{multiAlign}
\title{Generate an Alignment for Specific Alleles at Different Positions}
\usage{
multiAlign(alignType, alleles, positions)
}
\arguments{
\item{alignType}{The type of alignment being searched. Allowed values are "prot", codon", "nuc" and "gen".  Only one 'alignType' value is allowed.}

\item{alleles}{A vector of un-prefixed HLA allele names.}

\item{positions}{A list of vectors of nucleotide positions, exactly one vector for each allele, against which each corresponding allele will be aligned.}
}
\value{
A data frame of 'allele' and the corresponding nucleotide sequence for specified positions designated for an allele. A nucleotide sequence is not returned for a specific allele if input allele is not available in the ANHIG/IMGTHLA Github Repository. A position will be left empty if the specific allele does not have a variant at that position. If 'positions' includes values that do not exist in the specified alignment, FALSE is returned.
}
\description{
Generates a peptide, codon, coding nucleotide or genomic alignment for HLA alleles, allowing each allele to be aligned to a different set of positions.
}
\note{
For internal HLAtools use.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/GLStoUNI.R
\name{validateGLstring}
\alias{validateGLstring}
\title{Validate a GL String}
\usage{
validateGLstring(GLstring, version)
}
\arguments{
\item{GLstring}{A character string of allele names and operators in the GL String format.}

\item{version}{A character string identifying the GL String version for evaluation. Options are "1.0" and "1.1".}
}
\value{
A logical. TRUE is returned when all characters in a 'version' GL String are permitted. FALSE is returned when forbidden characters are present, or when an incorrect 'version' is specified.
}
\description{
Evaluates a GL String to identify unsupported characters for a specified GL String version.
}
\examples{
validateGLstring("HLA-A*02:01/HLA-A*02:02+HLA-A*03:01/HLA-A*03:02", version = "1.0")

}
\references{
Mack et al. HLA 2023;102(2):206-212 https://doi.org/10.1111/tan.15126
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{GLSC.ex}
\alias{GLSC.ex}
\title{Example Data Frame of Genotype List String Code Data}
\format{
A dataframe with 419 rows and 2 columns.
}
\source{
pould::hla.hap.demo
}
\usage{
data(GLSC.ex)
}
\description{
A two-column data frame (modified from pould::hla.hap.demo) including GL String data in GL String Code format.
Column one identifies each subject's status, while column two identifies each subject's HLA genotype in GL String Code format.
This data is provided for use in examples and demonstrations.
}
\references{
Mack et al. HLA 2023 Oct;102(4):501-507 https://doi.org/10.1111/tan.15145
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/UNItoGLS.R
\name{multiUNItoGLS}
\alias{multiUNItoGLS}
\title{Translate Multiple UNIFORMAT Strings to GL Strings}
\usage{
multiUNItoGLS(UniformatArray, prefix = "HLA-", pre = TRUE)
}
\arguments{
\item{UniformatArray}{A data frame or vector of UNIFORMAT formatted strings. If 'UniformatArray' is a data frame with more than one column, the first column should contain only identifiers. If 'UniformatgArray' is a vector, it should contain only UNIFORMAT strings.}

\item{prefix}{A character string of the desired locus prefix (default is "HLA-").}

\item{pre}{A logical. If 'pre' is TRUE, all allele names will be prefixed with 'prefix'. If 'pre' is FALSE, no allele names will be prefixed.}
}
\value{
A version of 'UniformatArray' in which the UNIFORMAT data have been converted to GL String format. If a 'UniformatArray' was a data frame, a data frame is returned. If 'UniformatArray' was a vector, a vector is returned.
}
\description{
Translate a data frame or vector of UNIFORMAT strings to GL Strings.
}
\note{
This function does not return the GL String "?" operator, as the "?" operator has no cognate in UNIFORMAT.
}
\examples{
multiUNItoGLS(unname(as.vector(UNIFORMAT.example[2]))[[1]])
multiUNItoGLS(UNIFORMAT.example[1:2])

}
\references{
Nunes Tissue Antigens 2007;69(s1):203-205 https://doi.org/10.1111/j.1399-0039.2006.00808.x

Mack et al. HLA 2023;102(2):206-212 https://doi.org/10.1111/tan.15126
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/atlasFull.R
\name{atlasFull}
\alias{atlasFull}
\title{Generate a Complete set of Protein, Coding nucleotide and Genomic Nucleotide Atlases}
\usage{
atlasFull(version = "Latest")
}
\arguments{
\item{version}{A character string identifying the version of the ANHIG/IMGTHLA Github repository used to generate atlases. The default value, "Latest", generates an atlas for the most recent IPD-IMGT/HLA Database release.}
}
\value{
A list object containing a list of data frames of protein (prot), coding nucleotide (nuc), and genomic nucleotide (gen) atlases for all genes in the IPD-IMGT/HLA Database release, along with a character vector identifying the pertinent IPD-IMGT/HLA Database release.
}
\description{
Applies atlasMaker() to build a set of 'atlases' identifying the gene-feature (exon, intron, and UTR) boundaries in the protein, coding nucleotide, and genomic nucleotide alignments of all HLA loci.
}
\note{
Data informing the atlases were downloaded from the ANHIG/IMGTHLA Github repository.

For internal HLAtools use.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/SearchandBuildAlignments.R
\name{variantTable}
\alias{variantTable}
\title{Builds a Data Frame of Variant Information for Specific Positions in an Alignment}
\usage{
variantTable(alignType, locus, position)
}
\arguments{
\item{alignType}{The type of alignment being searched. Allowed values are "prot", codon", "nuc" and "gen".  Only one 'alignType' value is allowed.}

\item{locus}{A locus supported by the IPD-IMGT/HLA Database.}

\item{position}{A numeric variant position.}
}
\value{
a data frame of three columns identifying the variants, and the number and frequency of each in the alignment.
}
\description{
Builds a data frame sequence variants at specific positions in specified alignments in HLAalignments, along with the counts and frequencies for each variant.
}
\note{
For internal HLAtools use.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/VersionValidation.R
\name{validateVersion}
\alias{validateVersion}
\title{Validate an IPD-IMGT/HLA Release Version}
\usage{
validateVersion(version)
}
\arguments{
\item{version}{A character string identifying the release version (branch) of the ANHIG/IMGTHLA Github repository (e.g. '3.53.0'). The value 'Latest' refers to the most recent release.}
}
\value{
A logical. TRUE indicates that data for 'version' is available in the ANHIG/IMGTHLA GitHub repository. FALSE means that data for 'version' is not available.
}
\description{
Determines if an IPD-IMGT/HLA Release Version is referenced in the AlleleListHistory file.
}
\note{
For internal HLAtools use.
}
\examples{
validateVersion(version = "3.31.1")
validateVersion(version = "3.13.0")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/SearchandBuildAlignments.R
\name{multiSearch}
\alias{multiSearch}
\title{Search Alignment Sequences at Multiple Positions for a Specified Allele}
\usage{
multiSearch(
  alignType,
  locus,
  allele,
  positions,
  prefix = TRUE,
  sep = "~",
  trimmed = FALSE
)
}
\arguments{
\item{alignType}{The type of alignment being searched. Allowed values are "prot", codon", "nuc" and "gen". Only one 'alignType' value is allowed.}

\item{locus}{A specific locus.}

\item{allele}{An allele name.}

\item{positions}{Nucleotide positions to search.}

\item{prefix}{A logical that indicates if the position number(s) should be included in the result.}

\item{sep}{Defines the separator between position sequences.}

\item{trimmed}{A logical identifying whether 'allele' is a two-field, 'trimmed' allele name (trimmed=TRUE), or a full-length name (trimmed=FALSE). The default value is FALSE.}
}
\value{
codons (possibly with position number) at 'positions' for 'allele'. If 'allele' does not have a codon at 'position', '' will be returned.
}
\description{
Generates a character string of multiple protein, codon or nucleotide or genomic sequences at the specified positions for the specified allele name.
}
\note{
For internal HLAtools use only.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/VersionValidation.R
\name{repoVersion}
\alias{repoVersion}
\title{Convert an AlleleListHistory Release Version to the GitHub Repository Version}
\usage{
repoVersion(version)
}
\arguments{
\item{version}{A character string of a validated IPD-IMGT/HLA Database release version.}
}
\value{
A character string of the release version appropriate for use in an IMGT/HLA GitHub repository URL.
}
\description{
Removes periods from IMGT/HLA Database release versions and converts release versions '3.00.0'-'3.09.0' to '300'-'390' to facilitate working with these GitHub repo branches.
}
\note{
For internal HLAtools use.
}
\examples{
repoVersion("3.05.0")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{alleleListHistory}
\alias{alleleListHistory}
\title{Allele Names Across All Release Versions}
\format{
A large list of 2 elements that identify the table of allele names for each IPD-IMGT/HLA reference Database release and the source release
\itemize{
\item(Version: the Date and IPD-IMGT HLA Release version under which the alleleListHistory object was made)
\item(AlleleListHistory: a dataframe version of the alleleListHistory.txt file)
}
}
\source{
https://raw.githubusercontent.com/ANHIG/IMGTHLA/Latest/Allelelist_history.txt
}
\usage{
data(alleleListHistory)
}
\description{
A large list object of two elements. The first element contains version information. The second element contains all releases of allele name lists in the IPD-IMGT/HLA Database.
This object is built by the updateAlleleListHistory() function.
}
\note{
The Allele List History file in IPD-IMGT/HLA Database release version 3.60.0 is too large to be included in the HLAtools package. The alleleListHistory object for release version 3.59.0 is included in the package. The alleleListHistory object in release versions 3.60.0 and later can be loaded via the updateAll() function.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/GLupdater.R
\name{GLV}
\alias{GLV}
\title{Retrieve Version from Input GL String}
\usage{
GLV(GLString)
}
\arguments{
\item{GLString}{A character string of HLA allele names and operators in GL String Code format, signifying their relation with one another and the pertinent IPD-IMGT/HLA Database release version.}
}
\value{
Returns A character string of the version of the release as it appears in the pertinent alleleListHistory$AlleleListHistory-column-header.
}
\description{
Extracts the version data from an input GL String Code, or provides appropriate options if version input is not present in the IPD-IMGT/HLA Database.
}
\note{
For internal use only.
}
\examples{
GLV("hla#3.25.0#HLA-B15:35")

}
\references{
Mack et al. HLA 2023;102(2):206-212 https://doi.org/10.1111/tan.15126

Mack et al. HLA 2023;102(4):501-507 https://doi.org/10.1111/tan.15145
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/locusValidator.R
\name{multiLocusValidation}
\alias{multiLocusValidation}
\title{Apply validateLocus() to Multiple Loci}
\usage{
multiLocusValidation(loci, source = "gDNA", verbose = TRUE)
}
\arguments{
\item{loci}{A character vector of locus names found in the current HLAgazetteer.}

\item{source}{A character vector describing the type of alignment 'loci' should be validated against. Allowed options are 'AA' (protein), 'cDNA' (nucleotide) and 'gDNA' (genomic). Only a single value should be provided. The default value is 'gen'.}

\item{verbose}{A logical value. If 'verbose' = TRUE, messages describing invalid 'loci' or 'source' values are generated. If 'verbose' = FALSE, no messages are generated.}
}
\value{
A character vector of locus names that are present in HLAgazetteer$gen. If none of the locus names are present in HLAgazetteer$gen, an Error is generated.
}
\description{
Applies validateLocus() to a vector of locus names, validates them against HLAgazetteer$gen (by default), and returns a vector of validated locus names.
}
\note{
The results of this check should only be considered valid for the IPD-IMGT/HLA Database release version of the current HLAgazetteer.
}
\examples{
multiLocusValidation(loci = c("DRB1","DPB1","DQB8"))
multiLocusValidation(loci = c("A","B","C","D","Q"))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/accessoryFunctions.R
\name{checkAlignType}
\alias{checkAlignType}
\title{Ensure that AlignType Values are Valid}
\usage{
checkAlignType(alignType)
}
\arguments{
\item{alignType}{A vector of character values specifying sequence alignment types to be used for a function.}
}
\value{
A character vector that includes only allowed 'alignType' values.
}
\description{
Evaluates 'alignType' values to ensure that only "prot", "nuc", "codon" and "gen" are passed to downstream functions. If any other values are entered, a message describing excluded values is generated. If no valid 'alignType' values are are present, an error is generated, and any calling function is ended.
}
\examples{
checkAlignType(c("nuc","prot","gDNA")) 

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/accessoryFunctions.R
\name{numFields}
\alias{numFields}
\title{Identify the Number of Fields in a Colon-Delimited Allele Name}
\usage{
numFields(allele)
}
\arguments{
\item{allele}{A character string of a colon-delimited HLA allele name.}
}
\value{
A numeric value describing the number of fields in the allele name.
}
\description{
Returns the number of fields in a colon-delimited HLA allele name. A value of 1 is returned for digit-delimited HLA allele names.
}
\examples{
numFields("HLA-A*01:01")
numFields("DRB1*04:03:01")
numFields("13:02:01:01")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/AlleleTrim.R
\name{getField}
\alias{getField}
\title{Trim Colon-Delimited HLA Allele Names by Field}
\usage{
getField(allele, res, append = FALSE)
}
\arguments{
\item{allele}{A character string representing an HLA allele.}

\item{res}{A numeric describing the resolution desired.}

\item{append}{A logical. When append = TRUE, the expression variant suffix of a full-length allele name is appended to a truncated allele name. The default value is FALSE.}
}
\value{
A version of the 'allele' character string that has been trimmed to 'res' resolution.
}
\description{
Trims a properly formatted colon-delimited HLA allele name to a desired number of fields.

If an allele name with an expression-variant suffix is truncated, the suffix can be appended to the end of the truncated allele name. If a resolution value greater than the number of fields in the submitted field is specified, the original allele is returned.
}
\note{
For internal HLAtools use.
}
\examples{
getField("HLA-A*01:01:01:01", 3)
getField("DRB1*11:01:01:12N", 2,TRUE)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/motifMatch.R
\name{motifMatch}
\alias{motifMatch}
\title{Identify Alleles that Share a Sequence Motif}
\usage{
motifMatch(motif, alignType, full = TRUE)
}
\arguments{
\item{motif}{A character string identifying a sequence variant motif in the following format: Locus*#$~#$~#$, where ## identifies a variant position, and $ identifies the sequence variant. Both nucleotide and peptide motifs can be provided, and any number of variants can be specified.}

\item{alignType}{A character string identifying the type of alignment being searched. Allowed values are "codon","gen", nuc" and "prot". Only one 'alignType' value is allowed.}

\item{full}{A logical value that specifies if full (TRUE) or truncated (FALSE) allele names should be returned.}
}
\value{
A character vector of allele names that contain the motif, or NA when no alleles contain the motif, NULL when no alignment is available for specified locus, and FALSE when the locus or motif is invalid.
}
\description{
Searches alignments for alleles that share a specific sequence motif.
}
\note{
This function requires an HLAalignments object that has been populated with alignments via alignmentFull().
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/UpdateAll.R
\name{updateAll}
\alias{updateAll}
\title{Update All Package Data Objects Derived from IPD-IMGT/HLA Database Resources}
\usage{
updateAll(updateType = "all", version = getLatestVersion())
}
\arguments{
\item{updateType}{A character vector of the names of data objects to be updated. By default, updateAll() builds all six data objects (updateType="all"). Alternatively specific data objects can be updated; e.g., updateType="alleleListHistory" or updateType=c("alleleListHistory","fragmentFeatureNames").}

\item{version}{A numeric value or character string identifying the version of the ANHIG/IMGTHLA Github repository to build these objects from. By default, updateAll() calls the getLatestVersion() function to identify the most recent IPD-IMGT/HLA Database release.}
}
\value{
No value is returned. The desired versions of the specified data objects are built into the environment that called updateAll().
}
\description{
Applies updateAlleleListHistory(), atlasFull(), buildGazetteer(), extractGeneTypes(), ffN() and buildGenomicReferences() to update the alleleListHistory, HLAatlas, HLAgazetteer, IMGTHLAGeneTypes, fragmentFeatureNames and IMGTGenomicReferences data objects.

A new alleleListHistory data object should be generated with each IPD-IMGT/HLA Database release. The other data objects will likely only change when new genes are added to the IPD-IMGT/HLA Database.
}
\note{
Generating a new HLAatlas can take 5 minutes or more to complete.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/GLupdater.R
\name{multiUpdateGL}
\alias{multiUpdateGL}
\title{Update a Vector of GL String Code Data to a Desired IPD-IMGT/HLA Database Version}
\usage{
multiUpdateGL(GLSCs, to, expand = FALSE, verbose = FALSE)
}
\arguments{
\item{GLSCs}{A vector of GL String Codes.}

\item{to}{A character string identifying the IPD-IMGT/HLA Database release version to translate 'GLSCs' to. Values can range from version 1.05.0 to the loaded version of the alleleListHistory object.}

\item{expand}{A logical that indicates if the lowest-numbered truncated allele names that match truncated allele names in 'GLSCs' should be returned (expand = FALSE), or if a slash-delimited string of all matching full-length allele names should be returned (expand = TRUE). The default value is FALSE.}

\item{verbose}{A logical that indicates if messages regarding the update process should be sent to the console (TRUE) or not (FALSE). The default value is FALSE.}
}
\value{
A version GLSCs (in a data frame) updated to the desired version. NA values are returned in place of alleles that are not present in 'to'.
}
\description{
Updates a column from a data frame in GL String Code format to a desired reference database version.
}
\examples{
multiUpdateGLs <- multiUpdateGL(GLSC.ex$GL.String.Code[1:50],"2.25.0")
multiUpdateGLs <- multiUpdateGL(GLSC.ex$GL.String.Code[1:50],"3.58.0")

}
\references{
Mack et al. HLA 2023;102(2):206-212 https://doi.org/10.1111/tan.15126

Mack et al. HLA 2023;102(4):501-507 https://doi.org/10.1111/tan.15145
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/GIANT.R
\name{GIANT}
\alias{GIANT}
\title{GL string code-Integrated Allele Name Translation}
\usage{
GIANT(data, transFrom, transTo)
}
\arguments{
\item{data}{A data frame or vector of HLA allele name character string data. Allele names in vectors must include locus prefixes. Data frames must include column headers identifying a single locus for the allele names in that column.}

\item{transFrom}{A dot-formatted character string identifying the IPD-IMGT/HLA Database version (e.g. 3.58.0) under which the HLA data were generated.}

\item{transTo}{A dot-formatted character string identifying the IPD-IMGT/HLA Database version (e.g. 2.25.0) to which which the HLA data are to be translated.}
}
\value{
A data frame or vector of HLA allele name data in the 'transTo' release.
}
\description{
Calls updateGL() to translate the HLA allele names in a vector or data frame between two IPD-IMGT/HLA Database release versions.
}
\note{
GIANT() will ignore the first two columns of BIGDAWG-formatted data frames, which include non-HLA data. All columns in non-BIGDAWG-formatted data frames are expected to contain HLA allele names.
}
\examples{
GIANT(c("A*01:01:01:01","DOA*01:01:01:01"),"3.56.0","2.20.0")
updsHLAdata <- GIANT(sHLAdata,"3.56.0","2.25.0")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/accessoryFunctions.R
\name{typeToSource}
\alias{typeToSource}
\title{Convert AlignType Values to Source Values}
\usage{
typeToSource(alignVector, toSource = TRUE)
}
\arguments{
\item{alignVector}{A vector of character values specifying kinds of four sequence alignment types ("prot","nuc","codon" and "gen"), or three source alignment files ("AA","cDNA", and "gDNA").}

\item{toSource}{A logical. If 'toSource' is true, 'alignType' values are converted to 'source' values. If 'toSource' is false, 'source' values are concerted to 'alignType' values}
}
\value{
A character string of the converted 'align' or 'source' value.
}
\description{
Converts between 'alignType' values, identifying four types of sequence alignments and 'source' values, identifying three kinds of alignment files.
}
\examples{
typeToSource(c("nuc","prot","gen"),TRUE) 

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{IMGTHLAGeneTypes}
\alias{IMGTHLAGeneTypes}
\title{Molecular characteristics of the Genes Curated by the IPD-IMGT/HLA Database}
\format{
A list object of two elements:
\itemize{
\item(GeneTypes: a data frame of three columns)
\item(version: a character string identifying the date that the source file was written)
}
}
\source{
@source https://hla.alleles.org/pages/genes/genes_list
}
\usage{
data(IMGTHLAGeneTypes)
}
\description{
A data frame of three columns identifying each gene supported by the IPD-IMGT/HLA Database, its molecular characteristics, and its status as a gene fragment or pseudogene.
This object is built by the BuildIMGTHLAGeneTypes() function.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/SearchandBuildAlignments.R
\name{queryPositions}
\alias{queryPositions}
\title{Extract Variant Information for Specific Positions in an Alignment}
\usage{
queryPositions(alignType, locus, positions, table = FALSE)
}
\arguments{
\item{alignType}{The type of alignment being searched. Allowed values are "prot", codon", "nuc" and "gen".  Only one 'alignType' value is allowed.}

\item{locus}{A locus supported by the IPD-IMGT/HLA Database.}

\item{positions}{A vector of numeric variant positions.}

\item{table}{A logical value specifying if a table of counts and frequencies for each variant at each value of 'positions' should be returned (table = TRUE). The default value is FALSE.}
}
\value{
When 'positions' specifies a single position, a character vector of variants is returned. When multiple positions are specified, a list of such character vectors is returned. When 'table' is TRUE, a list of data frames identifying the number (Count) and frequency (Frequency) of each variant is returned.
}
\description{
Identifies the sequence variants at specific positions in specified alignments in HLAalignments. When 'table' is TRUE, data frames of counts and frequencies for each variant at each position are returned.
}
\note{
This function requires that the HLAalignments object has been populated with alignments via the alignmentFull() function.

The '.' value in a vector of positions identifies absent sequence; e.g. following a nonsense mutation or in a region of sequence aligned to an insertion. The '*' value in a vector of positions identifies unknown sequence.
}
\examples{
\dontrun{
queryPositions("prot","DRB1",c(5,10,86,129))
queryPositions("prot","DRB1",c(5,10,86,129),TRUE)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fragmentFeatureNames.R
\name{ffN}
\alias{ffN}
\title{Identify and Annotate Gene Features in Pseudogenes and Gene Fragments.}
\usage{
ffN(version)
}
\arguments{
\item{version}{A character string identifying the pertinent IPD-IMGT/HLA Database release version (e.g., "3.55.0") under which the returned object is generated. This parameter does not impact the generation of the feature names or annotations, and is only included to provide IPD-IMGT/HLA Database release version context.}
}
\value{
A list object where each element is the name of a pseudogene or gene fragment. Each of these elements is a list of 'features' and 'annotation'. "Features" identifies the gene features for that gene, ordered from 5' to 3'. Because these genes are not-expressed, the three standard gene features (U, Untranslated Region (UTR); E, Exon; and I, Intron) may not always be present. In these cases, H, J, N and S features (defined below) are returned. Each feature identifier is followed by an identifying number (e.g. U.5 is the 5' UTR, and E.3 is Exon 3). "Annotation" provides the composition of the non-standard H, J, N and S features for each gene.
#'
}
\description{
HLA pseudogenes and gene fragments many not share all of the gene features of their expressed homologs. The ffN() function generates a list object identifying the features present in a given gene fragment or pseudogene with a genomic alignment file. This includes annotations of the differences between the pseudogenes and gene fragments, and their expressed homologs.
Pseudogenes and gene fragments are identified in the IMGTHLAGeneTypes data object.

Standard Features
\itemize{
\item E - Exon, a peptide-encoding sequence
\item I - Intron, an intervening sequence found between Exons
\item U - UTR, an untranslated region of sequence preceding the first Exon or following the last Exon
}

Additional Features used in these annotations, based on boundaries indicated in the sequence alignment, are:
\itemize{
\item H - Hybrid, a sequence that includes at least one known feature sequence and one nucleotide sequence that does not correspond to a known feature
\item J - Join, a sequence that includes two or more features that are separated by a feature boundary in the reference
\item N - Novel, a novel sequence that does correspond to a known feature sequence
\item S - Segment, a subset of a longer feature sequence
}

Reference Genes
\itemize{
\item HLA-C genomic sequence is used as the reference for class I pseudogenes and gene fragments.
\item HLA-DPA1 and -DPB1 genomic sequences are used as the references for class II pseudogenes -DPA2 and -DPB2, respectively.
}
}
\note{
Features and their annotations have been identified manually. Feature annotations will not change unless a new pseudogene or gene fragment is added in a future release, in which case new annotations will be generated.

The H, J, N and S features are described for class I pseudogenes and gene fragment sequences, all of which are described relative to the HLA-C reference sequence. Feature length differences for DPA2 and DPB2, relative to the DPA1 and DPB1 references, are noted in annotations of standard feature abbreviations (E, I and U).

No annotations are included for the DRB2, DRB6, DRB7, and DRB9 genes, as genomic alignments for these genes are not available.

For internal HLAtools use.
}
\examples{
fragmentFeatureNames <- ffN("3.35.0")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/buildAlignments.R
\name{buildAlignments}
\alias{buildAlignments}
\title{Build Amino Acid, cDNA and gDNA Alignments}
\usage{
buildAlignments(loci, source, version = "Latest", return_corr_table = FALSE)
}
\arguments{
\item{loci}{A character vector of HLA gene names (e.g., "DRB1", c("A","C")).}

\item{source}{A character vector of alignment types. The allowed values are "AA", "cDNA", and "gDNA". If 'source' is "cDNA", both codon and cDNA nucleotide alignments are generated. If source is 'AA' or 'gDNA', a single peptide or genomic nucleotide alignment is generated. Up to four alignments will be returned for a locus, as determined by its ability to be transcribed or translated.}

\item{version}{A character string describing the desired release version (branch) of the ANHIG/IMGTHLA Github repository (e.g. '3.53.0'). The default value ('Latest') returns alignments for the most recent release.}

\item{return_corr_table}{A logical. When return_corr_table = TRUE, correspondence tables are returned instead of the alignments. The default value is FALSE.}
}
\value{
A list object with a data frame of all allele names (and trimmed, two-field allele names) and their corresponding sequences (Amino Acid, codon, cDNA, or gDNA) for a specific locus, as well as version details for the returned information, when return_corr_table = FALSE. These alignments identify locations of feature boundaries in relation to amino acid, codon, cDNA, and gDNA sequences. When a three- or four-field allele name includes an expression variant suffix, that suffix is appended to the trimmed name. When return_corr_table = TRUE, correspondence tables, relating the named positions in an alignment to the ordinal numbers for each named position and the ordinal numbers of each insertion position, for the specified loci and source are returned, along with version details for the returned information.
}
\description{
Returns a list of data frames of amino acid, codon, nucleotide and genomic alignments, with version information.
}
\note{
For internal HLAtools use.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/GLupdater.R
\name{GLV2}
\alias{GLV2}
\title{Format GL String Code Version Number}
\usage{
GLV2(Version)
}
\arguments{
\item{Version}{A character string of the desired IPD-IMGT/HLA Database version, going back to version 1.05.0.}
}
\value{
Returns either a single, compressed character string-formatted release version, or a character vector of potential release versions.
}
\description{
Returns a compressed IPD-IMGT/HLA Database release version or a vector of dot-delimited release version options if the specified version is not present in the alleleListHistory object.
}
\note{
For internal use only.
}
\examples{
GLV2("3.34.0")
GLV2("3.0.0")

}
\references{
Mack et al. HLA 2023;102(2):206-212 https://doi.org/10.1111/tan.15126

Mack et al. HLA 2023;102(4):501-507 https://doi.org/10.1111/tan.15145
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{HLAgazetteer}
\alias{HLAgazetteer}
\title{Functional and Organizational Categories of Genes Supported by the IPD-IMGT/HLA Database}
\format{
A large list of 24 vectors that define specific categories of genes supported by the IPD-IMGT/HLA Database
\itemize{
.   \item(align: all genes with alignments in the IPD/IMGT-HLA GitHub Repository)
\item(gen: genes with genomic alignments in the IPD/IMGT-HLA GitHub Repository)
\item(nuc: genes with nucleotide alignments in the IPD/IMGT-HLA GitHub Repository)
\item(prot: genes with protein alignments in the IPD/IMGT-HLA GitHub Repository)
\item(nogen: genes with no genomic alignments)
\item(nonuc: genes with no nucleotide alignments)
\item(noprot: genes with no protein alignments)
\item(pseudo: pseudogenes)
\item(frag: gene fragments)
\item(hla: HLA genes)
\item(expressed: genes that are expressed)
\item(notexpressed: genes that are not expressed)
\item(classireg: genes found in the HLA class I region)
\item(classihla: class I HLA genes)
\item(classiireg: genes found in the HLA class II region)
\item(classiihla: class II HLA genes)
\item(classical: classical HLA genes)
\item(nonclassical: non-classical HLA genes)
\item(alpha: genes found in the alpha ancestral haplotype block)
\item(kappa: genes found in the kappa ancestral haplotype block)
\item(beta: genes found in the beta ancestral haplotype block)
\item(delta: genes found in the delta ancestral haplotype block)
\item(epsilon: genes found in the epsilon ancestral haplotype block)
\item(map: all genes organized by 5' to 3' map order on the genomic reference + strand)
\item(version: IPD-IMGT/HLA Database version used to build the HLAgazetteer)
}
}
\source{
https://hla.alleles.org/genes

https://github.com/ANHIG/IMGTHLA/tree/Latest/alignments

Andersson Front. Biosci. 1998, 3(4), 739–745. https://doi.org/10.2741/a317

Alexandrov et al. HLA 2023, Vol.102(5), p.599-606. https://doi.org/10.1111/tan.15180

Dawkins et al. Immunol Rev. 1999, Vol. 167, p.275-304. https://doi.org/10.1111/j.1600-065x.1999.tb01399.x

Kulski et al. Hum Genome Var. 2022 Dec 21;9(1):49. https://doi.org/10.1038/s41439-022-00226-5

https://www.ncbi.nlm.nih.gov/nucleotide/NC_000009.12
}
\usage{
data(HLAgazetteer)
}
\description{
A list object of nineteen vectors that identify genes in the HLA region that share specific commonalities.
This object is built by the buildGazetteer() function.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/VersionValidation.R
\name{expandVersion}
\alias{expandVersion}
\title{Add 'Dot' Delimiters to a Numeric Release Version}
\usage{
expandVersion(ver)
}
\arguments{
\item{ver}{A four character/numeral IPD-IMGT/HLA Database Release Version name (e.g., '3540' or 3450).}
}
\value{
A character vector of the expanded IPD-IMGT/HLA Database Release Version name.
}
\description{
Adds the '.' delimiters to a 'squashed' IPD-IMGT/HLA Database Release Version name
}
\note{
For internal HLAtools use.

This function assumes single-character-length first and third fields of IPD-IMGT/HLA Database Release Version names.
}
\examples{
expandVersion(3450)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/accessoryFunctions.R
\name{parseAlignmentHead}
\alias{parseAlignmentHead}
\title{Guides for Parsing the Header Blocks of Alignment Files}
\usage{
parseAlignmentHead(version)
}
\arguments{
\item{version}{A character string of a validated IPD-IMGT/HLA Database release version (e.g., '3.25.0' or '3250').}
}
\value{
Either FALSE if the version is not valid, or two-value numerical vector describing (1) the header line in which alignment version data appears and (2) the length of the character string in that line preceding version data.
}
\description{
Returns a vector describing the location of key information in the header blocks of alignment files.
}
\examples{
parseAlignmentHead("3.25.0")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/GLStoUNI.R
\name{GLStoUNI}
\alias{GLStoUNI}
\title{Translate GL String to UNIFORMAT}
\usage{
GLStoUNI(GLstring, prefix = "HLA-", pre = FALSE)
}
\arguments{
\item{GLstring}{A character string of HLA allele names and operators in the GL String format signifying their relation with one another.}

\item{prefix}{A character string of the pertinent gene-system prefix (default is "HLA-").}

\item{pre}{A logical that indicates whether returned allele names should contain 'prefix' (TRUE), or if 'prefix' should be excluded from returned names (FALSE).}
}
\value{
A version of 'GLstring' converted to UNIFORMAT format, or FALSE if 'GLstring' is invalid.
}
\description{
A wrapper function for GLtoUN, which translates strings from GL String format to UNIFORMAT format.
}
\note{
This function does not function with the "?" operator, as the "?" operator has no cognate in UNIFORMAT.
}
\examples{

GLStoUNI("HLA-A*02:01/HLA-A*02:02+HLA-A*03:01/HLA-A*03:02")
GLStoUNI("A+B/C~D^G|E^W+X/Y^Z+J")

}
\references{
Nunes Tissue Antigens 2007;69(s1):203-205 https://doi.org/10.1111/j.1399-0039.2006.00808.x

Mack et al. HLA 2023;102(2):206-212 https://doi.org/10.1111/tan.15126
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/relRisk.R
\name{relRisk}
\alias{relRisk}
\title{Calculate Relative Risk for Individual Alleles and Genotypes in BIGDAWG-formatted Non-Case-Control Datasets}
\usage{
relRisk(dataset, return = TRUE, save.path = tempdir())
}
\arguments{
\item{dataset}{A character string describing the name of a non-case-control genotype dataset using the BIGDAWG format. Here, "non-case-control" means that while two subject categories are required, the categories should not be patients and controls; instead, the categories may be, e.g., for a dataset of patients, either of two disease states, where one disease state is coded as 0 and the other is coded as 1 in the second column of the dataset. Either a tab-delimited file or a data frame can be specified.}

\item{return}{A logical identifying if the list object should be returned (return=TRUE), or if pairs of tab-delimited text files of results (one for alleles and one for genotypes) should be written to the working directory for each locus.}

\item{save.path}{A character string identifying the path in which to write the pair of files when return is FALSE. The default value is tempdir().}
}
\value{
A list object of two lists ("alleles" and "genotypes"), each of which contains a list of nine-column data frames containing results for each unique allele or genotype (in rows) at each locus. Column headers in each dataframe are, \emph{Locus}, \emph{Variant}, \emph{Status_1}, \emph{Status_0}, \emph{RelativeRisk}, \emph{CI.low}, \emph{CI.high}, \emph{p.value}, and \emph{Significant}.
}
\description{
Returns a list object containing relative risk, confidence interval and p-value data for the individual alleles and individual genotypes at each locus in a BIGDAWG-formatted non-case-control genotype data frame or file.
}
\examples{
rr <- relRisk(sHLAdata)

}
\references{
\href{https://CRAN.R-project.org/package=BIGDAWG/vignettes/BIGDAWG.html#input-data}{BIGDAWG Data Format}
}
\keyword{allele}
\keyword{genotype}
\keyword{relative}
\keyword{risk}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/SearchandBuildAlignments.R
\name{validatePositions}
\alias{validatePositions}
\title{Validate the Presence of Variant Positions in an Alignment}
\usage{
validatePositions(alignType, locus, positions)
}
\arguments{
\item{alignType}{The type of alignment being searched. Allowed values are "prot", codon", "nuc" and "gen".  Only one 'alignType' value is allowed.}

\item{locus}{A locus supported by the IPD-IMGT/HLA Database.}

\item{positions}{A vector of variant positions.}
}
\value{
A logical identifying if the position is present in the alignment (TRUE) or, if it is not in the alignment or is not valid (FALSE).
}
\description{
Returns TRUE when variant positions are present in a specified alignment in HLAalignments. Returns false when any variant position is not present in the specified alignment and generates a message identifying such positions.
}
\note{
For internal HLAtools use.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/GLupdater.R
\name{redec}
\alias{redec}
\title{Reintroduce Version Decimals}
\usage{
redec(Cname)
}
\arguments{
\item{Cname}{A character string describing a version name in AlleleListHistory column format.}
}
\value{
A character string describing a version name with decimals in appropriate places according to the IPD-IMGT/HLA Database.
}
\description{
Correctly put decimals back into an AlleleListHistory column-formatted version name.
}
\note{
For internal use only.
}
\examples{
redec("X3090")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/extractGenomicReferenceGenes.R
\name{buildGenomicReferences}
\alias{buildGenomicReferences}
\title{Generate a Table of Genomic Reference Alleles and Characteristics of HLA Genes Defined by the IPD-IMGT/HLA Database}
\usage{
buildGenomicReferences()
}
\value{
A list object of two elements -- GenomicReferences and Version.
}
\description{
buildGenomicReferences() extracts the table of genomic HLA reference alleles at 'www.ebi.ac.uk/ipd/imgt/hla/about/help/genomics/' and generates the IMGTGenomicReferences data object.
}
\note{
This function requires internet access to function.

For internal HLAtools use.
}
\references{
Barker et al. NAR (2025) 54:D1152–D1158 https://doi.org/10.1093/nar/gkaf1218
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/SearchandBuildAlignments.R
\name{uniAlign}
\alias{uniAlign}
\title{Generate an Alignment for Specific Alleles at Specific Positions}
\usage{
uniAlign(alignType, alleles, positions)
}
\arguments{
\item{alignType}{The type of alignment being searched. Allowed values are "prot", codon", "nuc" and "gen". Only one 'alignType' value is allowed.}

\item{alleles}{A vector of un-prefixed HLA allele names.}

\item{positions}{A vector of codon positions, against which all loci will be aligned.}
}
\value{
A data frame of allele names and the corresponding codon sequence for specified position. A codon sequence is not returned for alleles that are not present in the loaded HLAalignments object. A sequence position will be populated with whitespace if a specific allele has no sequence at the specified position. If 'positions' includes values that do not exist in the specified alignment, FALSE is returned.
}
\description{
Generates a peptide, codon, coding or genomic nucleotide alignment at a single set of positions for HLA alleles at one or more loci.
}
\note{
For internal HLAtools use.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{HLAatlas}
\alias{HLAatlas}
\title{Boundary Positions of Exons, Introns and UTRs in Amino Acid, cDNA and gDNA Alignments}
\format{
A list of 4 elements that include the gene features boundaries in each ANHIG/IMGTHLA sequence alignment.
\itemize{
\item(prot: peptide-alignment atlases)
\item(nuc: cDNA- and codon-alignment atlases)
\item(gen: gDNA-alignment atlases)
\item(version: The IPD-IMGT/HLA Database release version under which these data were generated)
}
}
\source{
https://github.com/ANHIG/IMGTHLA/tree/Latest/alignments
}
\usage{
data(HLAatlas)
}
\description{
A list object of two wlements. Sub-lists of R dataframes (atlases) for each locus with a protein (prot), cDNA (nuc), and gDNA (gen) alignment. Each atlas identifies the position of the exon (E), intron (I) or UTR (U) gene-feature boundaries in an alignment. Boundaries for non-standard hybrid (H), join (J), novel (N) and segment (S) features may be included in gene fragment and pseudogene atlases.
This object is built by the atlasFull() function.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ThirdPartyFunctions.R
\name{countSpaces}
\alias{countSpaces}
\title{Count the Spaces in a Character String}
\source{
https://www.reddit.com/r/rstats/comments/2th8ic/function_to_count_the_number_of_white_spaces/
}
\usage{
countSpaces(x)
}
\arguments{
\item{x}{A character string.}
}
\value{
A vector of numeric values identifying the number of instances of spaces (' ') in x, and the number of consecutive spaces in an instance.
}
\description{
Counts the number of spaces in a character string.
}
\examples{
countSpaces("abc def")
countSpaces("abc def  ghi   jkl mno")

}
\author{
Josh Bredeweg, Reddit user jbraids1421
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/AlleleTrim.R
\name{multiAlleleTrim}
\alias{multiAlleleTrim}
\title{Trim Multiple HLA Allele Names}
\usage{
multiAlleleTrim(alleles, resolution, version = 3, append = FALSE)
}
\arguments{
\item{alleles}{A vector of HLA allele names from a single nomenclature epoch.}

\item{resolution}{A number identifying the number of fields to which the alleles should be trimmed.}

\item{version}{A number identifying the HLA nomenclature epoch under which the allele was named. Epoch 1 allele names are found in IPD-IMGT/HLA Database releases 1.0.0 to 1.16.0. Epoch 2 allele names are found in IPD-IMGT/HLA Database releases 2.0.0 to 2.28.0. Epoch 3 allele names are found in IPD_IMGT/HLA Database releases 3.0.0 and onward.}

\item{append}{A logical. When append = TRUE, the expression variant suffix of a full-length allele name is appended to a truncated allele name. The default value is FALSE.}
}
\value{
A vector of the trimmed allele name, shortened according to the input parameters.
}
\description{
Trim a vector of allele names to a specified number of fields or digits.
}
\note{
Expression variant suffixes will not be removed from full-length allele names that have fewer than four fields or eight digits.
}
\examples{
alleles <- c("A*02:01:01:02L","DRB1*08:07", "DQB1*04:02:01:16Q")
multiAlleleTrim(alleles,2)
multiAlleleTrim(alleles,2,append = TRUE)

alleles <- c("A*01010102L","DRB1*1613N","HLA-Cw*0322Q")
multiAlleleTrim(alleles,1,2,TRUE)
multiAlleleTrim(alleles,2,2)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/checkgDNAStart.R
\name{checkgDNAstart}
\alias{checkgDNAstart}
\title{Identify gDNA Alignments in Which the First Feature Boundary is not Identified as Position +1.}
\usage{
checkgDNAstart(verbose = FALSE)
}
\arguments{
\item{verbose}{A logical indicating if loci with first feature boundary positions that are not 1 should be reported in the console (verbose = TRUE).}
}
\value{
A one-row data frame with one column for each locus with a gDNA alignment.
}
\description{
Checks the position of the first feature boundary (the 5' UTR - Exon 1 boundary in expressed genes) in each gDNA alignment, and identifies those alignments in which that position is not 1.
}
\details{
This function reviews the gDNA atlases in the HLAatlas object and returns a data frame of the first feature boundary position for each locus. For expressed genes and some pseudogenes, this is the position of the start of Exon 1.
}
\note{
This function requires that the HLAalignments object has been populated with alignments via the alignmentFull() function.

For internal HLAtools use.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/BDtoPyPop.R
\name{formatHead}
\alias{formatHead}
\title{Format PyPop Data Frame Headers}
\usage{
formatHead(colHead)
}
\arguments{
\item{colHead}{A vector of column names. PyPop format requires that paired locus/gene names should end in '_1' and '_2', respectively.}
}
\value{
A vector in which the locus names are suffixed with '_1' and '_2'.
}
\description{
Format the header of a PyPop-formatted data frame.
}
\note{
This function assumes that the first two elements the 'colHead' vector are not locus/gene names.
}
\examples{
formatHead(colHead = colnames(sHLAdata))

}
\references{
Lancaster et al. Front Immunol. 2024 Apr 2;15:1378512. https://pubmed.ncbi.nlm.nih.gov/38629078/

Pappas et al. Hum Immunol. 2016 Mar 77(3):283-287. https://pubmed.ncbi.nlm.nih.gov/26708359/
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/buildGazetteer.R
\name{buildGazetteer}
\alias{buildGazetteer}
\title{Define Categories of Genes Supported by the IPD-IMGT/HLA Database}
\source{
Andersson G. Evolution of the human HLA-DR region. Front. Biosci. 1998 Jul 27:3:d739-45.

Alexandrov et al. HLA-OLI: A new MHC class I pseudogene and HLA-Y are located on a 60 kb indel in the human MHC between HLA-W and HLA-J. HLA 2023 102(5):599-606.

Dawkins et al. Genomics of the major histocompatibility complex: haplotypes, duplication, retroviruses and disease. Immunological Reviews 1999 167:275-304.

Kulski et al. Human leukocyte antigen super-locus: nexus of genomic supergenes, SNPs, indels, transcripts, and haplotypes. Human Genome Variation 2022 9:1–15.
}
\usage{
buildGazetteer(version = getLatestVersion())
}
\arguments{
\item{version}{A string identifying the desired IPD-IMGT/HLA Database release version to which the gazetteer should be updated. The default value is most recent IPD-IMGT/HLA Database release version.}
}
\value{
A list object of vectors organizing the genes in the IPD-IMGT/HLA Database into specific categories.
}
\description{
Consumes information in the ANHIG/IMGTHLA GitHub repository and at hla.alleles.org/genes to define specific categories of genes supported by the IPD-IMGT/HLA Database, which are represented as nineteen elements of the HLAgazetteer object.

Elements:
\itemize{
\item All genes with alignments ($align)
\item Genes that that do and do not have amino acid ($prot/$noprot), nucleotide ($nuc/$nonuc), and genomic ($gen/$nogen) alignments
\item HLA genes ($hla), pseudogenes ($pseudo), gene fragments ($frag)
\item Genes that are expressed ($expressed) or not expressed ($notexpressed)
\item Genes in the Class I region ($classireg), Class I HLA genes ($classIhla), Genes in the Class II region ($classiireg), and Class II HLA genes ($classiihla)
\item Classical HLA genes ($classical) and non-classical expressed HLA genes ($nonclassical)
\item HLA region genes present in specific conserved ancestral HLA region haplotype blocks ($alpha, $kappa, $beta, $delta and $epsilon)
\item All genes presented in map order ($map)
}

The twentieth element ($version) identifies the IPD-IMGT/HLA Database version used to build the HLAgazetteer.
}
\note{
The \emph{$prot} and \emph{$nuc} vectors include a 'DRB' "gene". While 'DRB' is not a gene name, the DRB_prot.txt file includes combined alignments for the DRB1, DRB3, DRB4, and DRB5 genes, and the DRB_nuc.txt file includes combined alignments for the DRB1, DRB2, DRB3, DRB4, DRB5, DRB6, DRB7, DRB8, and DRB9 genes. 'DRB' is included in these vectors for the purpose of validation when these combined alignments are desired.

For interal HLAtools use.

These elements are constructed using data compiled from hla.alleles.org/genes/, github.com/ANHIG/IMGTHLA/tree/Latest/alignments, DOI:10.2741/a317, DOI:10.1111/tan.15180, DOI: 10.1038/s41439-022-00226-5, and Human Genome Assembly GRCh38.p14 reference assembly NC_000006.12.

Additional information about these genes can be found in HLAtools::IMGTHLAGeneTypes.

The $map element does not distinguish the order of the DRB3, DRB4 and DRB5 genes (DRB3/4/5) or the DR6 and DR7 (DR6/7) genes, as the individual elements of these gene sets are found on different DRB haplotypes.

This function requires internet access to function.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/BDtoPyPop.R
\name{pypopHeaders}
\alias{pypopHeaders}
\title{Convert BIGDAWG File Headers to PyPop Format}
\usage{
pypopHeaders(colHead)
}
\arguments{
\item{colHead}{A vector of column names. BIGDAWG format requires that the first two header values are the sample identifier character string and the subject status (0 or 1). All other header values should be paired locus/gene names.}
}
\value{
A PyPop-formatted column header vector.
}
\description{
Convert the header of a BIGDAWG-formatted data frame into a Python for Population Genomics (PyPop) formatted header.
}
\note{
This function returns unique locus names for each column. For example, "A_1" and "A_2" headers will be returned for two "A" locus columns. PyPop configuration (.ini) files allow this format to be customized.
}
\examples{
pyHead <- pypopHeaders(colnames(sHLAdata))

}
\references{
Lancaster et al. Front Immunol. 2024 Apr 2;15:1378512. https://pubmed.ncbi.nlm.nih.gov/38629078/

Pappas et al. Hum Immunol. 2016 Mar 77(3):283-287. https://pubmed.ncbi.nlm.nih.gov/26708359/
}
\seealso{
\href{http://pypop.org/docs/guide-chapter-usage.html#a-minimal-configuration-file}{PyPop Configuration File}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/BDtoPyPop.R
\name{convertAny}
\alias{convertAny}
\title{Convert Values Across an Entire Data Frame or Vector}
\usage{
convertAny(dataset, change.from = NA, change.to = "****")
}
\arguments{
\item{dataset}{A data frame or vector of values.}

\item{change.from}{The value present in the data frame or vector that should be changed. The default value is NA.}

\item{change.to}{The value to which 'change.from' should be changed to. The default value is "****".}
}
\value{
A data frame or vector in which 'change.from' has been converted to 'change.to'.
}
\description{
Converts all instances of a value in a data frame or vector to a specified value.
}
\examples{
dataset <- convertAny(sHLAdata)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/AlleleTrim.R
\name{alleleTrim}
\alias{alleleTrim}
\title{Trim All Versions of HLA Allele Names}
\usage{
alleleTrim(allele, resolution, version = 3, append = FALSE)
}
\arguments{
\item{allele}{A character string of an HLA allele name formatted as locus*allele_name, optionally including the "HLA-" prefix.}

\item{resolution}{A number identifying the number of fields to trim the allele name to.}

\item{version}{A number identifying the HLA nomenclature epoch under which the allele was named. Epoch 1 allele names are found in IPD-IMGT/HLA Database releases 1.0.0 to 1.16.0. Epoch 2 allele names are found in IPD-IMGT/HLA Database releases 2.0.0 to 2.28.0. Epoch 3 allele names are found in IPD_IMGT/HLA Database releases 3.0.0 and onward.}

\item{append}{A logical. When append = TRUE, the expression variant suffix of a full-length allele name is appended to a truncated allele name. The default value is FALSE.}
}
\value{
A character string of the trimmed allele name, shortened according to the input parameters.
}
\description{
Trims an HLA allele name to a specified number of fields or number of digits, depending on the nomenclature epoch. Expression variant suffixes in full-length allele names can be appended to truncated allele names.
}
\note{
Expression variant suffixes will not be removed from full-length allele names that have fewer than four fields or eight digits.
}
\examples{
alleleTrim(allele = "A*03:01:01", resolution = 2)
alleleTrim(allele = "A*030101", resolution = 2,version = 2)
alleleTrim(allele = "A*0303N",resolution = 1,version = 1, append = TRUE)
alleleTrim("HLA-A*24020102L",resolution = 3,version = 2, append = TRUE)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/accessoryFunctions.R
\name{checkSource}
\alias{checkSource}
\title{Ensure that Source Values are Valid}
\usage{
checkSource(source)
}
\arguments{
\item{source}{A vector of character values specifying sequence alignment file sources to be used for a function.}
}
\value{
A character vector that includes only allowed 'source' values.
}
\description{
Evaluates 'source' values to ensure that only "AA", "cDNA", and "gDNA" are passed to downstream functions. If any other values are entered, a message describing excluded values is generated.
}
\examples{
checkSource(c("AA","cDNA","codon")) 

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/accessoryFunctions.R
\name{verifyAllele}
\alias{verifyAllele}
\title{Determine if an Allele Name Ever Existed, and (if so) in Which IPD-IMGT/HLA Database Releases}
\usage{
verifyAllele(allele, version = FALSE, all = FALSE, first = FALSE, acc = FALSE)
}
\arguments{
\item{allele}{A character string of an HLA allele name. Colon-delimited and field-delimited names are both accepted.}

\item{version}{A logical that indicates if the most recent nomenclature release version in which that name was valid should be returned. The default value is FALSE.}

\item{all}{A logical that indicates if all of the nomenclature release versions in which that name was valid should be returned. The default value is FALSE.}

\item{first}{A logical that indicates if only the original nomenclature release version in which that name was valid should be returned. The default value is FALSE.}

\item{acc}{A logical that indicates if the IPD-IMGT/HLA accession number for an allele name should be returned. The default value is FALSE.}
}
\value{
A logical identifying if the allele name is found in AlleleListHistory (TRUE) or not (FALSE), c(TRUE,version) if version = TRUE, and c(TRUE,acc) or c(TRUE,version,acc) if acc = TRUE.
}
\description{
Returns TRUE if an allele name is present in AlleleListHistory or FALSE it is absent, c(TRUE,version), where 'version' can be the most recent, original, or all IPD-IMGT/HLA Database releases in which that name appears, and c(TRUE,version,acc), where 'acc' is the IPD-IMGT/HLA accession number for the allele.
}
\examples{
verifyAllele("A*01:01:01:01")
verifyAllele("A*01:01:01:01",TRUE)
verifyAllele("A*010101",TRUE,TRUE)
verifyAllele("A*0101",TRUE)
verifyAllele("A*02:01:01:01",FALSE,FALSE,FALSE,TRUE)
verifyAllele("A*02:01:01:01",TRUE,TRUE,FALSE,TRUE)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/GLupdater.R
\name{translateAllele}
\alias{translateAllele}
\title{Translate HLA Allele Names across IPD-IMGT/HLA Database Release Versions}
\usage{
translateAllele(allele, from, to, expand = FALSE, verbose = FALSE)
}
\arguments{
\item{allele}{A full or truncated HLA allele name.}

\item{from}{A character string identifying the IPD-IMGT/HLA Database release version of 'allele'.}

\item{to}{A character string identifying the IPD-IMGT/HLA Database release version to translate 'allele' to. Values can range from version 1.05.0 to the loaded version of the the alleleListHistory object.}

\item{expand}{A logical that indicates if the lowest-numbered truncated allele name that matches a truncated allele name should be returned (expand = FALSE), or if a slash-delimited string of all matching full-length allele names should be returned (expand = TRUE). The default value is FALSE.}

\item{verbose}{A logical that indicates if messages regarding allele translations should be sent to the console (TRUE) or not (FALSE). The default value is FALSE.}
}
\value{
A character string of either a single matching allele or all matching alleles in the 'to' release version. If an allele does not have a cognate match in 'to', NA is returned for that allele.
}
\description{
Translates a single HLA allele name across IPD-IMGT/HLA Database release versions.  Truncated allele names in one version can be expanded to a list of all allele names that match the provided allele name in a chosen version.
}
\examples{
translateAllele("A*01:01","3.01.0","2.20.0",FALSE)
translateAllele("A*0101","2.20.0","3.01.0",TRUE)
translateAllele("B*57:01","3.12.0","3.12.0",TRUE)

}
